/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.clientattribute;

import static org.springframework.data.mongodb.core.query.Criteria.*;
import static org.springframework.data.mongodb.core.query.Query.*;

import java.util.Objects;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.mongodb.core.MongoOperations;
import org.springframework.data.mongodb.core.query.Criteria;
import org.springframework.data.mongodb.core.query.Query;
import org.springframework.data.mongodb.core.query.Update;
import org.springframework.security.access.AccessDeniedException;
import org.springframework.stereotype.Repository;

import com.mongodb.client.result.UpdateResult;

import de.ozgcloud.vorgang.callcontext.CurrentUserService;
import de.ozgcloud.vorgang.clientattribute.ClientAttribute.AccessPermission;
import de.ozgcloud.vorgang.common.errorhandling.NotFoundException;
import de.ozgcloud.vorgang.vorgang.Vorgang;
import lombok.NonNull;
import lombok.extern.log4j.Log4j2;

@Repository
@Log4j2
class ClientAttributeRepository {

	private static final String FIELD_NAME_TEMPL = "clientAttributes.%s.%s";
	private static final String FIELD_IN_ATTRIBUTE_TEMPL = "clientAttributes.%s.%s.%s";

	@Autowired
	private MongoOperations mongoOperations;
	@Autowired
	private CurrentUserService userService;

	public ClientAttribute set(String vorgangId, ClientAttribute attribute) {
		var result = mongoOperations.updateFirst(createQueryByVorgangId(vorgangId), createUpdateToSetAttribute(attribute), Vorgang.COLLECTION_NAME);

		if (result.getModifiedCount() <= 0) {
			LOG.warn("No changes where made on persisting client attribute {}.", attribute.getAttributeName());
		}

		return attribute;
	}

	private Update createUpdateToSetAttribute(ClientAttribute attribute) {
		return Update.update(String.format(FIELD_NAME_TEMPL, attribute.getClientName(), attribute.getAttributeName()), attribute);
	}

	public void delete(ClientAttributeIdentificator id) {
		if (isCalledByOwningClient(id.getClientName())) {
			deleteForOwningClient(id);
		} else {
			deleteForOtherClient(id);
		}
	}

	private void deleteForOtherClient(ClientAttributeIdentificator id) {
		var result = mongoOperations.updateFirst(query(byVorgangIdAndWritable(id)), buildDeleteUpdate(id), Vorgang.COLLECTION_NAME);
		checkMissingPermissions(id, result);
	}

	private void deleteForOwningClient(ClientAttributeIdentificator id) {
		var result = mongoOperations.updateFirst(createQueryByVorgangId(id.getVorgangId()), buildDeleteUpdate(id), Vorgang.COLLECTION_NAME);
		if (result.getModifiedCount() < 1) {
			throw new NotFoundException(ClientAttribute.class, id);
		}
	}

	private Update buildDeleteUpdate(ClientAttributeIdentificator id) {
		var update = new Update();
		update.unset(String.format(FIELD_NAME_TEMPL, id.getClientName(), id.getAttributeName()));
		return update;
	}

	public boolean exists(ClientAttributeIdentificator identificator) {
		return mongoOperations.exists(query(existsByVorgangIdAndAttribute(identificator)), Vorgang.COLLECTION_NAME);
	}

	private Query createQueryByVorgangId(String vorgangId) {
		return query(byVorgangId(vorgangId));
	}

	private Criteria byVorgangId(String vorgangId) {
		return where(Vorgang.MONGODB_FIELDNAME_ID).is(vorgangId);
	}

	private Criteria byVorgangIdAndWritable(ClientAttributeIdentificator id) {
		return byVorgangId(id.getVorgangId())
				.and(String.format(FIELD_IN_ATTRIBUTE_TEMPL, id.getClientName(), id.getAttributeName(), ClientAttribute.FIELD_ACCESS))
				.is(AccessPermission.READ_WRITE);
	}

	public void update(String id, ClientAttribute attribute) {
		if (isCalledByOwningClient(attribute.getClientName())) {
			set(id, attribute);
		} else {
			updateForOtherClient(id, attribute);
		}
	}

	private boolean isCalledByOwningClient(@NonNull String attributeClientName) {
		return Objects.equals(attributeClientName, userService.getUser().getClientName());
	}

	private void updateForOtherClient(String vorgangId, ClientAttribute attribute) {
		var id = ClientAttributeIdentificator.from(vorgangId, attribute);
		var result = mongoOperations.updateFirst(query(byVorgangIdAndWritable(id)), buildUpdateToUpdateValue(attribute), Vorgang.COLLECTION_NAME);

		checkMissingPermissions(id, result);
	}

	private Update buildUpdateToUpdateValue(ClientAttribute attribute) {
		return Update.update(
				String.format(FIELD_IN_ATTRIBUTE_TEMPL, attribute.getClientName(), attribute.getAttributeName(), ClientAttribute.FIELD_VALUE),
				attribute.getValue());
	}

	private void checkMissingPermissions(ClientAttributeIdentificator id, UpdateResult result) {

		if ((result.getMatchedCount() == 0) && (result.getModifiedCount() < 1)) {
			if (mongoOperations.exists(query(existsByVorgangIdAndAttribute(id)), Vorgang.COLLECTION_NAME)) {
				throw new AccessDeniedException("Write-Access for client attribute denied: " + id);
			}
			throw new NotFoundException(ClientAttribute.class, id);
		}
	}

	private Criteria existsByVorgangIdAndAttribute(ClientAttributeIdentificator id) {
		return byVorgangId(id.getVorgangId())
				.and(String.format(FIELD_NAME_TEMPL, id.getClientName(), id.getAttributeName())).exists(true);
	}

}
