/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.command;

import static de.ozgcloud.vorgang.common.db.CriteriaUtil.*;
import static org.springframework.data.mongodb.core.query.Criteria.*;
import static org.springframework.data.mongodb.core.query.Query.*;

import java.time.ZonedDateTime;
import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.stream.Stream;

import org.apache.commons.collections.MapUtils;
import org.apache.commons.lang3.ArrayUtils;
import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.mongodb.core.FindAndModifyOptions;
import org.springframework.data.mongodb.core.MongoOperations;
import org.springframework.data.mongodb.core.aggregation.Aggregation;
import org.springframework.data.mongodb.core.aggregation.ComparisonOperators.Eq;
import org.springframework.data.mongodb.core.aggregation.ConditionalOperators;
import org.springframework.data.mongodb.core.aggregation.ConditionalOperators.Switch;
import org.springframework.data.mongodb.core.aggregation.ConditionalOperators.Switch.CaseOperator;
import org.springframework.data.mongodb.core.aggregation.SetOperation;
import org.springframework.data.mongodb.core.query.Criteria;
import org.springframework.data.mongodb.core.query.Query;
import org.springframework.data.mongodb.core.query.Update;
import org.springframework.data.mongodb.core.query.UpdateDefinition;
import org.springframework.stereotype.Repository;

import de.ozgcloud.command.Command;
import de.ozgcloud.command.CommandStatus;

@Repository
class CommandRepository {

	private static final String MONGODB_ID = "id";
	static final String MONGODB_STATUS = "status";
	private static final String MONGODB_REFERENCE_STATUS = "$status";
	private static final String MONGODB_FINISHED_AT = "finishedAt";
	private static final String MONGODB_RELATION_VERSION = "relationVersion";
	private static final String MONGODB_ERROR_MSG = "errorMessage";
	private static final String MONGODB_ORDER = "order";
	private static final String MONGODB_PREVIOUS_STATE = "previousState";
	private static final String MONGODB_CREATED_AT = "createdAt";
	private static final String MONGODB_CREATED_BY = "createdBy";
	private static final String MONGODB_ASSIGNED_TO = "body.assignedTo";
	private static final String MONGODB_CREATED_RESOURCE = "createdResource";
	private static final String MONGODB_PARENT_ID = PersistedCommand.FIELD_BODY_OBJECT + "." + PersistedCommand.PROPERTY_PARENT_ID;
	private static final String MONGODB_COMPLETE_IF_SUBS_COMPLETED = PersistedCommand.FIELD_BODY_OBJECT + "."
			+ PersistedCommand.PROPERTY_COMPLETE_IF_SUBS_COMPLETED;
	private static final String MONGODB_CLASS = "_class";

	@Autowired
	private MongoOperations mongoOperations;

	Command save(Command command) {
		return mongoOperations.save(command);
	}

	void finishCommand(String commandId) {
		finishCommand(commandId, new Update());
	}

	void finishCommand(String commandId, String createdResource) {
		finishCommand(commandId, new Update().set(MONGODB_CREATED_RESOURCE, createdResource));
	}

	private void finishCommand(String commandId, Update update) {
		mongoOperations.updateFirst(
				queryByIdAndPending(commandId),
				update.set(MONGODB_STATUS, CommandStatus.FINISHED).set(MONGODB_FINISHED_AT, ZonedDateTime.now()),
				Command.class);
	}

	private Query queryByIdAndPending(String commandId) {
		return query(new Criteria().andOperator(
				criteriaById(commandId),
				where(MONGODB_STATUS).in(CommandStatus.PENDING, CommandStatus.REVOKE_PENDING)));
	}

	void updateCommandStatus(String commandId, CommandStatus status) {
		mongoOperations.updateFirst(
				queryById(commandId),
				new Update().set(MONGODB_STATUS, status),
				Command.class);
	}

	public Command updateCommandStatusAndReturnPrevious(String commandId, CommandStatus status) {
		return mongoOperations.findAndModify(queryById(commandId), new Update().set(MONGODB_STATUS, status), Command.class);
	}

	void setErrorMessage(String commandId, String errorMsg) {
		mongoOperations.updateFirst(queryById(commandId),
				new Update()
						.set(MONGODB_STATUS, CommandStatus.ERROR)
						.set(MONGODB_ERROR_MSG, errorMsg)
						.set(MONGODB_FINISHED_AT, ZonedDateTime.now()),
				Command.class);
	}

	void updateCommandStatusAndVersion(String commandId, CommandStatus status, long relationVersion) {
		mongoOperations.updateFirst(
				queryById(commandId),
				new Update()
						.set(MONGODB_STATUS, status)
						.set(MONGODB_RELATION_VERSION, relationVersion),
				Command.class);
	}

	Optional<Command> findById(String commandId) {
		return Optional.ofNullable(mongoOperations.findById(commandId, Command.class));
	}

	boolean existsPendingCommands(String vorgangId) {
		return mongoOperations.exists(buildPendingCommandQuery(vorgangId), Command.class);
	}

	List<Command> getPendingCommands(String relationId) {
		return mongoOperations.find(buildPendingCommandQuery(relationId), Command.class);
	}

	private Query buildPendingCommandQuery(String vorgangId) {
		return query(new Criteria().andOperator(
				where(PersistedCommand.FIELD_VORGANG_ID).is(vorgangId),
				where(MONGODB_STATUS).in(CommandStatus.PENDING.name(), CommandStatus.REVOKE_PENDING.name())));
	}

	public List<Command> findCommands(String vorgangId, Collection<CommandStatus> status, Optional<Order> order) {
		return mongoOperations.find(buildCommandsQuery(vorgangId, status, order), Command.class);
	}

	private Query buildCommandsQuery(String vorgangId, Collection<CommandStatus> status, Optional<Order> commandOrder) {
		var result = query(where(PersistedCommand.FIELD_VORGANG_ID).is(vorgangId));

		if (!status.isEmpty()) {
			result.addCriteria(new Criteria(MONGODB_STATUS).in(status));
		}
		commandOrder.map(Order::name).ifPresent(order -> result.addCriteria(new Criteria(MONGODB_ORDER).is(order)));

		return result;
	}

	public Stream<Command> findCommands(Order order, CommandStatus status, ZonedDateTime createdBefore) {
		return mongoOperations.stream(buildCommandsQuery(order, status, createdBefore), Command.class);
	}

	private Query buildCommandsQuery(Order order, CommandStatus status, ZonedDateTime createdBefore) {
		var result = query(where(MONGODB_ORDER).is(order.name()).and(MONGODB_STATUS).is(status));
		result.addCriteria(new Criteria(MONGODB_CREATED_AT).lte(createdBefore));
		return result;
	}

	public void patch(String commandId, Map<String, Object> previousState) {
		mongoOperations.updateFirst(
				queryById(commandId),
				buildItemPatchUpdate(previousState),
				Command.class);
	}

	private Update buildItemPatchUpdate(Map<String, Object> valueMap) {
		Update update = new Update();
		for (Map.Entry<String, Object> entry : valueMap.entrySet()) {
			update.set("%s.%s".formatted(MONGODB_PREVIOUS_STATE, entry.getKey()), entry.getValue());
		}
		return update;
	}

	public void deleteAllByVorgang(String vorgangId) {
		mongoOperations.remove(query(where(PersistedCommand.FIELD_VORGANG_ID).is(vorgangId)), PersistedCommand.COLLECTION_NAME);
	}

	public boolean existsCommandWithUserId(String userId) {
		return mongoOperations.exists(query(buildCreatedByOrAssignedToCriteria(userId)), PersistedCommand.COLLECTION_NAME);
	}

	Criteria buildCreatedByOrAssignedToCriteria(String userId) {
		return new Criteria().orOperator(where(MONGODB_CREATED_BY).is(userId), where(MONGODB_ASSIGNED_TO).is(userId));
	}

	public void addSubCommands(String commandId, Map<String, Object> subCommandValues) {
		var update = new Update();
		subCommandValues.forEach((key, value) -> update.set("%s.%s".formatted(PersistedCommand.FIELD_BODY_OBJECT, key), value));
		mongoOperations.updateFirst(queryById(commandId), update, Command.class);
	}

	public boolean isCommandFailed(String commandId) {
		return mongoOperations.exists(query(new Criteria().andOperator(criteriaById(commandId), where(MONGODB_STATUS).is(CommandStatus.ERROR))),
				Command.class);
	}

	public Optional<String> getParentId(String commandId) {
		var query = queryById(commandId);
		query.fields().include(MONGODB_CLASS, MONGODB_PARENT_ID);
		return findOne(query).map(Command::getBodyObject).map(this::getParentIdProperty);
	}

	private Criteria criteriaById(String commandId) {
		return where(MONGODB_ID).is(commandId);
	}

	String getParentIdProperty(Map<String, Object> body) {
		return MapUtils.getString(body, PersistedCommand.PROPERTY_PARENT_ID);
	}

	Optional<Command> findOne(Query query) {
		return Optional.ofNullable(mongoOperations.findOne(query, Command.class));
	}

	public boolean isCompleteIfSubsCompleted(String commandId) {
		return mongoOperations.exists(buildQueryCompleteIfSubsCompleted(commandId), Command.class);
	}

	Query buildQueryCompleteIfSubsCompleted(String commandId) {
		return query(new Criteria().andOperator(isId(commandId), where(MONGODB_COMPLETE_IF_SUBS_COMPLETED).is(true)));
	}

	public boolean existsNotFinishedSubCommands(String parentId) {
		return mongoOperations.exists(buildQueryNotInStatus(parentId, CommandStatus.FINISHED, CommandStatus.REVOKED), Command.class);
	}

	Query buildQueryNotInStatus(String parentId, CommandStatus status, CommandStatus... statuses) {
		Object[] excludeStatuses = ArrayUtils.add(statuses, status);
		return query(new Criteria().andOperator(
				buildIsParentIdCriteria(parentId),
				where(MONGODB_STATUS).nin(excludeStatuses)));
	}

	public Stream<String> findSubCommandIds(String parentId) {
		return mongoOperations.stream(includeIdAndClass(query(buildIsParentIdCriteria(parentId))), Command.class).map(Command::getId);
	}

	private Criteria buildIsParentIdCriteria(String parentId) {
		return where(MONGODB_PARENT_ID).is(parentId);
	}

	Query includeIdAndClass(Query query) {
		query.fields().include(MONGODB_CLASS, MONGODB_ID);
		return query;
	}

	public Command setRevokeStatus(String id, String createdResource) {
		return mongoOperations.findAndModify(queryById(id), buildUpdateStatusRevoke(createdResource), FindAndModifyOptions.options().returnNew(true),
				Command.class);
	}

	private UpdateDefinition buildUpdateStatusRevoke(String createdResource) {
		var setOperation = SetOperation.set(MONGODB_STATUS).toValue(buildSetStatusCondition());
		if (StringUtils.isNotBlank(createdResource)) {
			setOperation = setOperation.and().set(MONGODB_CREATED_RESOURCE).toValue(buildSetCreatedResourceCondition(createdResource));
		}
		return Aggregation.newUpdate(setOperation);
	}

	private Switch buildSetStatusCondition() {
		return Switch.switchCases(
						CaseOperator.when(Eq.valueOf(MONGODB_STATUS).equalToValue(CommandStatus.NEW)).then(CommandStatus.CANCELED),
						CaseOperator.when(Eq.valueOf(MONGODB_STATUS).equalToValue(CommandStatus.PENDING)).then(CommandStatus.REVOKE_PENDING),
						CaseOperator.when(Eq.valueOf(MONGODB_STATUS).equalToValue(CommandStatus.FINISHED)).then(CommandStatus.REVOKE_PENDING))
				.defaultTo(MONGODB_REFERENCE_STATUS);
	}

	private ConditionalOperators.Cond buildSetCreatedResourceCondition(String createdResource) {
		return ConditionalOperators.Cond.when(Eq.valueOf(MONGODB_STATUS).equalToValue(CommandStatus.PENDING)).then(createdResource)
				.otherwiseValueOf(MONGODB_CREATED_RESOURCE);
	}

	public Optional<Command> setRevokeStatusIfNotPending(String id) {
		var updateCommand = mongoOperations.findAndModify(queryById(id), buildUpdateStatusRevokeIfNotPending(),
				FindAndModifyOptions.options().returnNew(true), Command.class);
		return Optional.ofNullable(updateCommand);
	}

	private Query queryById(String commandId) {
		return query(criteriaById(commandId));
	}

	private UpdateDefinition buildUpdateStatusRevokeIfNotPending() {
		var switchOperation = Switch.switchCases(
				CaseOperator.when(Eq.valueOf(MONGODB_STATUS).equalToValue(CommandStatus.NEW)).then(CommandStatus.CANCELED),
				CaseOperator.when(Eq.valueOf(MONGODB_STATUS).equalToValue(CommandStatus.FINISHED)).then(CommandStatus.REVOKE_PENDING)
		).defaultTo(MONGODB_REFERENCE_STATUS);
		return Aggregation.newUpdate(SetOperation.set(MONGODB_STATUS).toValue(switchOperation));
	}

}
