/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.command;

import static org.apache.commons.lang3.StringUtils.*;

import java.util.List;
import java.util.Optional;
import java.util.stream.Stream;

import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.ApplicationEventPublisher;

import de.ozgcloud.command.Command;
import de.ozgcloud.command.CommandExecutedEvent;
import de.ozgcloud.command.CommandStatus;
import de.ozgcloud.vorgang.command.CommandResponse.ResponseCode;
import de.ozgcloud.vorgang.common.errorhandling.NotFoundException;
import de.ozgcloud.vorgang.common.security.PolicyService;
import de.ozgcloud.vorgang.grpc.command.CommandServiceGrpc.CommandServiceImplBase;
import de.ozgcloud.vorgang.grpc.command.GrpcAddSubCommandsRequest;
import de.ozgcloud.vorgang.grpc.command.GrpcCommand;
import de.ozgcloud.vorgang.grpc.command.GrpcCommandResponse;
import de.ozgcloud.vorgang.grpc.command.GrpcCommandsResponse;
import de.ozgcloud.vorgang.grpc.command.GrpcCreateCommandRequest;
import de.ozgcloud.vorgang.grpc.command.GrpcEmpty;
import de.ozgcloud.vorgang.grpc.command.GrpcExistsPendingCommandsRequest;
import de.ozgcloud.vorgang.grpc.command.GrpcExistsPendingCommandsResponse;
import de.ozgcloud.vorgang.grpc.command.GrpcFindCommandsRequest;
import de.ozgcloud.vorgang.grpc.command.GrpcGetCommandRequest;
import de.ozgcloud.vorgang.grpc.command.GrpcGetPendingCommandsRequest;
import de.ozgcloud.vorgang.grpc.command.GrpcGetPendingCommandsResponse;
import de.ozgcloud.vorgang.grpc.command.GrpcRevokeCommandRequest;
import de.ozgcloud.vorgang.grpc.command.GrpcSetCommandExecutedRequest;
import io.grpc.stub.StreamObserver;
import lombok.extern.log4j.Log4j2;
import net.devh.boot.grpc.server.service.GrpcService;

@Log4j2
@GrpcService
public class GrpcCommandService extends CommandServiceImplBase {

	@Autowired
	private GrpcCreateCommandRequestMapper createCommandRequestMapper;
	@Autowired
	private CommandService commandService;

	@Autowired
	private GrpcCommandResponseMapper commandResponseMapper;
	@Autowired
	private GrpcCommandMapper grpcCommandMapper;

	@Autowired
	private PolicyService policyService;
	@Autowired
	private ApplicationEventPublisher publisher;
	@Autowired
	private GrpcAddSubCommandsRequestMapper addSubCommandsRequestMapper;

	@Override
	public void createCommand(GrpcCreateCommandRequest request, StreamObserver<GrpcCommandResponse> responseObserver) {
		policyService.checkPermission(request.getVorgangId());

		var mappedRequest = createCommandRequestMapper.fromGrpc(request);

		var command = commandService.createCommand(mappedRequest);

		responseObserver.onNext(mapToGrpcResponse(getCommand(command.getId())));
		responseObserver.onCompleted();
	}

	@Override
	public void getCommand(GrpcGetCommandRequest request, StreamObserver<GrpcCommand> responseObserver) {
		policyService.checkPermissionByCommand(request.getId());

		var command = getCommand(request.getId());

		responseObserver.onNext(grpcCommandMapper.toGrpc(command));
		responseObserver.onCompleted();
	}

	@Override
	public void revokeCommand(GrpcRevokeCommandRequest request, StreamObserver<GrpcCommandResponse> responseObserver) {
		policyService.checkPermissionByCommand(request.getId());
		var updatedCommand = commandService.revokeCommand(request.getId());
		responseObserver.onNext(mapToGrpcResponse(updatedCommand));
		responseObserver.onCompleted();
	}

	Command getCommand(String commandId) {
		return commandService.findById(commandId).orElseThrow(() -> new NotFoundException(Command.class, commandId));
	}

	private GrpcCommandResponse mapToGrpcResponse(Command response) {
		return commandResponseMapper.toGrpc(
				CommandResponse.builder()
						.command(response)
						.responseCode(ResponseCode.OK)
						.build());
	}

	@Override
	public void existsPendingCommands(GrpcExistsPendingCommandsRequest request, StreamObserver<GrpcExistsPendingCommandsResponse> responseObserver) {
		policyService.checkPermission(request.getVorgangId());

		boolean pendingCommandsExists = commandService.existsPendingCommands(request.getVorgangId());

		responseObserver.onNext(GrpcExistsPendingCommandsResponse.newBuilder().setExistsPendingCommands(pendingCommandsExists).build());
		responseObserver.onCompleted();
	}

	@Override
	public void getPendingCommands(GrpcGetPendingCommandsRequest request, StreamObserver<GrpcGetPendingCommandsResponse> responseObserver) {
		policyService.checkPermission(request.getVorgangId());

		var pendingCommands = commandService.getPendingCommands(request.getVorgangId());

		responseObserver.onNext(buildGrpcGetPendingCommandsResponse(pendingCommands));
		responseObserver.onCompleted();
	}

	private GrpcGetPendingCommandsResponse buildGrpcGetPendingCommandsResponse(List<Command> pendingCommands) {
		return GrpcGetPendingCommandsResponse.newBuilder()
				.addAllCommand(pendingCommands.stream()
						.map(grpcCommandMapper::toGrpc)
						.toList())
				.build();
	}

	@Override
	public void findCommands(GrpcFindCommandsRequest request, StreamObserver<GrpcCommandsResponse> responseObserver) {
		policyService.checkPermission(request.getVorgangId());

		var statusList = request.getStatusList().stream().map(CommandStatus::valueOf).toList();
		var order = Optional.of(request.getOrderString()).filter(StringUtils::isNotBlank).flatMap(Order::fromString);

		var commands = commandService.findCommands(request.getVorgangId(), statusList, order);

		responseObserver.onNext(buildCommandsResponse(commands));
		responseObserver.onCompleted();
	}

	@Override
	public void setCommandExecuted(GrpcSetCommandExecutedRequest request, StreamObserver<GrpcEmpty> responseObserver) {
		commandService.findById(request.getCommandId())
				.map(command -> new CommandExecutedEvent(command, trimToNull(request.getCreatedResource())))
				.ifPresentOrElse(publisher::publishEvent,
						() -> LOG.warn("Command with ID '%s' not found. Cannot mark as executed.".formatted(request.getCommandId())));

		responseObserver.onNext(GrpcEmpty.newBuilder().build());
		responseObserver.onCompleted();
	}

	@Override
	public void addSubCommands(GrpcAddSubCommandsRequest request, StreamObserver<GrpcCommandsResponse> responseObserver) {
		policyService.checkPermissionByCommand(request.getParentId());
		responseObserver.onNext(buildCommandsResponse(commandService.createSubCommands(addSubCommandsRequestMapper.from(request))));
		responseObserver.onCompleted();
	}

	GrpcCommandsResponse buildCommandsResponse(Stream<Command> commands) {
		var builder = GrpcCommandsResponse.newBuilder();
		commands.map(grpcCommandMapper::toGrpc).forEach(builder::addCommand);
		return builder.build();
	}

}
