/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.common.operator;

import static java.util.Objects.*;
import static org.apache.commons.lang3.StringUtils.*;

import java.lang.reflect.Field;
import java.lang.reflect.ParameterizedType;
import java.time.LocalDate;
import java.time.LocalDateTime;
import java.time.ZonedDateTime;
import java.util.Collection;
import java.util.Optional;

import org.apache.commons.lang3.StringUtils;

import de.ozgcloud.vorgang.common.GrpcQueryOperator;
import de.ozgcloud.vorgang.statistic.VorgangStatisticBadRequestException;
import de.ozgcloud.vorgang.vorgang.Vorgang;
import lombok.AccessLevel;
import lombok.NoArgsConstructor;
import lombok.SneakyThrows;
import lombok.extern.log4j.Log4j2;

@Log4j2
@NoArgsConstructor(access = AccessLevel.PRIVATE)
public class OperatorBuilder {

	private GrpcQueryOperator grpcOperator;
	private String fieldPath;
	private Object operand;

	public static OperatorBuilder from(GrpcQueryOperator grpcOperator) {
		var builder = new OperatorBuilder();
		builder.grpcOperator = grpcOperator;
		return builder;
	}

	public OperatorBuilder fieldPath(String fieldPath) {
		this.fieldPath = FieldPathProcessor.processToCriteriaPath(fieldPath);
		return this;
	}

	public OperatorBuilder operand(Object operand) {
		this.operand = operand;
		return this;
	}

	public Operator build() {
		if (isNull(grpcOperator)) {
			throw new NullPointerException("Operator is missing");
		}
		operand = getParsedOperand();
		return switch (grpcOperator) {
			case EQUAL -> EqualOperator.builder().fieldPath(fieldPath).operand(operand).build();
			case IS_EMPTY -> EqualOperator.builder().fieldPath(fieldPath).operand(StringUtils.EMPTY).build();
			case IS_NULL -> EqualOperator.builder().fieldPath(fieldPath).build();
			case UNEQUAL -> UnequalOperator.builder().fieldPath(fieldPath).operand(operand).build();
			case GREATER_THEN -> GreaterThenOperator.builder().fieldPath(fieldPath).operand(operand).build();
			case GREATER_THEN_OR_EQUAL_TO -> GreaterThenOrEqualOperator.builder().fieldPath(fieldPath).operand(operand).build();
			case LESS_THEN -> LessThenOperator.builder().fieldPath(fieldPath).operand(operand).build();
			case LESS_THEN_OR_EQUAL_TO -> LessThenOrEqualOperator.builder().fieldPath(fieldPath).operand(operand).build();
			case EXISTS -> ExistsOperator.builder().fieldPath(fieldPath).build();
			case NOT_EXIST -> ExistsOperator.builder().fieldPath(fieldPath).negate(true).build();
			case UNRECOGNIZED -> throw new VorgangStatisticBadRequestException("Unrecognized operator");
		};
	}

	@SneakyThrows
	Object getParsedOperand() {
		return Optional.ofNullable(getFieldClass()).map(this::parseValue).orElse(operand);
	}

	Class<?> getFieldClass() {
		if (isBlank(fieldPath)) {
			return null;
		}
		var fieldNames = fieldPath.split("\\.");
		Class<?> clazz = Vorgang.class;
		for (var fieldName : fieldNames) {
			try {
				clazz = getFieldClass(clazz.getDeclaredField(fieldName));
			} catch (NoSuchFieldException | ClassNotFoundException e) {
				LOG.trace("Cannot find field {} from {}", fieldName, fieldPath, e);
				return null;
			}
		}
		return clazz;
	}

	Class<?> getFieldClass(Field field) throws ClassNotFoundException {
		if (Collection.class.isAssignableFrom(field.getType())) {
			var actualTypeArgument = ((ParameterizedType) field.getGenericType()).getActualTypeArguments()[0];
			return Class.forName(actualTypeArgument.getTypeName());
		} else {
			return field.getType();
		}
	}

	Object parseValue(Class<?> clazz) {
		if (operand instanceof String valueString) {
			if (ZonedDateTime.class.equals(clazz)) {
				return ZonedDateTime.parse(valueString);
			}
			if (LocalDate.class.equals(clazz)) {
				return LocalDate.parse(valueString);
			}
			if (LocalDateTime.class.equals(clazz)) {
				return LocalDateTime.parse(valueString);
			}
		}
		return operand;
	}

}
