/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.common.search;

import java.text.MessageFormat;

import org.apache.commons.collections.MapUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.autoconfigure.condition.ConditionalOnProperty;
import org.springframework.context.event.EventListener;
import org.springframework.stereotype.Component;

import de.ozgcloud.command.Command;
import de.ozgcloud.command.VorgangAssignedEvent;
import de.ozgcloud.command.VorgangCreatedEvent;
import de.ozgcloud.command.status.StatusChangedEvent;
import de.ozgcloud.vorgang.command.CommandService;
import de.ozgcloud.vorgang.vorgang.SetAktenzeichenCompletedEvent;
import de.ozgcloud.vorgang.vorgang.VorgangDeletedEvent;
import de.ozgcloud.vorgang.vorgang.VorgangService;
import lombok.extern.log4j.Log4j2;

@Component
@ConditionalOnProperty(prefix = "ozgcloud.elasticsearch", name = "address")
@Log4j2
public class SearchEventListener {

	@Autowired
	private SearchService searchService;

	@Autowired
	private CommandService commandService;

	@Autowired
	private VorgangService vorgangService;

	@EventListener
	public void onNewVorgang(VorgangCreatedEvent vorgangCreatedEvent) {
		searchService.addVorgang(vorgangService.getById(vorgangCreatedEvent.getSource()));

	}

	@EventListener
	public void onVorgangAssigned(VorgangAssignedEvent vorgangAssignedEvent) {
		updateVorgang(vorgangAssignedEvent.getCommand());
	}

	@EventListener
	public void onVorgangStatusChanged(StatusChangedEvent event) {
		updateVorgangById(event.getSource());
	}

	private void updateVorgangById(String commandId) {
		commandService.findById(commandId).ifPresentOrElse(this::updateVorgang, () -> {
			throw new IllegalStateException(MessageFormat.format("Kein Command mit der ID {0} gefunden", commandId));
		});
	}

	void updateVorgang(Command command) {
		searchService.updateVorgang(vorgangService.getById(command.getVorgangId()));
	}

	@EventListener
	public void onVorgangDeleted(VorgangDeletedEvent event) {
		String vorgangId = event.getCommand().getVorgangId();
		try {
			searchService.deleteVorgang(vorgangId);
		} catch (RuntimeException e) {
			LOG.error("Error on deleting Vorgang (" + vorgangId + ") from search index.", e);
		}
	}

	@EventListener
	public void onSetAktenzeichenCompletedEvent(SetAktenzeichenCompletedEvent event) {
		var command = event.getCommand();
		var aktenzeichen = MapUtils.getString(command.getBodyObject(), VorgangService.BODY_OBJECT_AKTENZEICHEN);
		try {
			searchService.setAktenzeichen(command.getVorgangId(), aktenzeichen);
		} catch (RuntimeException e) {
			LOG.error("Error on setting new aktenzeichen ({}) for Vorgang ({}) in search index.", aktenzeichen, command.getVorgangId(), e);
		}
	}
}
