/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.files;

import java.util.ArrayList;
import java.util.List;
import java.util.Optional;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.mongodb.core.MongoTemplate;
import org.springframework.data.mongodb.core.aggregation.Aggregation;
import org.springframework.data.mongodb.core.aggregation.AggregationOperation;
import org.springframework.data.mongodb.core.query.Criteria;
import org.springframework.stereotype.Repository;

import de.ozgcloud.vorgang.common.db.CriteriaUtil;
import de.ozgcloud.vorgang.vorgang.IncomingFile;
import de.ozgcloud.vorgang.vorgang.Vorgang;

@Repository
class EingangFilesRepository {

	@Autowired
	private MongoTemplate mongoTemplate;

	private static final String DB_FIELDNAME_EINGANGS = "eingangs";
	private static final String FIELD_EINGANG_ID = DB_FIELDNAME_EINGANGS + "._id";
	private static final String DB_FIELDNAME_FILES = "files";
	private static final String FIELD_ATTACHMENT_FILES = DB_FIELDNAME_EINGANGS + ".attachments." + DB_FIELDNAME_FILES;
	private static final String FIELD_REPRESENTATION_FILES = DB_FIELDNAME_EINGANGS + ".representations";

	public List<IncomingFile> findAttachments(String vorgangId, Optional<String> eingangId) {
		return findFiles(buildSearchCriteria(vorgangId, eingangId), buildExtractAttachmentsAggregation());
	}

	private List<AggregationOperation> buildExtractAttachmentsAggregation() {
		return List.of(
				Aggregation.project().andExpression(FIELD_ATTACHMENT_FILES).as(DB_FIELDNAME_FILES), //
				// extract 'files' from deeply nested arrays within 'eingangs[].attachments[].files[]' and promote each file to a separate document
				// at the root level.
				Aggregation.unwind(DB_FIELDNAME_FILES),
				Aggregation.unwind(DB_FIELDNAME_FILES),
				Aggregation.unwind(DB_FIELDNAME_FILES),
				Aggregation.replaceRoot(DB_FIELDNAME_FILES)
		);
	}

	public List<IncomingFile> findRepresentations(String vorgangId, Optional<String> eingangId) {
		return findFiles(buildSearchCriteria(vorgangId, eingangId), buildFindRepresentationsAggregation());
	}

	private Criteria buildSearchCriteria(String vorgangId, Optional<String> eingangId) {
		return eingangId.map(eid -> buildIsVorganIdAndEingangIdCriteria(vorgangId, eid)).orElseGet(() -> CriteriaUtil.isId(vorgangId));
	}

	private Criteria buildIsVorganIdAndEingangIdCriteria(String vorgangId, String eingangId) {
		return new Criteria().andOperator(CriteriaUtil.isId(vorgangId), buildIsEingangIdCriteria(eingangId));
	}

	private Criteria buildIsEingangIdCriteria(String eingangId) {
		return Criteria.where(FIELD_EINGANG_ID).is(eingangId);
	}

	private List<AggregationOperation> buildFindRepresentationsAggregation() {
		return List.of(
				Aggregation.project().andExpression(FIELD_REPRESENTATION_FILES).as(DB_FIELDNAME_FILES),
				// extract 'files' from deeply nested arrays within 'eingangs[].representations[]' and promote each file to a separate document at
				// the root level.
				Aggregation.unwind(DB_FIELDNAME_FILES),
				Aggregation.unwind(DB_FIELDNAME_FILES),
				Aggregation.replaceRoot(DB_FIELDNAME_FILES));
	}

	private List<IncomingFile> findFiles(Criteria criteria, List<AggregationOperation> extractOperations) {
		List<AggregationOperation> operations = new ArrayList<>();
		operations.add(Aggregation.match(criteria));
		operations.addAll(extractOperations);
		return mongoTemplate.aggregate(Aggregation.newAggregation(operations), Vorgang.COLLECTION_NAME, IncomingFile.class).getMappedResults();
	}
}
