/*
 * Copyright (C) 2022 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.files;

import static org.springframework.data.mongodb.core.query.Criteria.*;
import static org.springframework.data.mongodb.core.query.Query.*;

import java.io.InputStream;
import java.util.Collection;
import java.util.Map;
import java.util.Objects;
import java.util.Optional;
import java.util.stream.Stream;

import org.bson.types.ObjectId;
import org.springframework.data.mongodb.core.MongoOperations;
import org.springframework.data.mongodb.core.query.Criteria;
import org.springframework.data.mongodb.core.query.Query;
import org.springframework.data.mongodb.core.query.Update;
import org.springframework.data.mongodb.core.query.UpdateDefinition;
import org.springframework.data.mongodb.gridfs.GridFsResource;
import org.springframework.data.mongodb.gridfs.GridFsTemplate;
import org.springframework.data.mongodb.gridfs.GridFsUpload;
import org.springframework.stereotype.Repository;

import com.mongodb.client.gridfs.model.GridFSFile;

import de.ozgcloud.vorgang.common.db.CollisionVerifier;
import de.ozgcloud.vorgang.common.db.CriteriaUtil;
import de.ozgcloud.vorgang.common.errorhandling.NotFoundException;
import lombok.RequiredArgsConstructor;

@Repository
@RequiredArgsConstructor
class OzgCloudFileRepository {

	static final String GRID_FS_FIELD_ID = "_id";

	private final CollisionVerifier collisionVerifier = new CollisionVerifier(this::existOzgCloudFile);

	private final GridFsTemplate gridFsTemplate;
	private final MongoOperations mongoOperations;

	public OzgCloudFile save(OzgCloudFile file) {
		return mongoOperations.save(setContentSize(file));
	}

	private OzgCloudFile setContentSize(OzgCloudFile file) {
		return file.toBuilder().size(getContentSize(file.getContentId())).build();
	}

	public void patch(FileId fileId, long version, Map<String, Object> patch) {
		var updateResult = mongoOperations.updateFirst(queryByIdAndVersion(fileId, version), buildUpdate(patch), OzgCloudFile.COLLECTION_NAME);
		collisionVerifier.verify(updateResult, fileId.toString());
	}

	private Query queryByIdAndVersion(FileId fileId, long version) {
		return query(CriteriaUtil.whereIdAndVersion(fileId.toString(), version));
	}

	private UpdateDefinition buildUpdate(Map<String, Object> patch) {
		var update = new Update();
		update.inc(OzgCloudFile.FIELD_VERSION);
		patch.forEach(update::set);
		Optional.ofNullable(patch.get(OzgCloudFile.FIELD_CONTENT_ID)).map(String.class::cast)
				.ifPresent(contentId -> update.set(OzgCloudFile.FIELD_SIZE, getContentSize(contentId)));
		return update;
	}

	private long getContentSize(String contentId) {
		if (Objects.isNull(contentId)) {
			return 0;
		}
		var contentGridFsFile = gridFsTemplate.find(queryByObjectId(contentId)).first();
		if (Objects.isNull(contentGridFsFile)) {
			throw new NotFoundException(GridFSFile.class, contentId);
		}
		return contentGridFsFile.getLength();
	}

	public String uploadContent(UploadedFilesReference ref, InputStream content) {
		return gridFsTemplate.store(buildGridFsUpload(ref, content)).toHexString();
	}

	GridFsUpload<ObjectId> buildGridFsUpload(UploadedFilesReference ref, InputStream content) {
		return GridFsUpload.fromStream(content).filename(createFilePath(ref)).build();
	}

	String createFilePath(UploadedFilesReference ref) {
		return String.format("%s/%s/%s/%s", ref.getVorgangId(), ref.getClient(), ref.getName(), ref.getFileName());
	}

	public InputStream getFileContent(FileId fileId) {
		return Optional.ofNullable(mongoOperations.findById(fileId.toString(), OzgCloudFile.class))
				.map(OzgCloudFile::getContentId)
				.map(this::getFile)
				.map(gridFsTemplate::getResource)
				.map(GridFsResource::getContent)
				.orElseThrow(() -> new NotFoundException(OzgCloudFile.class, fileId));
	}

	GridFSFile getFile(String contentId) {
		return gridFsTemplate.findOne(queryByObjectId(contentId));
	}

	public Stream<OzgCloudFile> findAll(Collection<FileId> fileIds) {
		var ids = fileIds.stream().map(FileId::toString).toList();
		return mongoOperations.stream(query(where(OzgCloudFile.FIELD_ID).in(ids)), OzgCloudFile.class);
	}

	public void deleteAllByVorgang(String vorgangId) {
		var deletedOzgCloudFile = mongoOperations.findAllAndRemove(query(where(OzgCloudFile.FIELD_VORGANG_ID).is(vorgangId)), OzgCloudFile.class);
		deletedOzgCloudFile.stream().map(OzgCloudFile::getContentId).filter(Objects::nonNull).forEach(this::deleteContent);
	}

	public void deleteContent(String contentId) {
		gridFsTemplate.delete(queryByObjectId(contentId));
	}

	private Query queryByObjectId(String contentId) {
		return Query.query(Criteria.where(GRID_FS_FIELD_ID).is(contentId));
	}

	private boolean existOzgCloudFile(String fileId) {
		return mongoOperations.exists(query(where(OzgCloudFile.FIELD_ID).is(fileId)), OzgCloudFile.class);
	}
}
