/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.statistic;

import static org.springframework.data.mongodb.core.aggregation.Aggregation.*;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Objects;

import org.bson.Document;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.mongodb.core.MongoOperations;
import org.springframework.data.mongodb.core.aggregation.AddFieldsOperation;
import org.springframework.data.mongodb.core.aggregation.AggregationExpression;
import org.springframework.data.mongodb.core.aggregation.AggregationOperation;
import org.springframework.data.mongodb.core.aggregation.AggregationResults;
import org.springframework.data.mongodb.core.aggregation.GroupOperation;
import org.springframework.data.mongodb.core.query.Query;
import org.springframework.stereotype.Repository;

import de.ozgcloud.vorgang.common.operator.Operator;
import de.ozgcloud.vorgang.vorgang.Vorgang;
import lombok.NonNull;

@Repository
class StatisticRepository {

	@Autowired
	private MongoOperations mongoOperations;

	public List<CountResult> countVorgangByStatus(StatisticFilter filter) {
		var aggregationOperations = initAggregationOperations(filter);
		var groupOperation = group(Vorgang.MONGODB_FIELDNAME_STATUS).count().as(CountResult.KEY_VALUE);
		aggregationOperations.add(groupOperation);
		var projectOperation = project(CountResult.KEY_VALUE).and(CountResult.KEY_NAME).previousOperation();
		aggregationOperations.add(projectOperation);

		return executeAggregation(CountResult.class, aggregationOperations).getMappedResults();
	}

	public List<CountResult> countVorgangByPath(CountByPath countByPaths, Operator operator, StatisticFilter filter) {
		var aggregationOperations = initAggregationOperations(filter);
		aggregationOperations.add(buildAddFieldOperation(countByPaths.getAlias(), operator.getAggregationExpression()));
		aggregationOperations.add(buildGroupOperation(countByPaths));

		var countByPathResults = executeAggregation(Document.class, aggregationOperations);
		return mapCountByPathResults(countByPathResults.getUniqueMappedResult(), countByPaths.getAlias());
	}

	List<AggregationOperation> initAggregationOperations(StatisticFilter filter) {
		var aggregationOperations = new ArrayList<AggregationOperation>();
		aggregationOperations.add(match(filter.getCriteria()));
		return aggregationOperations;
	}

	AddFieldsOperation buildAddFieldOperation(String alias, AggregationExpression aggregationExpression) {
		return addFields().addFieldWithValue(alias, aggregationExpression).build();
	}

	GroupOperation buildGroupOperation(CountByPath countByPaths) {
		return group().sum(countByPaths.getAlias()).as(countByPaths.getAlias());
	}

	List<CountResult> mapCountByPathResults(Document countByPathResults, String resultAlias) {
		if (Objects.isNull(countByPathResults)) {
			return List.of(buildZeroResult(resultAlias));
		}

		return mapCountByPathResults(countByPathResults);
	}

	private CountResult buildZeroResult(String resultAlias) {
		return CountResult.builder().name(resultAlias).value(0).build();
	}

	private List<CountResult> mapCountByPathResults(@NonNull Document countByPathResults) {
		return countByPathResults.entrySet().stream().filter(entry -> !"_id".equals(entry.getKey())).map(this::mapCountResult).toList();
	}

	CountResult mapCountResult(Map.Entry<String, Object> entry) {
		return CountResult.builder().name(entry.getKey()).value((Integer) entry.getValue()).build();
	}

	<O> AggregationResults<O> executeAggregation(Class<O> outputType, List<AggregationOperation> aggregationOperations) {
		return mongoOperations.aggregate(newAggregation(aggregationOperations), Vorgang.COLLECTION_NAME, outputType);
	}

	public boolean isFieldsExists(StatisticFilter filter) {
		return mongoOperations.exists(Query.query(filter.getCriteria()), Vorgang.COLLECTION_NAME);
	}

}
