/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.statistic;

import static org.apache.commons.collections4.CollectionUtils.*;

import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.EnumMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.Set;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import de.ozgcloud.vorgang.callcontext.CurrentUserService;
import de.ozgcloud.vorgang.common.operator.Operator;
import de.ozgcloud.vorgang.common.operator.OperatorFactory;
import de.ozgcloud.vorgang.vorgang.Vorgang;

@Service
class StatisticService {

	@Autowired
	private StatisticRepository statisticRepository;
	@Autowired
	private CurrentUserService currentUserService;
	@Autowired
	private OperatorFactory operatorFactory;

	public Statistic countVorgang(VorgangCountRequest request) {
		return Statistic.builder().countByStatus(countVorgangByStatus()).countResult(countVorgangByPathsOld(request.getCountByPaths())).build();
	}

	Map<String, Integer> countVorgangByPathsOld(List<CountByPath> countByPaths) {
		var organisationsEinheitIds = getOrganisationsEinheitIds();
		if (organisationsEinheitIds.isEmpty()) {
			return countByPaths.stream().collect(Collectors.toMap(CountByPath::getAlias, counterByPath -> 0));
		}
		var statisticFilter = StatisticFilter.builder().organisationsEinheitIds(organisationsEinheitIds.get()).build();
		return countByPaths.stream()
				.flatMap(
						countByPath -> statisticRepository.countVorgangByPath(countByPath, buildFieldExistsStatisticQuery(countByPath.getFieldPath()),
								statisticFilter).stream())
				.collect(Collectors.toMap(CountResult::getName, CountResult::getValue));
	}

	Operator buildFieldExistsStatisticQuery(String fieldPath) {
		return operatorFactory.newUnequalOperatorBuilder().fieldPath(fieldPath).operand(null).build();
	}

	Map<Vorgang.Status, Integer> countVorgangByStatus() {
		var organisationsEinheitIds = getOrganisationsEinheitIds();
		if (organisationsEinheitIds.isEmpty()) {
			return fillMissingStatusCounters(Collections.emptyMap());
		}
		var statisticFilter = StatisticFilter.builder().organisationsEinheitIds(organisationsEinheitIds.get()).build();
		List<CountResult> countResults = statisticRepository.countVorgangByStatus(statisticFilter);
		return fillMissingStatusCounters(mapCountResults(countResults));
	}

	Map<Vorgang.Status, Integer> mapCountResults(List<CountResult> countResults) {
		var countByStatus = new EnumMap<Vorgang.Status, Integer>(Vorgang.Status.class);
		countResults.forEach(countResult -> countByStatus.put(Vorgang.Status.valueOf(countResult.getName()), countResult.getValue()));
		return countByStatus;
	}

	Map<Vorgang.Status, Integer> fillMissingStatusCounters(Map<Vorgang.Status, Integer> countByStatus) {
		var resultMap = new EnumMap<Vorgang.Status, Integer>(Vorgang.Status.class);
		resultMap.putAll(countByStatus);
		Arrays.stream(Vorgang.Status.values()).forEach(status -> resultMap.putIfAbsent(status, 0));
		return resultMap;
	}

	public Statistic collectStatistic(Collection<VorgangStatisticRequest> statisticRequests) {
		return Statistic.builder()
				.countResult(countVorgangByPaths(getCountRequests(statisticRequests)))
				.existsResult(collectFieldExistsStatistic(getExistsRequests(statisticRequests)))
				.build();
	}

	Set<VorgangStatisticRequest> getCountRequests(Collection<VorgangStatisticRequest> statisticRequests) {
		return statisticRequests.stream().filter(request -> request.getStatisticGroupMethod() == StatisticGroupMethod.COUNT)
				.collect(Collectors.toSet());
	}

	Set<VorgangStatisticRequest> getExistsRequests(Collection<VorgangStatisticRequest> statisticRequests) {
		return statisticRequests.stream().filter(request -> request.getStatisticGroupMethod() == StatisticGroupMethod.EXISTS)
				.collect(Collectors.toSet());
	}

	Map<String, Integer> countVorgangByPaths(Collection<VorgangStatisticRequest> statisticRequests) {
		return statisticRequests.stream().flatMap(this::countVorgangByPath).collect(Collectors.toMap(CountResult::getName, CountResult::getValue));
	}

	Stream<CountResult> countVorgangByPath(VorgangStatisticRequest statisticRequest) {
		var organisationsEinheitIds = getOrganisationsEinheitIds();
		if (organisationsEinheitIds.isEmpty()) {
			return Stream.of(CountResult.builder().name(statisticRequest.getResultName()).value(0).build());
		}
		var statisticFilter = StatisticFilter.builder().organisationsEinheitIds(organisationsEinheitIds.get()).build();
		return statisticRepository
				.countVorgangByPath(CountByPath.fromStatisticRequest(statisticRequest), statisticRequest.getOperator(), statisticFilter).stream();
	}

	Map<String, Boolean> collectFieldExistsStatistic(Collection<VorgangStatisticRequest> fieldExistsRequests) {
		return fieldExistsRequests.stream().collect(Collectors.toMap(VorgangStatisticRequest::getResultName, this::isFieldExists));
	}

	boolean isFieldExists(VorgangStatisticRequest fieldExistsRequest) {
		var organisationsEinheitIds = getOrganisationsEinheitIds();
		if (organisationsEinheitIds.isEmpty()) {
			return false;
		}
		var filter = StatisticFilter.builder().filterCriteria(fieldExistsRequest.getOperator().getCriteria())
				.organisationsEinheitIds(organisationsEinheitIds.get()).build();
		return statisticRepository.isFieldsExists(filter);
	}

	Optional<Collection<String>> getOrganisationsEinheitIds() {
		var callContextUser = currentUserService.getUser();
		if (callContextUser.isOrganisationEinheitenIdCheckNecessary()) {
			if (isEmpty(callContextUser.getOrganisatorischeEinheitenIds())) {
				return Optional.empty();
			}
			return Optional.of(callContextUser.getOrganisatorischeEinheitenIds());
		}
		return Optional.of(Collections.emptyList());
	}

}
