/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.vorgang;

import static java.util.stream.Collectors.*;

import java.util.Collections;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.function.Predicate;

import org.apache.commons.collections4.MapUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.commons.lang3.tuple.Pair;
import org.springframework.stereotype.Component;

import lombok.NonNull;

@Component
class LabelProcessor {

	static final String CONTROL_DATA_KEY = "_kopControlData";

	static final String LABELS_KEY = "labels";
	static final String META_DATA = "metaData";

	public static final String VALUE_KEY = "value";
	public static final String LABEL_KEY = "label";

	private static final Predicate<Entry<String, Object>> IS_NOT_CONTROL_DATA = entry -> !StringUtils.equals(entry.getKey(), CONTROL_DATA_KEY);

	public Eingang moveLabelsToControlData(Eingang eingang) {
		var mappedFormData = moveLabelsToControlData(eingang.getFormData());
		return eingang.toBuilder().formData(mappedFormData).build();
	}

	Map<String, Object> moveLabelsToControlData(Map<String, Object> formData) {
		Map<String, Object> formDataWithoutLables = formData.entrySet().stream()
				.filter(IS_NOT_CONTROL_DATA)
				.map(entry -> Pair.of(entry.getKey(), entry.getValue()))
				.map(this::removeLabels)
				.map(this::doRecursionIfSubForm)
				.collect(toMap(Pair::getKey, Pair::getValue, this::merge, LinkedHashMap::new));

		return replaceControlData(formDataWithoutLables, addLabelsToControlData(formData));
	}

	private Pair<String, Object> removeLabels(Pair<String, Object> formEntry) {
		return Pair.of(formEntry.getKey(), getValueWithoutLabel(formEntry.getValue()));
	}

	Map<String, Object> addLabelsToControlData(Map<String, Object> formData) {
		var controlData = new HashMap<>(getControlData(formData));
		var labels = getLabelsFromFormData(formData);
		if (MapUtils.isNotEmpty(labels)) {
			controlData.put(LABELS_KEY, labels);
		}

		return controlData;
	}

	Map<String, Object> replaceControlData(Map<String, Object> formData, Map<String, Object> controlData) {
		if (controlData.isEmpty()) {
			return formData;
		}

		var editableFormData = new LinkedHashMap<String, Object>(formData);
		editableFormData.put(CONTROL_DATA_KEY, Collections.unmodifiableMap(controlData));

		return Collections.unmodifiableMap(editableFormData);
	}

	Pair<String, Object> doRecursionIfSubForm(Pair<String, Object> formEntry) {
		if (isSubFormValue(formEntry.getValue())) {
			return Pair.of(formEntry.getKey(), moveLabelsToControlData(castToSubFormMap(formEntry.getValue())));
		}
		return formEntry;
	}

	Map<String, String> getLabelsFromFormData(Map<String, Object> formData) {
		return formData.entrySet().stream()
				.filter(IS_NOT_CONTROL_DATA)
				.map(entry -> Pair.of(entry.getKey(), getLabelFromFieldValue(entry.getValue())))
				.filter(optPair -> optPair.getValue().isPresent())
				.collect(toMap(Pair::getKey, pair -> pair.getValue().get()));
	}

	Optional<String> getLabelFromFieldValue(Object value) {
		if (isMap(value)) {
			return Optional.ofNullable((String) castToSubFormMap(value).get(LABEL_KEY));
		}
		return Optional.empty();
	}

	Object getValueWithoutLabel(Object value) {
		if (isMap(value)) {
			return castToSubFormMap(value).getOrDefault(VALUE_KEY, value);
		}
		return value;
	}

	public Eingang addLabels(Eingang eingang) {
		var mappedFormData = addLabels(eingang.getFormData());
		return eingang.toBuilder().formData(mappedFormData).build();
	}

	Map<String, Object> addLabels(Map<String, Object> originFormData) {
		return originFormData.entrySet().stream().filter(IS_NOT_CONTROL_DATA)
				.map(entry -> Pair.of(entry.getKey(), addLabel(entry.getKey(), entry.getValue(), getLabels(originFormData))))
				.collect(toMap(Pair::getKey, Pair::getValue, this::merge, LinkedHashMap::new));
	}

	private Object merge(Object m1, Object m2) {
		throw new IllegalStateException("Collision in mapping function.");
	}

	@SuppressWarnings("unchecked")
	private Map<String, String> getLabels(@NonNull Map<String, Object> formData) {
		return (Map<String, String>) getControlData(formData).getOrDefault(LABELS_KEY, Collections.emptyMap());
	}

	private Map<String, Object> getControlData(@NonNull Map<String, Object> formData) {
		return castToSubFormMap(formData.getOrDefault(CONTROL_DATA_KEY, Collections.emptyMap()));
	}

	Map<String, Object> addLabel(String fieldName, Object fieldValue, Map<String, String> labels) {
		if (isFieldValue(fieldValue)) {
			return addLabelToFieldValue(fieldName, fieldValue, labels);
		} else {
			return addLabelsToSubForm(castToSubFormMap(fieldValue), Optional.ofNullable(labels.get(fieldName)));
		}
	}

	Map<String, Object> addLabelToFieldValue(String fieldName, Object fieldValue, Map<String, String> labels) {
		if (labels.containsKey(fieldName)) {
			return Map.of(
					LABEL_KEY, labels.get(fieldName),
					VALUE_KEY, fieldValue);
		} else {
			return Map.of(VALUE_KEY, fieldValue);
		}
	}

	Map<String, Object> addLabelsToSubForm(Map<String, Object> subForm, Optional<String> subFormLabel) {
		Map<String, Object> subFormMap = new LinkedHashMap<>();
		subFormLabel.ifPresent(label -> subFormMap.put(LABEL_KEY, label));

		subFormMap.put(VALUE_KEY, addLabels(subForm));
		return Collections.unmodifiableMap(subFormMap);
	}

	boolean isFieldValue(Object value) {
		if (isNotMap(value)) {
			return true;
		}

		var valueMap = castToSubFormMap(value);
		if (!valueMap.containsKey(LABEL_KEY)) {
			return false;
		}

		return isNotMap(valueMap.get(VALUE_KEY));
	}

	boolean isSubFormValue(Object value) {
		return !isFieldValue(value);
	}

	private boolean isMap(Object object) {
		return object instanceof Map<?, ?>;
	}

	Object getFieldValue(Object fieldValue) {
		if (isNotMap(fieldValue) || !castToSubFormMap(fieldValue).containsKey(LABEL_KEY)) {
			return fieldValue;
		}

		return castToSubFormMap(fieldValue).get(VALUE_KEY);
	}

	private boolean isNotMap(Object object) {
		return !isMap(object);
	}

	@SuppressWarnings("unchecked")
	private Map<String, Object> castToSubFormMap(Object value) {
		return (Map<String, Object>) value;
	}

}
