/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.vorgang;

import java.util.Collections;
import java.util.function.Predicate;

import org.springframework.context.ApplicationEventPublisher;
import org.springframework.context.event.EventListener;
import org.springframework.stereotype.Component;

import de.ozgcloud.command.Command;
import de.ozgcloud.command.CommandCreatedEvent;
import de.ozgcloud.command.CommandExecutedEvent;
import de.ozgcloud.command.CommandFailedEvent;
import de.ozgcloud.command.CommandRevokeFailedEvent;
import de.ozgcloud.command.CommandRevokedEvent;
import de.ozgcloud.command.RevokeCommandEvent;
import de.ozgcloud.command.VorgangLockedEvent;
import de.ozgcloud.command.VorgangUnlockedEvent;
import de.ozgcloud.vorgang.command.CommandService;
import de.ozgcloud.vorgang.command.Order;
import de.ozgcloud.vorgang.files.FileService;
import de.ozgcloud.vorgang.status.StatusService;
import de.ozgcloud.vorgang.vorgang.redirect.VorgangForwardFailedEvent;
import de.ozgcloud.vorgang.vorgang.redirect.VorgangRedirectedEvent;
import lombok.RequiredArgsConstructor;
import lombok.extern.log4j.Log4j2;

@Component
@RequiredArgsConstructor
@Log4j2
public class VorgangEventListener {

	private static final String IS_ASSIGN_COMMAND_CONDITION = "{T(de.ozgcloud.vorgang.vorgang.VorgangEventListener).IS_ASSIGN_COMMAND.test(event.getSource())}";
	public static final Predicate<Command> IS_ASSIGN_COMMAND = command -> Order.ASSIGN_USER.isMeant(command.getOrder());

	private static final String IS_LOESCHEN_COMMAND_CONDITION = "{T(de.ozgcloud.vorgang.vorgang.VorgangEventListener).IS_LOESCHEN_COMMAND.test(event.getSource())}";
	public static final Predicate<Command> IS_LOESCHEN_COMMAND = command -> Order.VORGANG_LOESCHEN.isMeant(command.getOrder());

	private static final String IS_SET_AKTENZEICHEN_COMMAND_CONDITION = "{T(de.ozgcloud.vorgang.vorgang.VorgangEventListener).IS_SET_AKTENZEICHEN_COMMAND.test(event.getSource())}";
	public static final Predicate<Command> IS_SET_AKTENZEICHEN_COMMAND = command -> Order.SET_AKTENZEICHEN.isMeant(command.getOrder());

	private static final String IS_LOCK_VORGANG_COMMAND = "{T(de.ozgcloud.vorgang.vorgang.VorgangEventListener).IS_LOCK_VORGANG_ORDER.test(event.getSource())}";
	public static final Predicate<Command> IS_LOCK_VORGANG_ORDER = command -> Order.LOCK_VORGANG.isMeant(command.getOrder());

	private static final String IS_UNLOCK_VORGANG_COMMAND_CONDITION = "{T(de.ozgcloud.vorgang.vorgang.VorgangEventListener).IS_UNLOCK_VORGANG_COMMAND.test(event.getSource())}";
	public static final Predicate<Command> IS_UNLOCK_VORGANG_COMMAND = command -> Order.UNLOCK_VORGANG.isMeant(command.getOrder());

	private final StatusService statusService;
	private final VorgangService vorgangService;
	private final VorgangHeaderService vorgangHeaderService;
	private final FileService fileService;
	private final CommandService commandService;
	private final LockMapper lockMapper;

	private final ApplicationEventPublisher publisher;

	@EventListener
	public void updateStatus(VorgangRedirectedEvent event) {
		statusService.setStatusToWeitergeleitet(event.getSource().getVorgangId(), event.getSource().getCreatedByCommand());
	}

	@EventListener
	public void updateStatus(VorgangForwardFailedEvent event) {
		statusService.setStatusToInBearbeitung(event.getVorgangId(), event.getSource());
	}

	@EventListener(condition = IS_ASSIGN_COMMAND_CONDITION)
	public void assignToUser(CommandCreatedEvent event) {
		vorgangService.assignToUser(event.getSource());
	}

	@EventListener(condition = IS_LOESCHEN_COMMAND_CONDITION)
	public void onVorgangLoeschen(CommandCreatedEvent event) {
		Command command = event.getSource();
		try {
			String vorgangId = command.getVorgangId();
			vorgangService.deleteVorgang(vorgangId);
			fileService.deleteAllByVorgang(vorgangId);

			publishEvent(new VorgangDeletedEvent(command));
		} catch (RuntimeException e) {
			LOG.error("Unexpected Error on deleting Vorgang.", e);
			publisher.publishEvent(new CommandFailedEvent(command.getId(), e.getMessage()));
		}
	}

	@EventListener(condition = IS_SET_AKTENZEICHEN_COMMAND_CONDITION)
	public void onSetAktenzeichen(CommandCreatedEvent event) {
		var command = event.getSource();
		try {
			updatePreviousState(command);
			vorgangService.setAktenzeichen(command);
			publisher.publishEvent(new SetAktenzeichenCompletedEvent(command));
		} catch (RuntimeException e) {
			LOG.error("Unexpected Error setting Aktenzeichen.", e);
			publisher.publishEvent(new CommandFailedEvent(command.getId(), e.getMessage()));
		}
	}

	@EventListener(condition = IS_LOCK_VORGANG_COMMAND)
	public void onLockVorgang(CommandCreatedEvent event) {
		var command = event.getSource();
		try {
			vorgangService.lockVorgang(lockMapper.fromCommand(command), command.getVorgangId(), command.getRelationVersion());
			publisher.publishEvent(new VorgangLockedEvent(command));
		} catch (RuntimeException e) {
			LOG.error("Error locking vorgang.", e);
			publisher.publishEvent(new CommandFailedEvent(command.getId(), e.getMessage()));
		}
	}

	@EventListener(condition = IS_LOCK_VORGANG_COMMAND)
	public void onRevokeLockVorgang(RevokeCommandEvent event) {
		var command = event.getSource();
		try {
			vorgangService.unlockVorgang(command);
			publisher.publishEvent(new CommandRevokedEvent(command));
		} catch (RuntimeException e) {
			LOG.error("Error unlocking vorgang.", e);
			publisher.publishEvent(new CommandRevokeFailedEvent(command.getId(), e.getMessage()));
		}
	}

	void publishEvent(CommandExecutedEvent event) {
		LOG.debug("Command {} executed", event.getSource());
		publisher.publishEvent(event);
	}

	private void updatePreviousState(Command command) {
		commandService.setPreviousState(command.getId(),
				Collections.singletonMap(VorgangService.BODY_OBJECT_AKTENZEICHEN, getAktenzeichen(command.getVorgangId())));
	}

	private String getAktenzeichen(String vorgangId) {
		return vorgangHeaderService.getById(vorgangId).getAktenzeichen();
	}

	@EventListener(condition = IS_UNLOCK_VORGANG_COMMAND_CONDITION)
	public void onUnlockVorgang(CommandCreatedEvent event) {
		var command = event.getSource();
		try {
			vorgangService.unlockVorgang(command);
			publisher.publishEvent(new VorgangUnlockedEvent(command));
		} catch (RuntimeException e) {
			LOG.error("Error unlocking vorgang.", e);
			publisher.publishEvent(new CommandFailedEvent(command.getId(), e.getMessage()));
		}
	}
}