/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.vorgang;

import static de.ozgcloud.vorgang.vorgang.Vorgang.*;

import java.util.Collections;
import java.util.LinkedList;
import java.util.List;
import java.util.Objects;
import java.util.Optional;

import org.apache.commons.collections4.CollectionUtils;
import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.domain.Page;
import org.springframework.data.domain.PageImpl;
import org.springframework.data.domain.PageRequest;
import org.springframework.data.domain.Pageable;
import org.springframework.data.domain.Sort;
import org.springframework.data.mongodb.core.MongoTemplate;
import org.springframework.data.mongodb.core.aggregation.Aggregation;
import org.springframework.data.mongodb.core.aggregation.AggregationOperation;
import org.springframework.data.mongodb.core.aggregation.AggregationPipeline;
import org.springframework.data.mongodb.core.aggregation.UnionWithOperation;
import org.springframework.data.mongodb.core.query.Criteria;
import org.springframework.data.mongodb.core.query.Query;

import de.ozgcloud.vorgang.common.db.CriteriaUtil;
import de.ozgcloud.vorgang.vorgang.FindVorgangRequest.OrderCriteria;
import lombok.RequiredArgsConstructor;
import lombok.extern.log4j.Log4j2;

@Log4j2
class VorgangHeaderRepositoryImpl implements CustomVorgangHeaderRepository {

	@Autowired
	private MongoTemplate template;
	@Autowired
	private QueryCriteriaBuilder queryCriteriaBuilder;

	@Override
	public Page<VorgangHeader> findAll(FindVorgangRequest request) {
		return checkAndExecuteRequest(request, buildFilterQuery(request.getFilterBy()));
	}

	private Page<VorgangHeader> checkAndExecuteRequest(FindVorgangRequest request, Query totalQuery) {
		if (hasNoPermissionByOrganisationseinheitenId(request.getFilterBy())) {
			return new PageImpl<>(Collections.emptyList());
		}
		return new PageImpl<>(executeRequest(request), buildPageable(request), countTotalVorgaenge(totalQuery));
	}

	private List<VorgangHeader> executeRequest(FindVorgangRequest request) {
		var aggregationBuilder = aggregationBuilder(buildPageable(request))
				.withFilterBy(request.getFilterBy());
		Optional.ofNullable(request.getFindVorgangQuery())
				.ifPresentOrElse(aggregationBuilder::withQuery, () -> aggregationBuilder.withSearchBy(StringUtils.trimToNull(request.getSearchBy())));
		return execAggration(aggregationBuilder.build(request.getOrderBy()));
	}

	private boolean hasNoPermissionByOrganisationseinheitenId(FilterCriteria filterCriteria) {
		return Objects.nonNull(filterCriteria)
				&& filterCriteria.getOrganisationseinheitIds().isEmpty()
				&& filterCriteria.isFilterByOrganisationseinheitenId();
	}

	Pageable buildPageable(FindVorgangRequest request) {
		return PageRequest.of(request.getOffset() / request.getLimit(), request.getLimit());
	}

	private AggregationBuilder aggregationBuilder(Pageable pageRequest) {
		return new AggregationBuilder(pageRequest);
	}

	private List<VorgangHeader> execAggration(Aggregation aggregation) {
		return template.aggregate(aggregation, Vorgang.COLLECTION_NAME, VorgangHeader.class).getMappedResults();
	}

	Query buildFilterQuery(FilterCriteria filterCriteria) {
		return Query.query(filterToCriteria(filterCriteria));
	}

	private Criteria filterToCriteria(FilterCriteria filterCriteria) {
		var organisationseinheitenIdCriteria = buildOrganisationseinheitenCriteria(filterCriteria);
		var assignedToCriteria = buildAssignedToCriteria(filterCriteria);
		var statusCriteria = buildStatusCriteria(filterCriteria);
		var hasNextWiedervorlageFrist = buildHasNextWiedervorlageCriteria(filterCriteria);

		return CriteriaUtil.vorgangNotInCreation().andOperator(organisationseinheitenIdCriteria, assignedToCriteria, statusCriteria,
				hasNextWiedervorlageFrist);
	}

	private Criteria buildOrganisationseinheitenCriteria(FilterCriteria filterCriteria) {
		return Optional.ofNullable(filterCriteria)
				.map(FilterCriteria::getOrganisationseinheitIds)
				.filter(CollectionUtils::isNotEmpty)
				.map(CriteriaUtil::inOrganisationseinheitenIds)
				.orElseGet(Criteria::new);
	}

	private Criteria buildAssignedToCriteria(FilterCriteria filterCriteria) {
		return Optional.ofNullable(filterCriteria)
				.filter(FilterCriteria::isFilterByAssignedTo)
				.map(FilterCriteria::getAssignedTo)
				.map(this::getAssigendToCriteria)
				.orElseGet(Criteria::new);
	}

	private Criteria getAssigendToCriteria(String assignedTo) {
		if (StringUtils.isNotEmpty(assignedTo)) {
			return CriteriaUtil.vorgangIsAssignedTo(assignedTo);
		} else {
			return CriteriaUtil.vorgangIsUnassigned();
		}
	}

	private Criteria buildStatusCriteria(FilterCriteria filterCriteria) {
		return Optional.ofNullable(filterCriteria)
				.map(FilterCriteria::getStatus)
				.filter(CollectionUtils::isNotEmpty)
				.map(CriteriaUtil::vorgangInStatus)
				.orElseGet(CriteriaUtil::isNotDeleted);
	}

	private Criteria buildHasNextWiedervorlageCriteria(FilterCriteria filterCriteria) {
		return Optional.ofNullable(filterCriteria)
				.filter(FilterCriteria::isHasNextWiedervorlageFrist)
				.map(criteria -> CriteriaUtil.nextWiedervorlageFristExists())
				.orElseGet(Criteria::new);
	}

	private long countTotalVorgaenge(Query query) {
		return template.count(query, Vorgang.COLLECTION_NAME);
	}

	@RequiredArgsConstructor
	private class AggregationBuilder {
		private final Pageable pageRequest;
		private Optional<String> searchBy = Optional.empty();
		private Optional<FindVorgangQuery> findVorgangQuery = Optional.empty();
		private Optional<FilterCriteria> filterCriteria = Optional.empty();

		AggregationBuilder withSearchBy(String searchBy) {
			this.searchBy = Optional.ofNullable(searchBy);
			return this;
		}

		AggregationBuilder withQuery(FindVorgangQuery findVorgangQuery) {
			this.findVorgangQuery = Optional.ofNullable(findVorgangQuery);
			return this;
		}

		AggregationBuilder withFilterBy(FilterCriteria filterCriteria) {
			this.filterCriteria = Optional.ofNullable(filterCriteria);
			return this;
		}

		Aggregation build(OrderCriteria orderBy) {
			switch (orderBy) {
			case EA_PRIORITY:
				return buildSortedByEAPriority();
			case PRIORITY:
				return buildSortedByPriority();
			case CREATED_AT_DESC:
				return buildSortedByCreatedAtDesc();
			case NEXT_WIEDERVORLAGE_FRIST:
				return buildSortedNextWiedervorlageFrist();
			default:
				LOG.warn("Unkown order by expression '{}'. Using default order.", orderBy);
				return buildSortedByPriority();
			}
		}

		private Aggregation buildSortedByEAPriority() {
			var operations = buildSearchAndFilterAggregation();

			operations.addAll(List.of(
					Aggregation.match(CriteriaUtil.vorgangStatusInBearbeitung()),
					Aggregation.sort(Sort.by(MONGODB_FIELDNAME_CREATED_AT)),
					unionWithVorgang().pipeline(buildZuBearbeitendePipelineEA()),
					unionWithVorgang().pipeline(buildBearbeitetPipeline()),
					unionWithVorgang().pipeline(buildRemainingStatusPipeline())));

			addPageableAggregations(operations);
			return Aggregation.newAggregation(operations);
		}

		private Aggregation buildSortedByPriority() {
			var operations = buildSearchAndFilterAggregation();

			operations.addAll(findAndSortExpiredWiedervorlagen());
			operations.addAll(List.of(
					unionWithVorgang().pipeline(buildZuBearbeitendePipeline()),
					unionWithVorgang().pipeline(buildOpenWiedervorlagePipeline()),
					unionWithVorgang().pipeline(buildBearbeitetPipeline()),
					unionWithVorgang().pipeline(buildRemainingStatusPipeline())));

			addPageableAggregations(operations);
			return Aggregation.newAggregation(operations);
		}

		private Aggregation buildSortedByCreatedAtDesc() {
			var operations = buildSearchAndFilterAggregation();

			operations.add(Aggregation.sort(Sort.by(MONGODB_FIELDNAME_CREATED_AT).descending()));

			addPageableAggregations(operations);
			return Aggregation.newAggregation(operations);
		}

		private Aggregation buildSortedNextWiedervorlageFrist() {
			var operations = buildSearchAndFilterAggregation();

			operations.add(Aggregation.sort(Sort.by(CriteriaUtil.CLIENT_ATTRIBUT_NEXT_WIEDERVORLAGE_FRIST_VALUE_KEY).ascending()));

			addPageableAggregations(operations);
			return Aggregation.newAggregation(operations);
		}

		private UnionWithOperation unionWithVorgang() {
			return UnionWithOperation.unionWith(Vorgang.COLLECTION_NAME);
		}

		private List<AggregationOperation> findAndSortExpiredWiedervorlagen() {
			return List.of(
					Aggregation.match(CriteriaUtil.nextWiedervorlageFristInThePast()),
					Aggregation.sort(Sort.by(CriteriaUtil.CLIENT_ATTRIBUT_NEXT_WIEDERVORLAGE_FRIST_VALUE_KEY).ascending()));
		}

		private AggregationPipeline buildOpenWiedervorlagePipeline() {
			var operations = buildSearchAndFilterAggregation();

			operations.addAll(List.of(
					Aggregation
							.match(CriteriaUtil.nextWiedervorlageFristInTheFuture()),
					Aggregation.sort(Sort.by(CriteriaUtil.CLIENT_ATTRIBUT_NEXT_WIEDERVORLAGE_FRIST_VALUE_KEY).ascending())));
			return Aggregation.newAggregation(operations).getPipeline();
		}

		private AggregationPipeline buildZuBearbeitendePipeline() {
			var operations = buildSearchAndFilterAggregation();
			operations.addAll(List.of(
					Aggregation.match(CriteriaUtil.withoutNextWiedervorlageFrist()),
					Aggregation.match(CriteriaUtil.vorgangStatusZuBearbeitende()),
					Aggregation.sort(Sort.by(MONGODB_FIELDNAME_CREATED_AT).ascending())));

			return Aggregation.newAggregation(operations).getPipeline();
		}

		private AggregationPipeline buildZuBearbeitendePipelineEA() {
			var operations = buildSearchAndFilterAggregation();
			operations.addAll(List.of(
					Aggregation.match(CriteriaUtil.vorgangStatusZuBearbeitendeEA()),
					Aggregation.sort(Sort.by(MONGODB_FIELDNAME_CREATED_AT))));

			return Aggregation.newAggregation(operations).getPipeline();
		}

		private AggregationPipeline buildBearbeitetPipeline() {
			var operations = buildSearchAndFilterAggregation();

			operations.addAll(List.of(
					Aggregation.match(CriteriaUtil.withoutNextWiedervorlageFrist()),
					Aggregation.match(CriteriaUtil.vorgangStatusBearbeitet()),
					Aggregation.sort(Sort.by(MONGODB_FIELDNAME_CREATED_AT).descending())));

			return Aggregation.newAggregation(operations).getPipeline();
		}

		private AggregationPipeline buildRemainingStatusPipeline() {
			var operations = buildSearchAndFilterAggregation();

			operations.addAll(List.of(
					Aggregation.match(CriteriaUtil.withoutNextWiedervorlageFrist()),
					Aggregation.match(CriteriaUtil.vorgangRemainingStatus()),
					Aggregation.sort(Sort.by(MONGODB_FIELDNAME_CREATED_AT).descending())));

			return Aggregation.newAggregation(operations).getPipeline();
		}

		private List<AggregationOperation> buildSearchAndFilterAggregation() {
			var operations = new LinkedList<AggregationOperation>();
			operations.add(buildInCreationCriteria());

			buildMatchOrganisationseinheitenId(filterCriteria).ifPresent(operations::add);
			buildMatchQuery().or(this::buildMatchSearchBy).ifPresent(operations::add);
			buildMatchStatus(filterCriteria).ifPresent(operations::add);
			buildMatchAssignedTo(filterCriteria).ifPresent(operations::add);
			buildHasNextWiedervorlagenFrist(filterCriteria).ifPresent(operations::add);

			return operations;
		}

		private void addPageableAggregations(List<AggregationOperation> aggregations) {
			aggregations.add(Aggregation.skip(pageRequest.getOffset()));
			aggregations.add(Aggregation.limit(pageRequest.getPageSize()));
			aggregations.add(Aggregation.project(VorgangHeader.class));
		}

		private AggregationOperation buildInCreationCriteria() {
			return Aggregation.match(CriteriaUtil.vorgangNotInCreation());
		}

		private Optional<AggregationOperation> buildMatchOrganisationseinheitenId(Optional<FilterCriteria> filterCriteria) {
			return filterCriteria.filter(FilterCriteria::isFilterByOrganisationseinheitenId)
					.map(FilterCriteria::getOrganisationseinheitIds)
					.map(organisationseinheitIds -> Aggregation.match(CriteriaUtil.inOrganisationseinheitenIds(organisationseinheitIds)));
		}

		private Optional<AggregationOperation> buildMatchQuery() {
			return findVorgangQuery.map(queryCriteriaBuilder::from).map(Aggregation::match);
		}

		private Optional<AggregationOperation> buildMatchSearchBy() {
			return searchBy.map(search -> Aggregation.match(CriteriaUtil.searchCriteria(search)));
		}

		private Optional<AggregationOperation> buildMatchStatus(Optional<FilterCriteria> filterCriteria) {
			return filterCriteria.map(FilterCriteria::getStatus)
					.filter(CollectionUtils::isNotEmpty)
					.map(status -> Aggregation.match(CriteriaUtil.vorgangInStatus(status)));
		}

		private Optional<AggregationOperation> buildMatchAssignedTo(Optional<FilterCriteria> filterCriteria) {
			return filterCriteria.filter(FilterCriteria::isFilterByAssignedTo)
					.map(FilterCriteria::getAssignedTo)
					.map(assignedTo -> Aggregation.match(getAssigendToCriteria(assignedTo)));
		}

		private Optional<AggregationOperation> buildHasNextWiedervorlagenFrist(Optional<FilterCriteria> filterCriteria) {
			return filterCriteria.filter(FilterCriteria::isHasNextWiedervorlageFrist)
					.map(hasNextWiedervorlageFrist -> Aggregation.match(CriteriaUtil.nextWiedervorlageFristExists()));
		}
	}
}