/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.vorgang;

import static de.ozgcloud.vorgang.common.db.CriteriaUtil.*;
import static org.springframework.data.mongodb.core.query.Query.*;

import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.stream.Stream;

import org.springframework.data.mongodb.core.MongoOperations;
import org.springframework.data.mongodb.core.query.Criteria;
import org.springframework.data.mongodb.core.query.Update;
import org.springframework.stereotype.Repository;

import de.ozgcloud.vorgang.common.db.CollisionVerifier;
import de.ozgcloud.vorgang.common.db.CriteriaUtil;
import lombok.RequiredArgsConstructor;

@Repository
@RequiredArgsConstructor
class VorgangRepository {

	private final MongoOperations mongoOperations;

	private final CollisionVerifier collisionVerifier = new CollisionVerifier(this::exists);

	public Optional<Vorgang> findById(String vorgangId) {
		return Optional.ofNullable(findOne(CriteriaUtil.isId(vorgangId)));
	}

	public Vorgang findByIdAndVersion(String vorgangId, long version) {
		return findOne(whereIdAndVersion(vorgangId, version));
	}

	private Vorgang findOne(Criteria queryObj) {
		return mongoOperations.findOne(query(queryObj.andOperator(CriteriaUtil.isNotDeleted())), Vorgang.class);
	}

	public Vorgang save(Vorgang vorgang) {
		return mongoOperations.save(vorgang);
	}

	public void saveStub(VorgangStub vorgangStub) {
		mongoOperations.save(vorgangStub);
	}

	public Stream<Vorgang> findAll() {
		return mongoOperations.stream(query(CriteriaUtil.vorgangNotInCreation().andOperator(CriteriaUtil.isNotDeleted())), Vorgang.class);
	}

	public void patch(String vorgangId, long version, Map<String, Object> patch) {
		updateFirst(vorgangId, whereIdAndVersion(vorgangId, version), createPatchUpdate(patch));
	}

	private Update createPatchUpdate(Map<String, Object> patch) {
		var update = createVersionIncUpdate();
		patch.forEach(update::set);
		return update;
	}

	public void removeLock(String vorgangId, long version) {
		updateFirst(vorgangId, whereIdAndVersion(vorgangId, version), createUnlockUpdate());
	}

	private Update createUnlockUpdate() {
		return createVersionIncUpdate().unset(VorgangService.KEY_HEADER_LOCK);
	}

	private Update createVersionIncUpdate() {
		return new Update().inc(Vorgang.MONGODB_FIELDNAME_VERSION, 1);
	}

	public void setAktenzeichen(String vorgangId, long version, String akteneinsicht) {
		updateFirst(vorgangId, whereIdAndVersion(vorgangId, version), new Update().set(Vorgang.FIELD_AKTENZEICHEN, akteneinsicht));
	}

	private void updateFirst(String vorgangId, Criteria queryObj, Update update) {
		var updateResult = mongoOperations.updateFirst(query(queryObj), update, Vorgang.class);
		collisionVerifier.verify(updateResult, vorgangId);
	}

	private Criteria whereIdAndVersion(String vorgangId, long version) {
		return CriteriaUtil.isId(vorgangId).and(Vorgang.MONGODB_FIELDNAME_VERSION).is(version);
	}

	public boolean exists(String vorgangId) {
		return mongoOperations.exists(query(isId(vorgangId)), Vorgang.class);
	}

	public boolean exists(String vorgangId, Collection<String> organisationEinheitenIds) {
		return mongoOperations.exists(
				query(isId(vorgangId)
						.andOperator(inOrganisationseinheitenIds(organisationEinheitenIds)
								.andOperator(isNotDeleted()))),
				Vorgang.class);
	}

	public boolean isVorgangLocked(String vorgangId) {
		return mongoOperations.exists(query(whereIdAndLock(vorgangId)), Vorgang.class);
	}

	Criteria whereIdAndLock(String vorgangId) {
		return new Criteria().andOperator(CriteriaUtil.isId(vorgangId), CriteriaUtil.isNotDeleted(),
				Criteria.where(VorgangService.KEY_HEADER_LOCK).ne(null));
	}

	public Stream<VorgangStub> findDeleted() {
		return mongoOperations.stream(query(CriteriaUtil.vorgangInStatus(List.of(Vorgang.Status.DELETED))), VorgangStub.class);
	}
}