/*
 * Copyright (C) 2025 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.vorgang.redirect;

import java.io.InputStream;
import java.util.List;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;
import java.util.function.BiFunction;
import java.util.function.Function;

import org.springframework.stereotype.Service;

import com.google.protobuf.ByteString;

import de.ozgcloud.common.binaryfile.GrpcFileUploadUtils;
import de.ozgcloud.common.binaryfile.GrpcFileUploadUtils.FileSender;
import de.ozgcloud.common.errorhandling.TechnicalException;
import de.ozgcloud.eingang.forwarder.RouteForwardingServiceGrpc;
import de.ozgcloud.eingang.forwarding.GrpcAttachment;
import de.ozgcloud.eingang.forwarding.GrpcFileContent;
import de.ozgcloud.eingang.forwarding.GrpcRepresentation;
import de.ozgcloud.eingang.forwarding.GrpcRouteForwardingRequest;
import de.ozgcloud.eingang.forwarding.GrpcRouteForwardingResponse;
import de.ozgcloud.vorgang.callcontext.VorgangManagerClientCallContextAttachingInterceptor;
import de.ozgcloud.vorgang.files.FileService;
import de.ozgcloud.vorgang.vorgang.Eingang;
import de.ozgcloud.vorgang.vorgang.IncomingFile;
import de.ozgcloud.vorgang.vorgang.IncomingFileGroup;
import de.ozgcloud.vorgang.vorgang.IncomingFileMapper;
import de.ozgcloud.vorgang.vorgang.VorgangService;
import io.grpc.stub.CallStreamObserver;
import io.grpc.stub.StreamObserver;
import lombok.RequiredArgsConstructor;
import net.devh.boot.grpc.client.inject.GrpcClient;

@Service
@RequiredArgsConstructor
class ForwardingRemoteService {

	private static final int TIMEOUT_MINUTES = 2;
	private final VorgangService vorgangService;
	private final ForwardingRequestMapper forwardingRequestMapper;
	@GrpcClient("forwarder")
	private final RouteForwardingServiceGrpc.RouteForwardingServiceStub serviceStub;
	private final FileService fileService;
	private final IncomingFileMapper incomingFileMapper;

	public void forward(ForwardingRequest request) {
		CompletableFuture<Void> responseFuture = new CompletableFuture<>();
		routeForwarding(request, new ForwardingResponseObserver(responseFuture));
		waitForCompletion(responseFuture);
	}

	void routeForwarding(ForwardingRequest request, ForwardingResponseObserver responseObserver) {
		var requestStreamObserver = serviceStub.withInterceptors(new VorgangManagerClientCallContextAttachingInterceptor())
				.routeForwarding(responseObserver);
		try {
			sendEingang(request, requestStreamObserver);
			requestStreamObserver.onCompleted();
		} catch (Exception e) {
			requestStreamObserver.onError(e);
			throw e;
		}
	}

	void sendEingang(ForwardingRequest request, StreamObserver<GrpcRouteForwardingRequest> requestStreamObserver) {
		var eingang = vorgangService.getById(request.getVorgangId()).getEingangs().getFirst();
		requestStreamObserver.onNext(buildRouteForwardingRequest(request, eingang));
		sendAttachments(eingang.getAttachments(), requestStreamObserver);
		sendRepresentations(eingang.getRepresentations(), requestStreamObserver);
	}

	GrpcRouteForwardingRequest buildRouteForwardingRequest(ForwardingRequest request, Eingang eingang) {
		var routeForwarding = forwardingRequestMapper.toGrpcRouteForwarding(request, eingang);
		return GrpcRouteForwardingRequest.newBuilder().setRouteForwarding(routeForwarding).build();
	}

	void sendAttachments(List<IncomingFileGroup> attachments, StreamObserver<GrpcRouteForwardingRequest> requestStreamObserver) {
		for (var attachment : attachments) {
			var groupName = attachment.getName();
			attachment.getFiles().forEach(file -> sendAttachmentFile(requestStreamObserver, groupName, file));
		}
	}

	private void sendAttachmentFile(StreamObserver<GrpcRouteForwardingRequest> requestStreamObserver, String groupName, IncomingFile file) {
		var fileContentStream = fileService.getUploadedFileStream(file.getId());
		createAttachmentFileSender(requestStreamObserver, groupName, file, fileContentStream).send();
	}

	FileSender<GrpcRouteForwardingRequest, GrpcRouteForwardingResponse> createAttachmentFileSender(
			StreamObserver<GrpcRouteForwardingRequest> requestStreamObserver, String groupName, IncomingFile file, InputStream fileContentStream) {
		return createSenderWithoutMetadata(this::buildAttachmentChunk, requestStreamObserver, fileContentStream)
				.withMetaData(buildGrpcAttachmentFile(groupName, file));
	}

	GrpcRouteForwardingRequest buildAttachmentChunk(byte[] chunk, int length) {
		return GrpcRouteForwardingRequest.newBuilder()
				.setAttachment(GrpcAttachment.newBuilder()
						.setContent(buildGrpcFileContent(chunk, length))
						.build())
				.build();
	}

	GrpcRouteForwardingRequest buildGrpcAttachmentFile(String name, IncomingFile file) {
		return GrpcRouteForwardingRequest.newBuilder()
				.setAttachment(GrpcAttachment.newBuilder()
						.setFile(incomingFileMapper.toAttachmentFile(name, file))
						.build())
				.build();
	}

	void sendRepresentations(List<IncomingFile> representations, StreamObserver<GrpcRouteForwardingRequest> requestObserver) {
		representations.forEach(representation -> {
			var fileContentStream = fileService.getUploadedFileStream(representation.getId());
			createRepresentationFileSender(requestObserver, representation, fileContentStream).send();
		});
	}

	FileSender<GrpcRouteForwardingRequest, GrpcRouteForwardingResponse> createRepresentationFileSender(
			StreamObserver<GrpcRouteForwardingRequest> requestStreamObserver, IncomingFile file, InputStream fileContentStream) {
		return createSenderWithoutMetadata(this::buildRepresentationChunk, requestStreamObserver, fileContentStream)
				.withMetaData(buildGrpcRepresentationFile(file));
	}

	FileSender<GrpcRouteForwardingRequest, GrpcRouteForwardingResponse> createSenderWithoutMetadata(
			BiFunction<byte[], Integer, GrpcRouteForwardingRequest> chunkBuilder,
			StreamObserver<GrpcRouteForwardingRequest> requestStreamObserver, InputStream fileContentStream) {
		return GrpcFileUploadUtils
				.createSender(chunkBuilder, fileContentStream, requestCallStreamObserverProvider(requestStreamObserver), false);
	}

	private Function<StreamObserver<GrpcRouteForwardingResponse>, CallStreamObserver<GrpcRouteForwardingRequest>> requestCallStreamObserverProvider(
			StreamObserver<GrpcRouteForwardingRequest> requestStreamObserver) {
		return response -> (CallStreamObserver<GrpcRouteForwardingRequest>) requestStreamObserver;
	}

	GrpcRouteForwardingRequest buildRepresentationChunk(byte[] chunk, int length) {
		return GrpcRouteForwardingRequest.newBuilder()
				.setRepresentation(GrpcRepresentation.newBuilder()
						.setContent(buildGrpcFileContent(chunk, length))
						.build())
				.build();
	}

	GrpcFileContent buildGrpcFileContent(byte[] chunk, int length) {
		var fileContentBuilder = GrpcFileContent.newBuilder();
		if (length <= 0) {
			fileContentBuilder.setIsEndOfFile(true);
		} else {
			fileContentBuilder.setContent(ByteString.copyFrom(chunk));
		}
		return fileContentBuilder.build();
	}

	GrpcRouteForwardingRequest buildGrpcRepresentationFile(IncomingFile file) {
		return GrpcRouteForwardingRequest.newBuilder()
				.setRepresentation(GrpcRepresentation.newBuilder()
						.setFile(incomingFileMapper.toRepresentationFile(file))
						.build())
				.build();
	}

	void waitForCompletion(CompletableFuture<Void> responseFuture) {
		try {
			responseFuture.get(TIMEOUT_MINUTES, TimeUnit.MINUTES);
		} catch (InterruptedException e) {
			Thread.currentThread().interrupt();
			throw new TechnicalException("Waiting for finishing file upload was interrupted.", e);
		} catch (ExecutionException e) {
			throw new TechnicalException("Error on uploading file content.", e);
		} catch (TimeoutException e) {
			throw new TechnicalException("Timeout on uploading file content.", e);
		}
	}

	@RequiredArgsConstructor
	static class ForwardingResponseObserver implements StreamObserver<GrpcRouteForwardingResponse> {
		private final CompletableFuture<Void> future;

		@Override
		public void onNext(GrpcRouteForwardingResponse value) {
			// noop
		}

		@Override
		public void onError(Throwable t) {
			future.completeExceptionally(t);
		}

		@Override
		public void onCompleted() {
			future.complete(null);
		}
	}
}
