/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.vorgang.redirect;

import java.io.IOException;
import java.io.StringWriter;
import java.time.ZoneOffset;
import java.time.ZonedDateTime;
import java.util.HashMap;
import java.util.Map;
import java.util.Optional;
import java.util.stream.Stream;

import org.apache.commons.io.IOUtils;
import org.apache.commons.lang3.ArrayUtils;
import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.context.ApplicationEventPublisher;
import org.springframework.stereotype.Service;
import org.springframework.util.MimeTypeUtils;

import de.ozgcloud.command.Command;
import de.ozgcloud.common.errorhandling.TechnicalException;
import de.ozgcloud.nachrichten.email.MailRecipient;
import de.ozgcloud.nachrichten.email.MailSendRequest;
import de.ozgcloud.nachrichten.email.MailSendRequest.MailAttachment;
import de.ozgcloud.nachrichten.email.MailService;
import de.ozgcloud.vorgang.command.PersistedCommand;
import de.ozgcloud.vorgang.status.StatusService;
import de.ozgcloud.vorgang.vorgang.Vorgang;
import de.ozgcloud.vorgang.vorgang.VorgangService;
import de.ozgcloud.vorgang.vorgang.redirect.Forwarding.Status;
import freemarker.template.Configuration;
import freemarker.template.Template;
import freemarker.template.TemplateException;
import jakarta.mail.util.ByteArrayDataSource;
import lombok.RequiredArgsConstructor;
import lombok.extern.log4j.Log4j2;

@Service
@Log4j2
@RequiredArgsConstructor
public class ForwardingService {

	static final String MAIL_TEMPLATE = "/mail/redirect.txt.ftlh";

	static final String FIELD_SHOW_DISCALAIMER = "showDisclaimer";

	static final String FIELD_ANTRAGSTELLER_ANREDE = "antragstellerAnrede";
	static final String FIELD_ANTRAGSTELLER_NAME = "antragstellerName";
	static final String FIELD_ANTRAGS_NAME = "antragsName";
	static final String FIELD_EINGANGS_KENNZEICHEN = "eingangsKennzeichen";
	static final String FIELD_AKTENZEICHEN = "aktenzeichen";
	static final String FIELD_ZUSTAENDIGE_STELLE = "zustaendigeStelle";
	static final String FIELD_CID_LOGO_EA = "cidLogoEa";
	static final String FIELD_CID_LOGO_ITVSH = "cidLogoItvsh";

	static final String IMAGES_PATH_TMPL = "/images/%s"; // NOSONAR
	static final String CID_LOGO_EA = "logo_ea_sh";
	static final String PATH_LOGO_EA = "logo-ea-sh.png";
	static final String CID_LOGO_ITVSH = "logo_itvsh";
	static final String PATH_LOGO_ITVSH = "logo-itvsh.png";

	static final String MAIL_CONFIRMATION_TEMPLATE = "/mail/redirect.confirmation.txt.ftlh";

	static final String SUBJECT_TEMPLATE = "%s - #%s#";
	static final String SUBJECT_BESTAETIGUNG_TEMPL = "Ihr Antrag auf %s - Az: %s";
	static final String ZIP_FILENAME_TMPL = "%s.zip";

	private final ForwardingRepository repository;
	private final Configuration freemarkerCfg;
	private final ZipBuilderService zipService;
	private final MailService emailService;
	private final VorgangService vorgangService;
	// TODO abhängigkeit entfernen
	private final StatusService statusService;
	private final ForwardingRemoteService forwardingRemoteService;
	private final ApplicationEventPublisher publisher;

	@Value("${ozgcloud.redirect.mail-from}")
	private String mailFrom;
	@Value("${ozgcloud.forwarding.reply-to:#{null}}")
	private Optional<String> replyTo;
	@Value("${ozgcloud.production}")
	private boolean production;

	public Vorgang forwardByCommand(Command command) {
		var vorgang = vorgangService.getById(command.getRelationId());
		if (vorgang.getVersion() == command.getRelationVersion()) {
			doRedirect(command, vorgang);
		} else {
			LOG.warn("concurrend modification on forwarding Vorgang");
			markAsFailed(command);
		}

		return vorgang;
	}

	void doRedirect(Command command, Vorgang vorgang) {
		statusService.setStatusToInBearbeitung(command.getRelationId(), command.getRelationVersion());
		redirectVorgang(vorgang, command);
	}

	void redirectVorgang(Vorgang vorgang, Command command) {
		var forwarding = saveForwarding(command);

		var mailRequest = createForwardMail(vorgang, forwarding, ((PersistedCommand) command).getRedirectRequest().getPassword().toCharArray());

		emailService.sendMail(mailRequest);
	}

	private Forwarding saveForwarding(Command command) {
		return repository.save(command.getRelationId(), buildForwarding(command));
	}

	Forwarding buildForwarding(Command command) {
		return Forwarding.builder()
				.vorgangId(command.getRelationId())
				.status(Status.NEW)
				.createdBy(command.getCreatedBy())
				.createdByName(command.getCreatedByName())
				.zustaendigeStelle(((PersistedCommand) command).getRedirectRequest().getEmail())
				.createdByCommand(command.getId())
				.build();
	}

	MailSendRequest createForwardMail(Vorgang vorgang, Forwarding forwarding, char[] password) {
		return createRedirectMailRequestBuilder()
				.toAddress(MailRecipient.builder().email(forwarding.getZustaendigeStelle()).build())
				.subject(buildForwardSubject(vorgang))
				.body(fillMailTemplate(vorgang))
				.receiptRequired(true)
				.requestReference(forwarding)
				.attachment(buildZipAttachment(vorgang, password, shouldEntcrypt(password)))
				.build();
	}

	boolean shouldEntcrypt(char[] password) {
		return ArrayUtils.isNotEmpty(password);
	}

	MailSendRequest.MailAttachment buildZipAttachment(Vorgang vorgang, char[] password, boolean encrypted) {
		var ds = new ByteArrayDataSource(zipService.buildVorgangZip(vorgang, password, encrypted).toByteArray(), "application/zip");
		ds.setName(buildZipFileName(vorgang));
		return MailAttachment.builder()
				.dataSource(ds)
				.build();
	}

	private String buildZipFileName(Vorgang vorgang) {
		var eingang = vorgang.getEingangs().get(0);
		return String.format(ZIP_FILENAME_TMPL, eingang.getHeader().getRequestId());
	}

	private String buildForwardSubject(Vorgang vorgang) {
		var eingang = vorgang.getEingangs().get(0);
		return String.format(SUBJECT_TEMPLATE, vorgang.getName(), eingang.getHeader().getRequestId());

	}

	String fillMailTemplate(Vorgang vorgang) {
		return fillTemplate(MAIL_TEMPLATE, buildValues(vorgang));
	}

	public void onRedirectMailSent(Forwarding forwarding) {
		setForwardingSent(forwarding);
		publisher.publishEvent(new VorgangRedirectedEvent(forwarding));

		sendRedirectConfirmationMail(forwarding);
	}

	private void setForwardingSent(Forwarding forwarding) {
		repository.patch(forwarding.getId(),
				Map.of(Forwarding.FIELD_STATUS, Status.SENT, Forwarding.FIELD_SENT_AT, ZonedDateTime.now(ZoneOffset.UTC)));
	}

	public void sendRedirectConfirmationMail(Forwarding forwarding) {
		var vorgang = getVorgang(forwarding);

		if (antragstellerEMailExists(vorgang)) {
			var mailRequest = createConfirmationMail(vorgang);

			emailService.sendMail(mailRequest);
		}
	}

	private Vorgang getVorgang(Forwarding forwarding) {
		return vorgangService.getById(forwarding.getVorgangId());
	}

	boolean antragstellerEMailExists(Vorgang vorgang) {
		String eMail = vorgang.getEingangs().get(0).getAntragsteller().getEmail();

		return StringUtils.isNotEmpty(eMail);
	}

	MailSendRequest createConfirmationMail(Vorgang vorgang) {
		return createRedirectMailRequestBuilder()
				.toAddress(MailRecipient.builder().email(vorgang.getEingangs().get(0).getAntragsteller().getEmail()).build())
				.subject(buildSubject(vorgang))
				.body(fillConfirmationTemplate(vorgang))
				.build();
	}

	public void onRedirectMailFailed(Forwarding forwarding, String errorMessage) {
		setForwardingSendError(forwarding, errorMessage);

		publisher.publishEvent(new VorgangRedirectFailedEvent(forwarding, errorMessage));
	}

	private void setForwardingSendError(Forwarding forwarding, String errorMessage) {
		repository.patch(forwarding.getId(), Map.of(Forwarding.FIELD_STATUS, Status.SEND_ERROR, Forwarding.FIELD_ERROR_MSG, errorMessage));
	}

	MailSendRequest.MailSendRequestBuilder createRedirectMailRequestBuilder() {
		var builder = MailSendRequest.builder()
				.fromAddress(mailFrom)
				.attachment(buildImageAttachment(PATH_LOGO_EA, CID_LOGO_EA))
				.attachment(buildImageAttachment(PATH_LOGO_ITVSH, CID_LOGO_ITVSH));
		replyTo.ifPresent(builder::replyTo);
		return builder;
	}

	private String buildSubject(Vorgang vorgang) {
		var eingang = vorgang.getEingangs().get(0);

		return String.format(SUBJECT_BESTAETIGUNG_TEMPL, eingang.getHeader().getFormName(), eingang.getHeader().getRequestId());
	}

	String fillConfirmationTemplate(Vorgang vorgang) {
		return fillTemplate(MAIL_CONFIRMATION_TEMPLATE, buildValues(vorgang));
	}

	Map<String, Object> buildValues(Vorgang vorgang) {
		var map = new HashMap<String, Object>();
		var eingang = vorgang.getEingangs().get(0);

		map.put(FIELD_SHOW_DISCALAIMER, !production);
		map.put(FIELD_ANTRAGS_NAME, eingang.getHeader().getFormName());
		map.put(FIELD_EINGANGS_KENNZEICHEN, eingang.getHeader().getRequestId());
		map.put(FIELD_AKTENZEICHEN, vorgang.getAktenzeichen());
		map.put(FIELD_ANTRAGSTELLER_ANREDE, eingang.getAntragsteller().getAnrede());
		map.put(FIELD_ANTRAGSTELLER_NAME, eingang.getAntragsteller().getNachname());
		map.put(FIELD_ZUSTAENDIGE_STELLE, eingang.getZustaendigeStelle().getBezeichnung());
		map.put(FIELD_CID_LOGO_EA, CID_LOGO_EA);
		map.put(FIELD_CID_LOGO_ITVSH, CID_LOGO_ITVSH);

		return map;
	}

	private String fillTemplate(String templateName, Object dataModel) {
		try {
			Template template = getTemplate(templateName);
			StringWriter stringWriter = new StringWriter();
			template.process(dataModel, stringWriter);
			return stringWriter.toString();

		} catch (IOException | TemplateException e) {
			throw new TechnicalException("Error filling mail template", e);
		}
	}

	Template getTemplate(String templateName) {
		try {
			return freemarkerCfg.getTemplate(templateName);
		} catch (IOException e) {
			throw new TechnicalException("Error loading mail template", e);
		}
	}

	MailAttachment buildImageAttachment(String name, String contentId) {
		ByteArrayDataSource dataSource;
		try {
			dataSource = new ByteArrayDataSource(IOUtils.resourceToByteArray(String.format(IMAGES_PATH_TMPL, name)), MimeTypeUtils.IMAGE_PNG_VALUE);
			dataSource.setName(name);
		} catch (IOException e) {
			throw new TechnicalException("Error read image for attachment", e);
		}

		return MailAttachment.builder()
				.dataSource(dataSource)
				.contentId(contentId)
				.build();
	}

	public void markAsSuccessfull(Command command) {
		repository.patch(command.getRelationId(), Map.of(Forwarding.FIELD_STATUS, Status.SUCCESSFULL));

		publisher.publishEvent(new VorgangForwardSuccessfullEvent(command));
	}

	public void markAsFailed(Command command) {
		repository.patch(command.getRelationId(), Map.of(Forwarding.FIELD_STATUS, Status.FAILED));

		publisher.publishEvent(new VorgangForwardFailedEvent(command));
	}

	public Stream<Forwarding> findForwardings(String vorgangId) {
		return repository.findByVorgangId(vorgangId).stream();
	}

	public void forward(ForwardingRequest request) {
		statusService.setStatusToWeitergeleitet(request.getVorgangId(), request.getVersion());
		forwardingRemoteService.forward(request);
	}
}