/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.vorgang.redirect;

import java.io.ByteArrayOutputStream;
import java.io.IOException;

import org.apache.commons.io.IOUtils;
import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import org.springframework.util.MimeTypeUtils;

import de.ozgcloud.common.errorhandling.TechnicalException;
import de.ozgcloud.vorgang.files.FileService;
import de.ozgcloud.vorgang.vorgang.Eingang;
import de.ozgcloud.vorgang.vorgang.IncomingFile;
import de.ozgcloud.vorgang.vorgang.IncomingFileGroup;
import de.ozgcloud.vorgang.vorgang.Vorgang;
import net.lingala.zip4j.io.outputstream.ZipOutputStream;
import net.lingala.zip4j.model.ZipParameters;
import net.lingala.zip4j.model.enums.AesKeyStrength;
import net.lingala.zip4j.model.enums.EncryptionMethod;

@Service
class ZipBuilderService {

	@Autowired
	private FileService fileService;

	public ByteArrayOutputStream buildVorgangZip(Vorgang vorgang, char[] password, boolean encrypted) {
		return new ZipBuilder(vorgang, encrypted).buildZip(password);
	}

	class ZipBuilder {

		private static final String FOLDER_NAME_XML = "xml";
		private static final String FILE_NAME_TEMPLATE = "%s/%s";

		private final Eingang eingang;
		private final boolean encrypted;

		private ZipOutputStream zipOut;

		ZipBuilder(Vorgang vorgang, boolean encrypted) {
			this.eingang = vorgang.getEingangs().get(0);
			this.encrypted = encrypted;
		}

		ByteArrayOutputStream buildZip(char[] password) {
			ByteArrayOutputStream out = new ByteArrayOutputStream(256);

			try (ZipOutputStream zOut = initZipOutputStream(out, password)) {
				this.zipOut = zOut;
				fillZip();

				zipOut.flush();
			} catch (IOException e) {
				throw new TechnicalException("Error writing Vorgang-ZIP", e);
			}

			return out;
		}

		private ZipOutputStream initZipOutputStream(ByteArrayOutputStream out, char[] password) {
			try {
				return encrypted ? new ZipOutputStream(out, password) : new ZipOutputStream(out);
			} catch (IOException e) {
				throw new TechnicalException("Error writing Vorgang-ZIP", e);
			}
		}

		private void fillZip() {
			addRepresentation();
			addAttachments();
		}

		private void addRepresentation() {
			eingang.getRepresentations().stream().forEach(inFile -> addFile(folderByContentType(inFile.getContentType()), inFile));
		}

		private String folderByContentType(String contentType) {
			return StringUtils.equals(contentType, MimeTypeUtils.APPLICATION_XML_VALUE) ? FOLDER_NAME_XML : "";
		}

		private void addAttachments() {
			eingang.getAttachments().stream().forEach(this::addFileGroup);
		}

		void addFileGroup(IncomingFileGroup fileGroup) {
			fileGroup.getFiles().stream().forEach(inFile -> addFile(fileGroup.getName(), inFile));
		}

		void addFile(String folder, IncomingFile file) {
			var fileName = StringUtils.isEmpty(folder) ? file.getName() : String.format(FILE_NAME_TEMPLATE, folder, file.getName());

			try {
				zipOut.putNextEntry(buildZipParameter(fileName));
				IOUtils.copy(fileService.getUploadedFileStream(file.getId()), zipOut);
				zipOut.closeEntry();
			} catch (IOException e) {
				throw new TechnicalException("Error writing zip entry", e);
			}

		}

		private ZipParameters buildZipParameter(String fileName) {
			var zipParameters = new ZipParameters();

			if (encrypted) {
				zipParameters.setEncryptFiles(true);
				zipParameters.setEncryptionMethod(EncryptionMethod.AES);
				zipParameters.setAesKeyStrength(AesKeyStrength.KEY_STRENGTH_256);
			}
			zipParameters.setFileNameInZip(fileName);

			return zipParameters;
		}
	}
}
