/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.document;

import static org.assertj.core.api.Assertions.*;
import static org.awaitility.Awaitility.*;
import static org.mockito.Mockito.*;

import java.util.Map;
import java.util.concurrent.TimeUnit;

import org.junit.jupiter.api.AfterAll;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.boot.test.mock.mockito.SpyBean;
import org.springframework.context.ApplicationEventPublisher;
import org.springframework.data.mongodb.core.MongoOperations;
import org.springframework.security.core.context.SecurityContextHolder;
import org.springframework.test.annotation.DirtiesContext;

import de.ozgcloud.apilib.common.command.grpc.CommandMapper;
import de.ozgcloud.command.Command;
import de.ozgcloud.command.CommandCreatedEvent;
import de.ozgcloud.command.CommandFailedEvent;
import de.ozgcloud.common.test.DataITCase;
import de.ozgcloud.vorgang.VorgangManagerServerApplication;
import de.ozgcloud.vorgang.attached_item.VorgangAttachedItem;
import de.ozgcloud.vorgang.attached_item.VorgangAttachedItemTestFactory;
import de.ozgcloud.vorgang.callcontext.CallContextAuthenticationTokenTestFactory;
import de.ozgcloud.vorgang.command.CommandTestFactory;

@SpringBootTest(classes = VorgangManagerServerApplication.class, properties = {
		"grpc.server.inProcessName=test",
		"grpc.server.port=-1",
		"grpc.client.ozgcloud-command-manager.address=in-process:test",
		"grpc.client.vorgang-manager.address=in-process:test"
})
@DataITCase
@DirtiesContext
class DocumentEventListenerITCase {

	private static final String FILE_ID = "file-id";

	@Autowired
	private ApplicationEventPublisher eventPublisher;
	@SpyBean
	private DocumentEventListener eventListener;
	@Autowired
	private CommandMapper mapper;
	@Autowired
	private MongoOperations mongoOperations;

	@Captor
	private ArgumentCaptor<BescheidDocumentCreatedEvent> bescheidDocumentCreatedEventCaptor;
	@Captor
	private ArgumentCaptor<CommandFailedEvent> commandFailedEventCaptor;

	@AfterAll
	static void cleanup() {
		SecurityContextHolder.clearContext();
	}

	@BeforeEach
	void init() {
		SecurityContextHolder.getContext().setAuthentication(CallContextAuthenticationTokenTestFactory.create());
		mongoOperations.dropCollection(VorgangAttachedItem.COLLECTION_NAME);
	}

	@Test
	void shouldCreateDocument() {
		var bescheidItem = mongoOperations.save(createBescheidItem("DRAFT"));
		var command = createCommand(bescheidItem.getId());

		eventPublisher.publishEvent(new CommandCreatedEvent(command));

		assertThat(loadBescheidDocument()).containsEntry(DocumentService.FIELD_DOCUMENT_FILE, FILE_ID);
	}

	@Test
	void shouldFailWhenNoBescheidFound() {
		var command = createCommand("non-existing-id");

		eventPublisher.publishEvent(new CommandCreatedEvent(command));

		await().atMost(60, TimeUnit.SECONDS).untilAsserted(() -> {
			verify(eventListener, timeout(5000)).publishEvent(commandFailedEventCaptor.capture());
			assertThat(commandFailedEventCaptor.getValue().getSource()).isEqualTo(command.getId());
		});
	}

	@Test
	void shouldFailWhenNotDraft() {
		var bescheidItem = mongoOperations.save(createBescheidItem("BESCHEID"));
		var command = createCommand(bescheidItem.getId());

		eventPublisher.publishEvent(new CommandCreatedEvent(command));

		await().atMost(60, TimeUnit.SECONDS).untilAsserted(() -> {
			verify(eventListener, timeout(5000)).publishEvent(commandFailedEventCaptor.capture());
			assertThat(commandFailedEventCaptor.getValue().getSource()).isEqualTo(command.getId());
		});
	}

	private VorgangAttachedItem createBescheidItem(String status) {
		return VorgangAttachedItemTestFactory.createBuilder().id(null).version(0)
				.item(Map.of("status", status))
				.build();
	}

	private Map<String, Object> loadBescheidDocument() {
		verify(eventListener, timeout(5000)).publishEvent(bescheidDocumentCreatedEventCaptor.capture());
		var bescheidDocumentId = bescheidDocumentCreatedEventCaptor.getValue().getCreatedResource();
		return mongoOperations.findById(bescheidDocumentId, VorgangAttachedItem.class).getItem();
	}

	private Command createCommand(String relationId) {
		return CommandTestFactory.createBuilder()
				.relationId(relationId)
				.order(DocumentEventListener.CREATE_BESCHEID_DOCUMENT_FROM_FILE_ORDER)
				.bodyObject(Map.of(DocumentService.FIELD_DOCUMENT_FILE, FILE_ID))
				.build();
	}
}