/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.document.bescheid;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.Mockito.*;

import java.util.Map;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.Mock;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.boot.test.mock.mockito.MockBean;
import org.springframework.data.mongodb.core.MongoOperations;
import org.springframework.test.annotation.DirtiesContext;

import de.ozgcloud.common.test.DataITCase;
import de.ozgcloud.document.bescheid.BescheidServiceGrpc.BescheidServiceBlockingStub;
import de.ozgcloud.document.bescheid.common.callcontext.CallContextUser;
import de.ozgcloud.document.bescheid.common.callcontext.CurrentUserService;
import de.ozgcloud.document.bescheid.common.user.UserProfile;
import de.ozgcloud.document.bescheid.common.user.UserProfileService;
import de.ozgcloud.document.common.attached_item.AttachedItemService;
import de.ozgcloud.vorgang.VorgangManagerServerApplication;
import de.ozgcloud.vorgang.attached_item.VorgangAttachedItem;
import de.ozgcloud.vorgang.attached_item.VorgangAttachedItemTestFactory;
import de.ozgcloud.vorgang.callcontext.CallContextUserTestFactory;
import de.ozgcloud.vorgang.vorgang.Vorgang;
import de.ozgcloud.vorgang.vorgang.VorgangTestFactory;
import net.devh.boot.grpc.client.inject.GrpcClient;

@SpringBootTest(classes = VorgangManagerServerApplication.class, properties = {
		"grpc.server.inProcessName=test",
		"grpc.server.port=-1",
		"grpc.client.inProzess.address=in-process:test",
		"grpc.client.vorgang-manager.address=in-process:test",
})
@DataITCase
@DirtiesContext
class BescheidGrpcServiceITCase {

	@GrpcClient("inProzess")
	private BescheidServiceBlockingStub bescheidServiceBlockingStub;

	@Autowired
	private MongoOperations mongoOperations;

	@MockBean
	private CurrentUserService currentUserService;
	@MockBean
	private UserProfileService userProfileService;
	@Mock
	private CallContextUser callContextUser;
	@Mock
	private UserProfile userProfile;

	@Nested
	class TestFindAll {

		@BeforeEach
		void init() {
			mongoOperations.dropCollection(VorgangAttachedItem.COLLECTION_NAME);
			mongoOperations.dropCollection(Vorgang.COLLECTION_NAME);
			when(currentUserService.getUser()).thenReturn(callContextUser);
			when(userProfileService.getUserProfile()).thenReturn(userProfile);
		}

		@Test
		void shouldReturnSortedResult() {
			var vorgangId = mongoOperations.save(VorgangTestFactory.createBuilder().id(null).build()).getId();
			var bescheidItemFirstId = mongoOperations.save(createBescheidAttachedItem(vorgangId, "2022-10-01")).getId();
			var bescheidItemLastId = mongoOperations.save(createBescheidAttachedItem(vorgangId, "2024-01-01")).getId();

			var response = bescheidServiceBlockingStub.getAll(GrpcGetAllBescheidRequest.newBuilder().setVorgangId(vorgangId).build());

			assertThat(response.getBescheidList()).hasSize(2);
			assertThat(response.getBescheidList().get(0).getId()).isEqualTo(bescheidItemLastId);
			assertThat(response.getBescheidList().get(1).getId()).isEqualTo(bescheidItemFirstId);
		}

		@Test
		void shouldReturnSentInfo() {
			var vorgangId = mongoOperations.save(VorgangTestFactory.createBuilder().id(null).build()).getId();
			var itemMap = createItemWithSentInfo();
			mongoOperations.save(createBescheidAttachedItem(vorgangId, itemMap));

			var response = bescheidServiceBlockingStub.getAll(GrpcGetAllBescheidRequest.newBuilder().setVorgangId(vorgangId).build());

			assertThat(response.getBescheidList()).hasSize(1).first().extracting(GrpcBescheid::getSentInfo)
					.isEqualTo(GrpcSentInfoTestFactory.create());
		}

		private Map<String, Object> createItemWithSentInfo() {
			return Map.of(
					Bescheid.FIELD_BEWILLIGT, "true",
					Bescheid.FIELD_STATUS, Bescheid.Status.SENT.name(),
					Bescheid.FIELD_BESCHIEDEN_AM, "2022-10-01",
					Bescheid.FIELD_SENT_INFO, Map.of(
							SentInfo.FIELD_SENT_AT, GrpcSentInfoTestFactory.SENT_AT_STR,
							SentInfo.FIELD_SENT_BY, CallContextUserTestFactory.ID));
		}

		private VorgangAttachedItem createBescheidAttachedItem(String vorgangId, String beschiedenAm) {
			var itemMap = Map.<String, Object>of(
					Bescheid.FIELD_BEWILLIGT, "true",
					Bescheid.FIELD_STATUS, Bescheid.Status.DRAFT.name(),
					Bescheid.FIELD_BESCHIEDEN_AM, beschiedenAm);
			return createBescheidAttachedItem(vorgangId, itemMap);
		}

		private VorgangAttachedItem createBescheidAttachedItem(String vorgangId, Map<String, Object> itemMap) {
			return VorgangAttachedItemTestFactory.createBuilder()
					.id(null)
					.version(0L)
					.vorgangId(vorgangId)
					.itemName(AttachedItemService.BESCHEID_ITEM_NAME)
					.client(BescheidCallContextAttachingInterceptor.BESCHEID_MANAGER_CLIENT_NAME)
					.item(itemMap)
					.build();
		}
	}

	@Nested
	class TestGetConfig {

		@Test
		void shouldReturnBescheidManagerConfig() {
			var config = bescheidServiceBlockingStub.getConfig(GrpcBescheidManagerConfigRequest.newBuilder().build());

			assertThat(config.getVersion()).isNotEmpty();
			assertThat(config.getJavaVersion()).isEqualTo(System.getProperty("java.version"));
			assertThat(config.getFeatures().getCanCreateBescheidDocument()).isFalse();
		}
	}
}