/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.nachrichten.antragraum;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.Collections;
import java.util.List;
import java.util.Map;

import org.apache.commons.lang3.StringUtils;
import org.apache.xml.security.stax.ext.XMLSecurityConstants.DIRECTION;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.Mock;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.boot.test.mock.mockito.MockBean;
import org.springframework.data.mongodb.core.MongoOperations;
import org.springframework.security.saml2.core.Saml2Error;
import org.springframework.security.test.context.support.WithMockUser;
import org.springframework.test.annotation.DirtiesContext;

import com.google.protobuf.ByteString;

import de.ozgcloud.apilib.common.datatypes.GenericId;
import de.ozgcloud.apilib.common.errorhandling.NotFoundException;
import de.ozgcloud.common.test.DataITCase;
import de.ozgcloud.common.test.TestUtils;
import de.ozgcloud.nachrichten.postfach.PostfachAddress;
import de.ozgcloud.nachrichten.postfach.PostfachNachricht;
import de.ozgcloud.nachrichten.postfach.PostfachNachrichtTestFactory;
import de.ozgcloud.nachrichten.postfach.osi.ReplyOption;
import de.ozgcloud.vorgang.VorgangManagerServerApplication;
import de.ozgcloud.vorgang.attached_item.VorgangAttachedItem;
import de.ozgcloud.vorgang.attached_item.VorgangAttachedItem.VorgangAttachedItemBuilder;
import de.ozgcloud.vorgang.attached_item.VorgangAttachedItemTestFactory;
import de.ozgcloud.vorgang.files.OzgCloudFileTestFactory;
import de.ozgcloud.vorgang.files.OzgCloudFileTestUtil;
import de.ozgcloud.vorgang.servicekonto.PostfachAddressTestFactory;
import de.ozgcloud.vorgang.servicekonto.ServiceKonto;
import de.ozgcloud.vorgang.servicekonto.ServiceKontoTestFactory;
import de.ozgcloud.vorgang.vorgang.Vorgang;
import de.ozgcloud.vorgang.vorgang.VorgangHeadTestFactory;
import de.ozgcloud.vorgang.vorgang.VorgangTestFactory;
import io.grpc.stub.CallStreamObserver;
import io.grpc.stub.StreamObserver;

@SpringBootTest(classes = { VorgangManagerServerApplication.class }, properties = {
		"grpc.server.in-process-name=antragraumitcase",
		"ozgcloud.antragraum.url=https://localhost/dummy/antragraum",
		"ozgcloud.antragraum.entityId=dummy-id",
		"ozgcloud.antragraum.metadataUri=http://localhost/dummy/antragraum/uri",
		"ozgcloud.antragraum.decryptionPrivateKey=dummyPrivateKey",
		"ozgcloud.antragraum.decryptionCertificate=dummyCertificate",
		"grpc.client.ozgcloud-command-manager.address=in-process:antragraumitcase",
		"grpc.client.command-manager.address=in-process:antragraumitcase",
		"grpc.client.info-manager.address=in-process:antragraumitcase",
		"grpc.client.vorgang-manager.address=in-process:antragraumitcase",
		"grpc.client.file-manager.address=in-process:antragraumitcase"
})
@WithMockUser
@DataITCase
@DirtiesContext(classMode = DirtiesContext.ClassMode.BEFORE_CLASS)
class AntragraumITCase {

	@Autowired
	private AntragraumGrpcService grpcService;
	@Autowired
	private MongoOperations mongoOperations;
	@Autowired
	private OzgCloudFileTestUtil ozgCloudFileTestUtil;

	@MockBean
	private Saml2Decrypter decrypter;
	@MockBean
	private Saml2Verifier verifier;

	@BeforeEach
	void prepareDatabase() {
		mongoOperations.dropCollection(Vorgang.COLLECTION_NAME);
		mongoOperations.dropCollection(VorgangAttachedItem.COLLECTION_NAME);
		mongoOperations.dropCollection("fs.files");
		mongoOperations.dropCollection("fs.chunks");
	}

	@DisplayName("Find rueckfragen")
	@Nested
	class TestFindRueckfragen {

		@Mock
		private StreamObserver<GrpcFindRueckfragenResponse> responseObserver;

		@Captor
		private ArgumentCaptor<GrpcFindRueckfragenResponse> captor;

		@BeforeEach
		void mock() {
			when(decrypter.decryptPostfachId(any()))
					.thenReturn(PostfachAddressTestFactory.STRING_BASED_IDENTIFIER_POSTFACH_ID_VALUE);
			when(verifier.verify(any())).thenReturn(Collections.emptyList());
		}

		@DisplayName("with nachrichten from different vorgaenge")
		@Nested
		class TestWithDifferentVorgangId {

			private Vorgang vorgangTrustLevel3;
			private Vorgang vorgangTrustLevel2;

			@BeforeEach
			void prepareDatabase() {
				vorgangTrustLevel3 = mongoOperations.save(createVorgang(TrustLevel.LEVEL_3), Vorgang.COLLECTION_NAME);
				mongoOperations.save(createPostfachNachrichtVorgangAttachedItem(vorgangTrustLevel3.getId()),
						VorgangAttachedItem.COLLECTION_NAME);

				vorgangTrustLevel2 = mongoOperations.save(createVorgang(TrustLevel.LEVEL_2), Vorgang.COLLECTION_NAME);
				mongoOperations.save(createPostfachNachrichtVorgangAttachedItem(vorgangTrustLevel2.getId()),
						VorgangAttachedItem.COLLECTION_NAME);
			}

			@Test
			void shouldHaveTrustLevelForVorgang() {
				when(decrypter.decryptTrustLevel(any())).thenReturn(TrustLevel.LEVEL_4.getValue());

				var rueckfragen = findRueckfragen().stream()
						.filter(rueckfrage -> StringUtils.equals(rueckfrage.getVorgangId(), vorgangTrustLevel3.getId()))
						.toList();

				assertThat(rueckfragen).isNotEmpty().hasSize(1);
				assertThat(rueckfragen.get(0).getTrustLevel()).isEqualTo(TrustLevel.LEVEL_3.getValue());
				assertThat(rueckfragen.get(0).getAccessible()).isTrue();
			}

			@Test
			void shouldHaveTrustLevelForOtherVorgang() {
				when(decrypter.decryptTrustLevel(any())).thenReturn(TrustLevel.LEVEL_1.getValue());

				var rueckfragen = findRueckfragen().stream()
						.filter(rueckfrage -> StringUtils.equals(rueckfrage.getVorgangId(), vorgangTrustLevel2.getId()))
						.toList();

				assertThat(rueckfragen).isNotEmpty().hasSize(1);
				assertThat(rueckfragen.get(0).getTrustLevel()).isEqualTo(TrustLevel.LEVEL_2.getValue());
				assertThat(rueckfragen.get(0).getAccessible()).isFalse();
			}
		}

		@DisplayName("with nachrichten from one vorgang")
		@Nested
		class TestWithOneVorgangId {

			@BeforeEach
			void prepareDatabase() {
				var vorgang = mongoOperations.save(createVorgang(TrustLevel.LEVEL_3), Vorgang.COLLECTION_NAME);
				mongoOperations.save(createPostfachNachrichtVorgangAttachedItem(vorgang.getId()),
						VorgangAttachedItem.COLLECTION_NAME);
			}

			@Test
			void shouldReturnOnMatchingTokenTrustLevel() {
				when(decrypter.decryptTrustLevel(any())).thenReturn(TrustLevel.LEVEL_3.getValue());

				var rueckfragen = findRueckfragen();

				assertThat(rueckfragen).isNotEmpty().hasSize(1);
				assertThat(rueckfragen.get(0).getTrustLevel()).isEqualTo(TrustLevel.LEVEL_3.getValue());
				assertThat(rueckfragen.get(0).getAccessible()).isTrue();
			}

			@Test
			void shouldReturnOnLowerTokenTrustLevel() {
				when(decrypter.decryptTrustLevel(any())).thenReturn(TrustLevel.LEVEL_1.getValue());

				var rueckfragen = findRueckfragen();

				assertThat(rueckfragen).isNotEmpty().hasSize(1);
				assertThat(rueckfragen.get(0).getTrustLevel()).isEqualTo(TrustLevel.LEVEL_3.getValue());
				assertThat(rueckfragen.get(0).getAccessible()).isFalse();
			}

			@Test
			void shouldReturnOnHigherTokenTrustLevel() {
				when(decrypter.decryptTrustLevel(any())).thenReturn(TrustLevel.LEVEL_4.getValue());

				var rueckfragen = findRueckfragen();

				assertThat(rueckfragen).isNotEmpty().hasSize(1);
				assertThat(rueckfragen.get(0).getTrustLevel()).isEqualTo(TrustLevel.LEVEL_3.getValue());
				assertThat(rueckfragen.get(0).getAccessible()).isTrue();
			}
		}

		private List<GrpcRueckfrageHead> findRueckfragen() {
			grpcService.findRueckfragen(
					GrpcFindRueckfrageRequestTestFactory.createBuilder()
							.setSamlToken(TestUtils.loadTextFile("SamlResponse.xml")).build(),
					responseObserver);
			verify(responseObserver, timeout(30000)).onNext(captor.capture());
			return captor.getValue().getRueckfrageHeadList();
		}
	}

	@DisplayName("Get rueckfrage")
	@Nested
	class TestGetRueckfrage {

		@Mock
		private StreamObserver<GrpcGetRueckfrageResponse> responseObserver;

		@Captor
		private ArgumentCaptor<GrpcGetRueckfrageResponse> captor;

		private Vorgang savedVorgang;
		private VorgangAttachedItem vorgangAttachedItem;

		@BeforeEach
		void prepareDatabase() {
			savedVorgang = mongoOperations.save(VorgangTestFactory.createBuilder().id(null).build(),
					Vorgang.COLLECTION_NAME);
			vorgangAttachedItem = mongoOperations.save(createPostfachNachrichtVorgangAttachedItem(savedVorgang.getId()),
					VorgangAttachedItem.COLLECTION_NAME);

			when(verifier.verify(any())).thenReturn(Collections.emptyList());
			when(decrypter.decryptTrustLevel(any())).thenReturn(ServiceKontoTestFactory.TRUST_LEVEL);
		}

		@Test
		void shouldReturnOnMatchingPostfachId() {
			when(decrypter.decryptPostfachId(any()))
					.thenReturn(PostfachAddressTestFactory.STRING_BASED_IDENTIFIER_POSTFACH_ID_VALUE);
			var request = GrpcGetRueckfrageRequestTestFactory.createBuilder()
					.setId(vorgangAttachedItem.getId())
					.setSamlToken(TestUtils.loadTextFile("SamlResponse.xml"))
					.build();

			grpcService.getRueckfrage(request, responseObserver);

			verify(responseObserver, timeout(30000)).onNext(captor.capture());
			assertThat(captor.getValue().getRueckfrage()).isNotNull();
		}

		@Test
		void shouldThrowExceptionOnMismatchingPostfachId() {
			when(decrypter.decryptPostfachId(any())).thenReturn("not-match-postfach-id");

			var request = GrpcGetRueckfrageRequestTestFactory.createBuilder()
					.setId(vorgangAttachedItem.getId())
					.setSamlToken(TestUtils.loadTextFile("SamlResponse.xml"))
					.build();
			assertThatThrownBy(() -> grpcService.getRueckfrage(request, responseObserver))
					.isInstanceOf(NotFoundException.class)
					.hasMessageContaining("PostfachNachricht")
					.hasMessageContaining(vorgangAttachedItem.getId());
		}
	}

	@Nested
	class TestGetAttachmentMetadata {

		private static final String SAML_TOKEN = TestUtils.loadTextFile("SamlResponse.xml");

		@Mock
		private StreamObserver<GrpcGetAttachmentMetadataResponse> responseObserver;

		@Captor
		private ArgumentCaptor<GrpcGetAttachmentMetadataResponse> captor;

		private VorgangAttachedItem vorgangAttachedItem;
		private VorgangAttachedItem vorgangAttachedItemDifferentAttachment;
		private String fileId;

		@BeforeEach
		void prepareDatabase() {
			fileId = ozgCloudFileTestUtil.saveOzgCloudFile().getId();
			var savedVorgang = mongoOperations.save(createVorgang(TrustLevel.LEVEL_3), Vorgang.COLLECTION_NAME);
			vorgangAttachedItem = mongoOperations.save(
					createPostfachNachrichtVorgangAttachedItem(savedVorgang.getId(), fileId),
					VorgangAttachedItem.COLLECTION_NAME);
			vorgangAttachedItemDifferentAttachment = mongoOperations.save(
					createPostfachNachrichtVorgangAttachedItem(savedVorgang.getId()),
					VorgangAttachedItem.COLLECTION_NAME);
		}

		@Test
		void shouldSendMetadata() {
			when(verifier.verify(any())).thenReturn(Collections.emptyList());
			when(decrypter.decryptTrustLevel(any())).thenReturn(TrustLevel.LEVEL_3.getValue());
			when(decrypter.decryptPostfachId(any()))
					.thenReturn(PostfachAddressTestFactory.STRING_BASED_IDENTIFIER_POSTFACH_ID_VALUE);
			var request = GrpcGetAttachmentMetadataRequest.newBuilder()
					.setSamlToken(SAML_TOKEN)
					.setNachrichtId(vorgangAttachedItem.getId())
					.setFileId(fileId)
					.build();
			var expectedMetadata = GrpcFileMetadata.newBuilder()
					.setContentType(OzgCloudFileTestFactory.CONTENT_TYPE)
					.setName(OzgCloudFileTestFactory.NAME)
					.setSize(OzgCloudFileTestFactory.CONTENT.length)
					.build();

			grpcService.getAttachmentMetadata(request, responseObserver);

			verify(responseObserver, timeout(30000)).onNext(captor.capture());
			assertThat(captor.getValue().getFileMetadata()).isEqualTo(expectedMetadata);
		}

		@Test
		void shouldFailDueToInvalidToken() {
			when(verifier.verify(any())).thenReturn(List.of(new Saml2Error("invalid_signature", "Signature missing")));
			when(decrypter.decryptTrustLevel(any())).thenReturn(TrustLevel.LEVEL_3.getValue());
			when(decrypter.decryptPostfachId(any()))
					.thenReturn(PostfachAddressTestFactory.STRING_BASED_IDENTIFIER_POSTFACH_ID_VALUE);
			var request = GrpcGetAttachmentMetadataRequest.newBuilder()
					.setSamlToken(SAML_TOKEN)
					.setNachrichtId(vorgangAttachedItem.getId())
					.setFileId(fileId)
					.build();

			assertThatThrownBy(() -> grpcService.getAttachmentMetadata(request, responseObserver))
					.isInstanceOf(SecurityException.class)
					.hasMessageContaining("Signature missing")
					.hasMessageContaining("invalid_signature");

		}

		@Test
		void shouldFailDueIncorrectPostfachId() {
			when(verifier.verify(any())).thenReturn(Collections.emptyList());
			when(decrypter.decryptTrustLevel(any())).thenReturn(TrustLevel.LEVEL_3.getValue());
			when(decrypter.decryptPostfachId(any())).thenReturn("wrong ID");
			var request = GrpcGetAttachmentMetadataRequest.newBuilder()
					.setSamlToken(SAML_TOKEN)
					.setNachrichtId(vorgangAttachedItem.getId())
					.setFileId(fileId)
					.build();

			assertThatThrownBy(() -> grpcService.getAttachmentMetadata(request, responseObserver))
					.isInstanceOf(NotFoundException.class)
					.hasMessageContaining("PostfachNachricht")
					.hasMessageContaining(vorgangAttachedItem.getId());
		}

		@Test
		void shouldFailDueToTrustLevelTooLow() {
			when(verifier.verify(any())).thenReturn(Collections.emptyList());
			when(decrypter.decryptTrustLevel(any())).thenReturn(TrustLevel.LEVEL_2.getValue());
			when(decrypter.decryptPostfachId(any()))
					.thenReturn(PostfachAddressTestFactory.STRING_BASED_IDENTIFIER_POSTFACH_ID_VALUE);
			var request = GrpcGetAttachmentMetadataRequest.newBuilder()
					.setSamlToken(SAML_TOKEN)
					.setNachrichtId(vorgangAttachedItem.getId())
					.setFileId(fileId)
					.build();

			assertThatThrownBy(() -> grpcService.getAttachmentMetadata(request, responseObserver))
					.isInstanceOf(NotFoundException.class)
					.hasMessageContaining("PostfachNachricht")
					.hasMessageContaining(vorgangAttachedItem.getId());
		}

		@Test
		void shouldFailDueToFileIdNotInPostfachNachricht() {
			when(verifier.verify(any())).thenReturn(Collections.emptyList());
			when(decrypter.decryptTrustLevel(any())).thenReturn(TrustLevel.LEVEL_3.getValue());
			when(decrypter.decryptPostfachId(any()))
					.thenReturn(PostfachAddressTestFactory.STRING_BASED_IDENTIFIER_POSTFACH_ID_VALUE);
			var request = GrpcGetAttachmentMetadataRequest.newBuilder()
					.setSamlToken(SAML_TOKEN)
					.setNachrichtId(vorgangAttachedItemDifferentAttachment.getId())
					.setFileId(fileId)
					.build();

			assertThatThrownBy(() -> grpcService.getAttachmentMetadata(request, responseObserver))
					.isInstanceOf(NotFoundException.class)
					.hasMessageContaining("PostfachNachricht")
					.hasMessageContaining(vorgangAttachedItemDifferentAttachment.getId());
		}

	}

	@Nested
	class TestGetAttachmentContent {

		private static final String SAML_TOKEN = TestUtils.loadTextFile("SamlResponse.xml");

		@Mock
		private CallStreamObserver<GrpcGetAttachmentContentResponse> responseObserver;

		@Captor
		private ArgumentCaptor<GrpcGetAttachmentContentResponse> captor;

		private VorgangAttachedItem vorgangAttachedItem;
		private VorgangAttachedItem vorgangAttachedItemDifferentAttachment;
		private String fileId;

		@BeforeEach
		void prepareDatabase() {
			fileId = ozgCloudFileTestUtil.saveOzgCloudFile().getId();
			var savedVorgang = mongoOperations.save(createVorgang(TrustLevel.LEVEL_3), Vorgang.COLLECTION_NAME);
			vorgangAttachedItem = mongoOperations.save(
					createPostfachNachrichtVorgangAttachedItem(savedVorgang.getId(), fileId),
					VorgangAttachedItem.COLLECTION_NAME);
			vorgangAttachedItemDifferentAttachment = mongoOperations.save(
					createPostfachNachrichtVorgangAttachedItem(savedVorgang.getId()),
					VorgangAttachedItem.COLLECTION_NAME);
		}

		@BeforeEach
		void mockObserver() {
			when(responseObserver.isReady()).thenReturn(true);
			doAnswer(invocation -> {
				((Runnable) invocation.getArgument(0)).run();
				return null;
			}).when(responseObserver).setOnReadyHandler(any());
		}

		@Test
		void shouldSendContent() {
			when(verifier.verify(any())).thenReturn(Collections.emptyList());
			when(decrypter.decryptTrustLevel(any())).thenReturn(TrustLevel.LEVEL_3.getValue());
			when(decrypter.decryptPostfachId(any()))
					.thenReturn(PostfachAddressTestFactory.STRING_BASED_IDENTIFIER_POSTFACH_ID_VALUE);
			var request = GrpcGetAttachmentContentRequest.newBuilder()
					.setSamlToken(SAML_TOKEN)
					.setNachrichtId(vorgangAttachedItem.getId())
					.setFileId(fileId)
					.build();
			var expectedContent = ByteString.copyFrom(OzgCloudFileTestFactory.CONTENT);

			grpcService.getAttachmentContent(request, responseObserver);

			verify(responseObserver, timeout(30000)).onNext(captor.capture());
			assertThat(captor.getValue().getFileContent()).isEqualTo(expectedContent);
		}

		@Test
		void shouldFailDueToInvalidToken() {
			when(verifier.verify(any())).thenReturn(List.of(new Saml2Error("invalid_signature", "Signature missing")));
			when(decrypter.decryptTrustLevel(any())).thenReturn(TrustLevel.LEVEL_3.getValue());
			when(decrypter.decryptPostfachId(any()))
					.thenReturn(PostfachAddressTestFactory.STRING_BASED_IDENTIFIER_POSTFACH_ID_VALUE);
			var request = GrpcGetAttachmentContentRequest.newBuilder()
					.setSamlToken(SAML_TOKEN)
					.setNachrichtId(vorgangAttachedItem.getId())
					.setFileId(fileId)
					.build();

			assertThatThrownBy(() -> grpcService.getAttachmentContent(request, responseObserver)).hasRootCauseInstanceOf(SecurityException.class);
		}

		@Test
		void shouldFailDueIncorrectPostfachId() {
			when(verifier.verify(any())).thenReturn(Collections.emptyList());
			when(decrypter.decryptTrustLevel(any())).thenReturn(TrustLevel.LEVEL_3.getValue());
			when(decrypter.decryptPostfachId(any())).thenReturn("wrong ID");
			var request = GrpcGetAttachmentContentRequest.newBuilder()
					.setSamlToken(SAML_TOKEN)
					.setNachrichtId(vorgangAttachedItem.getId())
					.setFileId(fileId)
					.build();

			assertThatThrownBy(() -> grpcService.getAttachmentContent(request, responseObserver)).hasRootCause(new NotFoundException(
					GenericId.from(request.getNachrichtId()), "PostfachNachricht"));
		}

		@Test
		void shouldFailDueToTrustLevelTooLow() {
			when(verifier.verify(any())).thenReturn(Collections.emptyList());
			when(decrypter.decryptTrustLevel(any())).thenReturn(TrustLevel.LEVEL_2.getValue());
			when(decrypter.decryptPostfachId(any()))
					.thenReturn(PostfachAddressTestFactory.STRING_BASED_IDENTIFIER_POSTFACH_ID_VALUE);
			var request = GrpcGetAttachmentContentRequest.newBuilder()
					.setSamlToken(SAML_TOKEN)
					.setNachrichtId(vorgangAttachedItem.getId())
					.setFileId(fileId)
					.build();

			assertThatThrownBy(() -> grpcService.getAttachmentContent(request, responseObserver)).hasRootCause(new NotFoundException(
					GenericId.from(request.getNachrichtId()), "PostfachNachricht"));
		}

		@Test
		void shouldFailDueToFileIdNotInPostfachNachricht() {
			when(verifier.verify(any())).thenReturn(Collections.emptyList());
			when(decrypter.decryptTrustLevel(any())).thenReturn(TrustLevel.LEVEL_3.getValue());
			when(decrypter.decryptPostfachId(any()))
					.thenReturn(PostfachAddressTestFactory.STRING_BASED_IDENTIFIER_POSTFACH_ID_VALUE);
			var request = GrpcGetAttachmentContentRequest.newBuilder()
					.setSamlToken(SAML_TOKEN)
					.setNachrichtId(vorgangAttachedItemDifferentAttachment.getId())
					.setFileId(fileId)
					.build();

			assertThatThrownBy(() -> grpcService.getAttachmentContent(request, responseObserver)).hasRootCause(new NotFoundException(
					GenericId.from(request.getNachrichtId()), "PostfachNachricht"));
		}
	}

	private Vorgang createVorgang(TrustLevel trustLevel) {
		return VorgangTestFactory.createBuilder()
				.id(null)
				.version(0)
				.header(VorgangHeadTestFactory.createBuilder().serviceKonto(createBayernIdServiceKonto(trustLevel))
						.build())
				.build();
	}

	private ServiceKonto createBayernIdServiceKonto(TrustLevel trustLevel) {
		return ServiceKontoTestFactory.createBuilder().type("BAYERN_ID").trustLevel(trustLevel.getValue()).build();
	}

	private VorgangAttachedItem createPostfachNachrichtVorgangAttachedItem(String vorgangId, String fileId) {
		return createPostfachNachrichtVorgangAttachedItemBuilder(vorgangId)
				.item(createBayernIdPostfachNachrichtItem(vorgangId, fileId))
				.build();
	}

	private VorgangAttachedItem createPostfachNachrichtVorgangAttachedItem(String vorgangId) {
		return createPostfachNachrichtVorgangAttachedItemBuilder(vorgangId).build();
	}

	private VorgangAttachedItemBuilder createPostfachNachrichtVorgangAttachedItemBuilder(String vorgangId) {
		return VorgangAttachedItemTestFactory.createBuilder()
				.id(null)
				.version(0)
				.vorgangId(vorgangId)
				.itemName("PostfachMail")
				.client("OzgCloud_NachrichtenManager")
				.item(createBayernIdPostfachNachrichtItem(vorgangId));
	}

	private Map<String, Object> createBayernIdPostfachNachrichtItem(String vorgangId, String fileId) {
		var nachrichtItem = createBayernIdPostfachNachrichtItem(vorgangId);
		nachrichtItem.put("attachments", List.of(fileId));
		return nachrichtItem;
	}

	private Map<String, Object> createBayernIdPostfachNachrichtItem(String vorgangId) {
		var nachrichtItem = PostfachNachrichtTestFactory.asMap();
		nachrichtItem.put(PostfachNachricht.FIELD_DIRECTION, DIRECTION.OUT.name());
		nachrichtItem.put(PostfachNachricht.FIELD_REPLY_OPTION, ReplyOption.POSSIBLE);
		nachrichtItem.put(PostfachNachricht.FIELD_VORGANG_ID, vorgangId);
		nachrichtItem.put(PostfachNachricht.POSTFACH_ADDRESS_FIELD, Map.of(
				PostfachAddress.TYPE_FIELD, 0,
				PostfachAddress.VERSION_FIELD, 1,
				PostfachAddress.IDENTIFIER_FIELD,
				Map.of(PostfachAddress.FIELD_POSTFACH_ID,
						PostfachAddressTestFactory.STRING_BASED_IDENTIFIER_POSTFACH_ID_VALUE),
				PostfachAddress.SERVICEKONTO_TYPE_FIELD, "BAYERN_ID"));
		return nachrichtItem;
	}
}