/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.nachrichten.postfach;

import static org.assertj.core.api.Assertions.*;
import static org.awaitility.Awaitility.*;

import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.UUID;
import java.util.concurrent.TimeUnit;

import org.apache.commons.collections.MapUtils;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.data.mongodb.core.MongoOperations;
import org.springframework.test.annotation.DirtiesContext;

import de.ozgcloud.common.test.DataITCase;
import de.ozgcloud.vorgang.VorgangManagerServerApplication;
import de.ozgcloud.vorgang.attached_item.VorgangAttachedItem;
import de.ozgcloud.vorgang.attached_item.VorgangAttachedItemTestFactory;
import de.ozgcloud.vorgang.servicekonto.PostfachAddressTestFactory;
import de.ozgcloud.vorgang.servicekonto.ServiceKontoTestFactory;
import de.ozgcloud.vorgang.vorgang.Vorgang;
import de.ozgcloud.vorgang.vorgang.VorgangTestFactory;
import lombok.SneakyThrows;

@SpringBootTest(classes = { VorgangManagerServerApplication.class }, properties = {
		"grpc.server.inProcessName=test",
		"grpc.client.vorgang-manager.address=in-process:test",
		"grpc.client.ozgcloud-command-manager.address=in-process:test",
		"grpc.client.pluto.address=in-process:test",
})
@DataITCase
@DirtiesContext
class PersistPostfachNachrichtServiceImplITCase {

	@Autowired
	private PersistPostfachNachrichtServiceImpl service;

	@Autowired
	private PostfachNachrichtMapper postfachNachrichtMapper;
	@Autowired
	private MongoOperations mongoOperations;

	private String vorgangId;

	@BeforeEach
	void init() {
		mongoOperations.dropCollection(VorgangAttachedItem.COLLECTION_NAME);
		mongoOperations.dropCollection(Vorgang.COLLECTION_NAME);
		vorgangId = mongoOperations.save(VorgangTestFactory.createBuilder().id(null).build(), Vorgang.COLLECTION_NAME).getId();
	}

	@Nested
	class TestPersistPostfachNachricht {

		@Test
		void shouldPersistPostfachNachricht() {
			var postfachNachricht = PostfachNachrichtTestFactory.createBuilder().id(null).vorgangId(vorgangId).build();
			service.persistNachricht(Optional.empty(), postfachNachricht);

			assertThat(loadPostfachNachrichts()).hasSize(1).first().usingRecursiveComparison().ignoringFields("id").isEqualTo(postfachNachricht);
		}

		@Test
		void shouldUpdatePersistedNachricht() {
			var persistedNachrichtId = mongoOperations.save(VorgangAttachedItemTestFactory.createBuilder().id(null).version(0)
					.itemName(PersistPostfachNachrichtServiceImpl.ITEM_NAME).item(PostfachNachrichtTestFactory.asMap()).build()).getId();
			var postfachNachricht = PostfachNachrichtTestFactory.createBuilder().id(persistedNachrichtId).vorgangId(vorgangId)
					.subject("updated subject").mailBody("updated body").build();

			service.persistNachricht(Optional.empty(), postfachNachricht);

			assertThat(loadPostfachNachrichts()).hasSize(1).first().usingRecursiveComparison().ignoringFields("id").isEqualTo(postfachNachricht);
		}

		private List<PostfachNachricht> loadPostfachNachrichts() {
			return await().atMost(60, TimeUnit.MINUTES)
					.until(() -> mongoOperations.findAll(VorgangAttachedItem.class), list -> !list.isEmpty()).stream()
					.map(VorgangAttachedItem::getItem).map(postfachNachrichtMapper::fromMapToPostfachMail).toList();
		}
	}

	@DisplayName("Find rueckfragen")
	@Nested
	class TestFindRueckfragen {

		@Test
		void shouldFindRueckfragen() {
			var postfachNachrichtId = mongoOperations.save(createPostfachNachrichtAsAttachedItem(), VorgangAttachedItem.COLLECTION_NAME).getId();

			var rueckfragen = service.findRueckfragen(ServiceKontoTestFactory.TYPE,
					PostfachAddressTestFactory.STRING_BASED_IDENTIFIER_POSTFACH_ID_VALUE);

			assertThat(rueckfragen).hasSize(1).first().extracting(PostfachNachricht::getId).isEqualTo(postfachNachrichtId);
		}
	}

	@DisplayName("Persist answer")
	@Nested
	class TestPersistAnswer {

		private final String referencedNachrichtId = UUID.randomUUID().toString();

		@SneakyThrows
		@Test
		void shouldPersistAnswer() {
			service.persistAnswer(referencedNachrichtId, createPostfachNachricht());

			await().atMost(60, TimeUnit.SECONDS).untilAsserted(() -> {
				var persistedAnswers = findAllVorgangAttachedItems();
				assertThat(persistedAnswers).hasSize(1).first()
						.extracting(VorgangAttachedItem::getItem)
						.extracting(this::getReferencedNachricht)
						.isEqualTo(referencedNachrichtId);
			});
		}

		private PostfachNachricht createPostfachNachricht() {
			return PostfachNachrichtTestFactory.createBuilder()
					.id(null)
					.vorgangId(vorgangId)
					.build();
		}

		private List<VorgangAttachedItem> findAllVorgangAttachedItems() {
			return mongoOperations.findAll(VorgangAttachedItem.class);
		}

		private String getReferencedNachricht(Map<String, Object> postfachNachricht) {
			return MapUtils.getString(postfachNachricht, PostfachNachricht.REFERENCED_NACHRICHT_FIELD);
		}
	}

	@DisplayName("Find answers")
	@Nested
	class TestFindAnswers {

		@BeforeEach
		void init() {
			var postfachNachricht = PostfachNachrichtTestFactory.asMap();
			postfachNachricht.remove(PostfachNachricht.REFERENCED_NACHRICHT_FIELD);
			mongoOperations.save(createPostfachNachrichtAsAttachedItem(postfachNachricht), VorgangAttachedItem.COLLECTION_NAME);
		}

		@Test
		void shouldFindByReferencedNachrichtId() {
			var referencedPostfachNachricht = mongoOperations.save(createPostfachNachrichtAsAttachedItem(), VorgangAttachedItem.COLLECTION_NAME)
					.getId();

			var answers = service.findAnswers(ServiceKontoTestFactory.TYPE, PostfachAddressTestFactory.STRING_BASED_IDENTIFIER_POSTFACH_ID_VALUE,
					GrpcPostfachMailTestFactory.REFERENCED_NACHRICHT_ID);

			assertThat(answers).hasSize(1).first().usingRecursiveComparison()
					.isEqualTo(PostfachNachrichtTestFactory.createBuilder().id(referencedPostfachNachricht).build());
		}
	}

	VorgangAttachedItem createPostfachNachrichtAsAttachedItem() {
		return createPostfachNachrichtAsAttachedItem(PostfachNachrichtTestFactory.asMap());
	}

	VorgangAttachedItem createPostfachNachrichtAsAttachedItem(Map<String, Object> item) {
		return VorgangAttachedItemTestFactory.createBuilder()
				.id(null)
				.version(0)
				.vorgangId(vorgangId)
				.client(AttachedItemRemoteService.CLIENT)
				.itemName(AttachedItemRemoteService.ITEM_NAME)
				.item(item)
				.build();
	}

}