/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.attached_item;

import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.Optional;
import java.util.stream.Stream;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import de.ozgcloud.vorgang.vorgangAttachedItem.GrpcFindVorgangAttachedItemRequest;
import de.ozgcloud.vorgang.common.security.PolicyService;
import de.ozgcloud.vorgang.vorgang.VorgangTestFactory;
import de.ozgcloud.vorgang.vorgangAttachedItem.GrpcFindVorgangAttachedItemResponse;
import de.ozgcloud.vorgang.vorgangAttachedItem.GrpcVorgangAttachedItemRequest;
import de.ozgcloud.vorgang.vorgangAttachedItem.GrpcVorgangAttachedItemResponse;
import io.grpc.stub.StreamObserver;

class GrpcVorgangAttachedItemServiceTest {

	@Spy
	@InjectMocks
	private GrpcVorgangAttachedItemService grpcService;
	@Mock
	private VorgangAttachedItemService service;
	@Mock
	private GrpcVorgangAttachedItemMapper mapper;
	@Mock
	private PolicyService policyService;

	@Nested
	class TestFind {

		@Mock
		private StreamObserver<GrpcFindVorgangAttachedItemResponse> responseObserver;
		private final GrpcFindVorgangAttachedItemRequest request = GrpcFindVorgangAttachedItemRequestTestFactory.create();
		private final VorgangAttachedItem vorgangAttachedItem = VorgangAttachedItemTestFactory.create();

		@BeforeEach
		void mockService() {
			when(service.find(any(), any(), any())).thenReturn(Stream.of(vorgangAttachedItem));
			when(mapper.mapTo(any(VorgangAttachedItem.class))).thenReturn(GrpcVorgangAttachedItemTestFactory.create());
		}

		@Test
		void shouldCallPolicyService() {
			find();

			verify(policyService).checkPermission(anyString());
		}

		@Test
		void shouldCallService() {
			find();

			verify(service).find(VorgangTestFactory.ID, Optional.of(VorgangAttachedItemTestFactory.CLIENT),
					Optional.of(VorgangAttachedItemTestFactory.ITEM_NAME));
		}

		@Test
		void shouldCallMapper() {
			find();

			verify(mapper).mapTo(vorgangAttachedItem);
		}

		@Test
		void shouldCallNext() {
			find();

			verify(responseObserver).onNext(any(GrpcFindVorgangAttachedItemResponse.class));
		}

		@Test
		void shouldCallOnCompleted() {
			find();

			verify(responseObserver).onCompleted();
		}

		private void find() {
			grpcService.find(request, responseObserver);
		}
	}

	@DisplayName("Get by id")
	@Nested
	class TestGetById {

		@Mock
		private StreamObserver<GrpcVorgangAttachedItemResponse> responseObserver;

		private GrpcVorgangAttachedItemRequest request = GrpcVorgangAttachedItemRequestTestFactory.create();

		@BeforeEach
		void init() {
			when(service.getById(any())).thenReturn(VorgangAttachedItemTestFactory.create());
			when(mapper.mapTo(any(VorgangAttachedItem.class))).thenReturn(GrpcVorgangAttachedItemTestFactory.create());
		}

		@Test
		void shouldCallPolicyService() {
			getById();

			verify(policyService).checkPermissionByVorgangAttachedItem(anyString());
		}

		@Test
		void shouldCallGetById() {
			getById();

			verify(service).getById(any());
		}

		@Test
		void shouldCallOnNext() {
			getById();

			verify(responseObserver).onNext(any());
		}

		@Test
		void shouldCallOnCompleted() {
			getById();

			verify(responseObserver).onCompleted();
		}

		private void getById() {
			grpcService.getById(request, responseObserver);
		}
	}
}