/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.attached_item;

import static de.ozgcloud.vorgang.attached_item.VorgangAttachedItemTestFactory.*;
import static org.assertj.core.api.Assertions.*;
import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.Map;
import java.util.Optional;
import java.util.UUID;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;
import org.springframework.context.ApplicationEventPublisher;

import de.ozgcloud.command.Command;
import de.ozgcloud.command.CommandRevokeFailedEvent;
import de.ozgcloud.command.CommandRevokedEvent;
import de.ozgcloud.command.VorgangAttachedItemCreatedEvent;
import de.ozgcloud.nachrichten.postfach.PostfachNachricht;
import de.ozgcloud.nachrichten.postfach.PostfachNachrichtTestFactory;
import de.ozgcloud.vorgang.command.CommandTestFactory;
import de.ozgcloud.vorgang.common.errorhandling.NotFoundException;
import de.ozgcloud.vorgang.vorgang.VorgangTestFactory;

class VorgangAttachedItemServiceTest {

	@InjectMocks
	@Spy
	private VorgangAttachedItemService service;
	@Mock
	private VorgangAttachedItemRepository repository;
	@Mock
	private ApplicationEventPublisher publisher;

	@Nested
	class TestCreateItem {

		private static final VorgangAttachedItem PREPARED_ITEM = VorgangAttachedItemTestFactory.create();
		private static final VorgangAttachedItem PERSISTED_ITEM = VorgangAttachedItemTestFactory.createBuilder()
				.id(CommandTestFactory.CREATED_RESOURCE).build();

		@Captor
		private ArgumentCaptor<VorgangAttachedItemCreatedEvent> eventCaptor;

		@BeforeEach
		void init() {
			when(repository.save(any())).thenReturn(PERSISTED_ITEM);
			doReturn(PREPARED_ITEM).when(service).prepareForCreation(any());
		}

		@Test
		void shouldCallPrepareForCreation() {
			var item = VorgangAttachedItemTestFactory.create();

			service.create(CommandTestFactory.create(), item);

			verify(service).prepareForCreation(item);
		}

		@Test
		void shouldCallRepository() {
			service.create(CommandTestFactory.create(), VorgangAttachedItemTestFactory.create());

			verify(repository).save(PREPARED_ITEM);
		}

		@Test
		void shouldCallBuildExecutedCommand() {
			var command = CommandTestFactory.create();

			service.create(command, VorgangAttachedItemTestFactory.create());

			verify(service).buildExecutedCommand(command, PERSISTED_ITEM);
		}

		@Test
		void shouldPublishEvent() {
			var executedCommand = CommandTestFactory.create();
			doReturn(executedCommand).when(service).buildExecutedCommand(any(), any());

			service.create(CommandTestFactory.create(), VorgangAttachedItemTestFactory.create());

			verify(publisher).publishEvent(eventCaptor.capture());
			assertThat(eventCaptor.getValue()).satisfies(event -> {
				assertThat(event.getCommand()).isSameAs(executedCommand);
				assertThat(event.getCreatedResource()).isEqualTo(CommandTestFactory.CREATED_RESOURCE);
			});
		}
	}

	@Nested
	class TestPrepareForCreation {
		@Test
		void shouldRemoveId() {
			var prepared = service.prepareForCreation(VorgangAttachedItemTestFactory.create());

			assertThat(prepared.getId()).isNull();
		}

		@Test
		void shouldSetVersionToZero() {
			var prepared = service.prepareForCreation(VorgangAttachedItemTestFactory.create());

			assertThat(prepared.getVersion()).isZero();
		}

		@Test
		void shouldRemoveIdFromItem() {
			var prepared = service.prepareForCreation(VorgangAttachedItemTestFactory.create());

			assertThat(prepared.getItem()).doesNotContainKey(VorgangAttachedItem.FIELDNAME_ID);
		}

		@Test
		void shouldSetIsDeletedToFalse() {
			var prepared = service.prepareForCreation(VorgangAttachedItemTestFactory.create());

			assertThat(prepared.isDeleted()).isFalse();
		}
	}

	@Nested
	class TestBuildExecutedCommand {

		private static final VorgangAttachedItem ITEM = VorgangAttachedItemTestFactory.create();

		@Test
		void shouldSetId() {
			var result = buildExecutedCommand();

			assertThat(result.getId()).isEqualTo(CommandTestFactory.ID);
		}

		@Test
		void shouldSetVorgangId() {
			var result = buildExecutedCommand();

			assertThat(result.getVorgangId()).isEqualTo(VorgangTestFactory.ID);
		}

		@Test
		void shouldSetRelationId() {
			var result = buildExecutedCommand();

			assertThat(result.getRelationId()).isEqualTo(CommandTestFactory.RELATION_ID);
		}

		@Test
		void shouldSetRelationVersion() {
			var result = buildExecutedCommand();

			assertThat(result.getRelationVersion()).isEqualTo(CommandTestFactory.RELATION_VERSION);
		}

		@Test
		void shouldSetOrder() {
			var result = buildExecutedCommand();

			assertThat(result.getOrder()).isEqualTo(CommandTestFactory.ORDER.name());
		}

		@Test
		void shouldSetCreatedAt() {
			var result = buildExecutedCommand();

			assertThat(result.getCreatedAt()).isEqualTo(CommandTestFactory.CREATED_AT);
		}

		@Test
		void shouldSetCreatedBy() {
			var result = buildExecutedCommand();

			assertThat(result.getCreatedBy()).isEqualTo(CommandTestFactory.CREATED_BY);
		}

		@Test
		void shouldSetStatus() {
			var result = buildExecutedCommand();

			assertThat(result.getStatus()).isEqualTo(CommandTestFactory.STATUS);
		}

		@Test
		void shouldSetCreatedResource() {
			var result = buildExecutedCommand();

			assertThat(result.getCreatedResource()).isEqualTo(VorgangAttachedItemTestFactory.ID);
		}

		@Test
		void shouldCallBuildBodyObject() {
			buildExecutedCommand();

			verify(service).buildBodyObject(ITEM);
		}

		@Test
		void shouldSetBodyObject() {
			var bodyObject = Map.<String, Object>of("key", "value");
			doReturn(bodyObject).when(service).buildBodyObject(any());

			var result = buildExecutedCommand();

			assertThat(result.getBodyObject()).isEqualTo(bodyObject);
		}

		private Command buildExecutedCommand() {
			return service.buildExecutedCommand(CommandTestFactory.create(), ITEM);
		}
	}

	@Nested
	class TestBuildBodyObject {

		@Test
		void shouldSetItemName() {
			var result = service.buildBodyObject(VorgangAttachedItemTestFactory.create());

			assertThat(result).containsEntry(VorgangAttachedItem.FIELDNAME_ITEM_NAME, ITEM_NAME);
		}

		@Test
		void shouldSetMailDirection() {
			var item = VorgangAttachedItemTestFactory.createBuilder()
					.item(Map.of(PostfachNachricht.FIELD_DIRECTION, PostfachNachricht.Direction.IN.name())).build();

			var result = service.buildBodyObject(item);

			assertThat(result).containsEntry(PostfachNachricht.FIELD_DIRECTION, PostfachNachricht.Direction.IN.name());
		}

		@Test
		void shouldHandleMissingDirection() {
			var result = service.buildBodyObject(VorgangAttachedItemTestFactory.create());

			assertThat(result).doesNotContainKey(PostfachNachricht.FIELD_DIRECTION);
		}
	}

	@DisplayName("Update item")
	@Nested
	class TestUpdateItem {

		@Captor
		private ArgumentCaptor<VorgangAttachedItemUpdatedEvent> eventCaptor;

		@Test
		void shouldCallDoUpate() {
			service.update(CommandTestFactory.ID, VorgangAttachedItemTestFactory.create());

			verify(service).doUpdate(VorgangAttachedItemTestFactory.ID, VERSION, STRING_MAP);
		}

		@Test
		void shouldPublishEvent() {
			service.update(CommandTestFactory.ID, VorgangAttachedItemTestFactory.create());

			verify(publisher).publishEvent(eventCaptor.capture());
			assertThat(eventCaptor.getValue()).extracting(VorgangAttachedItemUpdatedEvent::getSource).isEqualTo(CommandTestFactory.ID);
		}
	}

	@DisplayName("Force update")
	@Nested
	class TestForceUpdateItem {

		@Captor
		private ArgumentCaptor<VorgangAttachedItemUpdatedEvent> eventCaptor;

		@Test
		void shouldCallRepository() {
			service.forceUpdate(CommandTestFactory.ID, VorgangAttachedItemTestFactory.ID, VorgangAttachedItemTestFactory.STRING_MAP);

			verify(repository).forceUpdate(VorgangAttachedItemTestFactory.ID, VorgangAttachedItemTestFactory.STRING_MAP);
		}

		@Test
		void shouldPublishEvent() {
			service.forceUpdate(CommandTestFactory.ID, VorgangAttachedItemTestFactory.ID, VorgangAttachedItemTestFactory.STRING_MAP);

			verify(publisher).publishEvent(eventCaptor.capture());
			assertThat(eventCaptor.getValue()).extracting(VorgangAttachedItemUpdatedEvent::getSource).isEqualTo(CommandTestFactory.ID);
		}
	}

	@Nested
	class TestFindBy {

		@Test
		void shouldCallFindByIdAndDeleted() {
			service.findById(PostfachNachrichtTestFactory.ID);

			verify(repository).findByIdAndDeleted(PostfachNachrichtTestFactory.ID, VorgangAttachedItemService.NOT_DELETED);
		}
	}

	@Nested
	class TestFind {

		@Test
		void shouldCallFindByClientAndItemNameAndVorgangIdAndDeleted() {
			service.find(VorgangTestFactory.ID, Optional.of(CLIENT), Optional.of(ITEM_NAME));

			verify(repository).findByVorgangIdAndClientAndItemNameAndDeleted(VorgangTestFactory.ID, CLIENT, ITEM_NAME,
					VorgangAttachedItemService.NOT_DELETED);
		}

		@Test
		void shouldCallFindByClientAndVorgangIdAndDeleted() {
			service.find(VorgangTestFactory.ID, Optional.of(CLIENT), Optional.empty());

			verify(repository).findByVorgangIdAndClientAndDeleted(VorgangTestFactory.ID, CLIENT, VorgangAttachedItemService.NOT_DELETED);
		}

		@Test
		void shouldCallFindByItemNameAndVorgangIdAndDeleted() {
			service.find(VorgangTestFactory.ID, Optional.empty(), Optional.of(ITEM_NAME));

			verify(repository).findByVorgangIdAndItemNameAndDeleted(VorgangTestFactory.ID, ITEM_NAME, VorgangAttachedItemService.NOT_DELETED);
		}

		@Test
		void shouldCallFindByVorgangIdAndDeleted() {
			service.find(VorgangTestFactory.ID, Optional.empty(), Optional.empty());

			verify(repository).findByVorgangIdAndDeleted(VorgangTestFactory.ID, VorgangAttachedItemService.NOT_DELETED);
		}
	}

	@Nested
	class TestPatch {

		private final Command command = CommandTestFactory.create();

		@Captor
		private ArgumentCaptor<VorgangAttachedItemUpdatedEvent> eventCaptor;

		@Test
		void shouldCallDoPatch() {
			service.patch(command, VorgangAttachedItemTestFactory.STRING_MAP);

			verify(service).doPatch(command, CommandTestFactory.RELATION_VERSION, VorgangAttachedItemTestFactory.STRING_MAP);
		}

		@Test
		void shouldPublishEvent() {
			service.update(CommandTestFactory.ID, VorgangAttachedItemTestFactory.create());

			verify(publisher).publishEvent(eventCaptor.capture());
			assertThat(eventCaptor.getValue()).extracting(VorgangAttachedItemUpdatedEvent::getSource).isEqualTo(CommandTestFactory.ID);
		}
	}

	@Nested
	class TestForcePatchItem {

		@Test
		void shouldCallRepository() {
			Map<String, Object> fieldValueMap = Map.of(VorgangAttachedItemTestFactory.ITEM_FIELD_NAME,
					VorgangAttachedItemTestFactory.ITEM_FIELD_INT_VALUE);

			service.forcePatch(VorgangAttachedItemTestFactory.ID, fieldValueMap);

			verify(repository).forcePatch(VorgangAttachedItemTestFactory.ID, fieldValueMap);
		}
	}

	@Nested
	class TestGetById {

		@Test
		void shouldReturnAttachedItem() {
			var item = VorgangAttachedItemTestFactory.create();
			when(repository.findByIdAndDeleted(any(), anyBoolean())).thenReturn(Optional.of(item));

			var attachedItem = service.getById(VorgangAttachedItemTestFactory.ID);

			assertThat(attachedItem).usingRecursiveComparison().ignoringFields(VorgangAttachedItem.FIELDNAME_ITEM).isEqualTo(item);
		}

		@DisplayName("should throw an exception if no attachedItem was found")
		@Test
		void shouldThrowException() {
			when(repository.findByIdAndDeleted(any(), anyBoolean())).thenReturn(Optional.empty());

			assertThrows(NotFoundException.class, () -> service.getById(VorgangAttachedItemTestFactory.ID));
		}

		@Test
		void shouldCallFindByIdAndDeleted() {
			var item = VorgangAttachedItemTestFactory.create();
			when(repository.findByIdAndDeleted(any(), anyBoolean())).thenReturn(Optional.of(item));

			service.getById(VorgangAttachedItemTestFactory.ID);

			verify(repository).findByIdAndDeleted(VorgangAttachedItemTestFactory.ID, VorgangAttachedItemService.NOT_DELETED);
		}
	}

	@Test
	void shouldAddIdToItem() {
		String id = UUID.randomUUID().toString();

		var enrichedItem = service.addIdToItem(VorgangAttachedItemTestFactory.createBuilder().id(id).build());

		assertThat(enrichedItem.getItem()).containsEntry(VorgangAttachedItem.FIELDNAME_ID, id);
	}

	@Nested
	@DisplayName("Set deleted by ID and version")
	class TestMarkDeletedByIdAndVersion {

		@Test
		@DisplayName("should call updateDeletedByIdAndVersion to mark item as deleted")
		void shouldCallUpdateIsDeletedMethod() {
			var deletedFlag = true;

			service.markAsDeleteByIdAndVersion(VorgangAttachedItemTestFactory.ID, VorgangAttachedItemTestFactory.VERSION);

			verify(repository).updateDeleted(deletedFlag, VorgangAttachedItemTestFactory.ID, VorgangAttachedItemTestFactory.VERSION);
		}

		@Test
		@DisplayName("should call updateDeletedByIdAndVersion to mark item as not deleted")
		void shouldCallUpdateIsDeletedMethod1() {
			var deletedFlag = false;

			service.unmarkAsDeleteByIdAndVersion(VorgangAttachedItemTestFactory.ID, VorgangAttachedItemTestFactory.VERSION);

			verify(repository).updateDeleted(deletedFlag, VorgangAttachedItemTestFactory.ID, VorgangAttachedItemTestFactory.VERSION);
		}
	}

	@Nested
	@DisplayName("Delete item by vorgangId")
	class TestDeleteByVorgangId {

		@Test
		void shouldCallDeleteMethod() {
			service.deleteByVorgangId(VorgangAttachedItemTestFactory.ID);

			verify(repository).deleteByVorgangId(VorgangAttachedItemTestFactory.ID);
		}

	}

	@DisplayName("Delete")
	@Nested
	class TestDelete {

		@Test
		void shouldCallRepository() {
			service.delete(VorgangAttachedItemTestFactory.ID, VorgangAttachedItemTestFactory.VERSION);

			verify(repository).delete(VorgangAttachedItemTestFactory.ID, VorgangAttachedItemTestFactory.VERSION);
		}
	}

	@DisplayName("Do patch")
	@Nested
	class TestDoPatch {

		@Test
		void shouldCallRepository() {
			service.doPatch(CommandTestFactory.create(), CommandTestFactory.RELATION_VERSION, STRING_MAP);

			verify(repository).patch(CommandTestFactory.RELATION_ID, CommandTestFactory.RELATION_VERSION, STRING_MAP);
		}
	}

	@DisplayName("Do Update")
	@Nested
	class TestDoUpdate {

		@Test
		void shouldCallRepository() {
			service.doUpdate(VorgangAttachedItemTestFactory.ID, VorgangAttachedItemTestFactory.VERSION, STRING_MAP);

			verify(repository).update(VorgangAttachedItemTestFactory.ID, VorgangAttachedItemTestFactory.VERSION, STRING_MAP);
		}
	}

	@DisplayName("Revoke create")
	@Nested
	class TestRevokeCreate {

		private final Command command = CommandTestFactory.createBuilder().createdResource(CommandTestFactory.CREATED_RESOURCE).build();

		@Test
		void shouldCallHandleRevoke() {
			service.revokeCreate(command);

			verify(service).handleRevoke(eq(command), any(Runnable.class));
		}

		@Test
		void shouldCallDelete() {
			service.revokeCreate(command);

			verify(service).delete(CommandTestFactory.CREATED_RESOURCE, CommandTestFactory.RELATION_VERSION + 1);
		}
	}

	@DisplayName("Revoke delete")
	@Nested
	class TestRevokeDelete {

		private final Command command = CommandTestFactory.create();

		@Test
		void shouldCallHandleRevoke() {
			service.revokeDelete(command);

			verify(service).handleRevoke(eq(command), any(Runnable.class));
		}

		@Test
		void shouldCallDelete() {
			service.revokeDelete(command);

			verify(service).unmarkAsDeleteByIdAndVersion(CommandTestFactory.RELATION_ID, CommandTestFactory.RELATION_VERSION);
		}
	}

	@DisplayName("Revoke Patch")
	@Nested
	class TestRevokePatch {

		private final Command command = CommandTestFactory.create();

		@Test
		void shouldCallHandleRevoke() {
			service.revokePatch(command, STRING_MAP);

			verify(service).handleRevoke(eq(command), any(Runnable.class));
		}

		@Test
		void shouldCallPatch() {
			service.revokePatch(command, STRING_MAP);

			verify(service).doPatch(command, CommandTestFactory.RELATION_VERSION + 1, STRING_MAP);
		}
	}

	@DisplayName("Revoke Update")
	@Nested
	class TestRevokeUpdate {

		private final Command command = CommandTestFactory.create();

		@Test
		void shouldCallHandleRevoke() {
			service.revokeUpdate(command, STRING_MAP);

			verify(service).handleRevoke(eq(command), any(Runnable.class));
		}

		@Test
		void shouldCallPatch() {
			service.revokeUpdate(command, STRING_MAP);

			verify(service).doUpdate(CommandTestFactory.RELATION_ID, CommandTestFactory.RELATION_VERSION + 1, STRING_MAP);
		}
	}

	@DisplayName("Handle revoke")
	@Nested
	class TestHandleRevoke {

		@Captor
		private ArgumentCaptor<CommandRevokedEvent> commandRevokedEventCaptor;
		@Captor
		private ArgumentCaptor<CommandRevokeFailedEvent> commandRevokedFailedEventCaptor;

		@Mock
		private Runnable runnable;

		private final Command command = CommandTestFactory.create();

		@Test
		void shouldCallServiceFunction() {
			service.handleRevoke(command, runnable);

			verify(runnable).run();
		}

		@Test
		void shouldPublishCommandRevokedEvent() {
			service.handleRevoke(command, runnable);

			verify(publisher).publishEvent(commandRevokedEventCaptor.capture());
			assertThat(commandRevokedEventCaptor.getValue().getSource()).isEqualTo(command);
		}

		@Test
		void shouldPublishCommandRevokeFailedEventOnException() {
			doThrow(new RuntimeException()).when(runnable).run();

			service.handleRevoke(command, runnable);

			verify(publisher).publishEvent(commandRevokedFailedEventCaptor.capture());
			assertThat(commandRevokedFailedEventCaptor.getValue().getSource()).isEqualTo(CommandTestFactory.ID);
		}
	}
}
