/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.clientattribute;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.Set;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import de.ozgcloud.vorgang.callcontext.CallContextUserTestFactory;
import de.ozgcloud.vorgang.callcontext.CurrentUserService;
import de.ozgcloud.vorgang.common.security.PolicyService;
import de.ozgcloud.vorgang.grpc.clientAttribute.GrpcAcknowledgeResponse;
import de.ozgcloud.vorgang.grpc.clientAttribute.GrpcDeleteClientAttributeRequest;
import de.ozgcloud.vorgang.grpc.clientAttribute.GrpcSetClientAttributeRequest;
import de.ozgcloud.vorgang.grpc.clientAttribute.GrpcUpdateClientAttributeRequest;
import de.ozgcloud.vorgang.vorgang.VorgangTestFactory;
import io.grpc.stub.StreamObserver;

class GrpcClientAttributeServiceTest {

	@Spy
	@InjectMocks
	private GrpcClientAttributeService grpcService;
	@Mock
	private ClientAttributeMapper mapper;
	@Mock
	private ClientAttributeService service;
	@Mock
	private PolicyService policyService;
	@Mock
	private CurrentUserService userService;

	@Nested
	class TestSetClientAttribute {

		@Mock
		private StreamObserver<GrpcAcknowledgeResponse> responseObserver;

		private GrpcSetClientAttributeRequest request = GrpcSetClientAttributeRequestTestFactory.create();
		private ClientAttribute attribute = ClientAttributeTestFactory.create();

		@Captor
		private ArgumentCaptor<ClientAttribute> clientAttributeCaptor;

		@BeforeEach
		void mockMapper() {
			when(mapper.mapFromGrpc(any())).thenReturn(attribute);
		}

		@BeforeEach
		void mockUSerService() {
			when(userService.getUser()).thenReturn(CallContextUserTestFactory.create());
		}

		@Test
		void shouldCallPolicyService() {
			grpcService.set(request, responseObserver);

			verify(policyService).checkPermission(Set.of(request.getVorgangId()));
		}

		@Test
		void shouldCallUserService() {
			grpcService.set(request, responseObserver);

			verify(userService).getUser();
		}

		@Test
		void shouldCallMapper() {
			grpcService.set(request, responseObserver);

			verify(mapper).mapFromGrpc(request.getAttribute());
		}

		@Test
		void shouldCallService() {
			grpcService.set(request, responseObserver);

			verify(service).set(eq(VorgangTestFactory.ID), clientAttributeCaptor.capture());
			assertThat(clientAttributeCaptor.getValue()).usingRecursiveComparison().isEqualTo(attribute);
		}

		@Test
		void shouldSendAcknowledge() {
			grpcService.set(request, responseObserver);

			verify(responseObserver).onNext(notNull());
		}

		@Test
		void shouldCompleteCall() {
			grpcService.set(request, responseObserver);

			verify(responseObserver).onCompleted();
		}
	}

	@Nested
	class TestUpdateClientAttribute {
		@Mock
		private StreamObserver<GrpcAcknowledgeResponse> responseObserver;

		private GrpcUpdateClientAttributeRequest request = GrpcUpdateClientAttributeRequestTestFactory.create();
		private ClientAttribute attribute = ClientAttributeTestFactory.create();

		@Captor
		private ArgumentCaptor<ClientAttribute> clientAttributeCaptor;

		@BeforeEach
		void mockMapper() {
			when(mapper.mapFromGrpc(any())).thenReturn(attribute);
		}

		@Test
		void shouldCallMapper() {
			grpcService.update(request, responseObserver);

			verify(mapper).mapFromGrpc(request.getAttribute());
		}

		@Test
		void shouldCallService() {
			grpcService.update(request, responseObserver);

			verify(service).update(VorgangTestFactory.ID, attribute);
		}

		@Test
		void shouldNOTCallUserService() {
			grpcService.update(request, responseObserver);

			verify(userService, never()).getUser();
		}

		@Test
		void shouldSendAcknowledge() {
			grpcService.update(request, responseObserver);

			verify(responseObserver).onNext(notNull());
		}

		@Test
		void shouldCompleteCall() {
			grpcService.update(request, responseObserver);

			verify(responseObserver).onCompleted();
		}
	}

	@Nested
	class TestDeleteClientAttribute {

		@Mock
		private StreamObserver<GrpcAcknowledgeResponse> responseObserver;

		private GrpcDeleteClientAttributeRequest request = GrpcDeleteClientAttributeRequestTestFactory.create();

		@Captor
		private ArgumentCaptor<ClientAttributeIdentificator> identifactorCaptor;

		@Test
		void shouldCallPolicyService() {
			grpcService.delete(request, responseObserver);

			verify(policyService).checkPermission(Set.of(request.getVorgangId()));
		}

		@Test
		void shouldCallService() {
			grpcService.delete(request, responseObserver);

			verify(service).delete(identifactorCaptor.capture());
			assertThat(identifactorCaptor.getValue()).usingRecursiveComparison()
					.isEqualTo(ClientAttributeIdentificatorTestFactory.createBuilder().build());
		}

		@Test
		void shouldNotCallUserService() {
			grpcService.delete(request, responseObserver);

			verify(userService, never()).getUser();
		}

		@Test
		void shouldSendAcknowledge() {
			grpcService.delete(request, responseObserver);

			verify(responseObserver).onNext(notNull());
		}

		@Test
		void shouldCompleteCall() {
			grpcService.delete(request, responseObserver);

			verify(responseObserver).onCompleted();
		}
	}

}
