/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.command;

import static org.assertj.core.api.Assertions.*;
import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.time.ZonedDateTime;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import org.junit.jupiter.api.Assertions;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.EnumSource;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;
import org.springframework.context.ApplicationEventPublisher;

import de.ozgcloud.command.Command;
import de.ozgcloud.command.CommandExecutedEvent;
import de.ozgcloud.command.CommandStatus;
import de.ozgcloud.command.RevokeCommandEvent;
import de.ozgcloud.common.errorhandling.TechnicalException;
import de.ozgcloud.vorgang.callcontext.CallContextTestFactory;
import de.ozgcloud.vorgang.callcontext.CallContextUserTestFactory;
import de.ozgcloud.vorgang.callcontext.CurrentUserService;
import de.ozgcloud.vorgang.callcontext.UserTestFactory;
import de.ozgcloud.vorgang.common.errorhandling.NotFoundException;
import de.ozgcloud.vorgang.common.errorhandling.RevokeFailedException;
import de.ozgcloud.vorgang.vorgang.VorgangService;
import de.ozgcloud.vorgang.vorgang.VorgangTestFactory;

class CommandServiceTest {

	@Spy
	@InjectMocks
	private CommandService service;

	@Mock
	private CommandRepository repository;
	@Mock
	private ApplicationEventPublisher publisher;
	@Mock
	private CurrentUserService currentUserService;
	@Mock
	private VorgangService vorgangService;

	@Nested
	class TestCreateCommand {

		private final CreateCommandRequest request = CreateCommandRequestTestFactory.create();

		@Captor
		private ArgumentCaptor<Command> commandCaptor;

		@Test
		void shouldCallValidateVorgang() {
			service.createCommand(request);

			verify(service).validateVorgang(VorgangTestFactory.ID, CreateCommandRequestTestFactory.ORDER.name());
		}

		@Test
		void shouldCallOperationRepository() {
			service.createCommand(request);

			verify(repository).save(any());
		}

		@Test
		void shouldReturnCommand() {
			var command = (PersistedCommand) service.createCommand(request);

			assertThat(command.getOrder()).isEqualTo(Order.VORGANG_ANNEHMEN.name());
			assertThat(command.getId()).isNotNull();
			assertThat(command.getStatus()).isEqualTo(CommandStatus.PENDING);
			assertThat(command.getCreatedAt()).isNotNull();
			assertThat(command.getCreatedBy()).isEqualTo(UserTestFactory.ID);
			assertThat(command.getRelationId()).isEqualTo(CommandTestFactory.RELATION_ID);
			assertThat(command.getRelationVersion()).isEqualTo(CreateCommandRequestTestFactory.RELATION_VERSION);
		}

		@Test
		void shouldSetClientName() {
			when(currentUserService.findUser()).thenReturn(Optional.of(CallContextUserTestFactory.create()));

			var command = (PersistedCommand) service.createCommand(request);

			assertThat(command.getCreatedByClientName()).isEqualTo(CallContextTestFactory.CLIENT);
		}

		@Test
		void shouldHandleMissingCallContext() {
			var request = CreateCommandRequestTestFactory.createBuilder().callContext(null).build();

			var command = service.createCommand(request);

			assertThat(command).isNotNull();
		}

		@Nested
		class ShouldAddUser {

			@Test
			void fromCallContext() {
				when(currentUserService.findUser()).thenReturn(Optional.empty());

				var command = service.createCommand(request);

				assertThat(command.getCreatedBy()).isEqualTo(UserTestFactory.ID);
			}

			@Test
			void fromUserService() {
				var request = CreateCommandRequestTestFactory.createBuilder().callContext(null).build();
				when(currentUserService.findUser()).thenReturn(Optional.of(CallContextUserTestFactory.create()));

				var command = service.createCommand(request);

				assertThat(command.getCreatedBy()).isEqualTo(UserTestFactory.ID);
			}

			@Test
			void ifBothArePresentFromUserService() {
				when(currentUserService.findUser())
						.thenReturn(Optional.of(CallContextUserTestFactory.createBuilder().userId(Optional.of("other")).build()));

				var command = service.createCommand(request);

				assertThat(command.getCreatedBy()).isEqualTo("other");
			}

		}

		@Test
		void shouldHandleEmptyUser() {
			var request = CreateCommandRequestTestFactory.createBuilder().callContext(CallContextTestFactory.createBuilder().user(null).build())
					.build();

			service.createCommand(request);

			verify(repository).save(commandCaptor.capture());
			assertThat(commandCaptor.getValue().getCreatedBy()).isNull();
		}
	}

	@Nested
	class TestValidateVorgang {

		@Test
		void shouldCallIsVorgangLocked() {
			service.validateVorgang(VorgangTestFactory.ID, Order.ASSIGN_USER.name());

			verify(vorgangService).isVorgangLocked(VorgangTestFactory.ID);
		}

		@Test
		void shouldThrowExceptionIfVorgangLocked() {
			var order = Order.ASSIGN_USER.name();
			when(vorgangService.isVorgangLocked(anyString())).thenReturn(true);

			Assertions.assertThrows(TechnicalException.class, () -> service.validateVorgang(VorgangTestFactory.ID, order));
		}

		@Test
		void shouldNotCallIsVorgangLockedForUnlockVorgangOrder() {
			service.validateVorgang(VorgangTestFactory.ID, Order.UNLOCK_VORGANG.name());

			verify(vorgangService, never()).isVorgangLocked(any());
		}

		@Test
		void shouldNotThrowExceptionIfOrderIsUnlockVorgang() {
			Assertions.assertDoesNotThrow(() -> service.validateVorgang(VorgangTestFactory.ID, Order.UNLOCK_VORGANG.name()));
		}

	}

	@Nested
	class TestGetVorgangCommand {

		final String commandId = CommandTestFactory.ID;

		@BeforeEach
		void mockOperationRepository() {
			when(repository.findById(any())).thenReturn(Optional.of(CommandTestFactory.create()));
		}

		@Test
		void shouldCallOperationRepository() {
			service.findById(commandId);

			verify(repository).findById(any());
		}

		@Test
		void shouldReturnCommand() {
			Optional<Command> command = service.findById(commandId);

			assertThat(command).isPresent();
			assertThat(command.get().getId()).isEqualTo(commandId);
			assertThat(command.get().getCreatedAt()).isEqualTo(CommandTestFactory.CREATED_AT_STR);
			assertThat(command.get().getCreatedBy()).isEqualTo(CommandTestFactory.CREATED_BY);
			assertThat(command.get().getStatus()).isEqualTo(CommandTestFactory.STATUS);
			assertThat(command.get().getOrder()).isEqualTo(CommandTestFactory.ORDER.name());
			assertThat(command.get().getRelationId()).isEqualTo(CommandTestFactory.RELATION_ID);
		}
	}

	@Nested
	class TestFindCommands {

		@Test
		void shouldCallRepository() {
			var createdAfter = ZonedDateTime.now();

			service.findFinishedVorgangLoeschenCommandsOlderThen(createdAfter);

			verify(repository).findCommands(Order.VORGANG_LOESCHEN, CommandStatus.FINISHED, createdAfter);
		}
	}

	@Nested
	class TestSetCommandFinished {

		@Test
		void shouldThrowExceptionIfCommandNotFound() {
			doThrow(NotFoundException.class).when(service).getById(anyString());

			assertThrows(NotFoundException.class, () -> service.setCommandFinished(CommandTestFactory.ID));
		}

		@Nested
		class TestRevokePendingCommand {

			private static final Command REVOKE_PENDING_COMMAND = CommandTestFactory.createBuilder().status(CommandStatus.REVOKE_PENDING)
					.createdResource(CommandTestFactory.CREATED_RESOURCE).build();

			@BeforeEach
			void init() {
				when(repository.setRevokeStatus(anyString(), anyString())).thenReturn(REVOKE_PENDING_COMMAND);
				doReturn(true).when(service).shouldRevoke(any());
				doReturn(REVOKE_PENDING_COMMAND).when(service).getById(anyString());
			}

			@Test
			void shouldCallIsRevokeCommand() {
				setCommandFinished();

				verify(service).shouldRevoke(REVOKE_PENDING_COMMAND);
			}

			@Test
			void shouldCallRepository() {
				setCommandFinished();

				verify(repository).setRevokeStatus(CommandTestFactory.ID, CommandTestFactory.CREATED_RESOURCE);
			}

			@Test
			void shouldCallPublishRevokeCommandEvent() {
				setCommandFinished();

				verify(service).publishRevokeCommandEvent(REVOKE_PENDING_COMMAND);
			}

			@Test
			void shouldNotCallFinishCommandIfRevokePending() {
				setCommandFinished();

				verify(repository, never()).finishCommand(anyString());
				verify(repository, never()).finishCommand(anyString(), anyString());
			}

			void setCommandFinished() {
				service.setCommandFinished(CommandTestFactory.ID, CommandTestFactory.CREATED_RESOURCE);
			}
		}

		@Nested
		class TestFinishCommand {

			private static final Command FINISHED_COMMAND = CommandTestFactory.createBuilder().status(CommandStatus.FINISHED).build();

			@BeforeEach
			void init() {
				doReturn(false).when(service).isParentCommandFailed(any());
				doReturn(FINISHED_COMMAND).when(service).getById(anyString());
			}

			@Test
			void shouldCallSetCommandFinish() {
				service.setCommandFinished(CommandTestFactory.ID);

				verify(service).setCommandFinished(CommandTestFactory.ID, null);
			}

			@Test
			void shouldCallGetCommand() {
				setCommandFinished();

				verify(service).getById(CommandTestFactory.ID);
			}

			@Test
			void shouldCallRepositoryFinishCommand() {
				service.setCommandFinished(CommandTestFactory.ID, null);

				verify(repository).finishCommand(CommandTestFactory.ID);
			}

			@Test
			void shouldCallRepositoryFinishCommandWithCreatedResource() {
				setCommandFinished();

				verify(repository).finishCommand(CommandTestFactory.ID, CommandTestFactory.CREATED_RESOURCE);
			}

			@Test
			void shouldCallGetCompletableParentId() {
				setCommandFinished();

				verify(service).getCompletableParentId(FINISHED_COMMAND);
			}

			@Test
			void shouldCallPublishCommandExecutedEvent() {
				var parentId = "parent-id";
				doReturn(Optional.of(parentId)).when(service).getCompletableParentId(any());

				setCommandFinished();

				verify(service).publishCommandExecutedEvent(parentId);
			}

			@DisplayName("should not call publishCommandExecutedEvent when no completable parent id")
			@Test
			void shouldNotPublishWhenNoParentId() {
				doReturn(Optional.empty()).when(service).getCompletableParentId(any());

				setCommandFinished();

				verify(service, never()).publishCommandExecutedEvent(anyString());
			}

			void setCommandFinished() {
				service.setCommandFinished(CommandTestFactory.ID, CommandTestFactory.CREATED_RESOURCE);
			}
		}
	}

	@Nested
	class TestShouldRevoke {

		@Test
		void shouldReturnTrueWhenRevokePending() {
			var command = CommandTestFactory.createBuilder().status(CommandStatus.REVOKE_PENDING).build();

			var result = service.shouldRevoke(command);

			assertThat(result).isTrue();
		}

		@Test
		void shouldRevokeTrueWhenParentFailed() {
			doReturn(true).when(service).isParentCommandFailed(any());

			var result = service.shouldRevoke(CommandTestFactory.create());

			assertThat(result).isTrue();
		}
	}

	@Nested
	class TestIsParentCommandFailed {

		@Test
		void shouldCallGetParentId() {
			var command = CommandTestFactory.create();

			service.isParentCommandFailed(command);

			verify(service).getParentId(command);
		}

		@Test
		void shouldReturnFalseWhenParentMissing() {
			doReturn(Optional.empty()).when(service).getParentId(any());

			var result = isParentCommandFailed();

			assertThat(result).isFalse();
		}

		@Nested
		class TestWithParent {

			private static final String PARENT_ID = "parent-id";

			@BeforeEach
			void init() {
				doReturn(Optional.of(PARENT_ID)).when(service).getParentId(any());
			}

			@Test
			void shouldCallRepository() {
				isParentCommandFailed();

				verify(repository).isCommandFailed(PARENT_ID);
			}

			@Test
			void shouldReturnTrueWhenParentNotFailed() {
				when(repository.isCommandFailed(any())).thenReturn(true);

				var result = isParentCommandFailed();

				assertThat(result).isTrue();
			}

			@Test
			void shouldReturnFalseWhenParentFailed() {
				var result = isParentCommandFailed();

				assertThat(result).isFalse();
			}
		}

		boolean isParentCommandFailed() {
			return service.isParentCommandFailed(CommandTestFactory.create());
		}
	}

	@Nested
	class TestGetCompletableParentId {

		private static final Command COMMAND = CommandTestFactory.create();

		@Test
		void shouldCallGetParentId() {
			service.getCompletableParentId(COMMAND);

			verify(service).getParentId(COMMAND);
		}

		@Test
		void shouldReturnEmptyWhenNoParentId() {
			doReturn(Optional.empty()).when(service).getParentId(any());

			var result = service.getCompletableParentId(COMMAND);

			assertThat(result).isEmpty();
		}

		@Nested
		class TestCompleteIfSubsCompleted {

			private static final String PARENT_ID = "parent-id";

			@BeforeEach
			void init() {
				doReturn(Optional.of(PARENT_ID)).when(service).getParentId(any());
			}

			@Test
			void shouldCallIsCompleteIfSubsCompleted() {
				service.getCompletableParentId(COMMAND);

				verify(repository).isCompleteIfSubsCompleted(PARENT_ID);
			}

			@Test
			void shouldCallExistsNotFinishedSubCommands() {
				when(repository.isCompleteIfSubsCompleted(anyString())).thenReturn(true);

				service.getCompletableParentId(COMMAND);

				verify(repository).existsNotFinishedSubCommands(PARENT_ID);
			}

			@Test
			void shouldReturnParentId() {
				when(repository.isCompleteIfSubsCompleted(anyString())).thenReturn(true);

				var result = service.getCompletableParentId(COMMAND);

				assertThat(result).contains(PARENT_ID);
			}

			@Test
			void shouldReturnEmptyWhenNotCompletableBySubCommands() {
				var result = service.getCompletableParentId(COMMAND);

				assertThat(result).isEmpty();
			}
		}

	}

	@Nested
	class TestGetParentId {

		@Test
		void shouldReturnResult() {
			var parentId = "parent-id";
			var command = CommandTestFactory.createBuilder().bodyObject(Map.of(PersistedCommand.PROPERTY_PARENT_ID, parentId)).build();

			var result = service.getParentId(command);

			assertThat(result).contains(parentId);
		}

		@Test
		void shouldReturnEmptyWhenNoBodyObject() {
			var command = CommandTestFactory.createBuilder().bodyObject(null).build();

			var result = service.getParentId(command);

			assertThat(result).isEmpty();
		}

		@Test
		void shouldReturnEmptyWhenNoParentId() {
			var command = CommandTestFactory.createBuilder().bodyObject(Map.of()).build();

			var result = service.getParentId(command);

			assertThat(result).isEmpty();
		}
	}

	@Nested
	class TestPublishCommandExecutedEvent {

		@Captor
		private ArgumentCaptor<CommandExecutedEvent> eventCaptor;

		@Test
		void shouldCallFindCommand() {
			service.publishCommandExecutedEvent(CommandTestFactory.ID);

			verify(service).findById(CommandTestFactory.ID);
		}

		@Test
		void shouldCallPublishMethod() {
			var expectedCommand = CommandTestFactory.create();
			doReturn(Optional.of(expectedCommand)).when(service).findById(anyString());

			service.publishCommandExecutedEvent(CommandTestFactory.ID);

			verify(publisher).publishEvent(eventCaptor.capture());
			assertThat(eventCaptor.getValue().getSource()).isEqualTo(CommandTestFactory.ID);
		}

	}

	@Nested
	class TestSetCommandError {

		private static final String ERROR_MESSAGE = "error message";
		private static final String PARENT_ID = "parent-id";

		@Test
		void shouldCallRepositorySetErrorMessage() {
			service.setCommandError(CommandTestFactory.ID, ERROR_MESSAGE);

			verify(repository).setErrorMessage(CommandTestFactory.ID, ERROR_MESSAGE);
		}

		@Test
		void shouldCallRepositoryGetParentId() {
			service.setCommandError(CommandTestFactory.ID, ERROR_MESSAGE);

			verify(repository).getParentId(CommandTestFactory.ID);
		}

		@Nested
		class TestFailParentAndRevokeSubCommands {

			private static final Command COMMAND_BEFORE_SET_ERROR_STATUS = CommandTestFactory.createBuilder().id(PARENT_ID).build();

			@BeforeEach
			void init() {
				when(repository.getParentId(anyString())).thenReturn(Optional.of(PARENT_ID));
				doReturn(COMMAND_BEFORE_SET_ERROR_STATUS).when(service).setErrorStatus(anyString());
			}

			@Test
			void shouldCallSetErrorStatus() {
				service.setCommandError(CommandTestFactory.ID, ERROR_MESSAGE);

				verify(service).setErrorStatus(PARENT_ID);
			}

			@Test
			void shouldCallNotErrorStatus() {
				service.setCommandError(CommandTestFactory.ID, ERROR_MESSAGE);

				verify(service).notErrorStatus(COMMAND_BEFORE_SET_ERROR_STATUS);
			}

			@Test
			void shouldCallHandleCommandError() {
				service.setCommandError(CommandTestFactory.ID, ERROR_MESSAGE);

				verify(service).handleCommandError(CommandTestFactory.ID, PARENT_ID);
			}
		}

		@Nested
		class TestParentIdMissing {

			@BeforeEach
			void init() {
				doReturn(Optional.empty()).when(repository).getParentId(anyString());
			}

			@Test
			void shouldNotCallHandleCommandError() {
				service.setCommandError(CommandTestFactory.ID, ERROR_MESSAGE);

				verify(service, never()).handleCommandError(anyString(), anyString());
			}
		}

		@Nested
		class TestParentAlreadyFailed {

			@Mock
			private Command failedParentCommand;

			@BeforeEach
			void init() {
				when(repository.getParentId(anyString())).thenReturn(Optional.of(PARENT_ID));
				doReturn(failedParentCommand).when(service).setErrorStatus(anyString());
				doReturn(false).when(service).notErrorStatus(any());
			}

			@Test
			void shouldNotCallHandleCommandError() {
				service.setCommandError(CommandTestFactory.ID, ERROR_MESSAGE);

				verify(service, never()).handleCommandError(anyString(), anyString());
			}
		}

	}

	@Nested
	class TestSetErrorStatus {

		@Test
		void shouldCallRepositoryUpdateStatus() {
			when(repository.updateCommandStatusAndReturnPrevious(anyString(), any())).thenReturn(CommandTestFactory.create());

			service.setErrorStatus(CommandTestFactory.ID);

			verify(repository).updateCommandStatusAndReturnPrevious(CommandTestFactory.ID, CommandStatus.ERROR);
		}

		@Test
		void shouldReturnCommand() {
			var command = CommandTestFactory.create();
			when(repository.updateCommandStatusAndReturnPrevious(anyString(), any())).thenReturn(command);

			var result = service.setErrorStatus(CommandTestFactory.ID);

			assertThat(result).isSameAs(command);
		}

		@Test
		void shouldThrowExceptionIfNotFound() {
			assertThrows(NotFoundException.class, () -> service.setErrorStatus(CommandTestFactory.ID));
		}
	}

	@Nested
	class TestNotErrorStatus {

		@DisplayName("should return true when")
		@ParameterizedTest(name = "command status is {0}")
		@EnumSource(value = CommandStatus.class, names = { "ERROR" }, mode = EnumSource.Mode.EXCLUDE)
		void shouldReturnTrue(CommandStatus status) {
			var command = CommandTestFactory.createBuilder().status(status).build();

			var result = service.notErrorStatus(command);

			assertThat(result).isTrue();
		}

		@Test
		void shouldReturnFalse() {
			var command = CommandTestFactory.createBuilder().status(CommandStatus.ERROR).build();

			var result = service.notErrorStatus(command);

			assertThat(result).isFalse();
		}
	}

	@Nested
	class TestHandleCommandError {

		private static final String PARENT_ID = "parent-id";

		@Captor
		private ArgumentCaptor<String> errorMessageCaptor;

		@Test
		void shouldCallPublishCommandFailedEvent() {
			handleCommandError();

			verify(service).publishCommandFailedEvent(eq(PARENT_ID), errorMessageCaptor.capture());
			assertThat(errorMessageCaptor.getValue()).contains(PARENT_ID, CommandTestFactory.ID);
		}

		@Test
		void shouldCallRevokeSubCommands() {
			handleCommandError();

			verify(service).revokeSubCommands(PARENT_ID);
		}

		private void handleCommandError() {
			service.handleCommandError(CommandTestFactory.ID, PARENT_ID);
		}
	}

	@Nested
	class TestSetCommandRevoked {

		@Test
		void shouldCallRepositoryUpdateCommandStatus() {
			service.setCommandRevoked(CommandTestFactory.ID);

			verify(repository).updateCommandStatus(CommandTestFactory.ID, CommandStatus.REVOKED);
		}
	}

	@Nested
	class TestRevokeCommand {

		private static final Command REVOKE_PENDING_COMMAND = CommandTestFactory.createBuilder().status(CommandStatus.REVOKE_PENDING).build();

		@Test
		void shouldCallRevokeCommandSilent() {
			doReturn(Optional.of(REVOKE_PENDING_COMMAND)).when(service).revokeCommandSilent(anyString());

			service.revokeCommand(CommandTestFactory.ID);

			verify(service).revokeCommandSilent(CommandTestFactory.ID);
		}

		@Test
		void shouldCallIsRevokePending() {
			doReturn(Optional.of(REVOKE_PENDING_COMMAND)).when(service).revokeCommandSilent(anyString());

			service.revokeCommand(CommandTestFactory.ID);

			verify(service).isRevokePending(REVOKE_PENDING_COMMAND);
		}

		@Test
		void shouldReturnCommand() {
			doReturn(Optional.of(REVOKE_PENDING_COMMAND)).when(service).revokeCommandSilent(anyString());

			var result = service.revokeCommand(CommandTestFactory.ID);

			assertThat(result).isSameAs(REVOKE_PENDING_COMMAND);
		}

		@Test
		void shouldThrowNotFoundException() {
			doReturn(Optional.empty()).when(service).revokeCommandSilent(anyString());

			assertThrows(NotFoundException.class, () -> service.revokeCommand(CommandTestFactory.ID));
		}

		@Test
		void shouldThrowRevokeFailedException() {
			doReturn(Optional.of(CommandTestFactory.create())).when(service).revokeCommandSilent(anyString());
			doReturn(false).when(service).isRevokePending(any());

			assertThrows(RevokeFailedException.class, () -> service.revokeCommand(CommandTestFactory.ID));
		}
	}

	@Nested
	class TestRevokeCommandSilent {

		private static final Command COMMAND = CommandTestFactory.create();

		@BeforeEach
		void init() {
			when(repository.setRevokeStatusIfNotPending(anyString())).thenReturn(Optional.of(COMMAND));
		}

		@Test
		void shouldCallRepository() {
			service.revokeCommandSilent(CommandTestFactory.ID);

			verify(repository).setRevokeStatusIfNotPending(CommandTestFactory.ID);
		}

		@Test
		void shouldCallIsRevokePending() {
			service.revokeCommandSilent(CommandTestFactory.ID);

			verify(service).isRevokePending(COMMAND);
		}

		@Test
		void shouldCallPublishRevokeCommandEvent() {
			doReturn(true).when(service).isRevokePending(any());

			service.revokeCommandSilent(CommandTestFactory.ID);

			verify(service).publishRevokeCommandEvent(COMMAND);
		}
	}

	@Nested
	class TestIsRevokePending {

		@Test
		void shouldReturnTrue() {
			var command = CommandTestFactory.createBuilder().status(CommandStatus.REVOKE_PENDING).build();

			var result = service.isRevokePending(command);

			assertThat(result).isTrue();
		}

		@DisplayName("should return false when")
		@ParameterizedTest(name = "command status is {0}")
		@EnumSource(value = CommandStatus.class, names = { "REVOKE_PENDING" }, mode = EnumSource.Mode.EXCLUDE)
		void shouldReturnFalse(CommandStatus status) {
			var command = CommandTestFactory.createBuilder().status(status).build();

			var result = service.isRevokePending(command);

			assertThat(result).isFalse();
		}
	}

	@Nested
	class TestPublishRevokeCommandEvent {

		@Captor
		private ArgumentCaptor<RevokeCommandEvent> eventCaptor;

		@Test
		void shouldPublishEvent() {
			var command = CommandTestFactory.create();

			service.publishRevokeCommandEvent(command);

			verify(publisher).publishEvent(eventCaptor.capture());
			assertThat(eventCaptor.getValue().getSource()).isSameAs(command);
		}
	}

	@Nested
	class TestExistsPendingCommands {

		@BeforeEach
		void init() {
			when(repository.existsPendingCommands(any())).thenReturn(true);
		}

		@Test
		void shouldCallRepository() {
			assertThat(service.existsPendingCommands(VorgangTestFactory.ID)).isTrue();

			verify(repository).existsPendingCommands(VorgangTestFactory.ID);
		}
	}

	@Nested
	class TestGetPendingCommands {

		private final List<Command> pendingCommands = Stream.of(CommandTestFactory.create()).collect(Collectors.toList());

		@BeforeEach
		void init() {
			when(repository.getPendingCommands(any())).thenReturn(pendingCommands);
		}

		@Test
		void shouldCallRepository() {
			service.getPendingCommands(VorgangTestFactory.ID);

			verify(repository).getPendingCommands(VorgangTestFactory.ID);
		}

		@Test
		void shouldReturnValue() {
			var result = service.getPendingCommands(VorgangTestFactory.ID);

			assertThat(result).isEqualTo(pendingCommands);
		}
	}

	@Nested
	class TestCommandSetPreviousState {

		final String commandId = CommandTestFactory.ID;

		@Test
		void shouldSetPreviousState() {
			service.setPreviousState(commandId, CommandTestFactory.PREVIOUS_STATE);

			verify(repository).patch(anyString(), any());
		}
	}

	@Nested
	class TestDeleteAllByVorgang {
		@Test
		void shouldCallRepository() {
			service.deleteAllByVorgang(VorgangTestFactory.ID);

			verify(repository).deleteAllByVorgang(VorgangTestFactory.ID);
		}
	}

	@Nested
	class TestFilterUserIdsWithoutCommands {

		@Test
		void shouldCallRepository() {
			service.existsCommandWithUserId(UserTestFactory.ID);

			verify(repository).existsCommandWithUserId(UserTestFactory.ID);
		}

	}

	@Nested
	class TestCreateSubCommands {

		private static final Map<String, Object> SUB_COMMAND_VALUES = Map.of("key", "value");

		private static final Command PARENT_COMMAND = CommandTestFactory.create();

		@BeforeEach
		void init() {
			doReturn(SUB_COMMAND_VALUES).when(service).buildSubCommandValues(any());
			doReturn(PARENT_COMMAND).when(service).getPendingCommand(any());
		}

		@Test
		void shouldCallGetNotFailedCommand() {
			service.createSubCommands(CreateSubCommandsRequestTestFactory.create());

			verify(service).getPendingCommand(CreateSubCommandsRequestTestFactory.PARENT_ID);
		}

		@Test
		void shouldCallBuildSubCommandValues() {
			var request = CreateSubCommandsRequestTestFactory.create();

			service.createSubCommands(request);

			verify(service).buildSubCommandValues(request);
		}

		@Test
		void shouldCallAddSubCommands() {
			service.createSubCommands(CreateSubCommandsRequestTestFactory.create());

			verify(repository).addSubCommands(CommandTestFactory.ID, SUB_COMMAND_VALUES);
		}

		@Test
		void shouldCallBuildCreateCommandRequest() {
			var subCommand = CreateSubCommandRequestTestFactory.create();
			var request = CreateSubCommandsRequestTestFactory.createBuilder().clearSubCommands().subCommand(subCommand).build();

			service.createSubCommands(request).toList();

			verify(service).buildCreateCommandRequest(subCommand, PARENT_COMMAND);
		}

		@Test
		void shouldCreateCommand() {
			var createCommandRequest = CreateCommandRequestTestFactory.create();
			doReturn(createCommandRequest).when(service).buildCreateCommandRequest(any(), any());

			service.createSubCommands(CreateSubCommandsRequestTestFactory.create()).toList();

			verify(service).createCommand(createCommandRequest);
		}

		@Test
		void shouldReturnCreatedCommands() {
			Command createdCommand = CommandTestFactory.create();
			doReturn(createdCommand).when(service).createCommand(any());

			var result = service.createSubCommands(CreateSubCommandsRequestTestFactory.create());

			assertThat(result).containsExactly(createdCommand);
		}
	}

	@Nested
	class TestGetPendingCommand {

		@Test
		void shouldCallFindCommand() {
			doReturn(Optional.of(CommandTestFactory.create())).when(service).findById(anyString());

			service.getPendingCommand(CommandTestFactory.ID);

			verify(service).findById(CommandTestFactory.ID);
		}

		@DisplayName("should throw exception when")
		@ParameterizedTest(name = "command status is {0}")
		@EnumSource(value = CommandStatus.class, names = { "PENDING" }, mode = EnumSource.Mode.EXCLUDE)
		void shouldThrowExceptionIfCommandFailed(CommandStatus status) {
			var command = CommandTestFactory.createBuilder().status(status).build();
			doReturn(command).when(service).getById(anyString());

			assertThrows(TechnicalException.class, () -> service.getPendingCommand(CommandTestFactory.ID));
		}

		@Test
		void shouldReturnCommand() {
			var command = CommandTestFactory.create();
			doReturn(Optional.of(command)).when(service).findById(anyString());

			var result = service.getPendingCommand(CommandTestFactory.ID);

			assertThat(result).isSameAs(command);
		}
	}

	@Nested
	class TestBuildSubCommandValues {

		@Test
		void shouldSetExecutionMode() {
			var request = CreateSubCommandsRequestTestFactory.create();

			var result = service.buildSubCommandValues(request);

			assertThat(result).containsEntry(PersistedCommand.PROPERTY_EXECUTION_MODE, CreateSubCommandsRequestTestFactory.EXECUTION_MODE.name());
		}

		@Test
		void shouldSetCompleteIfSubsCompleted() {
			var request = CreateSubCommandsRequestTestFactory.create();

			var result = service.buildSubCommandValues(request);

			assertThat(result).containsEntry(PersistedCommand.PROPERTY_COMPLETE_IF_SUBS_COMPLETED, true);
		}
	}

	@Nested
	class TestBuildCreateCommandRequest {

		@Test
		void shouldSetVorgangId() {
			var result = buildCreateCommandRequest();

			assertThat(result.getVorgangId()).isEqualTo(VorgangTestFactory.ID);
		}

		@Test
		void shouldSetRelationId() {
			var result = buildCreateCommandRequest();

			assertThat(result.getRelationId()).isEqualTo(CreateCommandRequestTestFactory.RELATION_ID);
		}

		@Test
		void shouldSetRelationVersion() {
			var result = buildCreateCommandRequest();

			assertThat(result.getRelationVersion()).isEqualTo(CreateCommandRequestTestFactory.RELATION_VERSION);
		}

		@Test
		void shouldSetOrder() {
			var result = buildCreateCommandRequest();

			assertThat(result.getOrder()).isEqualTo(CreateCommandRequestTestFactory.ORDER.name());
		}

		@Test
		void shouldCallAddParentIdToBodyObject() {
			var bodyObj = Map.<String, Object>of("key", "value");
			var subCommand = CreateSubCommandRequestTestFactory.createBuilder().bodyObj(bodyObj).build();

			service.buildCreateCommandRequest(subCommand, CommandTestFactory.create());

			verify(service).addParentIdToBodyObject(bodyObj, CommandTestFactory.ID);
		}

		@Test
		void shouldSetBodyObject() {
			var expectedBodyObj = Map.<String, Object>of("key", "value");
			doReturn(expectedBodyObj).when(service).addParentIdToBodyObject(any(), anyString());

			var result = buildCreateCommandRequest();

			assertThat(result.getBodyObject()).isSameAs(expectedBodyObj);
		}

		private CreateCommandRequest buildCreateCommandRequest() {
			return service.buildCreateCommandRequest(CreateSubCommandRequestTestFactory.create(), CommandTestFactory.create());
		}
	}

	@Nested
	class TestAddParentIdToBodyObject {

		private static final Map<String, Object> BODY_OBJ = Map.of("key", "value");

		@Test
		void shouldAddParentId() {
			var result = service.addParentIdToBodyObject(BODY_OBJ, CommandTestFactory.ID);

			assertThat(result).containsEntry(PersistedCommand.PROPERTY_PARENT_ID, CommandTestFactory.ID);
		}

		@Test
		void shouldNotLooseValues() {
			var result = service.addParentIdToBodyObject(BODY_OBJ, CommandTestFactory.ID);

			assertThat(result).containsAllEntriesOf(BODY_OBJ);
		}
	}

}