/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.command;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.Collections;
import java.util.Optional;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;
import org.springframework.context.ApplicationEventPublisher;

import de.ozgcloud.command.Command;
import de.ozgcloud.command.CommandExecutedEvent;
import de.ozgcloud.command.CommandStatus;
import de.ozgcloud.vorgang.common.security.PolicyService;
import de.ozgcloud.vorgang.grpc.command.GrpcAddSubCommandsRequest;
import de.ozgcloud.vorgang.grpc.command.GrpcCommand;
import de.ozgcloud.vorgang.grpc.command.GrpcCommandResponse;
import de.ozgcloud.vorgang.grpc.command.GrpcCommandsResponse;
import de.ozgcloud.vorgang.grpc.command.GrpcCreateCommandRequest;
import de.ozgcloud.vorgang.grpc.command.GrpcEmpty;
import de.ozgcloud.vorgang.grpc.command.GrpcExistsPendingCommandsRequest;
import de.ozgcloud.vorgang.grpc.command.GrpcExistsPendingCommandsResponse;
import de.ozgcloud.vorgang.grpc.command.GrpcFindCommandsRequest;
import de.ozgcloud.vorgang.grpc.command.GrpcGetCommandRequest;
import de.ozgcloud.vorgang.grpc.command.GrpcGetPendingCommandsRequest;
import de.ozgcloud.vorgang.grpc.command.GrpcGetPendingCommandsResponse;
import de.ozgcloud.vorgang.grpc.command.GrpcRevokeCommandRequest;
import de.ozgcloud.vorgang.grpc.command.GrpcSetCommandExecutedRequest;
import de.ozgcloud.vorgang.vorgang.VorgangTestFactory;
import io.grpc.stub.StreamObserver;

class GrpcCommandServiceTest {

	@Spy
	@InjectMocks
	private GrpcCommandService service;
	@Mock
	private GrpcCreateCommandRequestMapper createCommandRequestMapper;
	@Mock
	private CommandService commandService;

	@Mock
	private GrpcCommandResponseMapper commandResponseMapper;
	@Mock
	private GrpcCommandMapper commandMapper;

	@Mock
	private PolicyService policyService;
	@Mock
	private ApplicationEventPublisher eventPublisher;
	@Mock
	private GrpcAddSubCommandsRequestMapper addSubCommandsRequestMapper;

	@Nested
	class TestCreateCommand {

		@Mock
		private StreamObserver<GrpcCommandResponse> responseObserver;
		@Captor
		private ArgumentCaptor<GrpcCommandResponse> responseCaptor;

		private final GrpcCreateCommandRequest request = GrpcCreateCommandRequestTestFactory.create();
		private final CreateCommandRequest createComandRequest = CreateCommandRequestTestFactory.create();
		private final GrpcCommandResponse response = GrpcCommandResponse.newBuilder().build();
		private final Command persistedCommand = CommandTestFactory.create();

		@BeforeEach
		void mockMapper() {
			when(createCommandRequestMapper.fromGrpc(any(GrpcCreateCommandRequest.class))).thenReturn(createComandRequest);
			when(commandService.createCommand(any(CreateCommandRequest.class))).thenReturn(persistedCommand);
			when(commandService.findById(anyString())).thenReturn(Optional.of(CommandTestFactory.create()));
			when(commandResponseMapper.toGrpc(any(CommandResponse.class))).thenReturn(response);
		}

		@Test
		void shouldCallPolicyService() {
			callCreateCommand();

			verify(policyService).checkPermission(VorgangTestFactory.ID);
		}

		@Test
		void shouldCallCommandRequestMapper() {
			callCreateCommand();

			verify(createCommandRequestMapper).fromGrpc(request);
		}

		@Test
		void shouldCallCommandService() {
			callCreateCommand();

			verify(commandService).createCommand(createComandRequest);
		}

		@Test
		void shouldReturnValue() {
			callCreateCommand();

			var values = responseCaptor.getValue();

			assertThat(values).isEqualTo(response);
		}

		private void callCreateCommand() {
			service.createCommand(request, responseObserver);

			verify(responseObserver).onNext(responseCaptor.capture());
		}
	}

	@DisplayName("Get command")
	@Nested
	class TestGetCommand {

		@Mock
		private StreamObserver<GrpcCommand> responseObserver;
		@Captor
		private ArgumentCaptor<GrpcCommand> responseCaptor;
		private final GrpcGetCommandRequest request = GrpcGetCommandRequest.newBuilder().setId(CommandTestFactory.ID).build();

		@BeforeEach
		void init() {
			when(commandService.findById(any())).thenReturn(Optional.of(CommandTestFactory.create()));
			when(commandMapper.toGrpc(any())).thenReturn(GrpcCommand.newBuilder().setId(CommandTestFactory.ID).build());
		}

		@Test
		void shouldCallPolicyService() throws Exception {
			callGetVorgangCommand();

			verify(policyService).checkPermissionByCommand(anyString());
		}

		@Test
		void shouldCallVorgangCommandService() throws Exception {
			callGetVorgangCommand();

			verify(commandService).findById(CommandTestFactory.ID);
		}

		@Test
		void shouldCallMapper() throws Exception {
			callGetVorgangCommand();

			verify(commandMapper).toGrpc(any());
		}

		@Test
		void shouldHaveReturnValue() throws Exception {
			callGetVorgangCommand();

			var values = responseCaptor.getValue();

			assertThat(values.getId()).isEqualTo(CommandTestFactory.ID);
		}

		private void callGetVorgangCommand() throws Exception {
			service.getCommand(request, responseObserver);

			verify(responseObserver).onNext(responseCaptor.capture());
		}
	}

	@Nested
	class TestRevokeCommand {

		private static final GrpcRevokeCommandRequest REVOKE_COMMAND_REQUEST = GrpcRevokeCommandRequest.newBuilder().setId(CommandTestFactory.ID)
				.build();

		@Mock
		private StreamObserver<GrpcCommandResponse> responseObserver;

		@Mock
		private GrpcCommandResponse commandsResponse;

		@Captor
		private ArgumentCaptor<CommandResponse> responseCaptor;

		private Command updatedCommand;

		private void callRevokeCommand() {
			service.revokeCommand(REVOKE_COMMAND_REQUEST, responseObserver);
		}

		@BeforeEach
		void init() {
			updatedCommand = CommandTestFactory.createBuilder().id(CommandTestFactory.ID).build();
		}

		@Test
		void shouldCallPolicyService() {
			callRevokeCommand();

			verify(policyService).checkPermissionByCommand(CommandTestFactory.ID);
		}

		@Test
		void shouldCallServiceRevokeCommand() {
			callRevokeCommand();

			verify(commandService).revokeCommand(CommandTestFactory.ID);
		}

		@Test
		void shouldCallMapper() {
			when(commandService.revokeCommand(any())).thenReturn(updatedCommand);

			callRevokeCommand();

			verify(commandResponseMapper).toGrpc(responseCaptor.capture());
			assertThat(responseCaptor.getValue()).extracting(CommandResponse::getCommand).isSameAs(updatedCommand);
		}

		@Test
		void shouldCallOnNext() {
			when(commandResponseMapper.toGrpc(any())).thenReturn(commandsResponse);

			callRevokeCommand();

			verify(responseObserver).onNext(commandsResponse);
		}

		@Test
		void shouldCloseStream() {
			callRevokeCommand();

			verify(responseObserver).onCompleted();
		}
	}

	@Nested
	class TestExistsPendingCommand {

		@Mock
		private StreamObserver<GrpcExistsPendingCommandsResponse> responseObserver;
		private final GrpcExistsPendingCommandsRequest request = GrpcExistsPendingCommandsRequest.newBuilder().setVorgangId(VorgangTestFactory.ID)
				.build();

		@BeforeEach
		void mockCommandService() {
			when(commandService.existsPendingCommands(any())).thenReturn(true);
		}

		@Test
		void shouldCallPolicyService() throws Exception {
			service.existsPendingCommands(request, responseObserver);

			verify(policyService).checkPermission(anyString());
		}

		@Test
		void shouldCallCommandService() {
			service.existsPendingCommands(request, responseObserver);

			verify(commandService).existsPendingCommands(VorgangTestFactory.ID);
		}
	}

	@Nested
	class TestGetPendingCommands {

		@Mock
		private StreamObserver<GrpcGetPendingCommandsResponse> responseObserver;
		@Captor
		private ArgumentCaptor<GrpcGetPendingCommandsResponse> responseCaptor;
		private final GrpcGetPendingCommandsRequest request = GrpcGetPendingCommandsRequest.newBuilder().setVorgangId(VorgangTestFactory.ID).build();
		private final Command command = CommandTestFactory.create();

		@BeforeEach
		void mockCommandService() {
			when(commandService.getPendingCommands(anyString())).thenReturn(Stream.of(command).collect(Collectors.toList()));
			when(commandMapper.toGrpc(any())).thenReturn(GrpcCommand.newBuilder().setId(CommandTestFactory.ID).build());
		}

		@Test
		void shouldCallPolicyService() throws Exception {
			callGetPendingCommands();

			verify(policyService).checkPermission(anyString());
		}

		@Test
		void shouldCallCommandService() throws Exception {
			callGetPendingCommands();

			verify(commandService).getPendingCommands(VorgangTestFactory.ID);
		}

		@Test
		void shouldCallMapper() throws Exception {
			callGetPendingCommands();

			verify(commandMapper).toGrpc(command);
		}

		@Test
		void shouldReturnValue() throws Exception {
			callGetPendingCommands();

			var response = responseCaptor.getValue();

			assertThat(response.getCommand(0).getId()).isEqualTo(CommandTestFactory.ID);
		}

		private void callGetPendingCommands() throws Exception {
			service.getPendingCommands(request, responseObserver);

			verify(responseObserver).onNext(responseCaptor.capture());
		}
	}

	@Nested
	class TestFindCommands {

		@Mock
		private StreamObserver<GrpcCommandsResponse> responseObserver;
		private final GrpcFindCommandsRequest.Builder requestBuilder = GrpcFindCommandsRequest.newBuilder().setVorgangId(VorgangTestFactory.ID);

		@Test
		void shouldCallPolicyService() throws Exception {
			callFindCommands(requestBuilder.build());

			verify(policyService).checkPermission(anyString());
		}

		@Test
		void shouldCallService() {
			callFindCommands(requestBuilder.build());

			verify(commandService).findCommands(VorgangTestFactory.ID, Collections.emptyList(), Optional.empty());
		}

		@Test
		void shouldCallServiceWithStatus() {
			callFindCommands(requestBuilder.addStatus(CommandStatus.FINISHED.name()).build());

			verify(commandService).findCommands(VorgangTestFactory.ID, Collections.singletonList(CommandStatus.FINISHED), Optional.empty());
		}

		@Test
		void shouldPreferOrderString() {
			callFindCommands(requestBuilder.setOrderString(Order.VORGANG_ABSCHLIESSEN.name()).build());

			verify(commandService).findCommands(VorgangTestFactory.ID, Collections.emptyList(), Optional.of(Order.VORGANG_ABSCHLIESSEN));
		}

		private void callFindCommands(GrpcFindCommandsRequest request) {
			service.findCommands(request, responseObserver);
		}
	}

	@Nested
	class TestSetCommandExecuted {

		@Mock
		private StreamObserver<GrpcEmpty> responseObserver;

		@BeforeEach
		void init() {
			when(commandService.findById(any())).thenReturn(Optional.of(CommandTestFactory.create()));
		}

		@Test
		void shouldLoadCommand() {
			service.setCommandExecuted(buildRequest(), responseObserver);

			verify(commandService).findById(CommandTestFactory.ID);
		}

		@Test
		void shouldPublishEvent() {
			service.setCommandExecuted(buildRequest(), responseObserver);

			verify(eventPublisher).publishEvent(any(CommandExecutedEvent.class));
		}

		@Test
		void shouldHandleMissingCommand() {
			when(commandService.findById(any())).thenReturn(Optional.empty());

			service.setCommandExecuted(buildRequest(), responseObserver);

			verifyNoInteractions(eventPublisher);
		}

		@Test
		void shouldCallOnNext() {
			service.setCommandExecuted(buildRequest(), responseObserver);

			verify(responseObserver).onNext(GrpcEmpty.newBuilder().build());
		}

		@Test
		void shouldCallOnCompleted() {
			service.setCommandExecuted(buildRequest(), responseObserver);

			verify(responseObserver).onCompleted();
		}

		private GrpcSetCommandExecutedRequest buildRequest() {
			return GrpcSetCommandExecutedRequest.newBuilder()
					.setCommandId(CommandTestFactory.ID)
					.build();
		}
	}

	@Nested
	class TestAddSubCommands {

		private final GrpcAddSubCommandsRequest request = GrpcAddSubCommandsRequestTestFactory.create();

		@Mock
		private StreamObserver<GrpcCommandsResponse> responseObserver;
		@Mock
		private GrpcCommandsResponse addSubCommandsResponse;

		@BeforeEach
		void inti() {
			doReturn(addSubCommandsResponse).when(service).buildCommandsResponse(any());
		}

		@Test
		void shouldCallPolicyService() {
			addSubCommands();

			verify(policyService).checkPermissionByCommand(GrpcAddSubCommandsRequestTestFactory.PARENT_ID);
		}

		@Test
		void shouldCallAddSubCommandsRequestMapper() {
			addSubCommands();

			verify(addSubCommandsRequestMapper).from(request);
		}

		@Test
		void shouldCallCreateSubCommands() {
			var createSubCommandsRequest = CreateSubCommandsRequestTestFactory.create();
			when(addSubCommandsRequestMapper.from(any())).thenReturn(createSubCommandsRequest);

			addSubCommands();

			verify(commandService).createSubCommands(createSubCommandsRequest);
		}

		@Test
		void shouldCallBuildCommandsResponse() {
			var createdCommandsStream = Stream.<Command>of(CommandTestFactory.create());
			doReturn(createdCommandsStream).when(commandService).createSubCommands(any());

			addSubCommands();

			verify(service).buildCommandsResponse(createdCommandsStream);
		}

		@Test
		void shouldCallOnNext() {
			addSubCommands();

			verify(responseObserver).onNext(addSubCommandsResponse);
		}

		@Test
		void shouldCallOnCompleted() {
			addSubCommands();

			verify(responseObserver).onCompleted();
		}

		private void addSubCommands() {
			service.addSubCommands(request, responseObserver);
		}
	}

	@Nested
	class TestBuildCommandsResponse {

		@Mock
		private GrpcCommand grpcCommand;

		private Command command;
		private Stream<Command> commands;

		@BeforeEach
		void init() {
			when(commandMapper.toGrpc(any())).thenReturn(grpcCommand);
			command = CommandTestFactory.create();
			commands = Stream.of(command);
		}

		@Test
		void shouldCallGrpcCommandMapper() {
			service.buildCommandsResponse(commands);

			verify(commandMapper).toGrpc(command);
		}

		@Test
		void shouldSetGrpcCommand() {
			var response = service.buildCommandsResponse(commands);

			assertThat(response.getCommandList()).containsExactly(grpcCommand);
		}
	}
}