/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.command;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.Map;
import java.util.Optional;
import java.util.stream.Stream;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;

import de.ozgcloud.nachrichten.file.AttachmentFileTestFactory;
import de.ozgcloud.nachrichten.postfach.GrpcPostfachMailTestFactory;
import de.ozgcloud.nachrichten.postfach.PostfachAddressTestFactory;
import de.ozgcloud.nachrichten.postfach.PostfachNachrichtTestFactory;
import de.ozgcloud.vorgang.attached_item.VorgangAttachedItemService;
import de.ozgcloud.vorgang.attached_item.VorgangAttachedItemTestFactory;
import de.ozgcloud.vorgang.vorgang.VorgangTestFactory;

class PersistPostfachNachrichtByCommandServiceTest {

	@InjectMocks
	private PersistPostfachNachrichtByCommandService service;
	@Mock
	private VorgangAttachedItemService attachedItemService;

	@DisplayName("Persist nachricht")
	@Nested
	class TestPersistNachricht {

		@Test
		void shouldThrowException() {
			var postfachNachricht = PostfachNachrichtTestFactory.create();
			var userId = Optional.<String>empty();

			assertThatThrownBy(() -> service.persistNachricht(userId, postfachNachricht))
					.isInstanceOf(UnsupportedOperationException.class);
		}
	}

	@Nested
	class TestFindById {

		@BeforeEach
		void init() {
			when(attachedItemService.findById(anyString()))
					.thenReturn(Optional.of(VorgangAttachedItemTestFactory.createBuilder().item(PostfachNachrichtTestFactory.asMap()).build()));
		}

		@Test
		void shouldCallItemService() {
			service.findById(PostfachNachrichtTestFactory.ID);

			verify(attachedItemService).findById(PostfachNachrichtTestFactory.ID);
		}

		@Test
		void shouldReturnNachrichtAsMap() {
			var found = service.findById(PostfachNachrichtTestFactory.ID);

			assertThat(found).isPresent().get().usingRecursiveComparison().isEqualTo(PostfachNachrichtTestFactory.asMap());
		}
	}

	@Nested
	class TestFindByVorgang {

		@BeforeEach
		void init() {
			when(attachedItemService.find(anyString(), any(), any()))
					.thenReturn(Stream.of(VorgangAttachedItemTestFactory.createBuilder().item(PostfachNachrichtTestFactory.asMap()).build()));
		}

		@Test
		void shouldCallItemService() {
			service.findByVorgangAsMap(VorgangTestFactory.ID);

			verify(attachedItemService).find(VorgangTestFactory.ID, Optional.of(PersistPostfachNachrichtByCommandService.CLIENT),
					Optional.of(PersistPostfachNachrichtByCommandService.ITEM_NAME));
		}

		@Test
		void shouldReturnNachrichtenAsMap() {
			var found = service.findByVorgangAsMap(VorgangTestFactory.ID);

			assertThat(found).first().usingRecursiveComparison().isEqualTo(PostfachNachrichtTestFactory.asMap());
		}
	}

	@Nested
	class TestPatch {

		@Test
		void shoulCallRepository() {
			Map<String, Object> propertyMap = Map.of(VorgangAttachedItemTestFactory.ITEM_FIELD_NAME,
					VorgangAttachedItemTestFactory.ITEM_FIELD_INT_VALUE);

			service.patch(VorgangAttachedItemTestFactory.ID, propertyMap);

			verify(attachedItemService).forcePatch(VorgangAttachedItemTestFactory.ID, propertyMap);
		}
	}

	@Nested
	class TestGetById {

		@BeforeEach
		void mockService() {
			when(attachedItemService.getById(anyString()))
					.thenReturn(VorgangAttachedItemTestFactory.createBuilder().item(PostfachNachrichtTestFactory.asMap()).build());
		}

		@Test
		void shouldCallService() {
			service.getById(PostfachNachrichtTestFactory.ID);

			verify(attachedItemService).getById(PostfachNachrichtTestFactory.ID);
		}

		@Test
		void shouldReturnItem() {
			Map<String, Object> item = service.getById(PostfachNachrichtTestFactory.ID);

			assertThat(item).isEqualTo(PostfachNachrichtTestFactory.asMap());
		}
	}

	@DisplayName("On Persisting Attachment")
	@Nested
	class TestPersistAttachment {

		@Test
		void shouldThrowException() {
			var attachmentFile = AttachmentFileTestFactory.create();

			assertThatThrownBy(() -> service.persistAttachment(VorgangTestFactory.ID, attachmentFile)).isInstanceOf(
					UnsupportedOperationException.class);
		}
	}

	@DisplayName("Persist answer")
	@Nested
	class TestPersistAnswers {

		@Test
		void shouldThrowException() {
			var postfachNachricht = PostfachNachrichtTestFactory.create();

			assertThatThrownBy(() -> service.persistAnswer(PostfachNachrichtTestFactory.ID, postfachNachricht))
					.isInstanceOf(UnsupportedOperationException.class);
		}
	}

	@DisplayName("Find answers")
	@Nested
	class TestFindAnswers {

		private final String serviceKontoType = "DummyServiceKontoType";

		@Test
		void shouldThrowException() {
			assertThatThrownBy(
					() -> service.findAnswers(serviceKontoType, PostfachAddressTestFactory.IDENTIFIER.toString(),
							GrpcPostfachMailTestFactory.REFERENCED_NACHRICHT_ID))
									.isInstanceOf(UnsupportedOperationException.class);
		}
	}
}
