/*
 * Copyright (C) 2025 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.common.grpc;

import static org.mockito.Mockito.*;

import java.util.function.Supplier;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.ValueSource;
import org.mockito.Mock;

import io.grpc.stub.StreamObserver;

class GrpcResponseBatchStreamerTest {

	private static final int BATCH_SIZE = 2;

	@Mock
	private GrpcResponseBuilder<DummyElement, DummyResponse> responseBuilder;
	@Mock
	private GrpcResponseBuilder<DummyElement, DummyResponse> responseBuilder2;
	@Mock
	private Supplier<GrpcResponseBuilder<DummyElement, DummyResponse>> responseBuilderSupplier;
	@Mock
	private StreamObserver<DummyResponse> responseObserver;
	private GrpcResponseBatchStreamer<DummyElement, DummyResponse> batchStreamer;

	@BeforeEach
	void init() {
		when(responseBuilderSupplier.get()).thenReturn(responseBuilder, responseBuilder2);
		batchStreamer = GrpcResponseBatchStreamer.create(responseBuilderSupplier, responseObserver);
		batchStreamer.setBatchSize(BATCH_SIZE);
	}

	@Test
	void shouldGetNewResponseBuilderDuringInitialization() {
		verify(responseBuilderSupplier).get();
	}

	@Nested
	class TestSend {

		private final DummyElement element = new DummyElement();

		@Test
		void shouldGetElementCount() {
			batchStreamer.send(element);

			verify(responseBuilder).getElementCount();
		}

		@Nested
		class BatchIsNotFull {

			@BeforeEach
			void init() {
				when(responseBuilder.getElementCount()).thenReturn(BATCH_SIZE - 1);
			}

			@Test
			void shouldAddElementToResponse() {
				batchStreamer.send(element);

				verify(responseBuilder).addElement(element);
			}

			@Test
			void shouldNotSendResponse() {
				batchStreamer.send(element);

				verifyNoInteractions(responseObserver);
			}
		}

		@Nested
		class BatchIsFull {

			private final DummyResponse response = new DummyResponse();

			@BeforeEach
			void init() {
				when(responseBuilder.getElementCount()).thenReturn(BATCH_SIZE);
				when(responseBuilder.build()).thenReturn(response);
			}

			@Test
			void shouldAddElementToResponse() {
				batchStreamer.send(element);

				verify(responseBuilder).addElement(element);
			}

			@Test
			void shouldSendResponse() {
				batchStreamer.send(element);

				verify(responseObserver).onNext(response);
			}

			@Test
			void shouldGetNewResponseBuilder() {
				batchStreamer.send(element);

				// first time during initialization, second time in send()
				verify(responseBuilderSupplier, times(2)).get();
			}

			@Test
			void shouldUseNewResponseBuilderNextTime() {
				batchStreamer.send(element);

				batchStreamer.send(element);

				verify(responseBuilder2).addElement(element);
			}
		}
	}

	@Nested
	class TestFinish {

		private final DummyResponse response = new DummyResponse();

		@Test
		void shouldGetElementCount() {
			batchStreamer.finish();

			verify(responseBuilder).getElementCount();
		}

		@Test
		void shouldNotSendResponseIfBatchIsEmpty() {
			when(responseBuilder.getElementCount()).thenReturn(0);

			batchStreamer.finish();

			verify(responseObserver, never()).onNext(any());
		}

		@Test
		void shouldSendResponseIfBatchIsNotEmpty() {
			when(responseBuilder.getElementCount()).thenReturn(1);
			when(responseBuilder.build()).thenReturn(response);

			batchStreamer.finish();

			verify(responseObserver).onNext(response);
		}

		@ParameterizedTest
		@ValueSource(ints = {0, 1})
		void shouldCallOnCompleted(int batchSize) {
			when(responseBuilder.getElementCount()).thenReturn(batchSize);

			batchStreamer.finish();

			verify(responseObserver).onCompleted();
		}
	}

	private static class DummyResponse {
	}

	private static class DummyElement {
	}
}
