/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.common.migration;

import static org.assertj.core.api.Assertions.*;

import java.util.Collections;
import java.util.List;
import java.util.Map;

import org.bson.Document;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.mockito.Spy;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.mongodb.core.MongoTemplate;
import org.springframework.data.mongodb.core.query.Query;

import de.ozgcloud.common.test.DataITCase;

@DataITCase
class M009_SetServiceKontoTrustLevelITCase {

	private static final String DEFAULT_TRUST_LEVEL = "STORK-QAA-Level-1";
	private static final String FIELD_HEADER = "header";
	private static final String FIELD_SERVICE_KONTO = "serviceKonto";
	private static final String FIELD_TRUST_LEVEL = "trustLevel";
	private static final String FIELD_DUMMY_KEY = "HEADER_FIELD";
	private static final String DUMMY_VALUE = "HEADER_VALUE";

	@Spy
	private M009_SetServiceKontoTrustLevel migration;

	@Autowired
	private MigrationDbTestUtils dbTestUtils;
	@Autowired
	private MongoTemplate template;

	@BeforeEach
	void init() {
		dbTestUtils.dropVorgangCollection();
	}

	@Test
	void shouldSetTrustLevel() {
		saveVorgangWithServiceKonto();

		migration.doMigration(template);

		assertThat(getServiceKonto()).containsEntry(FIELD_TRUST_LEVEL, DEFAULT_TRUST_LEVEL);
	}

	@Test
	void shouldUpdateAllVorgangs() {
		saveVorgangWithServiceKonto();
		saveVorgangWithServiceKonto();

		migration.doMigration(template);

		assertThat(loadVorgangs()).hasSize(2);
		loadVorgangs().forEach(vorgang -> assertThat(getServiceKonto(vorgang)).containsEntry(FIELD_TRUST_LEVEL, DEFAULT_TRUST_LEVEL));
	}

	@Test
	void shouldSkipWhenTrustLevelExists() {
		var expectedTrustLevel = "trust-level";
		saveVorgangWithServiceKonto(Map.of(FIELD_TRUST_LEVEL, expectedTrustLevel));

		migration.doMigration(template);

		assertThat(getServiceKonto()).containsEntry(FIELD_TRUST_LEVEL, expectedTrustLevel);
	}

	@Test
	void shouldKeepOtherValues() {
		saveVorgangWithServiceKonto();

		migration.doMigration(template);

		assertThat(getServiceKonto()).containsEntry(FIELD_DUMMY_KEY, DUMMY_VALUE);
	}

	@Test
	void shouldNotMigrateIfNoServiceKonto() {
		dbTestUtils.saveVorgang(new Document(Map.of("header", Map.of(FIELD_DUMMY_KEY, DUMMY_VALUE))));

		migration.doMigration(template);

		assertThat(getServiceKonto()).isNull();
	}

	private void saveVorgangWithServiceKonto() {
		saveVorgangWithServiceKonto(Collections.emptyMap());
	}

	private void saveVorgangWithServiceKonto(Map<String, String> serviceKontoValues) {
		var serviceKonto = new Document(serviceKontoValues);
		serviceKonto.append(FIELD_DUMMY_KEY, DUMMY_VALUE);
		dbTestUtils.saveVorgang(new Document(Map.of(FIELD_HEADER, Map.of(FIELD_SERVICE_KONTO, serviceKonto))));
	}

	private Document getServiceKonto() {
		var vorgangs = loadVorgangs();
		assertThat(vorgangs).hasSize(1);
		return getServiceKonto(vorgangs.get(0));
	}

	private Document getServiceKonto(Document vorgang) {
		return vorgang.get("header", Document.class).get("serviceKonto", Document.class);
	}

	private List<Document> loadVorgangs() {
		return dbTestUtils.findVorgang(new Query());
	}
}