/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.common.migration;

import static org.assertj.core.api.Assertions.*;

import java.util.List;
import java.util.Map;

import org.bson.Document;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.NullAndEmptySource;
import org.junit.jupiter.params.provider.ValueSource;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.mongodb.core.MongoTemplate;
import org.springframework.data.mongodb.core.query.Query;

import de.ozgcloud.common.test.DataITCase;

@DataITCase
class M012_UpdateBayernIdPostfachTypeITCase {

	private static final String FIELD_HEADER = "header";
	private static final String FIELD_SERVICE_KONTO = "serviceKonto";
	private static final String FIELD_TYPE = "type";
	private static final String FIELD_EINGANG = "eingangs";

	private final M012_UpdateBayernIdPostfachType migration = new M012_UpdateBayernIdPostfachType();

	@Autowired
	private MigrationDbTestUtils dbTestUtils;
	@Autowired
	private MongoTemplate template;

	@BeforeEach
	void init() {
		dbTestUtils.dropVorgangCollection();
	}

	@DisplayName("should migrate when")
	@ParameterizedTest(name = "header's serviceKonto type is \"{0}\"")
	@ValueSource(strings = { "BayernId", "BayernID" })
	void shouldMitgrateHeader(String type) {
		saveVorgangWithHeaderServiceKonto(createServiceKonto(type));

		migration.doMigration(template);

		assertThat(loadHeaderServiceKonto()).containsEntry(FIELD_TYPE, M012_UpdateBayernIdPostfachType.BAYERN_ID_TYPE_NEW);
	}

	@DisplayName("should migrate when")
	@ParameterizedTest(name = "eingang's serviceKonto type is \"{0}\"")
	@ValueSource(strings = { "BayernId", "BayernID" })
	void shouldMigrateEingang(String type) {
		saveVorgangWithEingangServiceKonto(createServiceKonto(type));

		migration.doMigration(template);

		assertThat(loadEingangServiceKonto()).containsEntry(FIELD_TYPE, M012_UpdateBayernIdPostfachType.BAYERN_ID_TYPE_NEW);
	}

	@DisplayName("should skip when")
	@ParameterizedTest(name = "header's serviceKonto type is \"{0}\"")
	@NullAndEmptySource
	@ValueSource(strings = { "another", "OSI" })
	void shouldSkipWhenServiceKontoMissingInHeader(String type) {
		saveVorgangWithHeaderServiceKonto(createServiceKonto(type));

		migration.doMigration(template);

		assertThat(loadHeaderServiceKonto()).containsEntry(FIELD_TYPE, type);
	}

	@DisplayName("should skip when")
	@ParameterizedTest(name = "eingang's serviceKonto type is \"{0}\"")
	@NullAndEmptySource
	@ValueSource(strings = { "another", "OSI" })
	void shouldSkipWhenServiceKontoMissingInEingang(String type) {
		saveVorgangWithEingangServiceKonto(createServiceKonto(type));

		migration.doMigration(template);

		assertThat(loadEingangServiceKonto()).containsEntry(FIELD_TYPE, type);
	}

	private Document createServiceKonto(String type) {
		var serviceKonto = new Document();
		serviceKonto.put(FIELD_TYPE, type);
		return serviceKonto;
	}

	private void saveVorgangWithHeaderServiceKonto(Document serviceKonto) {
		dbTestUtils.saveVorgang(new Document(Map.of(FIELD_HEADER, Map.of(FIELD_SERVICE_KONTO, serviceKonto))));
	}

	private void saveVorgangWithEingangServiceKonto(Document serviceKonto) {
		var eingang = new Document(Map.of(FIELD_HEADER, Map.of(FIELD_SERVICE_KONTO, serviceKonto)));
		dbTestUtils.saveVorgang(new Document(Map.of(FIELD_EINGANG, List.of(eingang))));
	}

	private Document loadHeaderServiceKonto() {
		var vorgangs = dbTestUtils.findVorgang(new Query());
		return getServiceKonto(vorgangs.getFirst());
	}

	private Document loadEingangServiceKonto() {
		var vorgangs = dbTestUtils.findVorgang(new Query());
		return getServiceKonto(vorgangs.getFirst().getList(FIELD_EINGANG, Document.class).getFirst());
	}

	private Document getServiceKonto(Document vorgang) {
		return vorgang.get(FIELD_HEADER, Document.class).get(FIELD_SERVICE_KONTO, Document.class);
	}
}