/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.common.operator;

import static org.assertj.core.api.Assertions.*;
import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.Mockito.*;

import java.time.LocalDate;
import java.time.LocalDateTime;
import java.time.ZonedDateTime;

import org.apache.commons.lang3.StringUtils;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.Mockito;

import de.ozgcloud.vorgang.common.GrpcQueryOperator;
import de.ozgcloud.vorgang.common.operator.EqualOperator;
import de.ozgcloud.vorgang.common.operator.GreaterThenOperator;
import de.ozgcloud.vorgang.common.operator.GreaterThenOrEqualOperator;
import de.ozgcloud.vorgang.common.operator.LessThenOperator;
import de.ozgcloud.vorgang.common.operator.LessThenOrEqualOperator;
import de.ozgcloud.vorgang.common.operator.Operator;
import de.ozgcloud.vorgang.common.operator.OperatorBuilder;
import de.ozgcloud.vorgang.common.operator.UnequalOperator;
import de.ozgcloud.vorgang.statistic.VorgangStatisticBadRequestException;
import de.ozgcloud.vorgang.vorgang.Vorgang;

class OperatorBuilderTest {

	private static final String QUERY_PATH = "Vorgang.fieldPath";
	private static final String FIELD_PATH = "fieldPath";
	private static final String OPERAND = "operand";

	@Test
	void shouldSetFieldSet() {
	var builder = OperatorBuilder.from(GrpcQueryOperator.UNEQUAL).fieldPath(QUERY_PATH);

		assertThat(builder).extracting("fieldPath").isEqualTo(FIELD_PATH);
	}

	@Test
	void shouldSetOperator() {
		var builder = OperatorBuilder.from(GrpcQueryOperator.UNEQUAL);

		assertThat(builder).extracting("grpcOperator").isEqualTo(GrpcQueryOperator.UNEQUAL);
	}

	@Test
	void shouldSetOperand() {
		var builder = OperatorBuilder.from(GrpcQueryOperator.UNEQUAL).operand(OPERAND);

		assertThat(builder).extracting("operand").isEqualTo(OPERAND);
	}

	@Nested
	@DisplayName("build operator")
	class TestBuild {

		@Test
		void shouldBuildEqualOperator() {
			var operator = buildOperator(GrpcQueryOperator.EQUAL);

			assertThat(operator).isInstanceOf(EqualOperator.class);
			assertThat(operator).extracting("fieldPath").isEqualTo(FIELD_PATH);
			assertThat(operator).extracting("operand").isEqualTo(OPERAND);
		}

		@Test
		void shouldBuildIsEmptyOperator() {
			var operator = buildOperator(GrpcQueryOperator.IS_EMPTY);

			assertThat(operator).isInstanceOf(EqualOperator.class);
			assertThat(operator).extracting("fieldPath").isEqualTo(FIELD_PATH);
			assertThat(operator).extracting("operand").isEqualTo(StringUtils.EMPTY);
		}

		@Test
		void shouldBuildIsNullOperator() {
			var operator = buildOperator(GrpcQueryOperator.IS_NULL);

			assertThat(operator).isInstanceOf(EqualOperator.class);
			assertThat(operator).extracting("fieldPath").isEqualTo(FIELD_PATH);
			assertThat(operator).extracting("operand").isNull();
		}

		@Test
		void shouldBuildUnequalOperator() {
			var operator = buildOperator(GrpcQueryOperator.UNEQUAL);

			assertThat(operator).isInstanceOf(UnequalOperator.class);
			assertThat(operator).extracting("fieldPath").isEqualTo(FIELD_PATH);
			assertThat(operator).extracting("operand").isEqualTo(OPERAND);
		}

		@Test
		void shouldBuildGreaterThenOperator() {
			var operator = buildOperator(GrpcQueryOperator.GREATER_THEN);

			assertThat(operator).isInstanceOf(GreaterThenOperator.class);
			assertThat(operator).extracting("fieldPath").isEqualTo(FIELD_PATH);
			assertThat(operator).extracting("operand").isEqualTo(OPERAND);
		}

		@Test
		void shouldBuildGreaterThenOrEqualOperator() {
			var operator = buildOperator(GrpcQueryOperator.GREATER_THEN_OR_EQUAL_TO);

			assertThat(operator).isInstanceOf(GreaterThenOrEqualOperator.class);
			assertThat(operator).extracting("fieldPath").isEqualTo(FIELD_PATH);
			assertThat(operator).extracting("operand").isEqualTo(OPERAND);
		}

		@Test
		void shouldBuildLessThenOperator() {
			var operator = buildOperator(GrpcQueryOperator.LESS_THEN);

			assertThat(operator).isInstanceOf(LessThenOperator.class);
			assertThat(operator).extracting("fieldPath").isEqualTo(FIELD_PATH);
			assertThat(operator).extracting("operand").isEqualTo(OPERAND);
		}

		@Test
		void shouldBuildLessThenOrEqualOperator() {
			var operator = buildOperator(GrpcQueryOperator.LESS_THEN_OR_EQUAL_TO);

			assertThat(operator).isInstanceOf(LessThenOrEqualOperator.class);
			assertThat(operator).extracting("fieldPath").isEqualTo(FIELD_PATH);
			assertThat(operator).extracting("operand").isEqualTo(OPERAND);
		}

		@Test
		void shouldThrowExceptionWhenUnrecognized() {
			assertThrows(VorgangStatisticBadRequestException.class, () -> buildOperator(GrpcQueryOperator.UNRECOGNIZED));
		}

		@Test
		void shouldBuildExistsOperator() {
			var operator = buildOperator(GrpcQueryOperator.EXISTS);

			assertThat(operator).isInstanceOf(ExistsOperator.class);
			assertThat(operator).extracting("fieldPath").isEqualTo(FIELD_PATH);
		}

		@Test
		void shouldBuildNotExistsOperator() {
			var operator = buildOperator(GrpcQueryOperator.NOT_EXIST);

			assertThat(operator).isInstanceOf(ExistsOperator.class);
			assertThat(operator).extracting("fieldPath").isEqualTo(FIELD_PATH);
			assertThat(operator).extracting("negate").isEqualTo(true);
		}

		private Operator buildOperator(GrpcQueryOperator grpcOperator) {
			return OperatorBuilder.from(grpcOperator).fieldPath(QUERY_PATH).operand(OPERAND).build();
		}
	}

	@Nested
	class TestParseOperand {

		private OperatorBuilder operatorBuilder = spy(OperatorBuilder.from(GrpcQueryOperator.EQUAL).fieldPath(QUERY_PATH).operand(OPERAND));

		@Test
		void shouldReturnUnmodified() {
			doReturn(null).when(operatorBuilder).getFieldClass();

			var parsedOperand = operatorBuilder.getParsedOperand();

			assertThat(parsedOperand).isEqualTo(OPERAND);
		}

		@Test
		void shouldCallGetFieldClass() {
			doReturn(String.class).when(operatorBuilder).getFieldClass();

			operatorBuilder.getParsedOperand();

			verify(operatorBuilder).getFieldClass();
		}

		@Test
		void shouldCallParseValue() {
			doReturn(ZonedDateTime.class).when(operatorBuilder).getFieldClass();
			doReturn(OPERAND).when(operatorBuilder).parseValue(ZonedDateTime.class);

			operatorBuilder.getParsedOperand();

			verify(operatorBuilder).parseValue(ZonedDateTime.class);
		}
	}

	@Nested
	class TestGetFieldClass {

		private OperatorBuilder operatorBuilder = OperatorBuilder.from(GrpcQueryOperator.EQUAL);

		@Test
		void shouldReturnNull() {
			var fieldClass = operatorBuilder.getFieldClass();

			assertThat(fieldClass).isNull();
		}

		@Test
		void shouldReturnNullIfFieldPathIsBlank() {
			operatorBuilder.fieldPath("Vorgang.");

			var fieldClass = operatorBuilder.getFieldClass();

			assertThat(fieldClass).isNull();
		}

		@Test
		void shouldReturnRootClass() {
			var fieldPath = "Vorgang.createdAt";
			operatorBuilder.fieldPath(fieldPath);

			var fieldClass = operatorBuilder.getFieldClass();

			assertThat(fieldClass).isEqualTo(ZonedDateTime.class);
		}

		@Test
		void shouldReturnNestedClass() {
			operatorBuilder.fieldPath("Vorgang.eingangs.header.serviceKonto.type");

			var fieldClass = operatorBuilder.getFieldClass();

			assertThat(fieldClass).isEqualTo(String.class);
		}

		@Test
		void shouldReturnNullWhenFieldNotExists() {
			operatorBuilder.fieldPath("Vorgang.notExists");

			var fieldClass = operatorBuilder.getFieldClass();

			assertThat(fieldClass).isNull();
		}
	}

	@Nested
	class TestParseValue {

		@Test
		void shouldReturnZonedDateTime() {
			var operatorBuilder = OperatorBuilder.from(GrpcQueryOperator.EQUAL).fieldPath(QUERY_PATH).operand("2024-01-01T00:00:00Z");

			var parsedValue = operatorBuilder.parseValue(ZonedDateTime.class);

			assertThat(parsedValue).isInstanceOf(ZonedDateTime.class);
		}

		@Test
		void shouldReturnLocalDate() {
			var operatorBuilder = OperatorBuilder.from(GrpcQueryOperator.EQUAL).fieldPath(QUERY_PATH).operand("2024-01-01");

			var parsedValue = operatorBuilder.parseValue(LocalDate.class);

			assertThat(parsedValue).isInstanceOf(LocalDate.class);
		}

		@Test
		void shouldReturnLocalDateTime() {
			var operatorBuilder = OperatorBuilder.from(GrpcQueryOperator.EQUAL).fieldPath(QUERY_PATH).operand("2024-01-01T00:00:00");

			var parsedValue = operatorBuilder.parseValue(LocalDateTime.class);

			assertThat(parsedValue).isInstanceOf(LocalDateTime.class);
		}

		@Test
		void shouldReturnUnmodified() {
			var operatorBuilder = OperatorBuilder.from(GrpcQueryOperator.EQUAL).fieldPath(QUERY_PATH).operand(1);

			var parsedValue = operatorBuilder.parseValue(String.class);

			assertThat(parsedValue).isInstanceOf(Integer.class);
		}
	}
}