/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.common.search;

import static de.ozgcloud.vorgang.common.search.IndexedVorgangTestFactory.*;
import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.assertj.core.data.MapEntry;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.ValueSource;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;
import org.springframework.data.domain.Sort;
import org.springframework.data.elasticsearch.client.elc.NativeQuery;
import org.springframework.data.elasticsearch.core.ElasticsearchOperations;
import org.springframework.data.elasticsearch.core.IndexOperations;
import org.springframework.data.elasticsearch.core.SearchHit;
import org.springframework.data.elasticsearch.core.SearchHits;
import org.springframework.data.elasticsearch.core.document.Document;
import org.springframework.data.elasticsearch.core.mapping.IndexCoordinates;
import org.springframework.data.elasticsearch.core.query.ScriptType;
import org.springframework.data.elasticsearch.core.query.UpdateQuery;

import de.ozgcloud.vorgang.callcontext.UserTestFactory;
import de.ozgcloud.vorgang.vorgang.FilterCriteriaTestFactory;
import de.ozgcloud.vorgang.vorgang.FindVorgangRequestTestFactory;
import de.ozgcloud.vorgang.vorgang.Vorgang.Status;
import de.ozgcloud.vorgang.vorgang.VorgangHeader;
import de.ozgcloud.vorgang.vorgang.VorgangHeaderTestFactory;
import de.ozgcloud.vorgang.vorgang.ZustaendigeStelleTestFactory;

class SearchVorgangCustomRepositoryImplTest {

	@Spy
	@InjectMocks
	private SearchVorgangCustomRepositoryImpl searchRepostitory;

	@Mock
	private IndexedVorgangMapper mapper;

	@Mock
	private SearchProperties properties;

	@Mock
	private ElasticsearchOperations operations;

	@Nested
	class TestSearch {

		@Mock
		private SearchHits<IndexedVorgang> searchHits;
		@Mock
		private SearchHit<IndexedVorgang> searchHit;

		@BeforeEach
		void init() {
			when(searchHit.getContent()).thenReturn(IndexedVorgangTestFactory.create());
			when(searchHits.get()).thenReturn(List.of(searchHit).stream());
			when(operations.search(any(NativeQuery.class), eq(IndexedVorgang.class))).thenReturn(searchHits);
		}

		@Test
		void shouldCallSearch() {
			searchRepostitory.searchBy(FindVorgangRequestTestFactory.create());

			verify(mapper, atLeastOnce()).toVorgangHeader(any());
			verify(operations).search(any(NativeQuery.class), eq(IndexedVorgang.class));
		}

		@Test
		void shouldMapResult() {
			when(mapper.toVorgangHeader(any())).thenReturn(VorgangHeaderTestFactory.create());

			var res = searchRepostitory.searchBy(FindVorgangRequestTestFactory.create());

			assertThat(res).isNotEmpty().hasAtLeastOneElementOfType(VorgangHeader.class);
		}
	}

	@Nested
	class TestQuery {

		@Test
		void shouldHavePaging() {
			var pageable = searchRepostitory.buildPageable(FindVorgangRequestTestFactory.create());

			assertThat(pageable).isNotNull();
			assertThat(pageable.getOffset()).isZero();
			assertThat(pageable.getPageSize()).isEqualTo(FindVorgangRequestTestFactory.LIMIT);
			assertThat(pageable.getSort()).isEqualTo(Sort.unsorted());
		}

		@Test
		void shouldHaveFirstPageNumber() {
			var pageable = searchRepostitory.buildPageable(FindVorgangRequestTestFactory.create());

			assertThat(pageable.getPageNumber()).isZero();
		}

		@Test
		void shouldHaveSecondPageNumber() {
			var pageable = searchRepostitory.buildPageable(
					FindVorgangRequestTestFactory.createBuilder()
							.offset(FindVorgangRequestTestFactory.LIMIT)
							.build());

			assertThat(pageable.getPageNumber()).isEqualTo(1);
		}

		@Test
		void shouldHaveFieldsWithWights() {
			var query = searchRepostitory.buildQueryBuilder(FindVorgangRequestTestFactory.create()).build();

			assertThat(query.getQuery()).isNotNull();
			assertThat(query.getQuery()).asString().contains("aktenzeichen^2.0");
			assertThat(query.getQuery()).asString().contains("antragstellerName^1.0");
			assertThat(query.getQuery()).asString().contains("antragstellerVorname^1.0");
			assertThat(query.getQuery()).asString().contains("vorgangName^0.5");
			assertThat(query.getQuery()).asString().contains("vorgangNummer^2.0");
		}

		@Test
		void shouldHaveSimpleQuery() {
			var request = FindVorgangRequestTestFactory.createBuilder().searchBy("vors").build();
			var query = searchRepostitory.buildQueryBuilder(request).build();

			assertThat(query.getQuery()).asString().contains("\"query\":\"(*vors*)\"");
		}

		@Test
		void shouldHaveQuery() {
			var request = FindVorgangRequestTestFactory.createBuilder().searchBy("vors test").build();
			var query = searchRepostitory.buildQueryBuilder(request).build();

			assertThat(query.getQuery()).asString().contains("\"query\":\"(*vors*) AND (*test*)\"");
		}

		@Nested
		class TestBuildSearchByQueryBuilder {

			@Test
			void shouldCallAdjustSearchTerm() {
				var request = FindVorgangRequestTestFactory.createBuilder().searchBy("vors test").build();

				searchRepostitory.buildSearchByQueryBuilder(request);

				verify(searchRepostitory, times(2)).adjustSearchTerm(any());
			}
		}

		@Nested
		class TestEscapingCharacters {

			private static final String SEARCH_TERM_PATTERN = "Ab%sT";
			private static final String ESCAPED_SEARCH_TERM_PATTERN = "(*Ab\\%sT*)";

			@ParameterizedTest(name = "should escape {0}")
			@ValueSource(strings = { "+", "-", "=", "&&", "||", "!", "(", ")", "{", "}", "[", "]", "^", "\"", "~", "*", "?", ":", "\\", "/" })
			void shouldFindWithSpecialCharacters(String character) {
				var result = searchRepostitory.adjustSearchTerm(SEARCH_TERM_PATTERN.formatted(character));

				assertThat(result).isEqualTo(ESCAPED_SEARCH_TERM_PATTERN.formatted(character));
			}

			@ParameterizedTest(name = "should replace {0}")
			@ValueSource(strings = { "<", ">" })
			void shouldRemoveSpecialCharacters(String character) {
				var result = searchRepostitory.adjustSearchTerm(SEARCH_TERM_PATTERN.formatted(character));

				assertThat(result).isEqualTo("(*" + SEARCH_TERM_PATTERN.formatted("?") + "*)");
			}

			@Test
			void shouldReplaceAllCharacters() {
				var searchTerm = "abc/123+z";
				var expectedResult = "(*abc\\/123\\+z*)";

				var result = searchRepostitory.adjustSearchTerm(searchTerm);

				assertThat(result).isEqualTo(expectedResult);
			}
		}

		@Nested
		class forRolePoststelle {
			@Test
			void shouldHaveStatusFilterOnly() {
				var request = FindVorgangRequestTestFactory.createBuilder().filterBy(
						FilterCriteriaTestFactory.createBuilder()
								.filterByOrganisationseinheitenId(false).clearOrganisationseinheitIds()
								.clearStatus().singleStatus(Status.ABGESCHLOSSEN)
								.build())
						.build();
				var query = searchRepostitory.buildQueryBuilder(request).build();

				assertThat(query.getFilter()).isNotNull();
				assertThat(query.getFilter()).asString().contains("status");
				assertThat(query.getFilter()).asString().contains(Status.ABGESCHLOSSEN.name());
				assertThat(query.getFilter()).asString().doesNotContain("organisationseinheitenId");
			}

			@Test
			void shouldHaveOrganisationseinheitenIdFilterOnly() {
				var request = FindVorgangRequestTestFactory.create();
				var query = searchRepostitory.buildQueryBuilder(request).build();

				assertThat(query.getFilter()).isNotNull();
				assertThat(query.getFilter()).asString().contains("organisationseinheitenId");
				assertThat(query.getFilter()).asString().contains(ZustaendigeStelleTestFactory.ORGANISATIONSEINHEIT_ID);
			}

			@Test
			void shouldHaveOrganisationseinheitenAndStatusFilter() {
				var request = FindVorgangRequestTestFactory.createBuilder().filterBy(
						FilterCriteriaTestFactory.createBuilder().clearStatus().singleStatus(Status.ABGESCHLOSSEN).build())
						.build();

				var query = searchRepostitory.buildQueryBuilder(request).build();

				assertThat(query.getFilter()).isNotNull();
				assertThat(query.getFilter()).asString().contains("organisationseinheitenId");
				assertThat(query.getFilter()).asString().contains(ZustaendigeStelleTestFactory.ORGANISATIONSEINHEIT_ID);
				assertThat(query.getFilter()).asString().contains("status");
				assertThat(query.getFilter()).asString().contains(Status.ABGESCHLOSSEN.name());
			}
		}

		@Nested
		class forOtherRoles {
			@Test
			void shouldHaveOrganisationseinheitenIdFilter() {
				var request = FindVorgangRequestTestFactory.create();
				var query = searchRepostitory.buildQueryBuilder(request).build();

				assertThat(query.getFilter()).isNotNull();
				assertThat(query.getFilter()).asString().contains("organisationseinheitenId");
				assertThat(query.getFilter()).asString().contains(ZustaendigeStelleTestFactory.ORGANISATIONSEINHEIT_ID);
			}

			@Test
			void shouldHaveAssignedToFilter() {
				var request = FindVorgangRequestTestFactory.create();
				var query = searchRepostitory.buildQueryBuilder(request).build();

				assertThat(query.getFilter()).isNotNull();
				assertThat(query.getFilter()).asString().contains("assignedTo");
				assertThat(query.getFilter()).asString().contains(UserTestFactory.ID);
			}
		}
	}

	@Nested
	class TestPartialUpdate {

		@Nested
		class TestUpdate {

			private static final Set<String> FIELDS_TO_REMOVE = Set.of(IndexedVorgang.FIELD_AKTENZEICHEN);
			private static final Map<String, Object> FIELDS_TO_UPDATE = Map.of(IndexedVorgang.FIELD_AKTENZEICHEN,
					AKTENZEICHEN);

			@Mock
			private UpdateQuery updateQuery;

			@Test
			void shouldCallExecuteUpdate() {
				doNothing().when(searchRepostitory).executeUpdate(any());
				doReturn(FIELDS_TO_REMOVE).when(searchRepostitory).getFieldsToRemove(any());
				doReturn(FIELDS_TO_UPDATE).when(searchRepostitory).getFieldsToUpdate(any());

				searchRepostitory.update(IndexedVorgangTestFactory.ID, Collections.emptyMap());

				verify(searchRepostitory, times(2)).executeUpdate(any());
			}

			@Test
			void shouldNotCallExecuteUpdate() {
				doReturn(Collections.emptySet()).when(searchRepostitory).getFieldsToRemove(any());
				doReturn(Collections.emptyMap()).when(searchRepostitory).getFieldsToUpdate(any());

				searchRepostitory.update(IndexedVorgangTestFactory.ID, createMap());

				verify(searchRepostitory, never()).executeUpdate(any());
			}

			Map<String, Object> createMap() {
				var map = new HashMap<String, Object>();
				map.put(IndexedVorgang.FIELD_ANTRAGSTELLER_NAME, IndexedVorgangTestFactory.ANTRAGSTELLER_NAME);
				map.put(IndexedVorgang.FIELD_AKTENZEICHEN, null);
				return map;
			}

			@Nested
			class TestUpdateWithScript {

				@BeforeEach
				void setup() {
					doNothing().when(searchRepostitory).executeUpdate(any());
					doReturn(FIELDS_TO_REMOVE).when(searchRepostitory).getFieldsToRemove(any());
				}

				@Test
				void shouldCallBuildRemoveFieldScript() {
					searchRepostitory.update(IndexedVorgangTestFactory.ID, createMap());

					verify(searchRepostitory).buildRemoveFieldScript(FIELDS_TO_REMOVE);
				}

				@Test
				void shouldCallBuildUpdateQueryWithScript() {
					var createdScript = "script";
					doReturn(createdScript).when(searchRepostitory).buildRemoveFieldScript(any());

					searchRepostitory.update(IndexedVorgangTestFactory.ID, createMap());

					verify(searchRepostitory).buildUpdateQueryWithScript(IndexedVorgangTestFactory.ID, createdScript);
				}

				@Test
				void shouldCallExecuteUpdate() {
					doReturn(updateQuery).when(searchRepostitory).buildUpdateQueryWithScript(any(), any());

					searchRepostitory.update(IndexedVorgangTestFactory.ID, createMap());

					verify(searchRepostitory).executeUpdate(updateQuery);
				}

				private Map<String, Object> createMap() {
					var map = new HashMap<String, Object>();
					map.put(IndexedVorgang.FIELD_AKTENZEICHEN, null);
					return map;
				}
			}

			@Nested
			class TestUpdateWithDocument {

				@BeforeEach
				void setup() {
					doNothing().when(searchRepostitory).executeUpdate(any());
					doReturn(FIELDS_TO_UPDATE).when(searchRepostitory).getFieldsToUpdate(any());
				}

				@Test
				void shouldCallBuildUpdateQueryWithDocument() {
					searchRepostitory.update(IndexedVorgangTestFactory.ID, FIELDS_TO_UPDATE);

					verify(searchRepostitory).buildUpdateQueryWithDocument(IndexedVorgangTestFactory.ID, FIELDS_TO_UPDATE);
				}

				@Test
				void shouldCallExecuteUpdate() {
					doReturn(updateQuery).when(searchRepostitory).buildUpdateQueryWithDocument(any(), any());

					searchRepostitory.update(IndexedVorgangTestFactory.ID, FIELDS_TO_UPDATE);

					verify(searchRepostitory).executeUpdate(updateQuery);
				}
			}

		}

		@Nested
		class TestGetFieldsToRemove {

			@Test
			void shouldReturnFieldsToRemove() {
				var fieldsToRemove = searchRepostitory.getFieldsToRemove(createMap(null));

				assertThat(fieldsToRemove).containsOnly(IndexedVorgang.FIELD_AKTENZEICHEN);
			}

			@Test
			void shouldReturnEmptyFieldsToRemove() {
				var fieldsToRemove = searchRepostitory.getFieldsToRemove(createMap(AKTENZEICHEN));

				assertThat(fieldsToRemove).isEmpty();
			}

			private Map<String, Object> createMap(Object value) {
				var map = new HashMap<String, Object>();
				map.put(IndexedVorgang.FIELD_AKTENZEICHEN, value);
				return map;
			}
		}

		@Nested
		class TestBuildRemoveFieldScript {

			@Test
			void shouldBuildScript() {
				var removeAktenzeichenScript = createScriptRemoveFields(IndexedVorgang.FIELD_AKTENZEICHEN);
				var removeAssignedToScript = createScriptRemoveFields(IndexedVorgang.FIELD_ASSIGNED_TO);

				var script = searchRepostitory.buildRemoveFieldScript(Set.of(IndexedVorgang.FIELD_AKTENZEICHEN, IndexedVorgang.FIELD_ASSIGNED_TO));

				assertThat(script).contains(removeAktenzeichenScript, removeAssignedToScript);
			}
		}

		@Nested
		class TestBuildUpdateQueryWithScript {

			@Mock
			private UpdateQuery.Builder updateQuery;

			@BeforeEach
			void setup() {
				when(updateQuery.withScript(any())).thenReturn(updateQuery);
				when(updateQuery.withScriptType(any())).thenReturn(updateQuery);
				doReturn(updateQuery).when(searchRepostitory).createUpdateBuilder(any());
			}

			@Test
			void shouldSetScriptType() {
				searchRepostitory.buildUpdateQueryWithScript(IndexedVorgangTestFactory.ID, "script");

				verify(updateQuery).withScriptType(ScriptType.INLINE);
			}

			@Test
			void shouldSetScript() {
				var script = "script";

				searchRepostitory.buildUpdateQueryWithScript(IndexedVorgangTestFactory.ID, script);

				verify(updateQuery).withScript(script);
			}

			@Test
			void shouldCallBuild() {
				searchRepostitory.buildUpdateQueryWithScript(IndexedVorgangTestFactory.ID, "script");

				verify(updateQuery).build();
			}
		}

		@Nested
		class TestGetFieldsToUpdate {

			@Test
			void shouldReturnFieldsToUpdate() {
				var fieldsToUpdate = searchRepostitory.getFieldsToUpdate(createMap(IndexedVorgang.FIELD_AKTENZEICHEN, AKTENZEICHEN));

				assertThat(fieldsToUpdate).containsOnly(MapEntry.entry(IndexedVorgang.FIELD_AKTENZEICHEN, AKTENZEICHEN));
			}

			@Test
			void shouldReturnEmptyFieldsToUpdate() {
				var fieldsToUpdate = searchRepostitory.getFieldsToUpdate(createMap(IndexedVorgang.FIELD_ASSIGNED_TO, null));

				assertThat(fieldsToUpdate).isEmpty();
			}

			private Map<String, Object> createMap(String key, Object value) {
				var map = new HashMap<String, Object>();
				map.put(key, value);
				return map;
			}
		}

		@Nested
		class TestBuildUpdateQueryWithDocument {

			@Mock
			private UpdateQuery.Builder updateQuery;
			@Mock
			private Document document;

			@BeforeEach
			void setup() {
				when(updateQuery.withDocument(any())).thenReturn(updateQuery);
				doReturn(updateQuery).when(searchRepostitory).createUpdateBuilder(any());
			}

			@Test
			void shouldSetDocument() {
				Map<String, Object> patch = Map.of(IndexedVorgang.FIELD_AKTENZEICHEN, AKTENZEICHEN);
				try(var fromMock = mockStatic(Document.class)) {
					fromMock.when(() -> Document.from(any())).thenReturn(document);

					searchRepostitory.buildUpdateQueryWithDocument(IndexedVorgangTestFactory.ID, patch);

					verify(updateQuery).withDocument(document);
				}
			}

			@Test
			void shouldCallBuild() {
				searchRepostitory.buildUpdateQueryWithDocument(IndexedVorgangTestFactory.ID, Map.of());

				verify(updateQuery).build();
			}
		}

		@Nested
		class TestExecuteUpdate {

			@Mock
			private UpdateQuery updateQuery;
			@Mock
			private IndexOperations indexOperations;
			@Mock
			private IndexCoordinates indexCoordinates;

			@BeforeEach
			void setup() {
				when(indexOperations.getIndexCoordinates()).thenReturn(indexCoordinates);
				when(operations.indexOps(any(Class.class))).thenReturn(indexOperations);
			}

			@Test
			void shouldCallUpdate() {
				searchRepostitory.executeUpdate(updateQuery);

				verify(operations).update(updateQuery, indexCoordinates);
			}
		}
	}
}
