/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.common.search;

import static org.awaitility.Awaitility.*;

import java.time.Duration;
import java.util.Optional;

import org.assertj.core.api.Assertions;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.ValueSource;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.data.mongodb.core.MongoOperations;
import org.springframework.data.mongodb.core.index.IndexInfo;

import de.ozgcloud.common.test.DataITCase;
import de.ozgcloud.vorgang.vorgang.AntragstellerTestFactory;
import de.ozgcloud.vorgang.vorgang.EingangTestFactory;
import de.ozgcloud.vorgang.vorgang.FilterCriteria;
import de.ozgcloud.vorgang.vorgang.FindVorgangRequest;
import de.ozgcloud.vorgang.vorgang.Vorgang;
import de.ozgcloud.vorgang.vorgang.VorgangTestFactory;

@SpringBootTest(properties = { "ozgcloud.mongodbsearch.enabled=true" })
@DataITCase
public class SearchVorgangCustomRepositoryMongoITCase {

	@Autowired
	private SearchVorgangCustomRepositoryMongo repostitory;

	@Autowired
	private MongoOperations mongoOperations;

	@Nested
	class TestSpecialCharacters {

		@BeforeEach
		void init() {
			await().atMost(Duration.ofSeconds(10)).until(() -> getTextIndex().isPresent());
			mongoOperations.remove(Vorgang.class).all();
		}

		@DisplayName("should find vorgang when vorgangName")
		@ParameterizedTest(name = "contains {0}")
		@ValueSource(strings = { "_", "+", "-", "=", "&&", "||", "!", "(", ")", "{", "}", "[", "]", "^", "\"", "~", "*", "?", ":", "\\", "/", "<",
				">", "$", "%", "&", "#", "@" })
		void shouldFindInVorgangName(String character) {
			var vorgangName = "AbC%sTest".formatted(character);
			mongoOperations.save(VorgangTestFactory.createBuilder().name(vorgangName).build());

			var vorgangHeaders = repostitory.searchBy(searchRequest(character));

			Assertions.assertThat(vorgangHeaders.getContent()).hasSize(1).first()
					.extracting("name").isEqualTo(vorgangName);
		}

		@Test
		void shouldFindInVorgangNummer() {
			mongoOperations.save(VorgangTestFactory.create());

			var vorgangHeaders = repostitory.searchBy(buildSearchRequest("VOR-GANG"));

			Assertions.assertThat(vorgangHeaders.getContent()).hasSize(1).first()
					.extracting("nummer").isEqualTo(VorgangTestFactory.VORGANG_NUMMER);
		}

		@Test
		void shouldFindInAntragstellerName() {
			var antragsteller = AntragstellerTestFactory.createBuilder().nachname("MusterMann").build();
			var eingang = EingangTestFactory.createBuilder().antragsteller(antragsteller).build();
			mongoOperations.save(VorgangTestFactory.createBuilder().clearEingangs().eingang(eingang).build());

			var vorgangHeaders = repostitory.searchBy(buildSearchRequest("mustermann"));

			Assertions.assertThat(vorgangHeaders.getContent()).hasSize(1);
		}

		@Test
		void shouldFindInAntragstellerVorname() {
			var antragsteller = AntragstellerTestFactory.createBuilder().vorname("Alexander").build();
			var eingang = EingangTestFactory.createBuilder().antragsteller(antragsteller).build();
			mongoOperations.save(VorgangTestFactory.createBuilder().clearEingangs().eingang(eingang).build());

			var vorgangHeaders = repostitory.searchBy(buildSearchRequest("alexand"));

			Assertions.assertThat(vorgangHeaders.getContent()).hasSize(1);
		}

		@Test
		void shouldFindInAktenzeichen() {
			var foundVorgang = mongoOperations.save(VorgangTestFactory.createBuilder().id(null).aktenzeichen("123-456").build());
			mongoOperations.save(VorgangTestFactory.createBuilder().id(null).build());

			var vorgangHeaders = repostitory.searchBy(buildSearchRequest("123-"));

			Assertions.assertThat(vorgangHeaders.getContent()).hasSize(1).first().extracting("id").isEqualTo(foundVorgang.getId());
		}

		@Test
		void shouldFindInMultipleFields() {
			var antragsteller = AntragstellerTestFactory.createBuilder().nachname("Müller").build();
			mongoOperations.save(VorgangTestFactory.createBuilder().id(null).name("Mäuse")
							.eingang(EingangTestFactory.createBuilder().antragsteller(antragsteller).build())
							.build(), Vorgang.COLLECTION_NAME);

			var result = repostitory.searchBy(buildSearchRequest("mull mau"));

			Assertions.assertThat(result.getContent()).hasSize(1);
		}

		private Optional<IndexInfo> getTextIndex() {
			return mongoOperations.indexOps(Vorgang.COLLECTION_NAME).getIndexInfo().stream()
					.filter(index -> MongodbFullTextSearchIndexInitializer.TEXT_INDEX_NAME.equals(index.getName())).findFirst();
		}

		private FindVorgangRequest searchRequest(String character) {
			return FindVorgangRequest.builder().filterBy(FilterCriteria.builder().build()).searchBy("abc%st".formatted(character)).build();
		}

		private FindVorgangRequest buildSearchRequest(String searchBy) {
			return FindVorgangRequest.builder().filterBy(FilterCriteria.builder().build()).searchBy(searchBy).build();
		}
	}
}
