/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.common.security;

import static org.assertj.core.api.Assertions.*;

import java.util.List;

import org.junit.jupiter.api.Assertions;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.mongodb.core.MongoOperations;

import de.ozgcloud.command.Command;
import de.ozgcloud.common.test.DataITCase;
import de.ozgcloud.vorgang.attached_item.VorgangAttachedItem;
import de.ozgcloud.vorgang.attached_item.VorgangAttachedItemTestFactory;
import de.ozgcloud.vorgang.command.CommandTestFactory;
import de.ozgcloud.vorgang.common.errorhandling.NotFoundException;
import de.ozgcloud.vorgang.files.OzgCloudFile;
import de.ozgcloud.vorgang.files.OzgCloudFileTestUtil;
import de.ozgcloud.vorgang.files.GridFsTestFactory;
import de.ozgcloud.vorgang.vorgang.Vorgang;
import de.ozgcloud.vorgang.vorgang.VorgangTestFactory;
import de.ozgcloud.vorgang.vorgang.ZustaendigeStelleTestFactory;

@DataITCase
class PolicyRepositoryITCase {

	@Autowired
	private PolicyRepository repository;
	@Autowired
	private MongoOperations mongoOperations;
	@Autowired
	private OzgCloudFileTestUtil ozgCloudFileTestUtil;

	@BeforeEach
	void prepareDatabase() {
		mongoOperations.dropCollection(Vorgang.COLLECTION_NAME);
		mongoOperations.dropCollection(OzgCloudFile.class);

	}

	@DisplayName("Match organisationEinheitenId")
	@Nested
	class TestMatchOrganisationEinheitId {

		@DisplayName("by command id")
		@Nested
		class TestByCommandId {

			@BeforeEach
			void initDatabase() {
				mongoOperations.dropCollection(Command.COLLECTION_NAME);

				mongoOperations.save(CommandTestFactory.create());
			}

			@Test
			void shouldReturnTrueIfMatch() {
				mongoOperations.save(VorgangTestFactory.create());

				var result = repository.existsByCommandId(CommandTestFactory.ID,
						List.of(ZustaendigeStelleTestFactory.ORGANISATIONSEINHEIT_ID));

				assertThat(result).isTrue();
			}

			@Test
			void shouldReturnFalseIfNotMatch() {
				mongoOperations.save(VorgangTestFactory.createWithOrganisationEinheitId("quatsch").toBuilder().id(null).build());

				var result = repository.existsByCommandId(CommandTestFactory.ID, List.of(ZustaendigeStelleTestFactory.ORGANISATIONSEINHEIT_ID));

				assertThat(result).isFalse();
			}
		}

		@DisplayName("by vorgang attached item id")
		@Nested
		class TestByVorgangAttachedItemId {

			private VorgangAttachedItem vorgangAttachedItem;

			@BeforeEach
			void initDatabase() {
				mongoOperations.dropCollection(VorgangAttachedItem.COLLECTION_NAME);

				vorgangAttachedItem = mongoOperations.save(VorgangAttachedItemTestFactory.createBuilder().version(0).build());
			}

			@Test
			void shouldReturnTrueIfMatch() {
				mongoOperations.save(VorgangTestFactory.create());

				var result = repository.existsByVorgangAttachedItem(vorgangAttachedItem.getId(),
						List.of(ZustaendigeStelleTestFactory.ORGANISATIONSEINHEIT_ID));

				assertThat(result).isTrue();
			}

			@Test
			void shouldReturnFalseIfNotMatch() {
				mongoOperations.save(VorgangTestFactory.createWithOrganisationEinheitId("quatsch").toBuilder().build());

				var result = repository.existsByVorgangAttachedItem(VorgangAttachedItemTestFactory.ID,
						List.of(ZustaendigeStelleTestFactory.ORGANISATIONSEINHEIT_ID));

				assertThat(result).isFalse();
			}

			@Test
			void shouldThrowNotFoundIfDeleted() {
				var vorgangAttachedItem = mongoOperations.save(
						VorgangAttachedItemTestFactory.createBuilder().id(null).deleted(true).version(0).build());
				var id = vorgangAttachedItem.getId();

				Assertions.assertThrows(NotFoundException.class, () -> existsByVorgangAttachedItem(id));
			}

			private void existsByVorgangAttachedItem(String vorgangAttachedItemId) {
				repository.existsByVorgangAttachedItem(vorgangAttachedItemId, List.of(ZustaendigeStelleTestFactory.ORGANISATIONSEINHEIT_ID));
			}
		}

		@DisplayName("by file id")
		@Nested
		class TestByFileId {

			private OzgCloudFile ozgCloudFile;

			@BeforeEach
			void initDatabase() {
				ozgCloudFile = ozgCloudFileTestUtil.saveOzgCloudFile();
			}

			@Test
			void shouldReturnTrueIfMatch() {
				mongoOperations.save(VorgangTestFactory.create());

				var result = repository.existsByFileId(ozgCloudFile.getId(),
						List.of(ZustaendigeStelleTestFactory.ORGANISATIONSEINHEIT_ID));

				assertThat(result).isTrue();
			}

			@Test
			void shouldReturnFalseIfNotMatch() {
				mongoOperations.save(VorgangTestFactory.createWithOrganisationEinheitId("quatsch").toBuilder().id(null).build());

				var result = repository.existsByFileId(GridFsTestFactory.ID.toHexString(),
						List.of(ZustaendigeStelleTestFactory.ORGANISATIONSEINHEIT_ID));

				assertThat(result).isFalse();
			}
		}

	}
}
