/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.files;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.Mockito.*;

import java.util.List;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.Mock;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.mongodb.core.MongoOperations;

import de.ozgcloud.common.test.DataITCase;
import de.ozgcloud.vorgang.grpc.file.GrpcGetAttachmentsRequest;
import de.ozgcloud.vorgang.grpc.file.GrpcGetAttachmentsResponse;
import de.ozgcloud.vorgang.grpc.file.GrpcGetRepresentationsRequest;
import de.ozgcloud.vorgang.grpc.file.GrpcGetRepresentationsResponse;
import de.ozgcloud.vorgang.vorgang.EingangTestFactory;
import de.ozgcloud.vorgang.vorgang.IncomingFile;
import de.ozgcloud.vorgang.vorgang.IncomingFileGroupTestFactory;
import de.ozgcloud.vorgang.vorgang.IncomingFileTestFactory;
import de.ozgcloud.vorgang.vorgang.Vorgang;
import de.ozgcloud.vorgang.vorgang.VorgangTestFactory;
import io.grpc.stub.StreamObserver;

@DataITCase
class FileITCase {

	@Autowired
	private GrpcFileService service;

	@Autowired
	private MongoOperations mongoOperations;

	private final static FileId ID2 = FileId.createNew();

	private final static List<IncomingFile> TEST_FILES = List.of(IncomingFileTestFactory.create(),
			IncomingFileTestFactory.createBuilder().id(ID2).build());

	private final Vorgang vorgang = VorgangTestFactory.createBuilder().clearEingangs()
			.eingang(EingangTestFactory.createBuilder().clearAttachments().clearRepresentations()
					.attachments(List.of(IncomingFileGroupTestFactory.createBuilder().clearFiles().files(TEST_FILES).build()))
					.representations(TEST_FILES).build())
			.build();

	@BeforeEach
	void init() {
		mongoOperations.dropCollection(Vorgang.class);
		mongoOperations.save(vorgang);
	}

	@Nested
	class TestGetAttachments {

		@Mock
		private StreamObserver<GrpcGetAttachmentsResponse> responseObserver;
		@Captor
		private ArgumentCaptor<GrpcGetAttachmentsResponse> responseCaptor;

		private GrpcGetAttachmentsRequest request;

		@Nested
		class TestByEingangId {

			@BeforeEach
			void init() {
				request = GrpcGetAttachmentsRequestTestFactory.create();
			}

			@Test
			void getAttachmentsShouldReturnNonEmptyGrpcResponse() {
				callServiceGetAttachments();

				var grpcResponse = responseCaptor.getValue();

				assertThat(grpcResponse).isNotNull();
			}

			@Test
			void eingangShouldHaveTwoAttachments() {
				callServiceGetAttachments();

				var grpcResponse = responseCaptor.getValue();

				assertThat(grpcResponse.getFileCount()).isEqualTo(2);
			}

			@Test
			void validateFirstAttachmentFileContent() {
				callServiceGetAttachments();

				var attachment1 = responseCaptor.getValue().getFile(0);

				assertThat(attachment1.getId()).isEqualTo(IncomingFileTestFactory.ID.toString());
				assertThat(attachment1.getName()).isEqualTo(IncomingFileTestFactory.NAME);
				assertThat(attachment1.getSize()).isEqualTo(IncomingFileTestFactory.SIZE);
			}

			@Test
			void validateSecondAttachmentFileContent() {
				callServiceGetAttachments();

				var attachment2 = responseCaptor.getValue().getFile(1);

				assertThat(attachment2.getId()).isEqualTo(ID2.toString());
			}
		}

		@Nested
		class TestByVorgangId {

			@BeforeEach
			void init() {
				request = GrpcGetAttachmentsRequestTestFactory.createBuilder().clearEingangId().build();
			}

			@Test
			void shouldGetAttachments() {
				callServiceGetAttachments();

				verify(responseObserver).onNext(responseCaptor.capture());
				assertThat(responseCaptor.getValue().getFileCount()).isEqualTo(2);
			}
		}

		@Nested
		class TestByVorgangAndEingangId {

			@BeforeEach
			void init() {
				request = GrpcGetAttachmentsRequestTestFactory.create();
			}

			@Test
			void shouldGetAttachments() {
				callServiceGetAttachments();

				verify(responseObserver).onNext(responseCaptor.capture());
				assertThat(responseCaptor.getValue().getFileCount()).isEqualTo(2);
			}
		}

		private void callServiceGetAttachments() {
			service.getAttachments(request, responseObserver);

			verify(responseObserver).onNext(responseCaptor.capture());
		}
	}

	@Nested
	class TestGetRepresentations {

		@Mock
		private StreamObserver<GrpcGetRepresentationsResponse> responseObserver;
		@Captor
		private ArgumentCaptor<GrpcGetRepresentationsResponse> responseCaptor;
		private final GrpcGetRepresentationsRequest request = GrpcGetRepresentationsRequestTestFactory.create();

		@Test
		void getRepresentationsShouldReturnNonEmptyGrpcResponse() {
			callServiceGetRepresentations();

			var grpcResponse = responseCaptor.getValue();

			assertThat(grpcResponse).isNotNull();
		}

		@Test
		void eingangShouldHaveTwoRepresentations() {
			callServiceGetRepresentations();

			var grpcResponse = responseCaptor.getValue();

			assertThat(grpcResponse.getFileCount()).isEqualTo(2);
		}

		@Test
		void validateFirstRepresentationFileContent() {
			callServiceGetRepresentations();

			var attachment1 = responseCaptor.getValue().getFile(0);

			assertThat(attachment1.getId()).isEqualTo(IncomingFileTestFactory.ID.toString());
			assertThat(attachment1.getName()).isEqualTo(IncomingFileTestFactory.NAME);
			assertThat(attachment1.getSize()).isEqualTo(IncomingFileTestFactory.SIZE);
		}

		@Test
		void validateSecondRepresentationFileContent() {
			callServiceGetRepresentations();

			var attachment1 = responseCaptor.getValue().getFile(1);

			assertThat(attachment1.getId()).isEqualTo(ID2.toString());
		}

		private void callServiceGetRepresentations() {
			service.getRepresentations(request, responseObserver);

			verify(responseObserver).onNext(responseCaptor.capture());
		}
	}
}