/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.files;

import static org.assertj.core.api.Assertions.*;
import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.io.BufferedInputStream;
import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.Random;
import java.util.Set;
import java.util.UUID;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.function.BiFunction;
import java.util.function.Consumer;
import java.util.function.Function;
import java.util.function.Predicate;
import java.util.stream.Stream;

import org.apache.commons.io.IOUtils;
import org.junit.jupiter.api.Assertions;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.ArgumentMatchers;
import org.mockito.Captor;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;
import org.springframework.test.util.ReflectionTestUtils;

import com.google.protobuf.ByteString;
import com.thedeanda.lorem.LoremIpsum;

import de.ozgcloud.common.errorhandling.TechnicalException;
import de.ozgcloud.vorgang.callcontext.CallContextTestFactory;
import de.ozgcloud.vorgang.callcontext.CallContextUserTestFactory;
import de.ozgcloud.vorgang.callcontext.CurrentUserService;
import de.ozgcloud.vorgang.callcontext.UserTestFactory;
import de.ozgcloud.vorgang.command.GrpcCallContextTestFactory;
import de.ozgcloud.vorgang.command.GrpcUserTestFactory;
import de.ozgcloud.vorgang.common.errorhandling.NotFoundException;
import de.ozgcloud.vorgang.common.security.PolicyService;
import de.ozgcloud.vorgang.grpc.binaryFile.GrpcFindFilesResponse;
import de.ozgcloud.vorgang.grpc.binaryFile.GrpcGetBinaryFileDataRequest;
import de.ozgcloud.vorgang.grpc.binaryFile.GrpcGetBinaryFileDataResponse;
import de.ozgcloud.vorgang.grpc.binaryFile.GrpcUpdateBinaryFileRequest;
import de.ozgcloud.vorgang.grpc.binaryFile.GrpcUpdateBinaryFileResponse;
import de.ozgcloud.vorgang.grpc.binaryFile.GrpcUploadBinaryFileRequest;
import de.ozgcloud.vorgang.grpc.binaryFile.GrpcUploadBinaryFileResponse;
import de.ozgcloud.vorgang.vorgang.VorgangTestFactory;
import io.grpc.stub.CallStreamObserver;
import io.grpc.stub.StreamObserver;
import lombok.SneakyThrows;

class GrpcBinaryFileServiceTest {

	private static final OzgCloudFile BINARY_FILE = OzgCloudFileTestFactory.create();
	private static final String NEW_CONTENT_FILE_ID = UUID.randomUUID().toString();

	@InjectMocks
	@Spy
	private GrpcBinaryFileService grpcService;
	@Mock
	private FileService service;
	@Mock
	private GrpcOzgFileMapper fileMapper;
	@Mock
	private PolicyService policyService;
	@Mock
	private CurrentUserService currentUserService;

	@Nested
	class TestFindOzgCloudFilesMetaData {

		@Mock
		private StreamObserver<GrpcFindFilesResponse> responseObserver;
		@Captor
		private ArgumentCaptor<Collection<FileId>> fileIdsCaptor;
		@Captor
		private ArgumentCaptor<GrpcFindFilesResponse> responseCaptor;

		@Test
		void shouldCallPolicyService() {
			findBinaryFilesMetaData();

			verify(policyService).checkPermission(ArgumentMatchers.<Set<String>>any());
		}

		@Test
		void shouldCallService() {
			findBinaryFilesMetaData();

			verify(service).findOzgCloudFiles(fileIdsCaptor.capture());
			assertThat(fileIdsCaptor.getValue()).contains(OzgCloudFileTestFactory.ID);
		}

		@Test
		void shouldReturnList() {
			findBinaryFilesMetaData();

			verify(responseObserver).onNext(responseCaptor.capture());
		}

		@Test
		void shouldCallComplete() {
			findBinaryFilesMetaData();

			verify(responseObserver).onCompleted();
		}

		private void findBinaryFilesMetaData() {
			grpcService.findBinaryFilesMetaData(GrpcBinaryFilesRequestTestFactory.create(), responseObserver);

			verify(responseObserver).onNext(responseCaptor.capture());
		}
	}

	@Nested
	class TestUploadOzgCloudFile {

		@Mock
		private OzgCloudFile updatedOzgCloudFile;

		@BeforeEach
		void init() {
			when(service.saveOzgCloudFile(any())).thenReturn(OzgCloudFileTestFactory.ID);
			doReturn(updatedOzgCloudFile).when(grpcService).updateContentId(any(), any());
		}

		@Test
		void shouldCallUpdateContentId() {
			uploadBinaryFile();

			verify(grpcService).updateContentId(BINARY_FILE, Optional.of(OzgCloudFileTestFactory.CONTENT_ID));
		}

		@Test
		void shouldCallUploadFile() {
			uploadBinaryFile();

			verify(service).saveOzgCloudFile(updatedOzgCloudFile);
		}

		@Test
		void shouldReturnFileId() {
			var result = uploadBinaryFile();

			assertThat(result).isEqualTo(OzgCloudFileTestFactory.ID);
		}

		private FileId uploadBinaryFile() {
			return grpcService.uploadOzgCloudFile(BINARY_FILE, Optional.of(OzgCloudFileTestFactory.CONTENT_ID));
		}

	}

	@Nested
	class TestUploadOzgCloudFileAsStream {

		private static final GrpcUploadBinaryFileRequest REQUEST = GrpcUploadBinaryFileRequestTestFactory.create();

		@Mock
		private StreamObserver<GrpcUploadBinaryFileResponse> responseObserver;
		@Mock
		private GrpcUploadBinaryFileResponse response;
		@Mock
		private GrpcUploadBinaryFileRequest requestMock;
		@Mock
		private ByteString content;
		@Captor
		private ArgumentCaptor<InputStream> inputStreamCaptor;

		@Test
		void shouldSetMetadataVerifier() {
			var uploadObserver = uploadBinaryFileAsStream();

			callMetadataVerifier(uploadObserver, REQUEST);
			verify(policyService).checkPermission(VorgangTestFactory.ID);
		}

		@Test
		void shouldSetHasMetadata() {
			var uploadObserver = uploadBinaryFileAsStream();

			callHasMetadata(uploadObserver, requestMock);
			verify(requestMock).hasMetadata();
		}

		@Test
		void shouldSetMetadataMapper() {
			doReturn(UserTestFactory.ID).when(grpcService).getUserId(any());
			doReturn(CallContextTestFactory.CLIENT).when(grpcService).getClient(any());

			var uploadObserver = uploadBinaryFileAsStream();

			callMetadataMapper(uploadObserver, REQUEST);
			verify(grpcService).buildOzgCloudFile(REQUEST);
		}

		@Test
		void shouldCallMetadataUploader() {
			var uploadObserver = uploadBinaryFileAsStream();

			callMetadataUploader(uploadObserver);
			verify(grpcService).uploadOzgCloudFile(BINARY_FILE, Optional.of(NEW_CONTENT_FILE_ID));
		}

		@Test
		void shouldCallFileContentUploader() {
			var uploadObserver = uploadBinaryFileAsStream();

			callFileContentUploader(uploadObserver);
			verify(grpcService).uploadFileContent(eq(BINARY_FILE), inputStreamCaptor.capture());
			assertThat(inputStreamCaptor.getValue()).hasBinaryContent(OzgCloudFileTestFactory.CONTENT);
		}

		@Test
		void shouldCallGetFileContent() {
			when(requestMock.getFileContent()).thenReturn(content);

			var uploadObserver = uploadBinaryFileAsStream();

			callGetFileContent(uploadObserver, requestMock);
			verify(requestMock).getFileContent();
			verify(content).toByteArray();
		}

		@Test
		void shouldCallBuildResponse() {
			var uploadObserver = uploadBinaryFileAsStream();

			callResponseConsumer(uploadObserver);
			verify(grpcService).buildResponse(OzgCloudFileTestFactory.ID);
		}

		@Test
		void shouldCallCompleteRequest() {
			doReturn(response).when(grpcService).buildResponse(any());

			var uploadObserver = uploadBinaryFileAsStream();

			callResponseConsumer(uploadObserver);
			verify(grpcService).completeRequest(response, responseObserver);
		}

		@Test
		void shouldCallDeleteContent() {
			var uploadObserver = uploadBinaryFileAsStream();

			callCleanupFileContent(uploadObserver);
			verify(service).deleteContent(OzgCloudFileTestFactory.CONTENT_ID);
		}

		private StreamObserver<GrpcUploadBinaryFileRequest> uploadBinaryFileAsStream() {
			return grpcService.uploadBinaryFileAsStream(responseObserver);
		}

	}

	@Nested
	class TestBuildOzgCloudFile {

		@Test
		void shouldBuildBinaryFile() {
			doReturn(OzgCloudFileTestFactory.CREATED_BY).when(grpcService).getUserId(any());
			doReturn(OzgCloudFileTestFactory.CLIENT).when(grpcService).getClient(any());

			var result = buildBinaryFile();

			assertThat(result).usingRecursiveComparison().ignoringFields("id", "version", "contentId", "size")
					.isEqualTo(OzgCloudFileTestFactory.create());
		}

		private OzgCloudFile buildBinaryFile() {
			return grpcService.buildOzgCloudFile(GrpcUploadBinaryFileRequestTestFactory.create());
		}
	}

	@Nested
	class TestGetUserId {

		@Nested
		class TestWithUploadRequest {

			private final GrpcUploadBinaryFileRequest fileUploadRequest = GrpcUploadBinaryFileRequestTestFactory.create();

			@Nested
			class TestContextUserId {

				@BeforeEach
				void init() {
					when(currentUserService.findUser()).thenReturn(Optional.of(CallContextUserTestFactory.create()));
				}

				@Test
				void shouldCallFindUser() {
					getUserId();

					verify(currentUserService).findUser();
				}

				@Test
				void shouldReturnContextUserId() {
					var result = getUserId();

					assertThat(result).contains(CallContextUserTestFactory.ID);
				}
			}

			@Nested
			class TestRequestUserId {

				@BeforeEach
				void init() {
					when(currentUserService.findUser()).thenReturn(
							Optional.of(CallContextUserTestFactory.createBuilder().userId(Optional.empty()).build()));
				}

				@Test
				void shouldCallFindUser() {
					getUserId();

					verify(currentUserService).findUser();
				}

				@Test
				void shouldReturnRequestUserId() {
					var result = getUserId();

					assertThat(result).contains(GrpcUserTestFactory.ID);
				}
			}

			private String getUserId() {
				return grpcService.getUserId(fileUploadRequest);
			}
		}

		@Nested
		class TestWithoutRequest {

			@Nested
			class TestUserExists {

				@BeforeEach
				void init() {
					when(currentUserService.getUser()).thenReturn(CallContextUserTestFactory.create());
				}

				@Test
				void shouldCallGetUser() {
					grpcService.getUserId();

					verify(currentUserService).getUser();
				}

				@Test
				void shouldReturnUserId() {
					var result = grpcService.getUserId();

					assertThat(result).isEqualTo(CallContextUserTestFactory.ID);
				}
			}

			@Nested
			class TestMissingUser {

				@Test
				void shouldThrowExceptionIfMissingUser() {
					when(currentUserService.getUser()).thenThrow(new IllegalStateException());

					Assertions.assertThrows(IllegalStateException.class, grpcService::getUserId);
				}

				@Test
				void shouldThrowExceptionIfMissingUserId() {
					when(currentUserService.getUser()).thenReturn(CallContextUserTestFactory.createBuilder().userId(Optional.empty()).build());

					assertThrows(IllegalStateException.class, grpcService::getUserId);
				}
			}
		}
	}

	@Nested
	class TestGetFileStream {
		private static final int FILE_SIZE_MB = 5;

		@Mock
		private StreamObserver<GrpcFindFilesResponse> responseObserver;

		@Mock
		private CallStreamObserver<GrpcGetBinaryFileDataResponse> dataResponseObserver;

		private GrpcGetBinaryFileDataRequest request;

		@BeforeEach
		void init() {
			request = GrpcGetBinaryFileDataRequest.newBuilder().setFileId(OzgCloudFileTestFactory.ID.toString()).build();
		}

		@Test
		void shouldCallService() {
			when(service.getUploadedFileStream(any(FileId.class)))
					.thenReturn(new ByteArrayInputStream(OzgCloudFileTestFactory.createContentInMiB(FILE_SIZE_MB)));

			grpcService.getBinaryFileContent(request, dataResponseObserver);

			verify(service, timeout(500)).getUploadedFileStream(any(FileId.class));
		}

		@Test
		void shouldCallSetOnReadyHander() {
			grpcService.getBinaryFileContent(request, dataResponseObserver);

			verify(dataResponseObserver, timeout(500)).setOnReadyHandler(any());
		}
	}

	@Nested
	class TestGetBinaryFileContent {

		private static final InputStream INPUT_STREAM = new BufferedInputStream(new ByteArrayInputStream(OzgCloudFileTestFactory.CONTENT),
				FileService.CHUNK_SIZE);

		@Mock
		private CallStreamObserver<GrpcGetBinaryFileDataResponse> responseObserver;

		@Captor
		private ArgumentCaptor<InputStream> inputStreamCaptor;

		@BeforeEach
		void mock() {
			when(service.getUploadedFileStream(any(FileId.class))).thenReturn(INPUT_STREAM);
		}

		@Test
		void shouldCallPolicyService() {
			getBinaryFileContent();

			verify(policyService).checkPermissionByFile(anyString());
		}

		@Test
		void shouldCallService() {
			getBinaryFileContent();

			verify(service).getUploadedFileStream(FileId.from(GrpcGetBinaryFileDataRequestTestFactory.FILE_ID));
		}

		@Test
		void shouldCallCallDownloadFile() {
			getBinaryFileContent();

			verify(grpcService).downloadFile(eq(responseObserver), inputStreamCaptor.capture());
			assertThat(inputStreamCaptor.getValue()).hasBinaryContent(OzgCloudFileTestFactory.CONTENT);
		}

		private void getBinaryFileContent() {
			grpcService.getBinaryFileContent(GrpcGetBinaryFileDataRequestTestFactory.create(), responseObserver);

		}
	}

	@Nested
	class TestBuildChunkResponse {
		@Test
		void shouldHaveFileContent() {
			GrpcGetBinaryFileDataResponse response = grpcService.buildChunkResponse(OzgCloudFileTestFactory.CONTENT,
					OzgCloudFileTestFactory.CONTENT.length);

			assertThat(response.getFileContent()).isNotEmpty();
		}
	}

	@Nested
	class TestBuildUploadResponse {

		@Test
		void shouldHaveFileId() {
			var fileId = "file-id";

			var response = grpcService.buildResponse(FileId.from(fileId));

			assertThat(response.getFileId()).isEqualTo(fileId);
		}
	}

	@Nested
	class TestDownloadFile {

		@Mock
		private CallStreamObserver<GrpcGetBinaryFileDataResponse> responseObserver;
		@Mock
		private InputStream contentStream;

		@Captor
		private ArgumentCaptor<Runnable> onReadyHandlerCaptor;
		@Captor
		private ArgumentCaptor<AtomicBoolean> doSendFileCaptor;

		@BeforeEach
		void init() {
			doNothing().when(grpcService).sendChunks(any(), any(), any());
		}

		@Test
		void shouldSetOnReadyHandler() {
			grpcService.downloadFile(responseObserver, contentStream);

			verify(responseObserver).setOnReadyHandler(onReadyHandlerCaptor.capture());
			onReadyHandlerCaptor.getValue().run();
			verify(grpcService, times(2)).sendChunks(eq(responseObserver), eq(contentStream), doSendFileCaptor.capture());
			assertThat(doSendFileCaptor.getValue()).isTrue();
		}

		@Test
		void shouldCallSendChunks() {
			grpcService.downloadFile(responseObserver, contentStream);

			verify(grpcService).sendChunks(eq(responseObserver), eq(contentStream), doSendFileCaptor.capture());
			assertThat(doSendFileCaptor.getValue()).isTrue();
		}
	}

	@Nested
	class TestSendChunk {

		private static final AtomicBoolean DO_SEND_FILE = new AtomicBoolean(true);

		@Mock
		private CallStreamObserver<GrpcGetBinaryFileDataResponse> responseObserver;
		@Mock
		private InputStream contentStream;

		@Test
		void shouldCallIsReady() {
			sendChunks();

			verify(responseObserver).isReady();
		}

		@Nested
		class TestSending {

			private static final int SENT_LAST_BYTES = new Random().nextInt(FileService.CHUNK_SIZE);

			@SneakyThrows
			@BeforeEach
			void init() {
				when(responseObserver.isReady()).thenReturn(true, false);
			}

			@Test
			void shouldCallSendNextChunk() {
				doReturn(FileService.CHUNK_SIZE).when(grpcService).sendNextChunk(any(), any());

				sendChunks();

				verify(grpcService).sendNextChunk(responseObserver, contentStream);
				verify(grpcService, never()).handleFileEndReached(any(), any(), any());
			}

			@Test
			void shouldCallHandleFileEndReached() {
				doReturn(SENT_LAST_BYTES).when(grpcService).sendNextChunk(any(), any());

				sendChunks();

				verify(grpcService).handleFileEndReached(responseObserver, contentStream, DO_SEND_FILE);
				verify(responseObserver, times(1)).isReady();
			}

		}

		@Nested
		class TestNotSending {

			@SneakyThrows
			@Test
			void shouldNotSendIfDownloadCompleted() {
				grpcService.sendChunks(responseObserver, contentStream, new AtomicBoolean(false));

				verify(grpcService, never()).sendNextChunk(any(), any());
			}

			@Test
			void shouldNotSendIfNotReady() {
				grpcService.sendChunks(responseObserver, contentStream, new AtomicBoolean(true));

				verify(grpcService, never()).sendNextChunk(any(), any());
			}
		}

		private void sendChunks() {
			grpcService.sendChunks(responseObserver, contentStream, DO_SEND_FILE);
		}
	}

	@Nested
	class TestSendNextChunk {

		private static final int READ_BYTES = new Random().nextInt(1, FileService.CHUNK_SIZE);
		private static final byte[] BUFFER = new byte[READ_BYTES];

		@Mock
		private CallStreamObserver<GrpcGetBinaryFileDataResponse> responseObserver;
		@Mock
		private GrpcGetBinaryFileDataResponse binaryFileDataGrpcResponse;
		@Mock
		private InputStream contentStream;

		@Nested
		class TestSendingSuccessfully {
			@Captor
			private ArgumentCaptor<byte[]> bufferCaptor;

			@SneakyThrows
			@BeforeEach
			void init() {
				doReturn(binaryFileDataGrpcResponse).when(grpcService).buildChunkResponse(any(), anyInt());
				new Random().nextBytes(BUFFER);
				when(contentStream.read(any())).thenAnswer(invocation -> {
					byte[] buffer = invocation.getArgument(0);
					System.arraycopy(BUFFER, 0, buffer, 0, READ_BYTES);
					return READ_BYTES;
				});
			}

			@Test
			void shouldCallBuildChuckResponse() {
				grpcService.sendNextChunk(responseObserver, contentStream);

				verify(grpcService).buildChunkResponse(bufferCaptor.capture(), eq(READ_BYTES));
				assertThat(bufferCaptor.getValue()).contains(BUFFER);
			}

			@Test
			void shouldCallOnNext() {
				grpcService.sendNextChunk(responseObserver, contentStream);

				verify(responseObserver).onNext(binaryFileDataGrpcResponse);
			}
		}

		@Nested
		class TestOnException {

			@SneakyThrows
			@Test
			void shouldThrowException() {
				doThrow(IOException.class).when(contentStream).read(any());

				assertThrows(TechnicalException.class, () -> grpcService.sendNextChunk(responseObserver, contentStream));
			}

			@SneakyThrows
			@Test
			void shouldCloseStreamOnException() {
				try (var ioUtilsMock = mockStatic(IOUtils.class)) {
					doThrow(IOException.class).when(contentStream).read(any());

					try {
						grpcService.sendNextChunk(responseObserver, contentStream);
					} catch (TechnicalException e) {
						// expected
					}

					ioUtilsMock.verify(() -> IOUtils.closeQuietly(eq(contentStream), any(Consumer.class)));
				}
			}
		}
	}

	@Nested
	class TestHandleFileEndReached {

		private static final AtomicBoolean DO_SEND_FILE = new AtomicBoolean(true);
		@Mock
		private CallStreamObserver<GrpcGetBinaryFileDataResponse> responseObserver;
		@Mock
		private InputStream contentStream;

		@Test
		void shouldSetDoSendFile() {
			handleFileEndReached();

			assertThat(DO_SEND_FILE).isFalse();
		}

		@Test
		void shouldCallCloseQuietly() {
			try (var ioUtilsMock = mockStatic(IOUtils.class)) {
				handleFileEndReached();

				ioUtilsMock.verify(() -> IOUtils.closeQuietly(eq(contentStream), any(Consumer.class)));
			}
		}

		@Test
		void shouldCallOnCompleted() {
			handleFileEndReached();

			verify(responseObserver).onCompleted();
		}

		private void handleFileEndReached() {
			grpcService.handleFileEndReached(responseObserver, contentStream, DO_SEND_FILE);
		}
	}

	@DisplayName("Upload filestream async")
	@Nested
	class TestUpdateOzgCloudFile {

		private static final GrpcUpdateBinaryFileRequest REQUEST = GrpcUpdateBinaryFileRequestTestFactory.create();

		@Mock
		private StreamObserver<GrpcUpdateBinaryFileResponse> responseObserver;
		@Mock
		private GrpcUpdateBinaryFileRequest requestMock;
		@Mock
		private ByteString content;
		@Captor
		private ArgumentCaptor<InputStream> inputStreamCaptor;

		@Test
		void shouldSetMetadataVerifier() {
			var uploadObserver = updateBinaryFile();

			callMetadataVerifier(uploadObserver, REQUEST);
			verify(grpcService).verifyUpdateRequest(REQUEST);
		}

		@Test
		void shouldSetHasMetadata() {
			var uploadObserver = updateBinaryFile();

			callHasMetadata(uploadObserver, requestMock);
			verify(requestMock).hasMetadata();
		}

		@Test
		void shouldSetMetadataMapper() {
			doReturn(BINARY_FILE).when(grpcService).buildOzgCloudFile(any(GrpcUpdateBinaryFileRequest.class));

			var uploadObserver = updateBinaryFile();

			callMetadataMapper(uploadObserver, REQUEST);
			verify(grpcService).buildOzgCloudFile(REQUEST);
		}

		@Test
		void shouldCallMetadataUploader() {
			doReturn(OzgCloudFileTestFactory.CREATED_BY).when(grpcService).getUserId();
			doReturn(OzgCloudFileTestFactory.CLIENT).when(grpcService).getClient();

			var uploadObserver = updateBinaryFile();

			callMetadataUploader(uploadObserver);
			verify(grpcService).patchOzgCloudFile(BINARY_FILE, Optional.of(NEW_CONTENT_FILE_ID));
		}

		@Test
		void shouldCallFileContentUploader() {
			var uploadObserver = updateBinaryFile();

			callFileContentUploader(uploadObserver);
			verify(grpcService).uploadFileContent(eq(BINARY_FILE), inputStreamCaptor.capture());
			assertThat(inputStreamCaptor.getValue()).hasBinaryContent(OzgCloudFileTestFactory.CONTENT);
		}

		@Test
		void shouldCallGetFileContent() {
			when(requestMock.getContent()).thenReturn(content);

			var uploadObserver = updateBinaryFile();

			callGetFileContent(uploadObserver, requestMock);
			verify(requestMock).getContent();
			verify(content).toByteArray();
		}

		@Test
		void shouldCallCompleteRequest() {
			var uploadObserver = updateBinaryFile();

			callResponseConsumer(uploadObserver);
			verify(grpcService).completeRequest(GrpcUpdateBinaryFileResponse.getDefaultInstance(), responseObserver);
		}

		@Test
		void shouldCallDeleteContent() {
			var uploadObserver = updateBinaryFile();

			callCleanupFileContent(uploadObserver);
			verify(service).deleteContent(OzgCloudFileTestFactory.CONTENT_ID);
		}

		private StreamObserver<GrpcUpdateBinaryFileRequest> updateBinaryFile() {
			return grpcService.updateBinaryFile(responseObserver);
		}

	}

	@DisplayName("Verify update request")
	@Nested
	class TestVerifyUpdateRequest {

		@BeforeEach
		void init() {
			doNothing().when(grpcService).validateMetadata(any());
			doNothing().when(policyService).checkPermissionByFile(any());
		}

		@Test
		void shouldCallValidateMetadata() {
			var request = GrpcUpdateBinaryFileRequestTestFactory.create();

			grpcService.verifyUpdateRequest(request);

			verify(grpcService).validateMetadata(request);
		}

		@Test
		void shouldCallVerifyPermissions() {
			var request = GrpcUpdateBinaryFileRequestTestFactory.create();

			grpcService.verifyUpdateRequest(request);

			verify(policyService).checkPermissionByFile(OzgCloudFileTestFactory.ID_STR);
		}
	}

	@DisplayName("Validate metadata of update binary file request")
	@Nested
	class TestValidateMetadata {

		@DisplayName("should validate metadata")
		@Test
		void shouldValidateRequest() {
			var request = GrpcUpdateBinaryFileRequestTestFactory.create();

			assertDoesNotThrow(() -> grpcService.validateMetadata(request));
		}

		@DisplayName("should throw exception when")
		@Nested
		class TestThrowException {

			@DisplayName("metadata missing")
			@Test
			void shouldThrowWhenMetadataMissing() {
				var request = GrpcUpdateBinaryFileRequestTestFactory.createBuilder().clearMetadata().build();

				assertThrows(IllegalArgumentException.class, () -> grpcService.validateMetadata(request));
			}

			@DisplayName("fileID missing")
			@Test
			void shouldThrowWhenFileIdMissing() {
				var request = GrpcUpdateBinaryFileRequestTestFactory.createBuilder()
						.setMetadata(GrpcBinaryFileMetadataTestFactory.createBuilder().clearFileId()).build();

				assertThrows(IllegalArgumentException.class, () -> grpcService.validateMetadata(request));
			}

			@DisplayName("file name missing")
			@Test
			void shouldThrowWhenFileNameMissing() {
				var request = GrpcUpdateBinaryFileRequestTestFactory.createBuilder()
						.setMetadata(GrpcBinaryFileMetadataTestFactory.createBuilder().clearFileName()).build();

				assertThrows(IllegalArgumentException.class, () -> grpcService.validateMetadata(request));
			}

			@DisplayName("content type missing")
			@Test
			void shouldThrowWhenContentTypeMissing() {
				var request = GrpcUpdateBinaryFileRequestTestFactory.createBuilder()
						.setMetadata(GrpcBinaryFileMetadataTestFactory.createBuilder().clearContentType()).build();

				assertThrows(IllegalArgumentException.class, () -> grpcService.validateMetadata(request));
			}

		}
	}

	@DisplayName("Build binary file from update request")
	@Nested
	class TestBuildOzgCloudFileUpdate {

		private static final GrpcUpdateBinaryFileRequest REQUEST = GrpcUpdateBinaryFileRequestTestFactory.create();
		private static final OzgCloudFile FOUND_BINARY_FILE = OzgCloudFileTestFactory.create();
		private static final OzgCloudFile UPDATED_BINARY_FILE = OzgCloudFileTestFactory.create();

		@Test
		void shouldThrowExceptionIfFileNotFound() {
			assertThrows(NotFoundException.class, this::buildBinaryFile);
		}

		@Nested
		class TestFileExists {

			@BeforeEach
			void init() {
				when(service.findOzgCloudFiles(anyList())).thenReturn(Stream.of(FOUND_BINARY_FILE));
				doReturn(UPDATED_BINARY_FILE).when(grpcService).updateOzgCloudFile(any(), any());
			}

			@Test
			void shouldCallFindFilesMetaData() {
				var fileid = FileId.from(GrpcBinaryFileMetadataTestFactory.FILE_ID);

				buildBinaryFile();

				verify(service).findOzgCloudFiles(List.of(fileid));
			}

			@Test
			void shouldCallUpdateBinaryFile() {
				buildBinaryFile();

				verify(grpcService).updateOzgCloudFile(FOUND_BINARY_FILE, REQUEST);
			}

			@Test
			void shouldReturnUpdatedBinaryFile() {
				var result = buildBinaryFile();

				assertThat(result).isSameAs(UPDATED_BINARY_FILE);
			}
		}

		private OzgCloudFile buildBinaryFile() {
			return grpcService.buildOzgCloudFile(REQUEST);
		}
	}

	@Nested
	class TestUpdateOzgCloudFileFields {

		private static final String NEW_FILE_NAME = LoremIpsum.getInstance().getWords(1);
		private static final String NEW_CONTENT_TYPE = LoremIpsum.getInstance().getWords(1);

		@Test
		void shouldUpdateFileName() {
			var result = updateBinaryFile();

			assertThat(result.getName()).isEqualTo(NEW_FILE_NAME);
		}

		@Test
		void shouldUpdateContentType() {
			var result = updateBinaryFile();

			assertThat(result.getContentType()).isEqualTo(NEW_CONTENT_TYPE);
		}

		@Test
		void shouldNotChangeOtherFields() {
			var result = updateBinaryFile();

			assertThat(result).usingRecursiveComparison().ignoringFields("name", "contentType").isEqualTo(OzgCloudFileTestFactory.create());
		}

		private OzgCloudFile updateBinaryFile() {
			return grpcService.updateOzgCloudFile(OzgCloudFileTestFactory.create(), GrpcUpdateBinaryFileRequestTestFactory.createBuilder()
					.setMetadata(GrpcBinaryFileMetadataTestFactory.createBuilder().setFileName(NEW_FILE_NAME).setContentType(NEW_CONTENT_TYPE))
					.build());
		}
	}

	@Nested
	class TestPatchFileMetadata {

		private static final OzgCloudFile BINARY_FILE = OzgCloudFileTestFactory.create();
		private static final OzgCloudFile UPDATED_BINARY_FILE = OzgCloudFileTestFactory.createBuilder().contentId(NEW_CONTENT_FILE_ID)
				.build();
		private static final Map<String, Object> PATCH_MAP = Map.of("key", "value");

		@BeforeEach
		void init() {
			doReturn(PATCH_MAP).when(grpcService).buildPatch(any());
		}

		@Test
		void shouldCallUpdateContentId() {
			patchBinaryFile();

			verify(grpcService).updateContentId(BINARY_FILE, Optional.of(NEW_CONTENT_FILE_ID));
		}

		@Test
		void shouldCallBuildPatch() {
			doReturn(UPDATED_BINARY_FILE).when(grpcService).updateContentId(any(), any());

			patchBinaryFile();

			verify(grpcService).buildPatch(UPDATED_BINARY_FILE);
		}

		@Test
		void shouldCallPatch() {
			patchBinaryFile();

			verify(service).patch(OzgCloudFileTestFactory.ID, OzgCloudFileTestFactory.VERSION, PATCH_MAP);
		}

		@Test
		void shouldCallDeleteContent() {
			patchBinaryFile();

			verify(service).deleteContent(OzgCloudFileTestFactory.CONTENT_ID);
		}

		@Test
		void shouldNotCallDeleteContent() {
			grpcService.patchOzgCloudFile(BINARY_FILE, Optional.empty());

			verify(service, never()).deleteContent(any());
		}

		@Test
		void shouldReturnFileId() {
			var result = patchBinaryFile();

			assertThat(result).isEqualTo(OzgCloudFileTestFactory.ID);
		}

		private FileId patchBinaryFile() {
			return grpcService.patchOzgCloudFile(BINARY_FILE, Optional.of(NEW_CONTENT_FILE_ID));
		}
	}

	@Nested
	class TestUpdateContentId {

		private static final OzgCloudFile BINARY_FILE = OzgCloudFileTestFactory.create();

		@Test
		void shouldUpdateContentId() {
			var result = updateContentId();

			assertThat(result.getContentId()).isEqualTo(NEW_CONTENT_FILE_ID);
		}

		@Test
		void shouldNotChangeOtherFields() {
			var result = updateContentId();

			assertThat(result).usingRecursiveComparison().ignoringFields("contentId").isEqualTo(BINARY_FILE);
		}

		@Test
		void shouldReturnUnchangedBinaryFile() {
			var result = grpcService.updateContentId(BINARY_FILE, Optional.empty());

			assertThat(result).isSameAs(BINARY_FILE);
		}

		private OzgCloudFile updateContentId() {
			return grpcService.updateContentId(BINARY_FILE, Optional.of(NEW_CONTENT_FILE_ID));
		}
	}

	@Nested
	class TestBuildPatch {

		private static final Map<String, Object> PATCH_MAP = Map.of(
				OzgCloudFile.FIELD_FILE_NAME, OzgCloudFileTestFactory.NAME,
				OzgCloudFile.FIELD_CONTENT_TYPE, OzgCloudFileTestFactory.CONTENT_TYPE,
				OzgCloudFile.FIELD_CONTENT_ID, OzgCloudFileTestFactory.CONTENT_ID,
				OzgCloudFile.FIELD_CREATED_BY, OzgCloudFileTestFactory.CREATED_BY,
				OzgCloudFile.FIELD_CLIENT, OzgCloudFileTestFactory.CLIENT);

		@BeforeEach
		void init() {
			doReturn(OzgCloudFileTestFactory.CREATED_BY).when(grpcService).getUserId();
			doReturn(CallContextTestFactory.CLIENT).when(grpcService).getClient();
		}

		@Test
		void shouldBuildPatch() {
			var result = grpcService.buildPatch(OzgCloudFileTestFactory.create());

			assertThat(result).containsExactlyEntriesOf(PATCH_MAP);
		}
	}

	@Nested
	class TestUploadFileContent {

		private static final OzgCloudFile BINARY_FILE = OzgCloudFileTestFactory.create();
		private static final UploadedFilesReference UPLOADED_FILES_REFERENCE = UploadedFilesReferenceTestFactory.create();
		private static final CompletableFuture<String> FILE_ID_FUTURE = CompletableFuture.completedFuture(OzgCloudFileTestFactory.CONTENT_ID);

		@Mock
		private InputStream inputStream;

		@BeforeEach
		void init() {
			doReturn(UPLOADED_FILES_REFERENCE).when(grpcService).buildUploadFilesReferences(any());
			when(service.uploadFileContentStreamAsync(any(), any())).thenReturn(FILE_ID_FUTURE);
		}

		@Test
		void shouldCallBuildUploadFilesReferences() {
			uploadFileContent();

			verify(grpcService).buildUploadFilesReferences(BINARY_FILE);
		}

		@Test
		void shouldCallUploadFileContent() {
			uploadFileContent();

			verify(service).uploadFileContentStreamAsync(UPLOADED_FILES_REFERENCE, inputStream);
		}

		@Test
		void shouldReturnFutureFileId() {
			var result = uploadFileContent();

			assertThat(result).isSameAs(FILE_ID_FUTURE);
		}

		private CompletableFuture<String> uploadFileContent() {
			return grpcService.uploadFileContent(BINARY_FILE, inputStream);
		}
	}

	@Nested
	class TestBuildUploadFilesReferences {

		@Test
		void shouldSetVorgangId() {
			var uploadFilesReference = buildUploadFilesReferences();

			assertThat(uploadFilesReference.getVorgangId()).isEqualTo(VorgangTestFactory.ID);
		}

		@Test
		void shouldSetName() {
			var uploadFilesReference = buildUploadFilesReferences();

			assertThat(uploadFilesReference.getName()).isEqualTo(UploadedFilesReferenceTestFactory.NAME);
		}

		@Test
		void shouldSetClient() {
			var uploadFilesReference = buildUploadFilesReferences();

			assertThat(uploadFilesReference.getClient()).isEqualTo(CallContextTestFactory.CLIENT);
		}

		@Test
		void shouldSetFileName() {
			var uploadFilesReference = buildUploadFilesReferences();

			assertThat(uploadFilesReference.getFileName()).isEqualTo(UploadedFilesReferenceTestFactory.FILE_NAME);
		}

		private UploadedFilesReference buildUploadFilesReferences() {
			return grpcService.buildUploadFilesReferences(OzgCloudFileTestFactory.create());
		}
	}

	@Nested
	class TestGetClient {

		@Nested
		class TestWithRequest {

			@Nested
			class TestGetFromContext {

				@BeforeEach
				void init() {
					doReturn(Optional.of(CallContextTestFactory.CLIENT)).when(grpcService).findContextClient();
				}

				@Test
				void shouldCallFindContextClient() {
					getClient();

					verify(grpcService).findContextClient();
				}

				@Test
				void shouldReturnResult() {
					var result = getClient();

					assertThat(result).contains(CallContextTestFactory.CLIENT);
				}
			}

			@Nested
			class TestFromRequest {

				@BeforeEach
				void init() {
					doReturn(Optional.empty()).when(grpcService).findContextClient();
					doReturn(Optional.of(GrpcCallContextTestFactory.CLIENT)).when(grpcService).getRequestClient(any());
				}

				@Test
				void shouldCallGetRequestClient() {
					var grpcRequest = GrpcUploadBinaryFileRequestTestFactory.create();

					grpcService.getClient(grpcRequest);

					verify(grpcService).getRequestClient(grpcRequest);
				}

				@Test
				void shouldReturnClientName() {
					var result = getClient();

					assertThat(result).isEqualTo(GrpcCallContextTestFactory.CLIENT);
				}
			}

			@Nested
			class TestNoClient {

				@Test
				void shouldThrowException() {
					doReturn(Optional.empty()).when(grpcService).findContextClient();
					doReturn(Optional.empty()).when(grpcService).getRequestClient(any());

					assertThrows(IllegalStateException.class, TestWithRequest.this::getClient);
				}
			}

			private String getClient() {
				return grpcService.getClient(GrpcUploadBinaryFileRequestTestFactory.create());
			}
		}

		@Nested
		class TestFromContext {

			@Test
			void shouldCallFindUser() {
				doReturn(Optional.of(CallContextTestFactory.CLIENT)).when(grpcService).findContextClient();

				grpcService.getClient();

				verify(grpcService).findContextClient();
			}

			@Test
			void shouldReturnClientName() {
				doReturn(Optional.of(CallContextTestFactory.CLIENT)).when(grpcService).findContextClient();

				var result = grpcService.getClient();

				assertThat(result).isEqualTo(CallContextTestFactory.CLIENT);
			}

			@Test
			void shouldThrowException() {
				doReturn(Optional.empty()).when(grpcService).findContextClient();

				assertThrows(IllegalStateException.class, grpcService::getClient);
			}
		}
	}

	@Nested
	class TestFindContextClient {

		@Nested
		class TestWithUser {

			@BeforeEach
			void init() {
				when(currentUserService.findUser()).thenReturn(Optional.of(CallContextUserTestFactory.create()));
			}

			@Test
			void shouldCallGetUser() {
				findContextClient();

				verify(currentUserService).findUser();
			}

			@Test
			void shouldReturnClientName() {
				var result = findContextClient();

				assertThat(result).contains(CallContextTestFactory.CLIENT);
			}
		}

		@Nested
		class TestWithoutUser {

			@BeforeEach
			void init() {
				when(currentUserService.findUser()).thenReturn(Optional.empty());
			}

			@Test
			void shouldReturnEmpty() {
				var result = findContextClient();

				assertThat(result).isEmpty();
			}
		}

		private Optional<String> findContextClient() {
			return grpcService.findContextClient();
		}
	}

	@Nested
	class TestGetRequestClient {

		@Test
		void shouldReturnClientName() {
			var result = grpcService.getRequestClient(GrpcUploadBinaryFileRequestTestFactory.create());

			assertThat(result).contains(GrpcCallContextTestFactory.CLIENT);
		}
	}

	@Nested
	class TestCompleteRequest {

		@Mock
		private StreamObserver<TestResponse> responseObserver;
		@Mock
		private TestResponse response;

		@Test
		void shouldCallOnNext() {
			completeRequest();

			verify(responseObserver).onNext(response);
		}

		@Test
		void shouldCallOnComplete() {
			completeRequest();

			verify(responseObserver).onCompleted();
		}

		private void completeRequest() {
			grpcService.completeRequest(response, responseObserver);
		}

		private interface TestResponse {
		}
	}

	@SuppressWarnings("unchecked")
	private <T> void callMetadataVerifier(StreamObserver<T> uploadObserver, T request) {
		var metadataVerifier = (Consumer<T>) ReflectionTestUtils.getField(uploadObserver, "requestVerifier");
		metadataVerifier.accept(request);
	}

	@SuppressWarnings("unchecked")
	private <T> void callHasMetadata(StreamObserver<T> uploadObserver, T request) {
		var hasMetadata = (Predicate<T>) ReflectionTestUtils.getField(uploadObserver, "hasMetadata");
		hasMetadata.test(request);
	}

	@SuppressWarnings("unchecked")
	private <T> void callMetadataMapper(StreamObserver<T> uploadObserver, T request) {
		var metadataMapper = (Function<T, OzgCloudFile>) ReflectionTestUtils.getField(uploadObserver, "metadataMapper");
		metadataMapper.apply(request);
	}

	@SuppressWarnings("unchecked")
	private <T> void callMetadataUploader(StreamObserver<T> uploadObserver) {
		var metadataUploader = (BiFunction<OzgCloudFile, Optional<String>, FileId>) ReflectionTestUtils.getField(uploadObserver, "metadataUploader");
		metadataUploader.apply(BINARY_FILE, Optional.of(NEW_CONTENT_FILE_ID));
	}

	@SuppressWarnings("unchecked")
	private <T> void callFileContentUploader(StreamObserver<T> uploadObserver) {
		var fileContentUploader = (BiFunction<OzgCloudFile, InputStream, CompletableFuture<FileId>>)
				ReflectionTestUtils.getField(uploadObserver, "fileContentUploader");
		fileContentUploader.apply(BINARY_FILE, new ByteArrayInputStream(OzgCloudFileTestFactory.CONTENT));
	}

	@SuppressWarnings("unchecked")
	private <T> void callGetFileContent(StreamObserver<T> uploadObserver, T request) {
		var getFileContent = (Function<T, byte[]>) ReflectionTestUtils.getField(uploadObserver, "getFileContent");
		getFileContent.apply(request);
	}

	@SuppressWarnings("unchecked")
	private <T> void callResponseConsumer(StreamObserver<T> uploadObserver) {
		var responseConsumer = (Consumer<FileId>) ReflectionTestUtils.getField(uploadObserver, "responseConsumer");
		responseConsumer.accept(OzgCloudFileTestFactory.ID);
	}

	@SuppressWarnings("unchecked")
	private <T> void callCleanupFileContent(StreamObserver<T> uploadObserver) {
		var cleanupFileContent = (Consumer<String>) ReflectionTestUtils.getField(uploadObserver, "cleanupFileContent");
		cleanupFileContent.accept(OzgCloudFileTestFactory.CONTENT_ID);
	}
}