/*
 * Copyright (C) 2022 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.files;

import static org.assertj.core.api.Assertions.*;

import java.io.ByteArrayInputStream;
import java.io.FileNotFoundException;
import java.util.Collection;
import java.util.ConcurrentModificationException;
import java.util.List;
import java.util.Map;
import java.util.Random;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.mongodb.core.MongoOperations;

import com.thedeanda.lorem.LoremIpsum;

import de.ozgcloud.common.test.DataITCase;
import de.ozgcloud.vorgang.common.errorhandling.NotFoundException;
import de.ozgcloud.vorgang.vorgang.IncomingFileTestFactory;
import de.ozgcloud.vorgang.vorgang.VorgangTestFactory;

@DataITCase
class OzgCloudFileRepositoryITCase {

	@Autowired
	private OzgCloudFileRepository repository;

	@Autowired
	private MongoOperations mongoOperations;
	@Autowired
	private OzgCloudFileTestUtil ozgCloudFileTestUtil;

	@BeforeEach
	void clearDB() {
		mongoOperations.dropCollection(OzgCloudFile.COLLECTION_NAME);
		mongoOperations.dropCollection("fs.files");
		mongoOperations.dropCollection("fs.chunks");
	}

	@Nested
	class TestSave {

		private OzgCloudFile ozgCloudFile;

		@BeforeEach
		void init() {
			var contentId = ozgCloudFileTestUtil.storeFileContent(OzgCloudFileTestFactory.CONTENT);
			ozgCloudFile = OzgCloudFileTestFactory.createBuilder().id(null).version(0).contentId(contentId).build();
		}

		@Test
		void shouldSaveFile() {
			var result = repository.save(ozgCloudFile);

			assertThat(result).usingRecursiveComparison().ignoringFields("id", OzgCloudFile.FIELD_VERSION).isEqualTo(ozgCloudFile);
			var savedOzgCloudFile = mongoOperations.findById(result.getId(), OzgCloudFile.class);
			assertThat(result).usingRecursiveComparison().isEqualTo(savedOzgCloudFile);
		}

		@Test
		void shouldIncreaseVersion() {
			var result = repository.save(ozgCloudFile);

			assertThat(result.getVersion()).isEqualTo(1);
		}

		@Test
		void shouldSetContentSize() {
			var ozgCloudFile = this.ozgCloudFile.toBuilder().size(new Random().nextLong()).build();

			var result = repository.save(ozgCloudFile);

			assertThat(result.getSize()).isEqualTo(OzgCloudFileTestFactory.SIZE);
		}

		@Test
		void shouldSetContentSizeToZero() {
			var ozgCloudFile = OzgCloudFileTestFactory.createBuilder().id(null).version(0).contentId(null).build();

			var result = repository.save(ozgCloudFile);

			assertThat(result.getSize()).isZero();
		}

		@Test
		void shouldFailIfContentMissing() {
			var ozgCloudFile = OzgCloudFileTestFactory.createBuilder().id(null).version(0).contentId("missing").build();

			assertThatThrownBy(() -> repository.save(ozgCloudFile)).isInstanceOf(NotFoundException.class);
		}
	}

	@Nested
	class TestUploadContent {

		@Test
		void shouldStoreFiles() {
			var content = IncomingFileTestFactory.getFilledChunk(45);

			uploadContent(content);

			assertThat(ozgCloudFileTestUtil.getFileContentByFileName()).hasBinaryContent(content);
		}

		private void uploadContent(byte[] content) {
			repository.uploadContent(UploadedFilesReferenceTestFactory.create(), new ByteArrayInputStream(content));
		}
	}

	@Nested
	class TestGetFileContent {

		@Test
		void shouldLoadFile() {
			var fileContent = IncomingFileTestFactory.getFilledChunk(45);
			var ozgCloudFileId = FileId.from(ozgCloudFileTestUtil.saveOzgCloudFile(fileContent).getId());

			var result = repository.getFileContent(ozgCloudFileId);

			assertThat(result).hasBinaryContent(fileContent);
		}

		@Test
		void shouldThrowExceptionWhenFileNotFound() {
			var fileId = FileId.from("not-existing");

			assertThatThrownBy(() -> repository.getFileContent(fileId))
					.isInstanceOf(NotFoundException.class)
					.hasMessageContaining(fileId.toString());
		}
	}

	@Nested
	class TestFindAll {

		private OzgCloudFile ozgCloudFile;

		@BeforeEach
		void initData() {
			ozgCloudFile = mongoOperations.save(OzgCloudFileTestFactory.createBuilder().id(null).version(0).build());
		}

		@Test
		void shouldReturnFile() {
			var result = findAll(List.of(ozgCloudFile.getId()));

			assertThat(result).hasSize(1).first().usingRecursiveComparison().isEqualTo(ozgCloudFile);
		}

		@Test
		void shouldReturnAllFiles() {
			var otherFileId = mongoOperations.save(OzgCloudFileTestFactory.createBuilder().id(null).version(0).build());

			var result = findAll(List.of(ozgCloudFile.getId(), otherFileId.getId()));

			assertThat(result).hasSize(2).usingRecursiveFieldByFieldElementComparator().containsExactly(ozgCloudFile, otherFileId);
		}

		private List<OzgCloudFile> findAll(Collection<String> fileIds) {
			return repository.findAll(fileIds.stream().map(FileId::from).toList()).toList();
		}
	}

	@Nested
	class TestDeleteAllByVorgang {

		@Test
		void shouldDeleteFiles() {
			ozgCloudFileTestUtil.saveOzgCloudFile();

			repository.deleteAllByVorgang(VorgangTestFactory.ID);

			var loaded = mongoOperations.findAll(OzgCloudFile.class);
			assertThat(loaded).isEmpty();
			assertThatThrownBy(ozgCloudFileTestUtil::getFileContentByFileName).cause().isInstanceOf(FileNotFoundException.class);
		}

		@Test
		void shouldNotDeleteOtherFiles() {
			var fileContent = IncomingFileTestFactory.getFilledChunk(1);
			var ozgCloudFile = ozgCloudFileTestUtil.saveOzgCloudFile(fileContent);

			repository.deleteAllByVorgang("other");

			var loaded = mongoOperations.findById(ozgCloudFile.getId(), OzgCloudFile.class);
			assertThat(loaded).usingRecursiveComparison().isEqualTo(ozgCloudFile);
			assertThat(ozgCloudFileTestUtil.getFileContentByFileName()).hasBinaryContent(fileContent);
		}
	}

	@Nested
	class TestPatch {

		private static final String NEW_NAME = "newName";
		private static final Map<String, Object> PATCH_MAP = Map.of(OzgCloudFile.FIELD_FILE_NAME, NEW_NAME);

		private OzgCloudFile ozgCloudFile;

		@BeforeEach
		void init() {
			ozgCloudFile = ozgCloudFileTestUtil.saveOzgCloudFile();
		}

		@Test
		void shouldPatchOzgCloudFile() {
			repository.patch(FileId.from(ozgCloudFile.getId()), ozgCloudFile.getVersion(), PATCH_MAP);

			var loaded = mongoOperations.findById(ozgCloudFile.getId(), OzgCloudFile.class);
			assertThat(loaded.getName()).isEqualTo(NEW_NAME);
		}

		@Test
		void shouldIncreaseVersion() {
			repository.patch(FileId.from(ozgCloudFile.getId()), ozgCloudFile.getVersion(), PATCH_MAP);

			var loaded = mongoOperations.findById(ozgCloudFile.getId(), OzgCloudFile.class);
			assertThat(loaded.getVersion()).isEqualTo(2);
		}

		@Test
		void shouldPatchContentSize() {
			var content = LoremIpsum.getInstance().getWords(23).getBytes();
			var newContentId = ozgCloudFileTestUtil.storeFileContent(content);
			var patchMap = Map.<String, Object>of(OzgCloudFile.FIELD_CONTENT_ID, newContentId);

			repository.patch(FileId.from(ozgCloudFile.getId()), ozgCloudFile.getVersion(), patchMap);

			var loaded = mongoOperations.findById(ozgCloudFile.getId(), OzgCloudFile.class);
			assertThat(loaded.getSize()).isEqualTo(content.length);
		}

		@Test
		void shouldThrowConcurrentModificationException() {
			var ozgCloudFile = ozgCloudFileTestUtil.saveOzgCloudFile();

			var id = FileId.from(ozgCloudFile.getId());
			var version = ozgCloudFile.getVersion() - 1;
			assertThatThrownBy(() -> repository.patch(id, version, PATCH_MAP)).isInstanceOf(ConcurrentModificationException.class);
		}
	}

	@Nested
	class TestDeleteContent {

		@Test
		void shouldDeleteContent() {
			var contentId = ozgCloudFileTestUtil.storeFileContent(OzgCloudFileTestFactory.CONTENT);

			repository.deleteContent(contentId);

			assertThatThrownBy(() -> ozgCloudFileTestUtil.getFileContentByFileName()).cause().isInstanceOf(FileNotFoundException.class);
		}
	}
}