/*
 * Copyright (C) 2022 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.files;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.Mockito.*;

import java.io.InputStream;

import org.bson.types.ObjectId;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;
import org.springframework.data.mongodb.gridfs.GridFsTemplate;
import org.springframework.data.mongodb.gridfs.GridFsUpload;

import com.thedeanda.lorem.LoremIpsum;

import de.ozgcloud.vorgang.callcontext.CallContextTestFactory;

class OzgCloudFileRepositoryTest {

	private static final UploadedFilesReference UPLOAD_REFERENCE = UploadedFilesReferenceTestFactory.create();

	@Spy
	@InjectMocks
	private OzgCloudFileRepository repository;

	@Mock
	private GridFsTemplate gridFsTemplate;
	@Mock
	private InputStream contentStream;

	@Nested
	class TestAddContentStream {

		private static final ObjectId OBJECT_ID = new ObjectId();

		@Mock
		private GridFsUpload<ObjectId> gridFsUpload;

		@BeforeEach
		void init() {
			doReturn(gridFsUpload).when(repository).buildGridFsUpload(any(), any());
			when(gridFsTemplate.store(any())).thenReturn(OBJECT_ID);
		}

		@Test
		void shouldCallBuildGridFsUpload() {
			addContentStream();

			verify(repository).buildGridFsUpload(UPLOAD_REFERENCE, contentStream);
		}

		@Test
		void shouldCallGridFsTemplateStore() {
			addContentStream();

			verify(gridFsTemplate).store(gridFsUpload);
		}

		@Test
		void shouldReturnFileId() {

			var result = addContentStream();

			assertThat(result).hasToString(OBJECT_ID.toString());
		}

		private String addContentStream() {
			return repository.uploadContent(UPLOAD_REFERENCE, contentStream);
		}
	}

	@Nested
	class TestBuildGridFsUpload {

		@Test
		void shouldSetContent() {

			var result = buildGridFsUpload();

			assertThat(result.getContent()).isSameAs(contentStream);
		}

		@Test
		void shouldCallCreateFilePath() {
			buildGridFsUpload();

			verify(repository).createFilePath(UPLOAD_REFERENCE);
		}

		@Test
		void shouldSetFilename() {
			var fileName = LoremIpsum.getInstance().getName();
			doReturn(fileName).when(repository).createFilePath(UPLOAD_REFERENCE);

			var result = buildGridFsUpload();

			assertThat(result.getFilename()).isEqualTo(fileName);
		}

		private GridFsUpload<ObjectId> buildGridFsUpload() {
			return repository.buildGridFsUpload(UPLOAD_REFERENCE, contentStream);
		}
	}

	@Nested
	class TestCreateFilePath {
		@Test
		void shouldCreatePath() {
			String path = repository.createFilePath(UploadedFilesReferenceTestFactory.create());

			assertThat(path).isEqualTo(
					UploadedFilesReferenceTestFactory.VORGANG_ID + "/" +
							CallContextTestFactory.CLIENT + "/" +
							UploadedFilesReferenceTestFactory.NAME + "/" +
							OzgCloudFileTestFactory.NAME);
		}
	}

}
