/*
 * Copyright (C) 2022 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.files;

import static org.assertj.core.api.Assertions.*;
import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.io.InputStream;
import java.io.PipedInputStream;
import java.io.PipedOutputStream;
import java.util.Optional;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;
import java.util.concurrent.atomic.AtomicReference;
import java.util.function.BiFunction;
import java.util.function.Consumer;
import java.util.function.Function;
import java.util.function.Predicate;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.ValueSource;
import org.mockito.Mock;
import org.mockito.Spy;
import org.springframework.test.util.ReflectionTestUtils;

import de.ozgcloud.common.errorhandling.TechnicalException;
import lombok.SneakyThrows;

class UploadServerStreamObserverTest {

	private static final String FIELD_CONTENT_ID_FUTURE = "contentIdFuture";
	private static final String FIELD_FILE_METADATA = "fileMetadata";
	private static final String FIELD_PIPED_OUTPUT = "pipedOutput";

	@Mock
	private Consumer<UploadTestRequest> requestVerifier;
	@Mock
	private Predicate<UploadTestRequest> hasMetadata;
	@Mock
	private BiFunction<MetadataTest, Optional<String>, FileId> metadataUploader;
	@Mock
	private BiFunction<MetadataTest, InputStream, CompletableFuture<String>> fileContentUploader;
	@Mock
	private Function<UploadTestRequest, byte[]> getFileContent;
	@Mock
	private Function<UploadTestRequest, MetadataTest> metadataMapper;
	@Mock
	private Consumer<FileId> responseConsumer;
	@Mock
	private Consumer<String> cleanupFileContent;
	@Mock
	private UploadTestRequest fileUploadRequest;

	private UploadServerStreamObserver<MetadataTest, UploadTestRequest> uploadStreamObserver;

	@BeforeEach
	void init() {
		// Ich habe es anders nicht hinbekommen den Observer zu initialisieren.
		// Wenn @InjectMocks verwendet wird, schafft mockito nicht die Mocks richtig zu initialisieren und \
		// wenn TestOnCompleted.shouldCallResponseConsumer gestartet wird, wird metadataVerifier als responseConsumer gesetzt.
		// Und in TestStoreMetadata#shouldCallMetadataVerifier passiert es umgekehrt.
		// In Debug-Modus funktioniert alles richtig.
		uploadStreamObserver = spy(UploadServerStreamObserver.<MetadataTest, UploadTestRequest>builder()
				.requestVerifier(requestVerifier)
				.hasMetadata(hasMetadata)
				.metadataUploader(metadataUploader)
				.fileContentUploader(fileContentUploader)
				.getFileContent(getFileContent)
				.metadataMapper(metadataMapper)
				.responseConsumer(responseConsumer)
				.cleanupFileContent(cleanupFileContent)
				.build());
	}

	private interface UploadTestRequest {
	}

	private interface MetadataTest {
	}

	@Nested
	class TestOnNext {

		@Nested
		class TestWithMetadata {

			@BeforeEach
			void init() {
				when(hasMetadata.test(any())).thenReturn(true);
				doNothing().when(uploadStreamObserver).setMetadata(any());
			}

			@Test
			void shouldCallSetMetadata() {
				onNext();

				verify(uploadStreamObserver).setMetadata(fileUploadRequest);
			}

			@Test
			void shouldNotCallStoreFileContent() {
				onNext();

				verify(uploadStreamObserver, never()).storeFileContent(any());
			}
		}

		@Nested
		class TestWithFileContent {

			private final CompletableFuture<String> contentIdFuture = CompletableFuture.completedFuture(OzgCloudFileTestFactory.CONTENT_ID);

			@BeforeEach
			void init() {
				when(hasMetadata.test(any())).thenReturn(false);
				doNothing().when(uploadStreamObserver).storeFileContent(any());
				doReturn(contentIdFuture).when(uploadStreamObserver).initContentUpload();
			}

			@Test
			void shouldCallStoreFileContent() {
				onNext();

				verify(uploadStreamObserver).storeFileContent(fileUploadRequest);
			}

			@Test
			void shouldNotCallStoreMetadata() {
				onNext();

				verify(uploadStreamObserver, never()).setMetadata(any());
			}

			@Test
			void shouldCallInitContentUploadOnce() {
				onNext();
				onNext();

				verify(uploadStreamObserver).initContentUpload();
			}

			@Test
			void shouldSetContentIdFuture() {
				onNext();

				assertThat(uploadStreamObserver).extracting(FIELD_CONTENT_ID_FUTURE, as(COMPLETABLE_FUTURE)).isCompletedWithValue(OzgCloudFileTestFactory.CONTENT_ID);
			}
		}

		private void onNext() {
			uploadStreamObserver.onNext(fileUploadRequest);
		}
	}

	@Nested
	class TestInitContentUpload {

		private final CompletableFuture<String> contentIdFuture = CompletableFuture.completedFuture(OzgCloudFileTestFactory.CONTENT_ID);

		@Mock
		private PipedInputStream pipedInputStream;
		@Mock
		private MetadataTest metadata;

		@BeforeEach
		void init() {
			when(fileContentUploader.apply(any(), any())).thenReturn(contentIdFuture);
			doReturn(pipedInputStream).when(uploadStreamObserver).initPipedStreams();
			ReflectionTestUtils.setField(uploadStreamObserver, FIELD_FILE_METADATA, new AtomicReference<>(metadata));
		}

		@Test
		void shouldCallInitPipedStreams() {
			initContentUpload();

			verify(uploadStreamObserver).initPipedStreams();
		}

		@Test
		void shouldCallFileContentUploader() {
			initContentUpload();

			verify(fileContentUploader).apply(metadata, pipedInputStream);
		}

		private CompletableFuture<String> initContentUpload() {
			return uploadStreamObserver.initContentUpload();
		}
	}

	@Nested
	class TestSetMetadata {

		@Mock
		private PipedInputStream pipedInputStream;
		@Mock
		private CompletableFuture<FileId> fileIdFuture;
		@Mock
		private MetadataTest metadata;

		@BeforeEach
		void init() {
			when(metadataMapper.apply(any())).thenReturn(metadata);
		}

		@Test
		void shouldCallMetadataVerifier() {
			setMetadata();

			verify(requestVerifier).accept(fileUploadRequest);
		}

		@Test
		void shouldCallMetadataMapper() {
			setMetadata();

			verify(metadataMapper).apply(fileUploadRequest);
		}

		@Test
		void shouldCallSetFileMetadata() {
			setMetadata();

			assertThat(uploadStreamObserver).extracting(FIELD_FILE_METADATA, as(ATOMIC_REFERENCE)).hasValue(metadata);
		}

		@Test
		void shouldThrowExceptionIfMetadataAlreadyStored() {
			setMetadata();

			assertThrows(OzgCloudFileUploadException.class, this::setMetadata);
			assertThrows(OzgCloudFileUploadException.class, this::setMetadata);
		}

		private void setMetadata() {
			uploadStreamObserver.setMetadata(fileUploadRequest);
		}
	}

	@Nested
	class TestStoreFileContent {

		@Nested
		class TestStore {

			@Mock
			private PipedOutputStream pipedOutputStream;
			@Mock
			private MetadataTest metadata;

			@BeforeEach
			void init() {
				ReflectionTestUtils.setField(uploadStreamObserver, FIELD_PIPED_OUTPUT, pipedOutputStream);
				ReflectionTestUtils.setField(uploadStreamObserver, FIELD_FILE_METADATA, new AtomicReference<>(metadata));
			}

			@Test
			void shouldCallGetFileContent() {
				storeFileContent();

				verify(getFileContent).apply(fileUploadRequest);
			}

			@SneakyThrows
			@Test
			void shouldWriteFileContent() {
				byte[] fileContent = new byte[] { 1, 2, 3 };
				doReturn(fileContent).when(getFileContent).apply(any());

				storeFileContent();

				verify(pipedOutputStream).write(fileContent);
			}
		}

		@Nested
		class TestWithException {

			@Test
			void shouldThrowExceptionIfNoMetadata() {
				assertThrows(OzgCloudFileUploadException.class, TestStoreFileContent.this::storeFileContent);
			}
		}

		private void storeFileContent() {
			uploadStreamObserver.storeFileContent(fileUploadRequest);
		}
	}

	@Nested
	class TestOnError {

		@Mock
		private Throwable throwable;
		@Mock
		private MetadataTest metadata;

		@BeforeEach
		void init() {
			ReflectionTestUtils.setField(uploadStreamObserver, FIELD_FILE_METADATA, new AtomicReference<>(metadata));
		}

		@Test
		void shouldCallCloseOutputPipe() {
			uploadStreamObserver.onError(throwable);

			verify(uploadStreamObserver).closeOutputPipe();
		}

		@Test
		void shouldCloseInputPipe() {
			uploadStreamObserver.onError(throwable);

			verify(uploadStreamObserver).closeInputPipe();
		}
	}

	@Nested
	class TestOnCompleted {

		@Mock
		private MetadataTest metadata;

		@BeforeEach
		void init() {
			ReflectionTestUtils.setField(uploadStreamObserver, FIELD_FILE_METADATA, new AtomicReference<>(metadata));
			doReturn(Optional.of(OzgCloudFileTestFactory.CONTENT_ID)).when(uploadStreamObserver).getContentId();
		}

		@Test
		void shouldCallGetContentId() {
			uploadStreamObserver.onCompleted();

			verify(uploadStreamObserver).getContentId();
		}

		@Test
		void shouldCallMetadataUploader() {
			uploadStreamObserver.onCompleted();

			verify(metadataUploader).apply(metadata, Optional.of(OzgCloudFileTestFactory.CONTENT_ID));
		}

		@Test
		void shouldCallResponseConsumer() {
			when(metadataUploader.apply(any(), any())).thenReturn(OzgCloudFileTestFactory.ID);

			uploadStreamObserver.onCompleted();

			verify(responseConsumer).accept(OzgCloudFileTestFactory.ID);
		}

		@Test
		void shouldThrowException() {
			when(metadataUploader.apply(any(), any())).thenThrow(new RuntimeException());

			assertThrows(TechnicalException.class, uploadStreamObserver::onCompleted);
		}

		@Test
		void shouldCallCleanupFileContent() {
			when(metadataUploader.apply(any(), any())).thenThrow(new RuntimeException());

			try {
				uploadStreamObserver.onCompleted();
			} catch (TechnicalException e) {
				// expected
			}

			verify(cleanupFileContent).accept(OzgCloudFileTestFactory.CONTENT_ID);
		}

		@Test
		void shouldCallCloseOutputPipe() {
			uploadStreamObserver.onCompleted();

			verify(uploadStreamObserver).closeOutputPipe();
		}

	}

	@Nested
	class TestGetContentId {

		@Nested
		class TestWithContentId {

			@Spy
			private CompletableFuture<String> contentIdFuture = CompletableFuture.completedFuture(OzgCloudFileTestFactory.CONTENT_ID);

			@BeforeEach
			@SneakyThrows
			void init() {
				ReflectionTestUtils.setField(uploadStreamObserver, FIELD_CONTENT_ID_FUTURE, contentIdFuture);
			}

			@SneakyThrows
			@Test
			void shouldCallGetContentIdFuture() {
				uploadStreamObserver.getContentId();

				verify(contentIdFuture).get(UploadServerStreamObserver.TIMEOUT_MINUTES, TimeUnit.MINUTES);
			}

			@Test
			void shouldCallCloseInputPipe() {
				uploadStreamObserver.getContentId();

				verify(uploadStreamObserver).closeInputPipe();
			}

			@Test
			void shouldReturnContentId() {
				var result = uploadStreamObserver.getContentId();

				assertThat(result).hasValue(OzgCloudFileTestFactory.CONTENT_ID);
			}

			@SneakyThrows
			@DisplayName("should throw exception and close input pipe when")
			@ParameterizedTest(name = "{0} thrown")
			@ValueSource(classes = { ExecutionException.class, TimeoutException.class, InterruptedException.class })
			void shouldThrowException(Class<? extends Exception> exceptionClass) {
				doThrow(exceptionClass).when(contentIdFuture).get(anyLong(), any());

				assertThrows(TechnicalException.class, uploadStreamObserver::getContentId);
				verify(uploadStreamObserver).closeInputPipe();
			}
		}

		@Test
		void shouldReturnWhenNoContentId() {
			var result = uploadStreamObserver.getContentId();

			assertThat(result).isEmpty();
		}

	}
}
