/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.statistic;

import static org.assertj.core.api.Assertions.*;
import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.Mockito.*;

import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.Spy;

class CountStatisticMapperTest {

	@Spy
	private CountStatisticMapper mapper = new CountStatisticMapper();

	@DisplayName("Map to")
	@Nested
	class TestMapTo {

		@DisplayName("path count result")
		@Nested
		class TestPathCountResult {

			@Test
			void shouldHaveSize() {
				var response = mapper.mapTo(StatisticTestFactory.create());

				assertThat(response.getPathCountResultList()).hasSize(1);
			}

			@Test
			void shouldHaveSetName() {
				var response = mapper.mapTo(StatisticTestFactory.create());

				assertThat(response.getPathCountResultList().get(0).getName()).isEqualTo(StatisticTestFactory.FIELD_PATH_ALIAS);
			}

			@Test
			void shouldHaveSetValue() {
				var response = mapper.mapTo(StatisticTestFactory.create());

				assertThat(response.getPathCountResultList().get(0).getValue()).isEqualTo(StatisticTestFactory.FIELD_PATH_VALUE);
			}
		}

		@Nested
		@DisplayName("by status")
		class TestMapByStatus {

			@Test
			@DisplayName("should map NEU status")
			void shouldMapNeu() {
				var response = mapper.mapTo(StatisticTestFactory.create());

				assertThat(response.getByStatus().getNeu()).isEqualTo(StatisticTestFactory.NEU_COUNT);
			}

			@Test
			@DisplayName("should map ANGENOMMEN status")
			void shouldMapAngenommen() {
				var response = mapper.mapTo(StatisticTestFactory.create());

				assertThat(response.getByStatus().getAngenommen()).isEqualTo(StatisticTestFactory.ANGENOMMEN_COUNT);
			}

			@Test
			@DisplayName("should map VERWORFEN status")
			void shouldMapVerworfen() {
				var response = mapper.mapTo(StatisticTestFactory.create());

				assertThat(response.getByStatus().getVerworfen()).isEqualTo(StatisticTestFactory.VERWORFEN_COUNT);
			}

			@Test
			@DisplayName("should map IN_BEARBEITUNG status")
			void shouldMapInBearbeitung() {
				var response = mapper.mapTo(StatisticTestFactory.create());

				assertThat(response.getByStatus().getInBearbeitung()).isEqualTo(StatisticTestFactory.IN_BEARBEITUNG_COUNT);
			}

			@Test
			@DisplayName("should map BESCHIEDEN status")
			void shouldMapBeschieden() {
				var response = mapper.mapTo(StatisticTestFactory.create());

				assertThat(response.getByStatus().getBeschieden()).isEqualTo(StatisticTestFactory.BESCHIEDEN_COUNT);
			}

			@Test
			@DisplayName("should map ABGESCHLOSSEN status")
			void shouldMapAbgeschlossen() {
				var response = mapper.mapTo(StatisticTestFactory.create());

				assertThat(response.getByStatus().getAbgeschlossen()).isEqualTo(StatisticTestFactory.ABGESCHLOSSEN_COUNT);
			}

			@Test
			@DisplayName("should map WEITERGELEITET status")
			void shouldMapWeitergeleitet() {
				var response = mapper.mapTo(StatisticTestFactory.create());

				assertThat(response.getByStatus().getWeitergeleitet()).isEqualTo(StatisticTestFactory.WEITERGELEITET_COUNT);
			}
		}
	}

	@DisplayName("Map from gRPC request")
	@Nested
	class TestMapToRequest {

		private final String countPath = CountResultTestFactory.NAME + ":" + CountPathRequestTestFactory.FIELD_PATH;

		@Test
		void shouldMapPath() {
			var countRequest = mapper.mapPathCountRequest(countPath);

			assertThat(countRequest).usingRecursiveComparison().isEqualTo(CountPathRequestTestFactory.create());
		}

		@Test
		void shouldMapGrpcRequest() {
			doReturn(CountPathRequestTestFactory.create()).when(mapper).mapPathCountRequest(anyString());
			var grpcRequest = GrpcVorgangCountRequest.newBuilder().addCountByPath(countPath).build();

			var request = mapper.mapFromVorgangCountRequest(grpcRequest);

			assertThat(request.getCountByPaths()).hasSize(1).first().usingRecursiveComparison().isEqualTo(CountPathRequestTestFactory.create());
		}

		@Nested
		@DisplayName("Extract request parts")
		class TestExtract {

			@Test
			void shouldExtractFieldPath() {
				var fieldPath = mapper.extractFieldPath(countPath);

				assertThat(fieldPath).isEqualTo(CountPathRequestTestFactory.FIELD_PATH);
			}

			@Test
			void shouldThrowExceptionOnInvalidPath() {
				assertThrows(VorgangStatisticBadRequestException.class, () -> mapper.extractFieldPath(""));
			}

			@Test
			void shouldExtractAlias() {
				var alias = mapper.extractAlias(countPath);

				assertThat(alias).isEqualTo(CountResultTestFactory.NAME);
			}

			@Test
			void shouldThrowExceptionOnInvalidAlias() {
				assertThrows(VorgangStatisticBadRequestException.class, () -> mapper.extractAlias("invalid"));
			}
		}
	}
}