/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.statistic;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import de.ozgcloud.vorgang.common.operator.OperandFunctionParser;

class StatisticMapperTest {

	@Spy
	@InjectMocks
	private StatisticMapper mapper;

	@Mock
	private OperandFunctionParser operandFunctionParser;

	@Nested
	class TestMapFromVorgangStatisticRequest {

		@Test
		void shouldCallMapFromVorgangStatisticQuery() {
			var statisticQuery = GrpcVorgangStatisticQuery.newBuilder().setPath(VorgangStatisticRequestTestFactory.REQUEST_FIELD_PATH)
					.setOperandIntValue(1).build();
			var grpcRequest = GrpcVorgangStatisticRequest.newBuilder().addQuery(statisticQuery).build();

			mapper.mapFromVorgangStatisticRequest(grpcRequest);

			verify(mapper).mapFromVorgangStatisticQuery(statisticQuery);
		}

		@Nested
		@DisplayName("Map from VorgangStatisticQuery")
		class TestMapFromVorgangStatisticQuery {

			@BeforeEach
			void setup() {
			}

			@Test
			void shouldSetResultName() {
				var vorgangStatisticRequest = mapper.mapFromVorgangStatisticQuery(GrpcVorgangStatisticQueryTestFactory.create());

				assertThat(vorgangStatisticRequest.getResultName()).isEqualTo(GrpcVorgangStatisticQueryTestFactory.RESULT_NAME);
			}

			@Test
			void shouldCallGetFieldPath() {
				var statisticQuery = GrpcVorgangStatisticQueryTestFactory.create();

				mapper.mapFromVorgangStatisticQuery(statisticQuery);

				verify(mapper, times(1)).getFieldPath(statisticQuery);
			}

			@Test
			void shouldSetGroupMethod() {
				var vorgangStatisticRequest = mapper.mapFromVorgangStatisticQuery(GrpcVorgangStatisticQueryTestFactory.create());

				assertThat(vorgangStatisticRequest.getStatisticGroupMethod()).isEqualTo(
						StatisticGroupMethod.from(GrpcVorgangStatisticQueryTestFactory.GROUP_METHOD));
			}

			@Test
			void shouldCallBuildStatisticFilter() {
				GrpcVorgangStatisticQuery statisticQuery = GrpcVorgangStatisticQueryTestFactory.create();

				mapper.mapFromVorgangStatisticQuery(statisticQuery);

				verify(mapper).buildOperator(statisticQuery);
			}

			@Nested
			class TestGetFieldPath {

				@Test
				void shouldReturnOriginPath() {
					var statisticQuery = GrpcVorgangStatisticQueryTestFactory.create();

					var fieldPath = mapper.getFieldPath(statisticQuery);

					assertThat(fieldPath).isEqualTo(GrpcVorgangStatisticQueryTestFactory.QUERY_PATH);
				}

				@Test
				@DisplayName("should extend field path by '.value' parameter begins with 'clientAttributes'")
				void shouldExtendPath() {
					var statisticQuery = GrpcVorgangStatisticQueryTestFactory.createBuilder().setPath("clientAttributes.path").build();

					var fieldPath = mapper.getFieldPath(statisticQuery);

					assertThat(fieldPath).endsWith(".value");
				}
			}
		}

		@Nested
		class TestBuildOperator {

			@Test
			void shouldSetFieldPath() {
				var operator = mapper.buildOperator(GrpcVorgangStatisticQueryTestFactory.create());

				assertThat(operator).extracting("fieldPath").isEqualTo(GrpcVorgangStatisticQueryTestFactory.PATH);
			}

			@Test
			void shouldExtendFieldPathWhenClientAttribute() {
				var statisticQuery = GrpcVorgangStatisticQueryTestFactory.createBuilder().setPath("ClientAttribute.path").build();

				var operator = mapper.buildOperator(statisticQuery);

				assertThat(operator).extracting("fieldPath").isEqualTo("clientAttributes.path.value");
			}

			@Test
			void shouldSetOperand() {
				var operator = mapper.buildOperator(GrpcVorgangStatisticQueryTestFactory.create());

				assertThat(operator).extracting("operand").isEqualTo(GrpcVorgangStatisticQueryTestFactory.OPERAND_STRING_VALUE);
			}

		}

		@Nested
		class TestGetValue {

			@Test
			void shouldReturnNullWhenNotSet() {
				var statisticQuery = GrpcVorgangStatisticQueryTestFactory.createBuilder().clearOperandStringValue().build();

				var value = mapper.getValue(statisticQuery);

				assertThat(value).isNull();
			}

			@Test
			void shouldCallGetStringValue() {
				var statisticQuery = GrpcVorgangStatisticQueryTestFactory.create();

				mapper.getValue(statisticQuery);

				verify(mapper).getStringValue(statisticQuery);
			}

			@Test
			void shouldReturnInteger() {
				var statisticQuery = GrpcVorgangStatisticQueryTestFactory.createBuilder().setOperandIntValue(1).build();

				var value = mapper.getValue(statisticQuery);

				assertThat(value).isEqualTo(1);
			}

			@Test
			void shouldReturnBoolean() {
				var statisticQuery = GrpcVorgangStatisticQueryTestFactory.createBuilder().setOperandBoolValue(true).build();

				var value = mapper.getValue(statisticQuery);

				assertThat(value).isEqualTo(true);
			}

			@Nested
			class TestGetStringValue {

				@Test
				void shouldCallOperandFunctionParser() {
					var functionValue = "function()";
					var statisticQuery = GrpcVorgangStatisticQueryTestFactory.createBuilder().setOperandStringValue(functionValue).build();

					mapper.getStringValue(statisticQuery);

					verify(operandFunctionParser).parse(functionValue);
				}

				@Test
				void shouldReturnNull() {
					var statisticQuery = GrpcVorgangStatisticQueryTestFactory.createBuilder().setOperandStringValue("null").build();

					var value = mapper.getStringValue(statisticQuery);

					assertThat(value).isNull();
				}

				@Test
				void shouldReturnStringValue() {
					var statisticQuery = GrpcVorgangStatisticQueryTestFactory.create();

					var value = mapper.getStringValue(statisticQuery);

					assertThat(value).isEqualTo(GrpcVorgangStatisticQueryTestFactory.OPERAND_STRING_VALUE);
				}
			}
		}
	}

	@Nested
	@DisplayName("Map to VorgangStatisticResponse")
	class TestMapToVorgangStatisticResponse {

		@Test
		void shouldCallMapCountByPathStatistic() {
			mapper.mapToVorgangStatisticResponse(StatisticTestFactory.create());

			verify(mapper).mapCountByPathStatistic(StatisticTestFactory.BY_PATH_MAP);
		}

		@Test
		void shouldCallMapIsFieldExistsStatistic() {
			mapper.mapToVorgangStatisticResponse(StatisticTestFactory.create());

			verify(mapper).mapIsFieldExistsStatistic(StatisticTestFactory.FIELD_EXISTS_MAP);
		}

		@Test
		void shouldSetStatisticResults() {
			var countByPathStatisticResult = GrpcVorgangStatisticResultTestFactory.create();
			var isFieldExistsStatisticResult = GrpcVorgangStatisticResultTestFactory.createBuilder().setResultBoolValue(true).build();
			init(countByPathStatisticResult, isFieldExistsStatisticResult);

			var response = mapper.mapToVorgangStatisticResponse(StatisticTestFactory.create());

			assertThat(response.getResultList()).containsExactly(countByPathStatisticResult, isFieldExistsStatisticResult);
		}

		private void init(GrpcVorgangStatisticResult countByPathStatisticResult, GrpcVorgangStatisticResult isFieldExistsStatisticResult) {
			var list = new ArrayList<GrpcVorgangStatisticResult>();
			list.add(countByPathStatisticResult);
			doReturn(list).when(mapper).mapCountByPathStatistic(anyMap());
			doReturn(List.of(isFieldExistsStatisticResult)).when(mapper).mapIsFieldExistsStatistic(anyMap());
		}

		@Nested
		@DisplayName("Map CountByPathStatistic")
		class TestMapByPathStatistic {

			@Test
			void shouldCallMapToCountResult() {
				mapper.mapCountByPathStatistic(StatisticTestFactory.BY_PATH_MAP);

				verify(mapper).mapToCountResult(StatisticTestFactory.FIELD_PATH_ALIAS, StatisticTestFactory.FIELD_PATH_VALUE);
			}

			@Test
			void shouldGetEmptyList() {
				var result = mapper.mapCountByPathStatistic(Collections.emptyMap());

				assertThat(result).isEmpty();
			}

			@Nested
			@DisplayName("Map to CountResult")
			class TestMapToCountResult {

				@Test
				void shouldSetName() {
					var result = mapper.mapToCountResult(StatisticTestFactory.FIELD_PATH_ALIAS, StatisticTestFactory.FIELD_PATH_VALUE);

					assertThat(result.getName()).isEqualTo(StatisticTestFactory.FIELD_PATH_ALIAS);
				}

				@Test
				void shouldSetIntValue() {
					var result = mapper.mapToCountResult(StatisticTestFactory.FIELD_PATH_ALIAS, StatisticTestFactory.FIELD_PATH_VALUE);

					assertThat(result.getResultIntValue()).isEqualTo(StatisticTestFactory.FIELD_PATH_VALUE);
				}
			}
		}

		@Nested
		@DisplayName("Map IsFieldExistsStatistic")
		class TestMapIsFieldExistsStatistic {

			@Test
			void shouldCallMapToIsFieldExistsResult() {
				mapper.mapIsFieldExistsStatistic(StatisticTestFactory.FIELD_EXISTS_MAP);

				verify(mapper).mapToIsFieldExistsResult(StatisticTestFactory.FIELD_PATH_ALIAS, true);
			}

			@Test
			void shouldGetEmptyList() {
				var result = mapper.mapIsFieldExistsStatistic(Collections.emptyMap());

				assertThat(result).isEmpty();
			}

			@Nested
			@DisplayName("Map to IsFieldExistsResult")
			class TestMapToIsFieldExistsResult {

				@Test
				void shouldSetName() {
					var result = mapper.mapToIsFieldExistsResult(StatisticTestFactory.FIELD_PATH_ALIAS, true);

					assertThat(result.getName()).isEqualTo(StatisticTestFactory.FIELD_PATH_ALIAS);
				}

				@Test
				void shouldSetBoolValue() {
					var result = mapper.mapToIsFieldExistsResult(StatisticTestFactory.FIELD_PATH_ALIAS, true);

					assertThat(result.getResultBoolValue()).isTrue();
				}
			}
		}
	}
}