/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.statistic;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.Set;

import org.assertj.core.data.MapEntry;
import org.bson.Document;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;
import org.springframework.data.mongodb.core.MongoOperations;
import org.springframework.data.mongodb.core.aggregation.AggregationResults;
import org.springframework.data.mongodb.core.query.Query;

import de.ozgcloud.vorgang.common.operator.Operator;
import de.ozgcloud.vorgang.vorgang.Vorgang;
import de.ozgcloud.vorgang.vorgang.ZustaendigeStelleTestFactory;

class StatisticRepositoryTest {

	@InjectMocks
	@Spy
	private StatisticRepository statisticRepository;

	@Mock
	private MongoOperations mongoOperations;

	private final StatisticFilter ORGANISATIONSEINHEIT_IDS_FILTER = StatisticFilter.builder()
			.organisationsEinheitId(ZustaendigeStelleTestFactory.ORGANISATIONSEINHEIT_ID).build();

	@Nested
	@DisplayName("Count vorgang")
	class TestCount {

		@Nested
		@DisplayName("by status")
		class TestByStatus {

			@Mock
			private AggregationResults aggregationResults;

			@BeforeEach
			void setUp() {
				doReturn(aggregationResults).when(statisticRepository).executeAggregation(any(), any());
			}

			@Test
			@DisplayName("should call initAggregationOperations")
			void shouldCallExecuteAggregationAll() {
				statisticRepository.countVorgangByStatus(ORGANISATIONSEINHEIT_IDS_FILTER);

				verify(statisticRepository).initAggregationOperations(ORGANISATIONSEINHEIT_IDS_FILTER);
			}

			@Test
			@DisplayName("should call executeAggregation")
			void shouldCallExecuteAggregation() {
				statisticRepository.countVorgangByStatus(ORGANISATIONSEINHEIT_IDS_FILTER);

				verify(statisticRepository).executeAggregation(any(), any());
			}

			@Test
			@DisplayName("should call get results")
			void shouldCallGetResults() {
				statisticRepository.countVorgangByStatus(ORGANISATIONSEINHEIT_IDS_FILTER);

				verify(aggregationResults).getMappedResults();
			}
		}

		@Nested
		@DisplayName("by path")
		class TestByPath {

			private final static CountByPath COUNT_PATH_REQUEST = CountPathRequestTestFactory.create();

			@Mock
			private Document document;
			@Mock
			private AggregationResults aggregationResults;

			private Operator statisticQuery = OperatorTestFactory.create();

			@BeforeEach
			void setup() {
				when(aggregationResults.getUniqueMappedResult()).thenReturn(document).getMock();
				doReturn(aggregationResults).when(statisticRepository).executeAggregation(any(), any());
			}

			@Test
			@DisplayName("should call initAggregationOperations")
			void shouldCallExecuteAggregationAll() {
				statisticRepository.countVorgangByPath(COUNT_PATH_REQUEST, statisticQuery, ORGANISATIONSEINHEIT_IDS_FILTER);

				verify(statisticRepository).initAggregationOperations(ORGANISATIONSEINHEIT_IDS_FILTER);
			}

			@Test
			@DisplayName("should call executeAggregation")
			void shouldCallExecuteAggregation() {
				statisticRepository.countVorgangByPath(COUNT_PATH_REQUEST, statisticQuery, ORGANISATIONSEINHEIT_IDS_FILTER);

				verify(statisticRepository).executeAggregation(any(), any());
			}

			@Test
			@DisplayName("should call get results")
			void shouldCallGetResults() {
				statisticRepository.countVorgangByPath(COUNT_PATH_REQUEST, statisticQuery, StatisticFilter.builder().build());

				verify(aggregationResults).getUniqueMappedResult();
			}

			@Test
			@DisplayName("should call map count results")
			void shouldCallMapCountResults() {
				statisticRepository.countVorgangByPath(COUNT_PATH_REQUEST, statisticQuery, StatisticFilter.builder().build());

				verify(statisticRepository).mapCountByPathResults(document, CountByPathTestFactory.ALIAS);
			}
		}
	}

	@Nested
	class TestMapByPathResults {

		@Mock
		private Document document;

		@Test
		void shouldFilterIdField() {
			when(document.entrySet()).thenReturn(Set.of(MapEntry.entry("_id", "id")));

			var result = statisticRepository.mapCountByPathResults(document, CountByPathTestFactory.ALIAS);

			assertThat(result).isEmpty();
		}

		@Test
		void shouldCallMapCountResult() {
			var entry = MapEntry.<String, Object>entry("key", 1);
			when(document.entrySet()).thenReturn(Set.of(entry));

			statisticRepository.mapCountByPathResults(document, CountByPathTestFactory.ALIAS);

			verify(statisticRepository, times(1)).mapCountResult(entry);
		}

		@Test
		void shouldReturnResult() {
			var entry = MapEntry.<String, Object>entry(CountResultTestFactory.NAME, CountResultTestFactory.VALUE);
			when(document.entrySet()).thenReturn(Set.of(entry));
			doReturn(CountResultTestFactory.create()).when(statisticRepository).mapCountResult(entry);

			var results = statisticRepository.mapCountByPathResults(document, CountByPathTestFactory.ALIAS);

			assertThat(results).containsExactly(CountResultTestFactory.create());
		}

		@Test
		void shouldZeroResultOnNull() {
			var results = statisticRepository.mapCountByPathResults(null, CountByPathTestFactory.ALIAS);

			assertThat(results).hasSize(1).first().usingRecursiveComparison().isEqualTo(CountResultTestFactory.createBuilder().value(0).build());
		}
	}

	@Nested
	class TestMapCountResults {

		@Test
		void shouldMapValue() {
			var entry = MapEntry.<String, Object>entry(CountResultTestFactory.NAME, CountResultTestFactory.VALUE);

			var result = statisticRepository.mapCountResult(entry);

			assertThat(result.getName()).isEqualTo(CountResultTestFactory.NAME);
		}

		@Test
		void shouldMapCount() {
			var entry = MapEntry.<String, Object>entry(CountResultTestFactory.NAME, CountResultTestFactory.VALUE);

			var result = statisticRepository.mapCountResult(entry);

			assertThat(result.getValue()).isEqualTo(CountResultTestFactory.VALUE);
		}
	}

	@Nested
	@DisplayName("Init aggregation operations")
	class TestInitAggregationOperation {

		@Test
		void addInCreationCriteria() {
			var result = statisticRepository.initAggregationOperations(StatisticFilter.builder().build());

			assertThat(result).hasSize(1);
		}

	}

	@Nested
	class TestIsFieldExists {

		@Spy
		private StatisticFilter filter = StatisticFilter.builder().build();

		@Test
		void shouldCallExists() {
			statisticRepository.isFieldsExists(filter);

			verify(mongoOperations).exists(any(Query.class), eq(Vorgang.COLLECTION_NAME));
		}

		@Test
		void shouldCallGetCriteria() {
			statisticRepository.isFieldsExists(filter);

			verify(filter).getCriteria();
		}
	}
}