/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.vorgang;

import static org.mockito.Mockito.*;

import java.util.Arrays;
import java.util.Optional;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mapstruct.factory.Mappers;
import org.mockito.InjectMocks;
import org.mockito.Mock;

import static org.assertj.core.api.Assertions.*;

import de.ozgcloud.vorgang.callcontext.CallContextUserTestFactory;
import de.ozgcloud.vorgang.callcontext.CurrentUserService;
import de.ozgcloud.vorgang.vorgang.Vorgang.Status;

class FilterByMapperTest {

	@InjectMocks
	private FilterByMapper mapper = Mappers.getMapper(FilterByMapper.class);
	@Mock
	private CurrentUserService currentUserService;

	@DisplayName("Map filter by")
	@Nested
	class TestMapFromFilterBy {

		@Nested
		class Organisationseinheiten {

			@Test
			void shouldMapFromRequestIgnoringUserService() {
				when(currentUserService.findUser()).thenReturn(Optional.of(CallContextUserTestFactory.create()));

				var grpcFilterBy = GrpcFilterBy.newBuilder().addOrganisationseinheitId("1").build();
				var filterBy = mapper.fromFilterBy(grpcFilterBy);

				assertThat(filterBy.getOrganisationseinheitIds()).hasSize(1);
				assertThat(filterBy.getOrganisationseinheitIds()).containsExactly("1");
			}

			@Test
			void shouldMapFromRequestOnEmptyUser() {
				when(currentUserService.findUser()).thenReturn(Optional.empty());

				var grpcFilterBy = GrpcFilterBy.newBuilder().addAllOrganisationseinheitId(Arrays.asList("1", "2")).build();
				var filterBy = mapper.fromFilterBy(grpcFilterBy);

				assertThat(filterBy.getOrganisationseinheitIds()).hasSize(2);
				assertThat(filterBy.getOrganisationseinheitIds()).containsExactly("1", "2");
			}

			@Test
			void shouldMapFromUserServiceOnEmptyRequestList() {
				when(currentUserService.findUser())
						.thenReturn(Optional.of(CallContextUserTestFactory.createBuilder().clearOrganisatorischeEinheitenIds()
								.organisatorischeEinheitenId(ZustaendigeStelleTestFactory.ORGANISATIONSEINHEIT_ID).build()));

				var grpcFilterBy = GrpcFilterBy.newBuilder().clearOrganisationseinheitId().build();
				var filterBy = mapper.fromFilterBy(grpcFilterBy);

				assertThat(filterBy.getOrganisationseinheitIds()).hasSize(1);
				assertThat(filterBy.getOrganisationseinheitIds()).containsExactly(ZustaendigeStelleTestFactory.ORGANISATIONSEINHEIT_ID);
			}
		}

		@DisplayName("status")
		@Nested
		class TestStatus {

			@BeforeEach
			void mockCurrentUserService() {
				when(currentUserService.findUser()).thenReturn(Optional.of(CallContextUserTestFactory.create()));
			}

			@Test
			void shouldMapList() {
				var grpcFilterBy = GrpcFilterBy.newBuilder().addAllStatus(Arrays.asList(Status.NEU.name(), Status.IN_BEARBEITUNG.name())).build();

				var filterBy = mapper.fromFilterBy(grpcFilterBy);

				assertThat(filterBy.getStatus()).hasSize(2);
				assertThat(filterBy.getStatus()).containsExactly(Status.NEU, Status.IN_BEARBEITUNG);
			}

			@Test
			void shouldMapSingle() {
				var grpcFilterBy = GrpcFilterBy.newBuilder().addStatus(Status.NEU.name()).build();

				var filterBy = mapper.fromFilterBy(grpcFilterBy);

				assertThat(filterBy.getStatus()).hasSize(1);
				assertThat(filterBy.getStatus()).containsExactly(Status.NEU);
			}
		}

		@DisplayName("is id check necessary")
		@Nested
		class TestIsOrganisationEinheitenIdCheckNecessary {

			@BeforeEach
			void mockCurrentUserService() {
				when(currentUserService.findUser())
						.thenReturn(Optional.of(CallContextUserTestFactory.createBuilder().organisationEinheitenIdCheckNecessary(true).build()));
			}

			@DisplayName("with given user check = true")
			@Nested
			class TestWithExistingUserCheck {

				@Test
				void shouldMapFalseIgnoringFilterValue() {
					var grpcFilterBy = GrpcFilterBy.newBuilder().setFilterByOrganisationseinheitenId(false).build();

					var filterBy = mapper.fromFilterBy(grpcFilterBy);

					assertThat(filterBy.isFilterByOrganisationseinheitenId()).isFalse();
				}

				@Test
				void shouldMapTrueFilterValue() {
					var grpcFilterBy = GrpcFilterBy.newBuilder().setFilterByOrganisationseinheitenId(true).build();

					var filterBy = mapper.fromFilterBy(grpcFilterBy);

					assertThat(filterBy.isFilterByOrganisationseinheitenId()).isTrue();
				}
			}

			@DisplayName("with given user check = false")
			@Nested
			class TestWithNonExistingUserCheck {

				@BeforeEach
				void mockCurrentUserService() {
					when(currentUserService.findUser())
							.thenReturn(Optional.of(CallContextUserTestFactory.createBuilder().organisationEinheitenIdCheckNecessary(false).build()));
				}

				@Test
				void shouldMapFalseConsideringFilterValue() {
					var grpcFilterBy = GrpcFilterBy.newBuilder().setFilterByOrganisationseinheitenId(false).build();

					var filterBy = mapper.fromFilterBy(grpcFilterBy);

					assertThat(filterBy.isFilterByOrganisationseinheitenId()).isFalse();
				}

				@Test
				void shouldMapTrueConsideringFilterValue() {
					var grpcFilterBy = GrpcFilterBy.newBuilder().setFilterByOrganisationseinheitenId(true).build();

					var filterBy = mapper.fromFilterBy(grpcFilterBy);

					assertThat(filterBy.isFilterByOrganisationseinheitenId()).isTrue();
				}
			}
		}
	}
}