/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.vorgang;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.Mockito.*;

import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mapstruct.factory.Mappers;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Mockito;

import de.ozgcloud.vorgang.common.GrpcQueryOperator;
import de.ozgcloud.vorgang.common.operator.OperandFunctionParser;
import de.ozgcloud.vorgang.common.operator.OperatorBuilder;
import de.ozgcloud.vorgang.common.operator.OperatorTestFactory;

class FindVorgangQueryMapperTest {

	@InjectMocks
	private FindVorgangQueryMapper mapper = spy(Mappers.getMapper(FindVorgangQueryMapper.class));

	@Mock
	private OperandFunctionParser operandFunctionParser;

	@Nested
	class TestFromFindVorgangQuery {

		@Test
		void shouldSetLogicalOperator() {
			var result = mapper.fromQuery(GrpcQueryTestFactory.create());

			assertThat(result.getLogicalOperator()).isEqualTo(FindVorgangQueryTestFactory.LOGICAL_OPERATOR);
		}

		@Test
		void shouldCallBuildOperator() {
			var grpcFindVorgangQuery = GrpcQueryTestFactory.create();
			var expectedExpression = grpcFindVorgangQuery.getExpressions(0);

			mapper.fromQuery(grpcFindVorgangQuery);

			verify(mapper).buildOperator(expectedExpression);
		}

		@Test
		void shouldSetOperator() {
			var expectedOperator = OperatorTestFactory.createEqualOperator();
			doReturn(expectedOperator).when(mapper).buildOperator(any());

			var result = mapper.fromQuery(GrpcQueryTestFactory.create());

			assertThat(result.getOperators()).containsExactly(expectedOperator);
		}

		@Test
		void shouldSkipMappingNestedQuery() {
			doReturn(false).when(mapper).hasSubQuery(any());

			mapper.fromQuery(GrpcQueryTestFactory.create());

			verify(mapper).fromQuery(any());
		}

		@Test
		void shouldSetNestedQuery() {
			var expectedNestedQuery = FindVorgangQueryTestFactory.createBuilder()
					.logicalOperator(FindVorgangQuery.LogicalOperator.OR)
					.clearOperators()
					.operator(OperatorTestFactory.createLessThenOperator())
					.build();
			var nestedQuery = GrpcQueryTestFactory.createBuilder().setLogicalOperator(GrpcLogicalOperator.OR)
					.clearExpressions()
					.addExpressions(GrpcVorgangQueryExpressionTestFactory.createBuilder()
							.setOperator(GrpcQueryOperator.LESS_THEN)
							.setOperandIntValue(OperatorTestFactory.INT_VALUE))
					.build();

			var result = mapper.fromQuery(GrpcQueryTestFactory.createBuilder().setNestedQuery(nestedQuery).build());

			assertThat(result.getNestedQuery()).usingRecursiveComparison().isEqualTo(expectedNestedQuery);
		}

	}

	@Nested
	class TestHasSubForm {

		@Test
		void shouldApproveSubQuery() {
			var result = mapper.hasSubQuery(GrpcQueryTestFactory.createBuilder().setNestedQuery(GrpcQueryTestFactory.create()).build());

			assertThat(result).isTrue();
		}

		@Test
		void shouldRejectIfNoSubQuery() {
			var result = mapper.hasSubQuery(GrpcQueryTestFactory.create());

			assertThat(result).isFalse();
		}
	}

	@Nested
	class TestFromLogicalOperator {

		@Test
		void shouldMapAnd() {
			var result = mapper.fromLogicalOperator(GrpcLogicalOperator.AND);

			assertThat(result).isEqualTo(FindVorgangQuery.LogicalOperator.AND);
		}

		@Test
		void shouldMapOr() {
			var result = mapper.fromLogicalOperator(GrpcLogicalOperator.OR);

			assertThat(result).isEqualTo(FindVorgangQuery.LogicalOperator.OR);
		}
	}

	@Nested
	class TestBuildOperator {

		@Mock
		private OperatorBuilder operatorBuilder;

		@Test
		void shouldSetFieldPath() {
			var operator = mapper.buildOperator(GrpcVorgangQueryExpressionTestFactory.create());

			assertThat(operator).extracting("fieldPath").isEqualTo(OperatorTestFactory.PATH);
		}

		@Test
		void shouldCallGetValue() {
			var queryExpression = GrpcVorgangQueryExpressionTestFactory.create();

			mapper.buildOperator(queryExpression);

			verify(mapper).getValue(queryExpression);
		}

		@Test
		void shouldSetOperand() {
			doReturn(OperatorTestFactory.STRING_VALUE).when(mapper).getValue(any());

			var operator = mapper.buildOperator(GrpcVorgangQueryExpressionTestFactory.create());

			assertThat(operator).extracting("operand").isEqualTo(OperatorTestFactory.STRING_VALUE);
		}

		@Test
		void shouldCallOperatorBuilder() {
			try(var mockOperatorBuilder = Mockito.mockStatic(OperatorBuilder.class)) {
				mockOperatorBuilder.when(() -> OperatorBuilder.from(any())).thenReturn(operatorBuilder);
				mockOperatorBuilder.when(() -> operatorBuilder.fieldPath(any())).thenReturn(operatorBuilder);
				mockOperatorBuilder.when(() -> operatorBuilder.operand(any())).thenReturn(operatorBuilder);

				mapper.buildOperator(GrpcVorgangQueryExpressionTestFactory.create());

				mockOperatorBuilder.verify(() -> OperatorBuilder.from(GrpcVorgangQueryExpressionTestFactory.OPERATOR));
			}
		}

	}

	@Nested
	class TestGetValue {

		@Test
		void shouldReturnNullWhenNotSet() {
			var statisticQuery = GrpcVorgangQueryExpressionTestFactory.createBuilder().clearOperandStringValue().build();

			var value = mapper.getValue(statisticQuery);

			assertThat(value).isNull();
		}

		@Test
		void shouldCallGetStringValue() {
			var statisticQuery = GrpcVorgangQueryExpressionTestFactory.create();

			mapper.getValue(statisticQuery);

			verify(mapper).getStringValue(statisticQuery);
		}

		@Test
		void shouldReturnInteger() {
			var statisticQuery = GrpcVorgangQueryExpressionTestFactory.createBuilder().setOperandIntValue(1).build();

			var value = mapper.getValue(statisticQuery);

			assertThat(value).isEqualTo(1);
		}

		@Test
		void shouldReturnBoolean() {
			var statisticQuery = GrpcVorgangQueryExpressionTestFactory.createBuilder().setOperandBoolValue(true).build();

			var value = mapper.getValue(statisticQuery);

			assertThat(value).isEqualTo(true);
		}

		@Nested
		class TestGetStringValue {

			@Test
			void shouldCallOperandFunctionParser() {
				var functionValue = "function()";
				var statisticQuery = GrpcVorgangQueryExpressionTestFactory.createBuilder().setOperandStringValue(functionValue).build();

				mapper.getStringValue(statisticQuery);

				verify(operandFunctionParser).parse(functionValue);
			}

			@Test
			void shouldReturnNull() {
				var statisticQuery = GrpcVorgangQueryExpressionTestFactory.createBuilder().setOperandStringValue("null").build();

				var value = mapper.getStringValue(statisticQuery);

				assertThat(value).isNull();
			}

			@Test
			void shouldReturnStringValue() {
				var statisticQuery = GrpcVorgangQueryExpressionTestFactory.create();

				var value = mapper.getStringValue(statisticQuery);

				assertThat(value).isEqualTo(OperatorTestFactory.STRING_VALUE);
			}
		}
	}
}