/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.vorgang;

import static de.ozgcloud.vorgang.vorgang.EingangTestFactory.*;
import static de.ozgcloud.vorgang.vorgang.LabelProcessor.*;
import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;

import org.apache.commons.lang3.tuple.Pair;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mockito;
import org.mockito.Spy;

class LabelProcessorTest {

	@Spy
	@InjectMocks
	private LabelProcessor mapper;

	private Map<String, String> singleFieldLabelMap = Map.of(SINGLE_FIELD_NAME, SINGLE_FIELD_LABEL);
	private Map<String, String> subFormLabelMap = Map.of(SUBFORM_NAME, SUBFORM_LABEL);
	private Map<String, String> fullFormLabelMap = Map.of(SINGLE_FIELD_NAME, SINGLE_FIELD_LABEL, SUBFORM_NAME, SUBFORM_LABEL);

	private Map<String, Object> singleFieldWithLabel = Map.of(SINGLE_FIELD_NAME, SINGLE_FIELD_VALUE, CONTROL_DATA_KEY, singleFieldLabelMap);

	@Nested
	class TestMoveLabelsToControlData {

		private Map<String, Object> singleFieldMap = Map.of(VALUE_KEY, SINGLE_FIELD_VALUE, LABEL_KEY, SINGLE_FIELD_LABEL);

		@Nested
		class MoveLabels {
			@BeforeEach
			void init() {
				doReturn(Optional.of(SINGLE_FIELD_LABEL)).when(mapper).getLabelFromFieldValue(any());

			}

			@Test
			void shouldAddLabelToControlData() {
				doReturn(SINGLE_FIELD_VALUE).when(mapper).getValueWithoutLabel(any());

				var result = mapper.moveLabelsToControlData(Map.of(SINGLE_FIELD_NAME, singleFieldMap));

				assertThat(result).contains(entry(CONTROL_DATA_KEY, Map.of(LABELS_KEY, singleFieldLabelMap)));
			}

			@Test
			void shouldRemoveLabelFormValue() {
				doReturn(SINGLE_FIELD_VALUE).when(mapper).getValueWithoutLabel(any());

				var result = mapper.moveLabelsToControlData(Map.of(SINGLE_FIELD_NAME, singleFieldMap));

				assertThat(result).contains(entry(SINGLE_FIELD_NAME, SINGLE_FIELD_VALUE));
			}

			@Test
			void shouldProcessSubForm() {
				mapper.moveLabelsToControlData(Map.of(SUBFORM_NAME, SUBFORM));

				verify(mapper).doRecursionIfSubForm(Pair.of(SUBFORM_NAME, SUBFORM_FLAT));
				verify(mapper).moveLabelsToControlData(SUBFORM_FLAT);
			}
		}

		@Nested
		class GetLables {
			@Test
			void shouldNotCallGetLabelForControlData() {
				mapper.getLabelsFromFormData(Map.of(CONTROL_DATA_KEY, ""));

				verify(mapper, never()).getLabelFromFieldValue(any());
			}

			@Test
			void shouldReturnFieldLable() {
				doReturn(Optional.of(SINGLE_FIELD_LABEL)).when(mapper).getLabelFromFieldValue(any());

				var labels = mapper.getLabelsFromFormData(singleFieldWithLabel);

				assertThat(labels).containsExactly(entry(SINGLE_FIELD_NAME, SINGLE_FIELD_LABEL));
				verify(mapper).getLabelFromFieldValue(SINGLE_FIELD_VALUE);
			}

			@Test
			void shouldOmitMissingLabel() {
				doReturn(Optional.empty()).when(mapper).getLabelFromFieldValue(any());

				var labels = mapper.getLabelsFromFormData(singleFieldWithLabel);

				assertThat(labels).isEmpty();
			}

			@Nested
			class FromFieldValue {
				@Test
				void shouldReturnLabel() {
					var label = mapper.getLabelFromFieldValue(Map.of(VALUE_KEY, SINGLE_FIELD_VALUE, LABEL_KEY, SINGLE_FIELD_LABEL));

					assertThat(label).contains(SINGLE_FIELD_LABEL);
				}

				@Test
				void shouldReturnEmptyForValue() {
					var label = mapper.getLabelFromFieldValue(SINGLE_FIELD_VALUE);

					assertThat(label).isEmpty();
				}

				@Test
				void shouldReturnLableForSubform() {
					var label = mapper.getLabelFromFieldValue(Map.of(VALUE_KEY, SUBFORM, LABEL_KEY, SUBFORM_LABEL));

					assertThat(label).contains(SUBFORM_LABEL);
				}
			}
		}

		@Nested
		class RemoveLabel {
			@Test
			void shouldReturnSingleValue() {
				var result = mapper.getValueWithoutLabel(SINGLE_FIELD_VALUE);

				assertThat(result).isEqualTo(SINGLE_FIELD_VALUE);
			}

			@Test
			void shouldReturnSingleValueWithoutLabel() {
				var result = mapper.getValueWithoutLabel(Map.of(VALUE_KEY, SINGLE_FIELD_VALUE, LABEL_KEY, SINGLE_FIELD_LABEL));

				assertThat(result).isEqualTo(SINGLE_FIELD_VALUE);
			}

			@Test
			void shouldReturnSubForm() {
				var result = mapper.getValueWithoutLabel(SUBFORM);

				assertThat(result).isEqualTo(Map.of(SUBFORM_FIELD_NAME, SUBFORM_FIELD));
			}

			@Test
			void shouldReturnSubFormWithoutLabel() {
				var result = mapper.getValueWithoutLabel(Map.of(VALUE_KEY, SUBFORM, LABEL_KEY, SUBFORM_LABEL));

				assertThat(result).isEqualTo(SUBFORM);
			}
		}

		@Nested
		class AddLabelsToControlData {
			@Test
			void shouldAddLables() {
				doReturn(singleFieldLabelMap).when(mapper).getLabelsFromFormData(any());

				var result = mapper.addLabelsToControlData(Map.of(SINGLE_FIELD_NAME, singleFieldMap));

				assertThat(result).isEqualTo(Map.of(LABELS_KEY, Map.of(SINGLE_FIELD_NAME, SINGLE_FIELD_LABEL)));
			}

			@Test
			void shouldAddNothingIfNoLabel() {
				doReturn(Collections.emptyMap()).when(mapper).getLabelsFromFormData(any());

				var result = mapper.addLabelsToControlData(Map.of(SINGLE_FIELD_NAME, singleFieldMap));

				assertThat(result).isEmpty();
			}
		}

		@Nested
		class ReplaceControlData {
			@Test
			void shouldReplaceControlData() {
				Map<String, Object> newControlData = Map.of("test", "test");

				var result = mapper.replaceControlData(Map.of(CONTROL_DATA_KEY, Map.of("old", "old")), newControlData);

				assertThat(result).containsOnly(entry(CONTROL_DATA_KEY, newControlData));
			}

			@Test
			void shouldAddNothingIfEmpty() {
				var result = mapper.replaceControlData(Collections.emptyMap(), Collections.emptyMap());

				assertThat(result).isEmpty();
			}
		}

		@Test
		void shouldMapEmptyFormData() {
			var eingang = EingangTestFactory.createBuilder().formData(Collections.emptyMap()).build();

			var mapped = mapper.moveLabelsToControlData(eingang);

			assertThat(mapped.getFormData()).isEqualTo(Collections.emptyMap());
		}
	}

	@Nested
	class TestMapKopControlDataToLabels {

		@DisplayName("Mapping a single Field should call mapValue()")
		@Test
		void shouldCallMapValueForSingleField() {
			addLabels(SINGLE_FIELD_NAME, SINGLE_FIELD_VALUE);

			verify(mapper).addLabel(SINGLE_FIELD_NAME, SINGLE_FIELD_VALUE, Collections.emptyMap());
		}

		@DisplayName("Should not map kopControlData")
		@Test
		void shouldNOTCallMapValueForControlData() {
			addLabels(LabelProcessor.CONTROL_DATA_KEY, "");

			verify(mapper, never()).addLabel(any(), any(), any());
		}

		private void addLabels(String key, Object value) {
			mapper.addLabels(Map.of(key, value));
		}

		@Nested
		class AddLabel {

			@Test
			void shouldCallAddToFieldValue() {
				mapper.addLabel(SINGLE_FIELD_NAME, SINGLE_FIELD_VALUE, singleFieldLabelMap);

				verify(mapper).addLabelToFieldValue(SINGLE_FIELD_NAME, SINGLE_FIELD_VALUE, singleFieldLabelMap);
			}

			@Test
			void shouldJoinSingleFieldMappings() {
				Map<String, List<Object>> mappingResult = Map.of(SINGLE_FIELD_NAME, Collections.emptyList());
				doReturn(mappingResult).when(mapper).addLabelToFieldValue(any(), any(), any());

				var mapped = mapper.addLabel(SINGLE_FIELD_NAME, SINGLE_FIELD_VALUE, singleFieldLabelMap);

				assertThat(mapped).containsAllEntriesOf(mappingResult);
			}

			@Nested
			class ToFieldValue {
				@Test
				void shouldAddLabel() {
					var mapped = mapper.addLabelToFieldValue(SINGLE_FIELD_NAME, SINGLE_FIELD_VALUE, Map.of(SINGLE_FIELD_NAME, SINGLE_FIELD_LABEL));

					assertThat(mapped).isEqualTo(
							Map.of(LabelProcessor.LABEL_KEY, SINGLE_FIELD_LABEL, LabelProcessor.VALUE_KEY, SINGLE_FIELD_VALUE));
				}

				@Test
				void shouldHandleMissingLabel() {
					var mapped = mapper.addLabelToFieldValue(SINGLE_FIELD_NAME, SINGLE_FIELD_VALUE, Collections.emptyMap());

					assertThat(mapped).isEqualTo(Map.of(LabelProcessor.VALUE_KEY, SINGLE_FIELD_VALUE));
				}
			}

			@Test
			void shouldCallAddLabelToSubForm() {
				mapper.addLabel(SUBFORM_NAME, SUBFORM, Map.of(SUBFORM_NAME, SUBFORM_LABEL));

				verify(mapper).addLabelsToSubForm(SUBFORM, Optional.of(SUBFORM_LABEL));
			}

			@Test
			void shouldHandleMissingLabel() {
				mapper.addLabel(SUBFORM_NAME, SUBFORM, Collections.emptyMap());

				verify(mapper).addLabelsToSubForm(SUBFORM, Optional.empty());
			}

			@Nested
			class ToSubForm {
				@Test
				void shouldAddLabel() {
					var mapped = mapper.addLabelsToSubForm(SUBFORM, Optional.of(SUBFORM_LABEL));

					assertThat(mapped).containsEntry(LabelProcessor.LABEL_KEY, SUBFORM_LABEL);
				}

				@Test
				void shouldIgnoreMissingLable() {
					var mapped = mapper.addLabelsToSubForm(SUBFORM, Optional.empty());

					assertThat(mapped).doesNotContainKey(LabelProcessor.LABEL_KEY);
				}

				@Test
				void shouldMapSubForm() {
					mapper.addLabelsToSubForm(SUBFORM, Optional.of(SUBFORM_LABEL));

					verify(mapper).addLabels(SUBFORM);
				}

				@Test
				void shouldAddMappedValue() {
					var mappedValue = Collections.emptyMap();
					doReturn(mappedValue).when(mapper).addLabels(Mockito.<Map<String, Object>>any());

					var mapped = mapper.addLabelsToSubForm(SUBFORM, Optional.of(SUBFORM_LABEL));

					assertThat(mapped).containsEntry(LabelProcessor.VALUE_KEY, mappedValue);
				}
			}

			@Nested
			class MergedProcessedValues {
				private Map<String, Object> singleFieldMap = Map.of(VALUE_KEY, SINGLE_FIELD_VALUE, LABEL_KEY, SINGLE_FIELD_LABEL);

				private Map<String, Object> subFormFieldMap = Map.of(VALUE_KEY, SUBFORM_FIELD_VALUE, LABEL_KEY, SUBFORM_FIELD_LABEL);

				private Map<String, Object> subFormMap = Map.of(LabelProcessor.VALUE_KEY, subFormFieldMap, LABEL_KEY, SUBFORM_LABEL);

				@Test
				void shouldContainFieldValue() {
					doReturn(singleFieldMap).when(mapper).addLabelToFieldValue(any(), any(), any());

					var mapped = mapper.addLabels(
							Map.of(SINGLE_FIELD_NAME, SINGLE_FIELD_VALUE, CONTROL_DATA_KEY, singleFieldLabelMap));

					assertThat(mapped).containsExactly(entry(SINGLE_FIELD_NAME, singleFieldMap));
				}

				@Test
				void shouldContainSubFormValue() {
					doReturn(subFormMap).when(mapper).addLabelsToSubForm(any(), any());

					var mapped = mapper.addLabels(
							Map.of(CONTROL_DATA_KEY, subFormLabelMap, SUBFORM_NAME, SUBFORM));

					assertThat(mapped).containsExactly(entry(SUBFORM_NAME, subFormMap));
				}

				@Test
				void shouldContainsAllLabels() {
					doReturn(singleFieldMap).when(mapper).addLabelToFieldValue(any(), any(), any());
					doReturn(subFormMap).when(mapper).addLabelsToSubForm(any(), any());

					var mapped = mapper.addLabels(
							Map.of(SINGLE_FIELD_NAME, SINGLE_FIELD_VALUE, CONTROL_DATA_KEY, fullFormLabelMap, SUBFORM_NAME, SUBFORM));

					assertThat(mapped).isEqualTo(buildFullFormMap());
				}

				private Map<String, Object> buildFullFormMap() {
					var map = new LinkedHashMap<String, Object>();
					map.put(SINGLE_FIELD_NAME, singleFieldMap);
					map.put(SUBFORM_NAME, subFormMap);

					return map;
				}
			}
		}

		@Test
		void shouldMapEmptyFormData() {
			var eingang = EingangTestFactory.createBuilder().formData(Collections.emptyMap()).build();

			var mapped = mapper.moveLabelsToControlData(eingang);

			assertThat(mapped.getFormData()).isEqualTo(Collections.emptyMap());
		}
	}

	@Nested
	class TestIsFieldValue {

		@Test
		void shouldReturnTrueForSimpleValue() {
			var field = SINGLE_FIELD_VALUE;

			assertThat(mapper.isFieldValue(field)).isTrue();
		}

		@Test
		void shouldReturnTrueForValueWithLabel() {
			var field = Map.of(
					LabelProcessor.LABEL_KEY, SINGLE_FIELD_LABEL,
					LabelProcessor.VALUE_KEY, SINGLE_FIELD_VALUE);

			assertThat(mapper.isFieldValue(field)).isTrue();
		}

		@Test
		void shouldReturnFalseForSubFormValueWithoutLabel() {
			var field = Map.of(SINGLE_FIELD_NAME, SINGLE_FIELD_VALUE);

			assertThat(mapper.isFieldValue(field)).isFalse();
		}

		@Test
		void shouldReturnFalseForSubFormValueWithLabel() {
			var field = Map.of(
					LabelProcessor.LABEL_KEY, SUBFORM_LABEL,
					LabelProcessor.VALUE_KEY, Map.of(SINGLE_FIELD_NAME, SINGLE_FIELD_VALUE));

			assertThat(mapper.isFieldValue(field)).isFalse();
		}

		@Test
		void shouldReturnFalseForEmptySubFormValue() {
			var field = Map.of(
					LabelProcessor.LABEL_KEY, SUBFORM_LABEL,
					LabelProcessor.VALUE_KEY, Collections.emptyMap());

			assertThat(mapper.isFieldValue(field)).isFalse();
		}
	}

	@Nested
	class TestGetFieldValue {

		@Test
		void shouldReturnValueForSimpleField() {
			var fieldValue = SINGLE_FIELD_VALUE;

			assertThat(mapper.getFieldValue(fieldValue)).isEqualTo(fieldValue);
		}

		@Test
		void shouldReturnValueForFieldWithLabel() {
			var fieldValue = Map.of(
					LabelProcessor.LABEL_KEY, SINGLE_FIELD_LABEL,
					LabelProcessor.VALUE_KEY, SINGLE_FIELD_VALUE);

			assertThat(mapper.getFieldValue(fieldValue)).isEqualTo(SINGLE_FIELD_VALUE);
		}

		@Test
		void shouldReturnValueForSubForm() {
			var fieldValue = Map.of(SINGLE_FIELD_NAME, SINGLE_FIELD_VALUE);

			assertThat(mapper.getFieldValue(fieldValue)).isEqualTo(fieldValue);
		}

		@Test
		void shouldReturnValueForSubFormWithLabel() {
			var fieldValue = Map.of(
					LabelProcessor.LABEL_KEY, SUBFORM_LABEL,
					LabelProcessor.VALUE_KEY, SUBFORM);

			assertThat(mapper.getFieldValue(fieldValue)).isEqualTo(SUBFORM);
		}
	}
}