/*
 * Copyright (C) 2025 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.vorgang;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.Mockito.*;

import java.time.ZoneOffset;
import java.time.ZonedDateTime;
import java.time.temporal.ChronoUnit;
import java.util.HashMap;
import java.util.Map;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.NullAndEmptySource;
import org.junit.jupiter.params.provider.ValueSource;
import org.mapstruct.factory.Mappers;
import org.mockito.Spy;

import com.thedeanda.lorem.LoremIpsum;

import de.ozgcloud.command.Command;
import de.ozgcloud.vorgang.command.CommandTestFactory;

class LockMapperTest {

	@Spy
	private LockMapper mapper = Mappers.getMapper(LockMapper.class);

	@Nested
	class TestFromCommand {

		private final Command command = CommandTestFactory.createBuilder().bodyObject(Map.of(Lock.FIELD_REASON, LockTestFactory.REASON)).build();

		@BeforeEach
		void init() {
			doReturn(LockTestFactory.REASON).when(mapper).getReason(any());
			doReturn(LockTestFactory.LOCKED_SINCE).when(mapper).getLockedSince();
		}

		@Test
		void shouldGetLockedSince() {
			map();

			verify(mapper).getLockedSince();
		}

		@Test
		void shouldGetReason() {
			map();

			verify(mapper).getReason(command.getBodyObject());
		}

		@Test
		void shouldMap() {
			var mapped = map();

			assertThat(mapped).usingRecursiveComparison().isEqualTo(LockTestFactory.create());
		}

		private Lock map() {
			return mapper.fromCommand(command);
		}
	}

	@Nested
	class TestGetReason {

		@Test
		void shouldReturnNullIfKeyNotFound() {
			var mapped = mapper.getReason(Map.of("dummyKey", "dummyValue"));

			assertThat(mapped).isNull();
		}

		@Test
		void shouldReturnNullIfValueIsNull() {
			Map<String, Object> bodyObject = new HashMap<>();
			bodyObject.put(Lock.FIELD_REASON, null); // need to put, because Map.of() does not accept null values

			var mapped = mapper.getReason(bodyObject);

			assertThat(mapped).isNull();
		}

		@Test
		void shouldReturnReason() {
			var reason = LoremIpsum.getInstance().getWords(2);
			Map<String, Object> bodyObject = Map.of(Lock.FIELD_REASON, reason);

			var mapped = mapper.getReason(bodyObject);

			assertThat(mapped).isEqualTo(reason);
		}

		@ParameterizedTest
		@ValueSource(strings = { " " })
		@NullAndEmptySource
		void shouldTrimToNull(String reason) {
			Map<String, Object> bodyObject = new HashMap<>();
			bodyObject.put(Lock.FIELD_REASON, reason); // need to put, because Map.of() does not accept null values

			var mapped = mapper.getReason(bodyObject);

			assertThat(mapped).isNull();
		}
	}

	@Nested
	class TestLockedSince {

		@Test
		void shouldReturnCurrentTime() {
			var lockedSince = mapper.getLockedSince();

			assertThat(lockedSince).isCloseTo(ZonedDateTime.now(ZoneOffset.UTC), within(2, ChronoUnit.SECONDS));
		}
	}
}
