/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.vorgang;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.Mockito.*;

import java.util.ArrayList;
import java.util.List;

import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Spy;
import org.springframework.data.mongodb.core.query.Criteria;

import de.ozgcloud.vorgang.common.operator.OperatorTestFactory;
import de.ozgcloud.vorgang.vorgang.FindVorgangQuery.LogicalOperator;

class QueryCriteriaBuilderTest {

	@Spy
	@InjectMocks
	private QueryCriteriaBuilder queryCriteriaBuilder;

	private Criteria criteria = Criteria.where(OperatorTestFactory.PATH).is(OperatorTestFactory.STRING_VALUE);

	@Nested
	class TestFrom {

		@Test
		void shouldBuildCriteriaFromOperators() {
			var result = queryCriteriaBuilder.from(FindVorgangQueryTestFactory.create());

			assertThat(result).isEqualTo(criteria);
		}

		@Test
		void shouldBuildFromNestedQuery() {
			var nestedQuery = FindVorgangQueryTestFactory.createBuilder().logicalOperator(LogicalOperator.OR).build();
			var query = FindVorgangQueryTestFactory.createBuilder().clearOperators().nestedQuery(nestedQuery).build();

			var result = queryCriteriaBuilder.from(query);

			assertThat(result).isEqualTo(criteria);
		}

		@Test
		void shouldCallAddFromOperators() {
			var query = FindVorgangQueryTestFactory.create();

			queryCriteriaBuilder.from(query);

			verify(queryCriteriaBuilder).addFromOperators(any(), eq(query.getOperators()));
		}

		@Test
		void shouldCallConnect() {
			var query = FindVorgangQueryTestFactory.create();

			queryCriteriaBuilder.from(query);

			verify(queryCriteriaBuilder).connect(eq(query.getLogicalOperator()), any());
		}
	}

	@Nested
	class TestAddFromOperators {

		@Test
		void shouldAddOperators() {
			var resultList = new ArrayList<Criteria>();

			queryCriteriaBuilder.addFromOperators(resultList, List.of(OperatorTestFactory.createEqualOperator()));

			assertThat(resultList).containsExactly(criteria);
		}

	}

	@Nested
	class TestConnect {

		private Criteria criteria1 = Criteria.where(OperatorTestFactory.PATH).is(OperatorTestFactory.STRING_VALUE);

		@Test
		void shouldConnectWithAnd() {
			var result = queryCriteriaBuilder.connect(LogicalOperator.AND, List.of(criteria, criteria1));

			assertThat(result.getCriteriaObject()).containsKey("$and");
		}

		@Test
		void shouldConnectWithOr() {
			var result = queryCriteriaBuilder.connect(LogicalOperator.OR, List.of(criteria, criteria1));

			assertThat(result.getCriteriaObject()).containsKey("$or");
		}

		@Test
		void shouldReturnEmptyCriteria() {
			var result = queryCriteriaBuilder.connect(LogicalOperator.AND, List.of());

			assertThat(result.getCriteriaObject()).isEmpty();
		}

		@Test
		void shouldReturnSingleCriteria() {
			var result = queryCriteriaBuilder.connect(LogicalOperator.AND, List.of(criteria));

			assertThat(result).isEqualTo(criteria);
		}
	}

}