/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.vorgang;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.Collections;
import java.util.List;
import java.util.stream.Stream;

import org.junit.jupiter.api.AfterEach;
import org.junit.jupiter.api.Assertions;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.MockedStatic;
import org.mockito.Spy;
import org.springframework.data.domain.Page;

import de.ozgcloud.nachrichten.common.vorgang.GrpcVorgangHeaderTestFactory;
import de.ozgcloud.vorgang.collaboration.CollaborationService;
import de.ozgcloud.vorgang.collaboration.CreateCollaborationVorgangBadRequestException;
import de.ozgcloud.vorgang.collaboration.CreateCollaborationVorgangRequest;
import de.ozgcloud.vorgang.collaboration.CreateCollaborationVorgangRequestMapper;
import de.ozgcloud.vorgang.collaboration.CreateCollaborationVorgangRequestTestFactory;
import de.ozgcloud.vorgang.collaboration.GrpcCreateCollaborationRequestDataTestFactory;
import de.ozgcloud.vorgang.collaboration.GrpcCreateCollaborationVorgangResponseTestFactory;
import de.ozgcloud.vorgang.common.grpc.GrpcResponseBatchStreamer;
import io.grpc.stub.StreamObserver;

class VorgangGrpcServiceTest {

	private static final int LIMIT = 10;
	private static final int OFFSET = 0;
	private static final long TOTAL = 42L;

	@Spy
	@InjectMocks
	private VorgangGrpcService service;
	@Mock
	private VorgangService vorgangService;
	@Mock
	private VorgangHeaderService headerService;
	@Mock
	private VorgangHeaderMapper vorgangHeaderMapper;
	@Mock
	private VorgangStubMapper vorgangStubMapper;

	@Mock
	private EingangMapper eingangMapper;
	@Mock
	private VorgangWithEingangMapper vorgangWithEingangMapper;
	@Mock
	private FindVorgangRequestMapper findVorgangRequestMapper;

	@Mock
	private FilterByMapper filterByMapper;

	@Mock
	private IncomingFileMapper incomingFileMapper;
	@Mock
	private IncomingFileGroupMapper incomingFileGroupMapper;
	@Mock
	private CollaborationService collaborationService;
	@Mock
	private CreateCollaborationVorgangRequestMapper createCollaborationVorgangRequestMapper;

	@Nested
	class TestCreate {

		@Mock
		private StreamObserver<GrpcCreateVorgangResponse> responseObserver;
		@Captor
		private ArgumentCaptor<GrpcCreateVorgangResponse> responseCaptor;

		private final GrpcCreateVorgangRequest request = GrpcCreateVorgangRequest.newBuilder()
				.setEingang(GrpcEingangTestFactory.create())
				.build();
		private final Eingang eingang = EingangTestFactory.create();

		@BeforeEach
		void mockMapperReturnValues() {
			when(eingangMapper.fromGrpc(any())).thenReturn(eingang);
		}

		@Nested
		class TestStartCreation {

			@BeforeEach
			void mockVorgangService() {
				when(vorgangService.startCreation(any())).thenReturn(VorgangTestFactory.create());
			}

			@Test
			void shouldCallEingangMapper() {
				callStartCreation();

				verify(eingangMapper).fromGrpc(GrpcEingangTestFactory.create());
			}

			@Test
			void shouldReturnMessage() {
				callStartCreation();

				assertThat(responseCaptor.getValue().getMessage()).isEqualTo("ok");
			}

			@Test
			void shouldReturnVorgangId() {
				callStartCreation();

				assertThat(responseCaptor.getValue().getVorgangId()).isEqualTo(VorgangTestFactory.ID);
			}

			private void callStartCreation() {
				service.startCreation(request, responseObserver);

				verify(responseObserver).onNext(responseCaptor.capture());
			}
		}
	}

	@Nested
	class TestFinishCreation {

		@Mock
		private StreamObserver<GrpcFinishCreationResponse> responseObserver;
		@Captor
		private ArgumentCaptor<GrpcFinishCreationResponse> responseCaptor;

		@Test
		void shouldCallFileGroupMapper() {
			callService();

			verify(incomingFileGroupMapper).mapGrpc(GrpcFinishCreationRequestTestFactory.attachment);
		}

		@Test
		void shouldCallFileMapper() {
			callService();

			verify(incomingFileMapper).mapGrpc(GrpcFinishCreationRequestTestFactory.representation);
		}

		@Nested
		class TestCallService {

			private final IncomingFileGroup fileGroup = IncomingFileGroupTestFactory.create();
			private final IncomingFile file = IncomingFileTestFactory.create();

			@BeforeEach()
			void mockMapper() {
				when(incomingFileGroupMapper.mapGrpc(any(GrpcIncomingFileGroup.class))).thenReturn(fileGroup);
				when(incomingFileMapper.mapGrpc(any(GrpcIncomingFile.class))).thenReturn(file);
			}

			@Test
			void shouldCallService() {
				callService();

				verify(vorgangService).finishCreation(VorgangTestFactory.ID, List.of(fileGroup), List.of(file));
			}
		}

		@Test
		void shouldCallOnNextWithResponse() {
			callService();

			verify(responseObserver).onNext(responseCaptor.capture());
			assertThat(responseCaptor.getValue().getMessage()).isEqualTo("ok");
		}

		@Test
		void shouldComplete() {
			callService();

			verify(responseObserver).onCompleted();
		}

		private void callService() {
			service.finishCreation(GrpcFinishCreationRequestTestFactory.create(), responseObserver);
		}
	}

	@Nested
	class TestFindVorgange {

		@Mock
		private StreamObserver<GrpcFindVorgangResponse> responseObserver;
		@Captor
		private ArgumentCaptor<GrpcFindVorgangResponse> responseCaptor;
		@Mock
		private Page<VorgangHeader> page;
		@Captor
		private ArgumentCaptor<VorgangHeader> headerCaptor;
		@Captor
		private ArgumentCaptor<FindVorgangRequest> findVorgangRequestCaptor;

		private final GrpcVorgangHeader grpcVorgangHeader = GrpcVorgangHeader.newBuilder().build();
		private final GrpcFindVorgangRequest request = GrpcFindVorgangRequest.newBuilder().setLimit(LIMIT).setOffset(OFFSET)
				.build();

		@BeforeEach
		void initMockReturnValues() {
			when(page.getContent()).thenReturn(Collections.singletonList(VorgangHeaderTestFactory.create()));
			when(page.getTotalElements()).thenReturn(TOTAL);
			when(headerService.findAll(any(FindVorgangRequest.class))).thenReturn(page);
			when(vorgangHeaderMapper.toGrpcVorgangHeader(any())).thenReturn(grpcVorgangHeader);
			when(findVorgangRequestMapper.fromFindVorgangRequest(any(GrpcFindVorgangRequest.class)))
					.thenReturn(FindVorgangRequestTestFactory.create());
		}

		@Test
		void shouldCallVorgangHeaderService() {
			callFindVorgaenge();

			verify(headerService).findAll(findVorgangRequestCaptor.capture());
		}

		@Test
		void shouldCallVorgangHeaderMapper() {
			callFindVorgaenge();

			verify(vorgangHeaderMapper).toGrpcVorgangHeader(any());
		}

		@Test
		void shouldContainVorgangData() {
			callFindVorgaenge();

			verify(vorgangHeaderMapper).toGrpcVorgangHeader(headerCaptor.capture());
			assertThat(headerCaptor.getValue().getId()).isEqualTo(VorgangTestFactory.ID);
			assertThat(headerCaptor.getValue().getCreatedAt()).isEqualTo(VorgangTestFactory.INITIAL_DATE_STR);
		}

		@Test
		void shouldSetTotal() {
			callFindVorgaenge();

			var response = responseCaptor.getValue();

			assertThat(response.getTotal()).isEqualTo(TOTAL);
		}

		private void callFindVorgaenge() {
			service.findVorgang(request, responseObserver);

			verify(responseObserver).onNext(responseCaptor.capture());
		}
	}

	@DisplayName("Find vorgang with eingang")
	@Nested
	class TestFindVorgangWithEingang {

		@Mock
		private StreamObserver<GrpcFindVorgangWithEingangResponse> responseObserver;
		@Captor
		private ArgumentCaptor<GrpcFindVorgangWithEingangResponse> responseCaptor;

		private final FilterCriteria filterCriteria = FilterCriteriaTestFactory.create();
		private final Vorgang vorgang = VorgangTestFactory.create();
		private final GrpcVorgangWithEingang vorgangWithEingang = GrpcVorgangWithEingangTestFactory.create();
		private final GrpcFindVorgangWithEingangRequest request = GrpcFindVorgangWithEingangRequest.newBuilder()
				.setId(VorgangTestFactory.ID)
				.setFilterBy(GrpcFilterByTestFactory.create())
				.build();

		@BeforeEach
		void mock() {
			when(vorgangService.getById(anyString(), any())).thenReturn(vorgang);
			when(vorgangWithEingangMapper.toVorgangWithEingang(any())).thenReturn(vorgangWithEingang);
		}

		@DisplayName("service")
		@Nested
		class TestFindVorgangWithEingangService {

			@Captor
			private ArgumentCaptor<Vorgang> vorgangArgumentCaptor;

			@BeforeEach
			void mock() {
				when(filterByMapper.fromFilterBy(any(GrpcFilterBy.class))).thenReturn(filterCriteria);
			}

			@Test
			void shouldCallHeaderService() {
				callFindVorgangWithEingang();

				verify(vorgangService).getById(VorgangTestFactory.ID, filterCriteria);
			}

			@Test
			void shouldCallFindVorgangRequestMapper() {
				callFindVorgangWithEingang();

				verify(filterByMapper).fromFilterBy(GrpcFilterByTestFactory.create());
			}

			@Test
			void shouldCallVorgangWithEingangMapper() {
				callFindVorgangWithEingang();

				verify(vorgangWithEingangMapper).toVorgangWithEingang(vorgang);
			}

			@Test
			void shouldContainVorgangData() {
				callFindVorgangWithEingang();

				assertThat(responseCaptor.getValue().getVorgangWithEingang()).usingRecursiveComparison()
						.isEqualTo(GrpcVorgangWithEingangTestFactory.create());
			}

		}

		private void callFindVorgangWithEingang() {
			service.findVorgangWithEingang(request, responseObserver);

			verify(responseObserver).onNext(responseCaptor.capture());
		}
	}

	@Nested
	class TestCreateCollaborationVorgangEndpoint {

		@Nested
		class TestCollaborationRequest {

			private static final GrpcCreateCollaborationRequestData GRPC_COLLABORATION_REQUEST = GrpcCreateCollaborationRequestDataTestFactory.create();
			private static final GrpcCreateCollaborationVorgangRequest REQUEST = GrpcCreateCollaborationVorgangRequest.newBuilder()
					.setRequestData(GRPC_COLLABORATION_REQUEST).build();
			private static final GrpcCreateCollaborationVorgangResponse GRPC_COLLABORATION_RESPONSE = GrpcCreateCollaborationVorgangResponseTestFactory
					.create();

			@Mock
			private StreamObserver<GrpcCreateCollaborationVorgangResponse> responseObserver;

			@Captor
			private ArgumentCaptor<GrpcCreateCollaborationVorgangResponse> responseCaptor;

			@BeforeEach
			void init() {
				doReturn(GRPC_COLLABORATION_RESPONSE).when(service).createCollaborationVorgang(any());
			}

			@Test
			void shouldCallCreateCollaborationVorgang() {
				createCollaborationVorgang();

				verify(service).createCollaborationVorgang(GRPC_COLLABORATION_REQUEST);
			}

			@Test
			void shouldCallOnNext() {
				createCollaborationVorgang();

				verify(responseObserver).onNext(responseCaptor.capture());
				assertThat(responseCaptor.getValue().getVorgangId()).isEqualTo(GrpcCreateCollaborationVorgangResponseTestFactory.VORGANG_ID);
			}

			@Test
			void shouldCallOnCompleted() {
				createCollaborationVorgang();

				verify(responseObserver).onCompleted();
			}

			private void createCollaborationVorgang() {
				service.createCollaborationVorgang(REQUEST, responseObserver);
			}
		}

		@Nested
		class TestEmptyCollaborationRequest {

			@Mock
			private StreamObserver<GrpcCreateCollaborationVorgangResponse> responseObserver;

			@Test
			void shouldThrowException() {
				Assertions.assertThrows(CreateCollaborationVorgangBadRequestException.class, this::createCollaborationVorgang);
			}

			private void createCollaborationVorgang() {
				service.createCollaborationVorgang(GrpcCreateCollaborationVorgangRequest.newBuilder().build(), responseObserver);
			}
		}
	}

	@Nested
	class TestCreateCollaborationVorgang {

		private static final GrpcCreateCollaborationRequestData GRPC_COLLABORATION_REQUEST = GrpcCreateCollaborationRequestDataTestFactory.create();
		private static final CreateCollaborationVorgangRequest CREATE_COLLABORATION_VORGANG_REQUEST = CreateCollaborationVorgangRequestTestFactory
				.create();
		private static final Vorgang COLLABORATION_VORGANG = VorgangTestFactory.create();

		@BeforeEach
		void init() {
			doReturn(CREATE_COLLABORATION_VORGANG_REQUEST).when(service).buildCreateCollaborationVorgangRequest(any());
			when(collaborationService.createCollaborationVorgang(any())).thenReturn(COLLABORATION_VORGANG);
		}

		@Test
		void shouldCallBuildCreateCollaborationVorgangRequest() {
			createCollaborationVorgang();

			verify(service).buildCreateCollaborationVorgangRequest(GRPC_COLLABORATION_REQUEST);
		}

		@Test
		void shouldCallCreateCollaborationVorgang() {
			createCollaborationVorgang();

			verify(collaborationService).createCollaborationVorgang(CREATE_COLLABORATION_VORGANG_REQUEST);
		}

		@Test
		void shouldCallBuildResponse() {
			createCollaborationVorgang();

			verify(service).buildCreateCollaborationVorgangResponse(COLLABORATION_VORGANG);
		}

		@Test
		void shouldReturnResponse() {
			var response = GrpcCreateCollaborationVorgangResponseTestFactory.create();
			doReturn(response).when(service).buildCreateCollaborationVorgangResponse(any());

			var result = createCollaborationVorgang();

			assertThat(result).isSameAs(response);
		}

		private GrpcCreateCollaborationVorgangResponse createCollaborationVorgang() {
			return service.createCollaborationVorgang(GRPC_COLLABORATION_REQUEST);
		}
	}

	@Nested
	class TestBuildCreateCollaborationVorgangRequest {

		private static final GrpcCreateCollaborationRequestData GRPC_COLLABORATION_REQUEST = GrpcCreateCollaborationRequestDataTestFactory.create();
		private static final CreateCollaborationVorgangRequest CREATE_COLLABORATION_VORGANG_REQUEST = CreateCollaborationVorgangRequestTestFactory
				.create();
		private static final Vorgang VORGANG = VorgangTestFactory.create();

		@BeforeEach
		void init() {
			when(createCollaborationVorgangRequestMapper.mapFrom(any())).thenReturn(CREATE_COLLABORATION_VORGANG_REQUEST);
			when(vorgangService.getById(anyString())).thenReturn(VORGANG);
		}

		@Test
		void shouldCallGetVorgang() {
			buildCreateCollaborationVorgangRequest();

			verify(vorgangService).getById(VorgangTestFactory.ID);
		}

		@Test
		void shouldCallRequestMapper() {
			buildCreateCollaborationVorgangRequest();

			verify(createCollaborationVorgangRequestMapper).mapFrom(GRPC_COLLABORATION_REQUEST);
		}

		@Test
		void shouldSetVorgang() {
			var result = buildCreateCollaborationVorgangRequest();

			assertThat(result.getVorgang()).isSameAs(VORGANG);
		}

		private CreateCollaborationVorgangRequest buildCreateCollaborationVorgangRequest() {
			return service.buildCreateCollaborationVorgangRequest(GRPC_COLLABORATION_REQUEST);
		}
	}

	@Nested
	class TestBuildCreateCollaborationVorgangResponse {

		@Test
		void shouldSetVorgangId() {
			var result = service.buildCreateCollaborationVorgangResponse(VorgangTestFactory.create());

			assertThat(result.getVorgangId()).isEqualTo(VorgangTestFactory.ID);
		}
	}

	@Nested
	class TestFindDeletedVorgang {

		private final GrpcFindDeletedVorgangRequest request = GrpcFindDeletedVorgangRequestTestFactory.create();
		private MockedStatic<GrpcResponseBatchStreamer> mockedStatic;
		@Mock
		private GrpcResponseBatchStreamer<GrpcVorgangHeader, GrpcFindVorgangResponse> batchStreamer;
		@Mock
		private StreamObserver<GrpcFindDeletedVorgangResponse> responseObserver;

		private final VorgangStub vorgangStub = VorgangStubTestFactory.create();
		private final GrpcVorgangHeader grpcVorgangHeader = GrpcVorgangHeaderTestFactory.create();

		@BeforeEach
		void init() {
			mockedStatic = mockStatic(GrpcResponseBatchStreamer.class);
			mockedStatic.when(() -> GrpcResponseBatchStreamer.create(any(), any())).thenReturn(batchStreamer);
			when(vorgangService.findDeleted()).thenReturn(Stream.of(vorgangStub));
			when(vorgangStubMapper.toGrpcVorgangHeader(vorgangStub)).thenReturn(grpcVorgangHeader);
		}

		@AfterEach
		void cleanup() {
			mockedStatic.close();
		}

		@Test
		void shouldFindDeleted() {
			findDeletedVorgang();

			verify(vorgangService).findDeleted();
		}

		@Test
		void shouldMapToGrpc() {
			findDeletedVorgang();

			verify(vorgangStubMapper).toGrpcVorgangHeader(vorgangStub);
		}

		@Test
		void shouldSendToBatchStreamer() {
			findDeletedVorgang();

			verify(batchStreamer).send(grpcVorgangHeader);
		}

		@Test
		void shouldCallFinish() {
			findDeletedVorgang();

			verify(batchStreamer).finish();
		}

		@Test
		void shouldNotCallStreamObserverDirectly() {
			findDeletedVorgang();

			verifyNoInteractions(responseObserver);
		}

		@Test
		void shouldNotCatchExceptions() {
			var exception = new RuntimeException();
			when(vorgangStubMapper.toGrpcVorgangHeader(vorgangStub)).thenThrow(exception);

			assertThatThrownBy(this::findDeletedVorgang).isSameAs(exception);
		}

		private void findDeletedVorgang() {
			service.findDeletedVorgang(request, responseObserver);
		}
	}
}