/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.vorgang;

import static org.assertj.core.api.Assertions.*;

import java.time.LocalDate;
import java.time.temporal.ChronoUnit;
import java.util.Optional;
import java.util.UUID;

import org.apache.commons.lang3.StringUtils;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.mongodb.core.MongoOperations;
import org.springframework.data.mongodb.core.MongoTemplate;
import org.springframework.data.mongodb.core.query.Query;

import de.ozgcloud.common.test.DataITCase;
import de.ozgcloud.vorgang.clientattribute.ClientAttribute;
import de.ozgcloud.vorgang.clientattribute.ClientAttributeTestFactory;
import de.ozgcloud.vorgang.clientattribute.ClientAttributesMap;
import de.ozgcloud.vorgang.clientattribute.ClientAttributesMapTestFactory;
import de.ozgcloud.vorgang.common.db.CriteriaUtil;
import de.ozgcloud.vorgang.vorgang.Vorgang.Status;

@DataITCase
class VorgangHeaderRepositoryImplITCase {

	@Autowired
	private VorgangHeaderRepositoryImpl repositoryImpl;
	@Autowired
	private MongoOperations mongoOperations;
	@Autowired
	private MongoTemplate template;

	@BeforeEach
	void clearVorgaenge() {
		mongoOperations.dropCollection(Vorgang.class);
	}

	@Nested
	class TestBuildFilterQuery {

		private static final String USER_ID = UUID.randomUUID().toString();

		@Test
		void shouldIgnoreVorgangInCreation() {
			mongoOperations.save(VorgangTestFactory.createBuilder().id(null).inCreation(true).build());

			var query = buildFilterQuery(FilterCriteriaTestFactory.create());

			assertThat(countTotalVorgaenge(query)).isZero();
		}

		@Test
		void shouldIgnoreDELETEDVorgang() {
			mongoOperations.save(VorgangTestFactory.createBuilder().status(Status.DELETED).build());

			var query = buildFilterQuery(FilterCriteriaTestFactory.createBuilder().clearStatus().build());

			assertThat(countTotalVorgaenge(query)).isZero();
		}

		@Nested
		class TestWithOrganisationseinheitenIdFilter {

			@BeforeEach
			void initData() {
				mongoOperations.save(buildWithOrganisationseinheitId("not matched"));
				mongoOperations.save(VorgangTestFactory.createBuilder().id(null).build());
			}

			@Test
			void shouldReturnTotalOnOrganisationseinheit() {
				var filterCriteria = FilterCriteria.builder()
						.filterByOrganisationseinheitenId(true)
						.organisationseinheitId(FilterCriteriaTestFactory.ORGANISATIONSEINHEIT_ID).build();
				var query = buildFilterQuery(filterCriteria);

				assertThat(countTotalVorgaenge(query)).isEqualTo(1);
			}

			private Vorgang buildWithOrganisationseinheitId(String organisationseinheitenId) {
				var eingang = buildEingangWithOrganisationseinheitenId(organisationseinheitenId);
				return VorgangTestFactory.createBuilder().id(null).clearEingangs().eingang(eingang).build();
			}

			private Eingang buildEingangWithOrganisationseinheitenId(String organisationseinheitenId) {
				var zustaendigeStelle = ZustaendigeStelleTestFactory.createBuilder().organisationseinheitenId(organisationseinheitenId).build();
				return EingangTestFactory.createBuilder().zustaendigeStelle(zustaendigeStelle).build();
			}
		}

		@Nested
		class TestWithAssignedToFilter {

			private final FilterCriteria.FilterCriteriaBuilder filterCriteriaBuilder = FilterCriteriaTestFactory.createBuilder()
					.clearStatus().clearOrganisationseinheitIds();

			@BeforeEach
			void initData() {
				mongoOperations.save(VorgangTestFactory.createBuilder().id(null).assignedTo(USER_ID).status(Status.NEU).build());
				mongoOperations.save(VorgangTestFactory.createBuilder().id(null).assignedTo(null).build());
			}

			@Test
			void shouldReturnTotalOnAssignedTo() {
				var query = buildFilterQuery(filterCriteriaBuilder.assignedTo(USER_ID).build());

				assertThat(countTotalVorgaenge(query)).isEqualTo(1);
			}

			@Test
			void shouldReturnUnfilteredList() {
				var query = buildFilterQuery(filterCriteriaBuilder.filterByAssignedTo(false).assignedTo(StringUtils.EMPTY).build());

				assertThat(countTotalVorgaenge(query)).isEqualTo(2);
			}
		}

		@Nested
		class TestWithStatusToFilter {

			@BeforeEach
			void initData() {
				mongoOperations.save(VorgangTestFactory.createBuilder().id(null).status(Status.NEU).build());
				mongoOperations.save(VorgangTestFactory.createBuilder().id(null).status(Status.IN_BEARBEITUNG).build());
			}

			@Test
			void shouldReturnTotalOnStatus() {
				var query = buildFilterQuery(FilterCriteria.builder().singleStatus(Status.NEU).build());

				assertThat(countTotalVorgaenge(query)).isEqualTo(1);
			}

			@Test
			void shouldReturnUnfilteredList() {
				var query = buildFilterQuery(FilterCriteria.builder().clearStatus().build());

				assertThat(countTotalVorgaenge(query)).isEqualTo(2);
			}
		}

		@Nested
		class TestWithHasNextWiedervorlageFristFilter {

			@BeforeEach
			void initData() {
				mongoOperations.save(VorgangTestFactory.createBuilder().id("1")
						.clientAttributes(createClientAttributesMapWithWiedervorlageNextFrist(LocalDate.now().plus(1, ChronoUnit.DAYS))).build());
				mongoOperations.save(VorgangTestFactory.createBuilder().id(null).status(Status.NEU).build());
			}

			private ClientAttributesMap createClientAttributesMapWithWiedervorlageNextFrist(LocalDate nextFrist) {
				return ClientAttributesMapTestFactory.createWithAttribute(buildWiedervorlageNextFristClientAttribute(nextFrist));
			}

			private ClientAttribute buildWiedervorlageNextFristClientAttribute(LocalDate nextFrist) {
				return ClientAttributeTestFactory.createBuilder()
						.clientName(CriteriaUtil.DEFAULT_CLIENT)
						.attributeName(CriteriaUtil.CLIENT_ATTRIBUT_NEXT_WIEDERVORLAGE_FRIST)
						.stringValue(Optional.of(nextFrist.toString())).build();
			}

			@Test
			void shouldReturnTotalOnStatus() {
				var query = buildFilterQuery(FilterCriteria.builder().hasNextWiedervorlageFrist(true).build());

				assertThat(countTotalVorgaenge(query)).isEqualTo(1);
			}

			@Test
			void shouldReturnUnfilteredList() {
				var query = buildFilterQuery(FilterCriteria.builder().clearStatus().build());

				assertThat(countTotalVorgaenge(query)).isEqualTo(2);
			}
		}

		private Query buildFilterQuery(FilterCriteria filterCriteria) {
			return repositoryImpl.buildFilterQuery(filterCriteria);

		}

		private long countTotalVorgaenge(Query query) {
			return template.count(query, Vorgang.COLLECTION_NAME);
		}
	}

}