/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.vorgang;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.List;

import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.security.test.context.support.WithMockUser;
import org.springframework.test.context.bean.override.mockito.MockitoBean;

import de.ozgcloud.common.test.DataITCase;
import de.ozgcloud.notification.antragsteller.AntragstellerNotificationEventListener;
import de.ozgcloud.notification.user.UserNotificationEventListener;
import de.ozgcloud.vorgang.command.CommandService;
import de.ozgcloud.vorgang.servicekonto.ServiceKontoTestFactory;
import jakarta.validation.ConstraintViolationException;

@DataITCase
@WithMockUser
class VorgangServiceITCase {

	@Autowired
	private VorgangService vorgangService;

	@MockitoBean
	private CommandService commandService;
	@MockitoBean
	private AntragstellerNotificationEventListener antragstellerNotificationEventListener;
	@MockitoBean
	private UserNotificationEventListener userNotificationEventListener;

	@DisplayName("Test creating a new Vorgang")
	@Nested
	class TestVorgangCreation {
		@Test
		void shouldStartCreateVorgang() {
			var vorgang = vorgangService.startCreation(EingangTestFactory.create());

			assertThat(vorgang.getId()).isNotNull();
			assertThat(vorgang.isInCreation()).isTrue();
		}

		@Test
		void shouldFinishCreateVorgang() {
			var vorgang = vorgangService.startCreation(EingangTestFactory.create());

			vorgangService.finishCreation(vorgang.getId(), List.of(), List.of());

			vorgang = vorgangService.getById(vorgang.getId());

			assertThat(vorgang.isInCreation()).isFalse();
			verify(antragstellerNotificationEventListener, timeout(60000)).onVorgangCreated(any());
			verify(userNotificationEventListener, timeout(60000)).onVorgangCreated(any());
		}

		@DisplayName("validation")
		@Nested
		class TestValidation {

			@DisplayName("should be invalid on missing postfachAddress in serviceKonto")
			@Test
			void shouldBeInvalidOnMissingPostfachAddress() {
				var eingang = EingangTestFactory.createBuilder().header(EingangHeaderTestFactory.createBuilder()
								.serviceKonto(ServiceKontoTestFactory.createBuilder().clearPostfachAddresses().build()).build())
						.build();

				assertThatThrownBy(() -> vorgangService.startCreation(eingang)).isInstanceOf(ConstraintViolationException.class);
			}
		}
	}

	@Nested
	class TestLockVorgang {

		@Nested
		class TestValidation {

			@Test
			void shouldBeInvalidOnMissingClientName() {
				var lock = LockTestFactory.createBuilder().clientName(null).build();

				assertThatThrownBy(() -> vorgangService.lockVorgang(lock, VorgangTestFactory.ID, VorgangTestFactory.VERSION)).isInstanceOf(
						ConstraintViolationException.class);
			}

			@Test
			void shouldBeInvalidOnMissingLockedSince() {
				var lock = LockTestFactory.createBuilder().lockedSince(null).build();

				assertThatThrownBy(() -> vorgangService.lockVorgang(lock, VorgangTestFactory.ID, VorgangTestFactory.VERSION)).isInstanceOf(
						ConstraintViolationException.class);
			}

			@Test
			void shouldBeInvalidOnMissingReason() {
				var lock = LockTestFactory.createBuilder().reason(null).build();

				assertThatThrownBy(() -> vorgangService.lockVorgang(lock, VorgangTestFactory.ID, VorgangTestFactory.VERSION)).isInstanceOf(
						ConstraintViolationException.class);
			}

			@Test
			void shouldBeValidWhenRequiredFieldsAreNotNull() {
				assertThatNoException().isThrownBy(
						() -> vorgangService.lockVorgang(LockTestFactory.create(), VorgangTestFactory.ID, VorgangTestFactory.VERSION));
			}
		}
	}
}