/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.vorgang.redirect;

import static org.assertj.core.api.Assertions.*;
import static org.awaitility.Awaitility.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.concurrent.TimeUnit;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.EnumSource;
import org.junit.jupiter.params.provider.EnumSource.Mode;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.ApplicationEventPublisher;
import org.springframework.mail.javamail.JavaMailSender;
import org.springframework.test.context.bean.override.mockito.MockitoBean;

import de.ozgcloud.command.Command;
import de.ozgcloud.common.test.ITCase;
import de.ozgcloud.nachrichten.email.MailSentEventTestFactory;
import de.ozgcloud.vorgang.command.CommandCreatedEventTestFactory;
import de.ozgcloud.vorgang.command.CommandService;
import de.ozgcloud.vorgang.command.CommandTestFactory;
import de.ozgcloud.vorgang.command.Order;
import de.ozgcloud.vorgang.vorgang.VorgangService;
import de.ozgcloud.vorgang.vorgang.VorgangTestFactory;

@ITCase
class ForwardingEventListenerITCase {

	@Autowired
	private ApplicationEventPublisher publisher;

	@MockitoBean
	private ForwardingService service;
	@MockitoBean
	private JavaMailSender mailSender;
	@MockitoBean
	private VorgangService vorgangService;
	@MockitoBean
	private CommandService commandService;

	@Nested
	class TestRedirectdVorgang {

		@Test
		void shouldCallForwardingService() {
			var command = CommandTestFactory.createBuilder().order(Order.REDIRECT_VORGANG.name()).build();

			publisher.publishEvent(CommandCreatedEventTestFactory.create(command));

			await().atMost(5, TimeUnit.SECONDS).untilAsserted(() -> verify(service).forwardByCommand(command));
		}

	}

	@Nested
	class TestMailSentEvent {
		@Test
		void shouldCallForwardingService() {
			Forwarding forwarding = ForwardingTestFactory.create();

			publisher.publishEvent(MailSentEventTestFactory.create(forwarding));

			await().atMost(5, TimeUnit.SECONDS).untilAsserted(() -> verify(service).onRedirectMailSent(forwarding));
		}

		@ParameterizedTest
		@EnumSource(mode = Mode.EXCLUDE, names = { "REDIRECT_VORGANG" })
		void shouldNOTCallServiceOnAnyOtherOrder(Order order) {
			Command command = CommandTestFactory.createBuilder().order(order.name()).build();

			publisher.publishEvent(MailSentEventTestFactory.create(command));

			verify(service, never()).onRedirectMailSent(any());
		}
	}

	@Nested
	class CommandCreatedEvent {

		@Test
		void shouldCallForwardingService() {
			var command = CommandTestFactory.createBuilder().order(Order.FORWARD_SUCCESSFULL.name()).build();

			publisher.publishEvent(CommandCreatedEventTestFactory.create(command));

			await().atMost(5, TimeUnit.SECONDS).untilAsserted(() -> verify(service).markAsSuccessfull(command));
		}
	}

	@Nested
	class TestForwardVorgang {

		private Command command = CommandTestFactory.createBuilder().order(Order.FORWARD_VORGANG.name()).build();
		@Captor
		private ArgumentCaptor<ForwardingRequest> requestCaptor;

		@BeforeEach
		void init() {
			when(vorgangService.getById(any())).thenReturn(VorgangTestFactory.create());
		}

		@Test
		void shouldCallForwardingService() {
			publishEvent();

			await().atMost(5, TimeUnit.SECONDS).untilAsserted(() -> verify(service).forward(requestCaptor.capture()));
			assertThat(requestCaptor.getValue().getVorgangId()).isEqualTo(command.getVorgangId());
		}

		@Test
		void shouldSetCommandFinished() {
			publishEvent();

			await().atMost(5, TimeUnit.SECONDS).untilAsserted(() -> verify(commandService).setCommandFinished(command.getId(), null));
		}

		@Test
		void shouldSetCommandFailed() {
			doThrow(RuntimeException.class).when(service).forward(any());

			publishEvent();

			await().atMost(5, TimeUnit.SECONDS).untilAsserted(() -> verify(commandService).setCommandError(eq(command.getId()), any()));
		}

		private void publishEvent() {
			publisher.publishEvent(CommandCreatedEventTestFactory.create(command));
		}
	}

}
