/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.vorgang.redirect;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.ConcurrentModificationException;
import java.util.Map;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;
import org.springframework.context.ApplicationEventPublisher;

import com.thedeanda.lorem.LoremIpsum;

import de.ozgcloud.command.Command;
import de.ozgcloud.command.CommandCreatedEvent;
import de.ozgcloud.command.CommandFailedEvent;
import de.ozgcloud.command.VorgangLockedEvent;
import de.ozgcloud.common.errorhandling.TechnicalException;
import de.ozgcloud.nachrichten.email.MailSentEventTestFactory;
import de.ozgcloud.vorgang.command.CommandCreatedEventTestFactory;
import de.ozgcloud.vorgang.command.CommandTestFactory;
import de.ozgcloud.vorgang.command.Order;
import de.ozgcloud.vorgang.vorgang.Lock;
import de.ozgcloud.vorgang.vorgang.LockMapper;
import de.ozgcloud.vorgang.vorgang.LockTestFactory;
import de.ozgcloud.vorgang.vorgang.VorgangService;
import de.ozgcloud.vorgang.vorgang.VorgangTestFactory;

class ForwardingEventListenerTest {

	@Spy
	@InjectMocks // NOSONAR
	private ForwardingEventListener listener;

	@Mock
	private ForwardingService forwardingService;
	@Mock
	private ApplicationEventPublisher publisher;
	@Mock
	private ForwardingRequestMapper forwardingRequestMapper;
	@Mock
	private VorgangService vorgangService;
	@Mock
	private LockMapper lockMapper;

	@Nested
	class OnRedirectVorgangCommand {

		@Test
		void shouldCallService() {
			listener.onRedirectVorgangOrder(CommandCreatedEventTestFactory.create());

			verify(forwardingService).forwardByCommand(notNull());
		}

		@Test
		void shouldPublishFailedEventOnError() {
			when(forwardingService.forwardByCommand(any())).thenThrow(ConcurrentModificationException.class);

			listener.onRedirectVorgangOrder(CommandCreatedEventTestFactory.create());

			verify(publisher).publishEvent(any(CommandFailedEvent.class));
		}
	}

	@Nested
	class OnMailSentEvent {

		private final Forwarding forwarding = ForwardingTestFactory.create();

		@Test
		void shouldCallRedirectService() {
			listener.triggerRedirectMailSent(MailSentEventTestFactory.create(forwarding));

			verify(forwardingService).onRedirectMailSent(forwarding);
		}

	}

	@Nested
	class TestOnForwardVorgangOrder {

		private final CommandCreatedEvent event = CommandCreatedEventTestFactory.create();

		@Captor
		private ArgumentCaptor<CommandFailedEvent> failedEventCaptor;

		@Test
		void shouldHandleForwardVorgangCommand() {
			doNothing().when(listener).handleForwardVorgangCommand(any());

			onForwardVorgangOrder();

			verify(listener).handleForwardVorgangCommand(event.getSource());
		}

		@Test
		void shouldPublishFailedEventOnError() {
			doThrow(TechnicalException.class).when(listener).handleForwardVorgangCommand(any());

			onForwardVorgangOrder();

			verify(publisher).publishEvent(failedEventCaptor.capture());
			assertThat(failedEventCaptor.getValue().getSource()).isEqualTo(event.getSource().getId());
		}

		private void onForwardVorgangOrder() {
			listener.onForwardVorgangOrder(event);
		}
	}

	@Nested
	class TestHandleForwardVorgangCommand {

		private final Command command = CommandTestFactory.createBuilder().bodyObject(commandBody()).build();
		private final Lock lock = LockTestFactory.create();
		@Captor
		private ArgumentCaptor<Lock> lockCaptor;
		@Captor
		private ArgumentCaptor<VorgangLockedEvent> lockedEventCaptor;
		@Captor
		private ArgumentCaptor<ForwardingRequest> requestCaptor;

		@BeforeEach
		void init() {
			when(forwardingRequestMapper.fromCommand(any())).thenReturn(ForwardingRequestTestFactory.create());
			when(lockMapper.fromCommand(any())).thenReturn(lock);
		}

		@Test
		void shouldMapToLock() {
			handleForwardVorgangCommand();

			verify(lockMapper).fromCommand(command);
		}

		@Test
		void shouldLockVorgang() {
			handleForwardVorgangCommand();

			verify(vorgangService).lockVorgang(lockCaptor.capture(), eq(VorgangTestFactory.ID), eq(CommandTestFactory.RELATION_VERSION));
			assertThat(lockCaptor.getValue()).usingRecursiveComparison()
					.isEqualTo(LockTestFactory.createBuilder().reason("Vorgang was forwarded").build());
		}

		@Test
		void shouldMapCommandToForwardingRequest() {
			handleForwardVorgangCommand();

			verify(forwardingRequestMapper).fromCommand(command);
		}

		@Test
		void shouldCallForwardingService() {
			handleForwardVorgangCommand();

			verify(forwardingService).forward(requestCaptor.capture());
			assertThat(requestCaptor.getValue()).usingRecursiveComparison().ignoringFields("version")
					.isEqualTo(ForwardingRequestTestFactory.create());
		}

		@Test
		void shouldIncrementVersionInRequest() {
			handleForwardVorgangCommand();

			verify(forwardingService).forward(requestCaptor.capture());
			assertThat(requestCaptor.getValue().getVersion()).isEqualTo(VorgangTestFactory.VERSION + 1);
		}

		@Test
		void shouldPublishVorgangLockedEvent() {
			handleForwardVorgangCommand();

			verify(publisher).publishEvent(lockedEventCaptor.capture());
			assertThat(lockedEventCaptor.getValue().getCommand()).isSameAs(command);
		}

		private void handleForwardVorgangCommand() {
			listener.handleForwardVorgangCommand(command);
		}

		private static Map<String, Object> commandBody() {
			return Map.of(LoremIpsum.getInstance().getWords(1), LoremIpsum.getInstance().getWords(1));
		}
	}

	@Nested
	class TestMarkAsSuccessfull {

		@Test
		void shouldCallService() {
			Command command = CommandTestFactory.createBuilder().order(Order.FORWARD_SUCCESSFULL.name()).build();

			listener.markAsSuccessfull(CommandCreatedEventTestFactory.create(command));

			verify(forwardingService).markAsSuccessfull(command);
		}

		@Test
		void shouldPublishFailedEventOnError() {
			doThrow(RuntimeException.class).when(forwardingService).markAsSuccessfull(any());

			listener.markAsSuccessfull(CommandCreatedEventTestFactory.create());

			verify(publisher).publishEvent(any(CommandFailedEvent.class));
		}

	}

	@Nested
	class TestMarkAsFailed {

		@Test
		void shouldCallService() {
			Command command = CommandTestFactory.createBuilder().order(Order.FORWARD_FAILED.name()).build();

			listener.markAsFailed(CommandCreatedEventTestFactory.create(command));

			verify(forwardingService).markAsFailed(command);
		}

	}
}