/*
 * Copyright (C) 2025 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.vorgang.redirect;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.Mockito.*;

import java.util.Map;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mapstruct.factory.Mappers;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import de.ozgcloud.eingang.forwarding.GrpcRouteCriteria;
import de.ozgcloud.eingang.forwarding.GrpcRouteForwarding;
import de.ozgcloud.vorgang.command.CommandTestFactory;
import de.ozgcloud.vorgang.vorgang.Eingang;
import de.ozgcloud.vorgang.vorgang.EingangMapper;
import de.ozgcloud.vorgang.vorgang.EingangTestFactory;
import de.ozgcloud.vorgang.vorgang.GrpcEingang;
import de.ozgcloud.vorgang.vorgang.GrpcEingangTestFactory;
import de.ozgcloud.vorgang.vorgang.redirect.ForwardingRequestMapper.CommandBodyFields;

class ForwardingRequestMapperTest {

	@Mock
	private EingangMapper eingangMapper;
	@Spy
	@InjectMocks
	private ForwardingRequestMapper mapper = Mappers.getMapper(ForwardingRequestMapper.class);
	private final Map<String, Object> bodyObject = Map.of(
			CommandBodyFields.ORGANISATION_EINHEIT_ID, OrganisationEinheitTestFactory.ID,
			CommandBodyFields.NAME, OrganisationEinheitTestFactory.NAME,
			CommandBodyFields.STRASSE, AddressTestFactory.STREET,
			CommandBodyFields.HAUSNUMMER, AddressTestFactory.HOUSE_NUMBER,
			CommandBodyFields.PLZ, AddressTestFactory.ZIP_CODE,
			CommandBodyFields.ORT, AddressTestFactory.CITY
	);

	@Nested
	class TestFromCommand {

		@BeforeEach
		void init() {
			doReturn(OrganisationEinheitTestFactory.create()).when(mapper).toOrganisationEinheit(any());
		}

		@Test
		void shouldMapToOrganisationEinheit() {
			map();

			verify(mapper).toOrganisationEinheit(bodyObject);
		}

		@Test
		void shouldMapForwardingRequestFields() {
			var mapped = map();

			assertThat(mapped).usingRecursiveComparison().isEqualTo(ForwardingRequestTestFactory.create());
		}

		private ForwardingRequest map() {
			return mapper.fromCommand(CommandTestFactory.createBuilder().bodyObject(bodyObject).build());
		}
	}

	@Nested
	class TestToOrganisationEinheit {

		@BeforeEach
		void init() {
			doReturn(AddressTestFactory.create()).when(mapper).toAddress(any());
		}

		@Test
		void shouldMapToAddress() {
			map();

			verify(mapper).toAddress(bodyObject);
		}

		@Test
		void shouldMapOrganisationEinheitFields() {
			var mapped = map();

			assertThat(mapped).usingRecursiveComparison().isEqualTo(OrganisationEinheitTestFactory.create());
		}

		private OrganisationEinheit map() {
			return mapper.toOrganisationEinheit(bodyObject);
		}
	}

	@Nested
	class TestToAddress {

		@Test
		void shouldMapAddressFields() {
			var mapped = map();

			assertThat(mapped).usingRecursiveComparison().isEqualTo(AddressTestFactory.create());
		}

		private Address map() {
			return mapper.toAddress(bodyObject);
		}
	}

	@Nested
	class TestToGrpcRouteForwarding {

		private final ForwardingRequest request = ForwardingRequestTestFactory.create();
		private final Eingang eingang = EingangTestFactory.create();
		private final GrpcEingang grpcEingang = GrpcEingangTestFactory.create();

		@BeforeEach
		void init() {
			doReturn(GrpcRouteCriteriaTestFactory.create()).when(mapper).toGrpcRouteCriteria(any());
			when(eingangMapper.toGrpc(any())).thenReturn(grpcEingang);
		}

		@Test
		void shouldCallEingangMapper() {
			map();

			verify(eingangMapper).toGrpc(eingang);
		}

		@Test
		void shouldMapToGrpcRouteCriteria() {
			map();

			verify(mapper).toGrpcRouteCriteria(request);
		}

		@Test
		void shouldMapToGrpcEingangStub() {
			map();

			verify(mapper).toGrpcEingangStub(grpcEingang);
		}

		@Test
		void shouldMapToGrpcRouteForwarding() {
			var mapped = map();

			assertThat(mapped).isEqualTo(GrpcRouteForwardingTestFactory.create());
		}

		private GrpcRouteForwarding map() {
			return mapper.toGrpcRouteForwarding(request, eingang);
		}
	}

	@Nested
	class TestToGrpcRouteCriteria {

		@Test
		void shouldMap() {
			var mapped = map();

			assertThat(mapped).isEqualTo(GrpcRouteCriteriaTestFactory.create());
		}

		private GrpcRouteCriteria map() {
			return mapper.toGrpcRouteCriteria(ForwardingRequestTestFactory.create());
		}
	}
}
