/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.vorgang.redirect;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.zip.ZipEntry;

import org.apache.commons.io.IOUtils;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.springframework.test.util.ReflectionTestUtils;

import de.ozgcloud.vorgang.files.FileId;
import de.ozgcloud.vorgang.files.FileService;
import de.ozgcloud.vorgang.vorgang.IncomingFileGroupTestFactory;
import de.ozgcloud.vorgang.vorgang.IncomingFileTestFactory;
import de.ozgcloud.vorgang.vorgang.VorgangTestFactory;
import de.ozgcloud.vorgang.vorgang.redirect.ZipBuilderService.ZipBuilder;
import net.lingala.zip4j.ZipFile;
import net.lingala.zip4j.io.outputstream.ZipOutputStream;
import net.lingala.zip4j.model.ZipParameters;
import net.lingala.zip4j.model.enums.EncryptionMethod;

class ZipBuilderServiceTest {

	@InjectMocks // NOSONAR
	private ZipBuilderService service;

	private ZipBuilder builder;

	@Mock
	private ZipOutputStream zipOut;

	@Mock
	private FileService fileService;

	@BeforeEach
	void init() {
		when(fileService.getUploadedFileStream(any(FileId.class))).thenReturn(new ByteArrayInputStream(IncomingFileTestFactory.CONTENT));
		builder = spy(service.new ZipBuilder(VorgangTestFactory.create(), true));

		ReflectionTestUtils.setField(builder, "zipOut", zipOut);
	}

	@Nested
	class TestAddFile {

		@Captor // NOSONAR
		private ArgumentCaptor<ZipEntry> entryCaptor;
		@Captor
		private ArgumentCaptor<ZipParameters> parametersCaptor;

		@Test
		void shouldAddEntry() throws IOException {
			builder.addFile("", IncomingFileTestFactory.create());

			verify(zipOut).putNextEntry(parametersCaptor.capture());
			assertThat(parametersCaptor.getValue().getFileNameInZip()).isEqualTo(IncomingFileTestFactory.NAME);
		}

		@Test
		void shouldEnableEncryption() throws IOException {
			builder.addFile("", IncomingFileTestFactory.create());

			verify(zipOut).putNextEntry(parametersCaptor.capture());
			assertThat(parametersCaptor.getValue().isEncryptFiles()).isTrue();
			assertThat(parametersCaptor.getValue().getEncryptionMethod()).isEqualTo(EncryptionMethod.AES);
		}

		@Test
		void shouldPutNextEntry() throws Exception {
			builder.addFile("", IncomingFileTestFactory.create());

			verify(zipOut).putNextEntry(notNull());
		}

		@Test
		void shouldCloseEntry() throws IOException {
			builder.addFile("", IncomingFileTestFactory.create());

			verify(zipOut).closeEntry();
		}

	}

	@Nested
	class TestAddFileGroup {
		@Test
		void shouldAddInFolderNamedByGroup() {
			builder.addFileGroup(IncomingFileGroupTestFactory.create());

			verify(builder).addFile(startsWith(IncomingFileGroupTestFactory.NAME), same(IncomingFileGroupTestFactory.FILE));
		}
	}

	@Nested
	class TestBuildZip {

		@Nested
		class TestWithEntcryption {

			@BeforeEach
			void init() {
				builder = spy(service.new ZipBuilder(VorgangTestFactory.create(), true));
			}

			@Test
			void shouldValidZipFile() throws IOException {
				var zipFile = new ZipFile(buildZippedFile());

				assertThat(zipFile.isValidZipFile()).isTrue();
			}

			@Test
			void shouldBeEncrypted() throws IOException {
				var zipFile = new ZipFile(buildZippedFile());

				assertThat(zipFile.isEncrypted()).isTrue();
			}
		}

		@Nested
		class TestWithNoEntcryption {

			@BeforeEach
			void init() {
				builder = spy(service.new ZipBuilder(VorgangTestFactory.create(), false));
			}

			@Test
			void shouldValidZipFile() throws IOException {
				var zipFile = new ZipFile(buildZippedFile());

				assertThat(zipFile.isValidZipFile()).isTrue();
			}

			@Test
			void shouldBeEncrypted() throws IOException {
				var zipFile = new ZipFile(buildZippedFile());

				assertThat(zipFile.isEncrypted()).isFalse();
			}

		}

		private File buildZippedFile() throws FileNotFoundException, IOException {
			byte[] zipData = builder.buildZip(RedirectRequestTestFactory.PASSWORD.toCharArray()).toByteArray();

			File outFile = File.createTempFile("ea_", ".zip");
			outFile.deleteOnExit();
			IOUtils.write(zipData, new FileOutputStream(outFile));

			return outFile;
		}
	}
}
