/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.common.grpc;

import static de.ozgcloud.vorgang.common.grpc.FormDataMappingUtils.*;

import java.util.Collection;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Objects;

import org.apache.commons.lang3.StringUtils;
import org.mapstruct.CollectionMappingStrategy;
import org.mapstruct.Mapper;
import org.mapstruct.NullValueCheckStrategy;
import org.mapstruct.NullValuePropertyMappingStrategy;
import org.mapstruct.ReportingPolicy;

import de.ozgcloud.vorgang.vorgang.GrpcFormData;
import de.ozgcloud.vorgang.vorgang.GrpcFormField;
import de.ozgcloud.vorgang.vorgang.GrpcSubForm;

@Mapper(unmappedTargetPolicy = ReportingPolicy.WARN, //
		unmappedSourcePolicy = ReportingPolicy.WARN, //
		nullValuePropertyMappingStrategy = NullValuePropertyMappingStrategy.IGNORE, //
		nullValueCheckStrategy = NullValueCheckStrategy.ALWAYS, //
		collectionMappingStrategy = CollectionMappingStrategy.ADDER_PREFERRED)
public interface GrpcFormDataMapper {

	default GrpcFormData mapToFormData(Map<String, Object> formData) {
		return GrpcFormData.newBuilder()
				.addAllField(mapToAllFields(formData))
				.addAllForm(mapToSubForms(formData))
				.addAllField(mapStringListsToFields(formData))
				.addAllForm(mapObjectListsToFields(formData)).build();
	}

	default List<GrpcFormField> mapToAllFields(Map<String, Object> formData) {
		return formData.entrySet().stream()
				.filter(entry -> isValue(entry.getValue()))
				.map(entry -> mapToField(entry.getKey(), entry.getValue()))
				.toList();
	}

	default List<GrpcFormField> mapStringListsToFields(Map<String, Object> formData) {
		return formData.entrySet().stream()
				.filter(entry -> getValueObject(entry.getValue()) instanceof List)
				.map(this::mapListStringElementsToFormFields)
				.flatMap(List::stream)
				.toList();
	}

	@SuppressWarnings("unchecked")
	private List<GrpcFormField> mapListStringElementsToFormFields(Entry<String, Object> entry) {
		return ((List<Object>) getValueObject(entry.getValue())).stream()
				.filter(FormDataMappingUtils::isValue)
				.map(valueContainer -> mapToField(entry.getKey(), valueContainer))
				.map(field -> getLabel(entry.getValue()).map(label -> field.toBuilder().setLabel(label).build()).orElse(field))
				.toList();
	}

	default List<GrpcSubForm> mapObjectListsToFields(Map<String, Object> formData) {
		return formData.entrySet().stream()
				.filter(entry -> getValueObject(entry.getValue()) instanceof List)
				.map(this::mapListObjectElementsToFormFields)
				.flatMap(List::stream)
				.toList();
	}

	@SuppressWarnings("unchecked")
	default List<GrpcSubForm> mapListObjectElementsToFormFields(Entry<String, Object> entry) {
		return ((List<Object>) getValueObject(entry.getValue())).stream()
				.filter(FormDataMappingUtils::isSubForm)
				.map(Map.class::cast)
				.map(valueMap -> buildSubForm(entry.getKey(), valueMap))
				.toList();
	}

	@SuppressWarnings("unchecked")
	default List<GrpcSubForm> mapToSubForms(Map<String, Object> formData) {
		return formData.entrySet().stream()
				.filter(entry -> isSubForm(entry.getValue()))
				.map(entry -> buildSubForm(entry.getKey(), (Map<String, Object>) entry.getValue()))
				.toList();
	}

	default GrpcSubForm buildSubForm(String name, Map<String, Object> map) {
		Map<String, Object> formMap = getValueMap(map);

		return GrpcSubForm.newBuilder().setTitle(name)
				.setLabel(getLabel(map).orElse(StringUtils.EMPTY))
				.addAllField(mapToAllFields(formMap))
				.addAllSubForm(mapToSubForms(formMap))
				.addAllField(mapStringListsToFields(formMap))
				.addAllSubForm(mapObjectListsToFields(formMap)).build();
	}

	default GrpcFormField mapToField(String name, Object valueContainer) {
		var fieldBuilder = GrpcFormField.newBuilder().setName(name).setValue(getValue(valueContainer));
		getLabel(valueContainer).ifPresent(fieldBuilder::setLabel);
		return fieldBuilder.build();
	}

	default Map<String, Object> mapFromFormData(GrpcFormData formData) {
		Map<String, Object> result = new LinkedHashMap<>();
		result.putAll(mapSubFormFields(formData.getFieldList()));
		result.putAll(mapFormData(formData.getFormList()));
		return result;
	}

	default Map<String, Object> mapFormData(List<GrpcSubForm> subForms) {
		var subFormMap = subForms.stream()
				.map(subForm -> new FormDataContainer(subForm.getTitle(), subForm.getLabel(), mapSubForm(subForm)))
				.collect(LinkedHashMap::new, this::formDataAccumulator, this::combiner);

		return removeLeadingValueKey(subFormMap);
	}

	private Map<String, Object> formDataAccumulator(Map<String, Object> map, FormDataContainer container) {
		return addToMap(map, container.name, container.formData, container.label);
	}

	default Map<String, Object> mapSubForm(GrpcSubForm subForm) {
		Map<String, Object> result = new LinkedHashMap<>();
		result.putAll(mapSubFormFields(subForm.getFieldList()));
		result.putAll(mapFormData(subForm.getSubFormList()));

		return result;
	}

	default Map<String, Object> mapSubFormFields(List<GrpcFormField> fields) {
		var fieldMap = fields.stream().collect(LinkedHashMap::new, this::accumulator, this::combiner);

		return removeLeadingValueKey(fieldMap);
	}

	@SuppressWarnings("unchecked")
	private Map<String, Object> removeLeadingValueKey(Map<String, Object> formMap) {
		if (formMap.containsKey(VALUE_KEY)) {
			return (Map<String, Object>) formMap.get(VALUE_KEY);
		}
		return formMap;
	}

	private Map<String, Object> accumulator(Map<String, Object> map, GrpcFormField field) {
		return addToMap(map, field.getName(), field.getValue(), field.getLabel());
	}

	default Map<String, Object> combiner(Map<String, Object> map1, Map<String, Object> map2) {
		for (Entry<String, Object> entry : map2.entrySet()) {
			addToMap(map1, entry.getKey(), entry.getValue(), null);
		}

		return map1;
	}

	@SuppressWarnings("unchecked")
	private Map<String, Object> addToMap(Map<String, Object> map, String name, Object value, String label) {
		var valueInMap = map.get(name);
		if (Objects.isNull(valueInMap)) {
			map.put(name, buildFieldMap(value, label));
		} else if (value instanceof Collection) {
			addCollectionValue(map, name, (Collection<Object>) value, label);
		} else {
			addNonCollectionValue(map, name, value, label);
		}

		return map;
	}

	private Map<String, Object> buildFieldMap(Object value, String label) {
		if (StringUtils.isNotBlank(label)) {
			return Map.of(LABEL_KEY, label, VALUE_KEY, value);
		}
		return Map.of(VALUE_KEY, value);
	}

	@SuppressWarnings("unchecked")
	private Map<String, Object> addCollectionValue(Map<String, Object> map, String name, Collection<Object> value, String label) {
		var valueInMap = getValueObject(map.get(name));

		var newList = new LinkedList<Object>();
		if (valueInMap instanceof Collection) {
			newList.addAll((Collection<? extends Object>) valueInMap);
			newList.addAll(value);
		} else {
			newList.add(valueInMap);
			newList.addAll(value);
		}
		map.put(name, buildWithLabel(newList, label));
		return map;
	}

	@SuppressWarnings("unchecked")
	private Map<String, Object> addNonCollectionValue(Map<String, Object> map, String name, Object value, String label) {
		var valueInMap = getValueObject(map.get(name));

		if (valueInMap instanceof Collection) {
			var newList = new LinkedList<>((Collection<Object>) valueInMap);
			newList.add(value);
			map.put(name, buildWithLabel(newList, label));
		} else {
			map.put(name, buildWithLabel(List.of(valueInMap, value), label));
		}

		return map;
	}

	private Map<String, Object> buildWithLabel(Object value, String label) {
		if (StringUtils.isBlank(label)) {
			return Map.of(VALUE_KEY, value);
		}
		return Map.of(LABEL_KEY, label, VALUE_KEY, value);
	}

	record FormDataContainer(String name, String label, Map<String, Object> formData) {

	}
}
