/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.common.grpc;

import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.function.Predicate;

import org.apache.commons.lang3.StringUtils;
import org.apache.commons.lang3.tuple.Pair;
import org.mapstruct.Mapper;

import de.ozgcloud.vorgang.common.GrpcObject;
import de.ozgcloud.vorgang.common.GrpcProperty;
import de.ozgcloud.vorgang.common.GrpcSubObject;

@Mapper
public interface GrpcObjectMapper {

	Predicate<Entry<String, Object>> IS_VALUE_INSTANCEOF_MAP = entry -> entry.getValue() instanceof Map;

	default GrpcObject fromMap(Map<String, Object> map) {
		return GrpcObject.newBuilder()
				.addAllProperty(mapToAllProperties(map))
				.addAllSubObject(mapToSubObjects(map))
				.build();
	}

	@SuppressWarnings("unchecked")
	default List<GrpcSubObject> mapToSubObjects(Map<String, Object> propertyMap) {
		return propertyMap.entrySet().stream()
				.filter(IS_VALUE_INSTANCEOF_MAP)
				.map(entry -> buildSubObject(entry.getKey(), (Map<String, Object>) entry.getValue()))
				.toList();
	}

	default GrpcSubObject buildSubObject(String name, Map<String, Object> map) {
		return GrpcSubObject.newBuilder().setName(name)
				.addAllProperty(mapToAllProperties(map))
				.addAllSubObject(mapToSubObjects(map))
				.build();
	}

	default List<GrpcProperty> mapToAllProperties(Map<String, Object> propertyMap) {
		return propertyMap.entrySet().stream()
				.filter(IS_VALUE_INSTANCEOF_MAP.negate())
				.map(entry -> mapToProperty(entry.getKey(), entry.getValue()))
				.toList();
	}

	default GrpcProperty mapToProperty(String name, Object value) {
		if (value instanceof Collection<?> valueList) {
			return buildPropertyForValueList(name, valueList);
		}
		return buildPropertyForValueObject(name, value);
	}

	private GrpcProperty buildPropertyForValueList(String name, Collection<?> valueList) {
		return GrpcProperty.newBuilder().setName(name).addAllValue(valueList.stream().map(this::getStringValue).toList()).build();
	}

	private GrpcProperty buildPropertyForValueObject(String name, Object value) {
		return GrpcProperty.newBuilder().setName(name).addValue(getStringValue(value)).build();
	}

	// TODO do null value handling externally configurable
	private String getStringValue(Object value) {
		return Optional.ofNullable(value).map(Object::toString).orElse(StringUtils.EMPTY);
	}

	default Map<String, Object> mapFromGrpc(GrpcObject bodyObject) {
		var map = new HashMap<String, Object>();
		bodyObject.getPropertyList().stream().map(this::mapProperty).forEach(entry -> map.put(entry.getKey(), entry.getValue()));
		bodyObject.getSubObjectList().forEach(subObject -> map.put(subObject.getName(), mapSubObject(subObject)));

		return map;
	}

	default Map<String, Object> mapSubObject(GrpcSubObject object) {
		var map = new HashMap<String, Object>();
		object.getPropertyList().stream().map(this::mapProperty).forEach(entry -> map.put(entry.getKey(), entry.getValue()));
		object.getSubObjectList().forEach(subObject -> map.put(subObject.getName(), mapSubObject(subObject)));

		return map;
	}

	default Pair<String, Object> mapProperty(GrpcProperty property) {
		return switch (property.getValueCount()) {
		case 0 -> Pair.of(property.getName(), null);
		case 1 -> Pair.of(property.getName(), property.getValue(0));
		default -> Pair.of(property.getName(), property.getValueList().stream().map(StringUtils::trimToNull).toList());
		};
	}
}